import React from 'react';
import PropTypes from 'prop-types';
import classNames from 'classnames';
import { observer } from 'mobx-react';
import { withStyles } from '@material-ui/core/styles';
import Table from '@material-ui/core/Table';
import TableBody from '@material-ui/core/TableBody';
import TableCell from '@material-ui/core/TableCell';
import TableHead from '@material-ui/core/TableHead';
import TableRow from '@material-ui/core/TableRow';
import Menu from '@material-ui/core/Menu';
import MenuItem from '@material-ui/core/MenuItem';
import InputBase from '@material-ui/core/InputBase';
import TextField from '@material-ui/core/TextField';
import Button from '@material-ui/core/Button';
import Tooltip from '@material-ui/core/Tooltip';
import FocusInput from '../utilities/FocusInput';

const styles = theme => ({
  root: {
    fontSize: 11,
    width: 275,
  },
  disabled: {
    opacity: 0.4,
    pointerEvents: 'none',
  },
  table: {
    tableLayout: 'fixed',
    '& tr:nth-child(even)': {
      backgroundColor: '#fafafa',
    },
  },
  tableRowStyle: {
    height: 'auto',
  },
  tableHeaderCellStyle: {
    fontWeight: 'bold',
    height: 16,
    lineHeight: '16px',
    fontSize: 11,
    padding: 4,
    margin: 0,
    color: 'rgb(26, 50, 115)',
    backgroundColor: '#e7f0d8',
    textTransform: 'capitalize',
    width: 118,
    '&:first-child': {
      width: 40,
      textAlign: 'center',
    },
    '&:last-child': {
      paddingRight: 4,
    },
  },
  tableEditCellStyle: {
    height: 22,
    padding: 0,
    margin: 0,
    color: 'rgb(26, 50, 115)',
    '&:first-child': {
      width: 40,
      textAlign: 'center',
    },
    '&:last-child': {
      paddingRight: 0,
    },
  },
  editField: {
    fontSize: 11,
    height: 22,
    lineHeight: '22px',
    margin: 0,
    padding: 0,
  },
  input: {
    textAlign: 'right',
    padding: '0px 4px 0px 0px',
    margin: 0,
    '&:hover': {
      borderBottom: `1px solid ${theme.palette.primary.main}`,
      marginBottom: -1,
    },
  },
  tableCellRoot: {
    borderBottom: 'none',
  },
  copyPasteDiv: {
    position: 'relative',
  },
  copyPasteInput: {
    fontSize: 11,
    lineHeight: '18px',
    '& ::selection': {
      backgroundColor: 'rgb(0, 0, 0, 0.14)',
    },
  },
  tooltip: {
    position: 'absolute',
    top: 0,
    right: 0,
  },
  tooltipInner: {
    fontSize: 11,
  },
});

@withStyles(styles)
@observer
class DataTable extends React.Component {
  static propTypes = {
    classes: PropTypes.object.isRequired,
    dataArray: PropTypes.object,
    onChange: PropTypes.func,
    disabled: PropTypes.bool,
  };

  static defaultProps = {
    disabled: false,
    onChange: null,
    dataArray: null,
  };

  state = {
    contextMenuEvent: null,
    displayMode: 'STRATIFIED',
  };

  componentDidUpdate = (prevProps, prevState) => {
    const { displayMode } = this.state;
    if (
      prevState.displayMode !== 'COPY' &&
      displayMode === 'COPY' &&
      this.copyField
    ) {
      FocusInput(this.copyField, 1);
    } else if (
      prevState.displayMode !== 'PASTE' &&
      displayMode === 'PASTE' &&
      this.pasteField
    ) {
      FocusInput(this.pasteField, 1);
    }
  };

  handleTableRightClick = event => {
    event.preventDefault();
    event.persist();
    this.setState({
      contextMenuEvent: event,
    });
  };

  handleCloseContextMenu = event => {
    event.preventDefault();
    event.persist();
    this.setState({
      contextMenuEvent: null,
    });
  };

  handleTableCellClick = event => {
    event.preventDefault();
    event.persist();
    if (event.target.select) {
      event.target.select();
    }
  };

  handleTableCellChange = event => {
    const { onChange } = this.props;
    event.preventDefault();
    event.persist();

    const row = event.target.attributes.row.value;
    const col = event.target.attributes.col.value;
    const val = event.target.value;
    const { dataArray } = this.props;

    dataArray.setValue(row, col, val);

    if (onChange) {
      onChange(dataArray);
    }
  };

  handleDataArrayFill = () => {
    const { contextMenuEvent } = this.state;
    const { onChange } = this.props;

    if (contextMenuEvent) {
      const fillValue = contextMenuEvent.nativeEvent.target.value;
      const { dataArray } = this.props;
      dataArray.fillWithValue(fillValue);

      if (onChange) {
        onChange(dataArray);
      }
    }
    this.handleCloseContextMenu(contextMenuEvent);
  };

  handleViewSwitch = view => () => {
    const { contextMenuEvent } = this.state;
    if (contextMenuEvent) {
      this.handleCloseContextMenu(contextMenuEvent);
    }

    this.setState({
      displayMode: view,
    });
  };

  getHeaderCell = val => val;

  getBodyCell = (val, i, j, classes) => (
    <InputBase
      inputProps={{ row: i, col: j }}
      value={val}
      margin="dense"
      className={classes.editField}
      classes={{
        input: classes.input,
      }}
      onClick={this.handleTableCellClick}
      onChange={this.handleTableCellChange}
    />
  );

  getTableRow = (i, dataRow, classes, cellClassName, getCell) => {
    const cells = dataRow.rowVals.map((val, j) => {
      const key = j;
      return (
        <TableCell
          align="right"
          key={key}
          className={cellClassName}
          classes={{
            root: classes.tableCellRoot,
          }}
        >
          {getCell(val, i, j, classes)}
        </TableCell>
      );
    });

    return (
      <TableRow key={dataRow.rowName} className={classes.tableRowStyle}>
        <TableCell
          className={cellClassName}
          classes={{
            root: classes.tableCellRoot,
          }}
        >
          {dataRow.rowName}
        </TableCell>
        {cells}
      </TableRow>
    );
  };

  render() {
    const { contextMenuEvent, displayMode } = this.state;
    const { classes, dataArray, disabled } = this.props;

    if (!dataArray) {
      return null;
    }

    let contextMenuPosition;
    let contextMenuAnchor;
    let fillMenuOption = null;
    if (contextMenuEvent !== null) {
      contextMenuAnchor = contextMenuEvent.target;
      contextMenuPosition = {
        left: contextMenuEvent.clientX,
        top: contextMenuEvent.clientY,
      };
      if (contextMenuEvent.target.localName === 'input') {
        fillMenuOption = (
          <MenuItem onClick={this.handleDataArrayFill}>
            Fill table with this value
          </MenuItem>
        );
      }
    } else {
      contextMenuAnchor = null;
      contextMenuPosition = {
        left: 0,
        top: 0,
      };
    }

    let content;
    if (displayMode === 'STRATIFIED') {
      let tableHeaderRow = [];
      let tableBodyRows = [];
      if (dataArray) {
        const dataObject = dataArray.asObject;

        const dimNames = Object.keys(dataArray.dimOrder);
        if (dimNames.length === 2) {
          tableHeaderRow = this.getTableRow(
            0,
            dataObject.headRow,
            classes,
            classes.tableHeaderCellStyle,
            this.getHeaderCell,
            disabled,
          );
          tableBodyRows = dataObject.bodyRows.map((bodyRow, i) =>
            this.getTableRow(
              i,
              bodyRow,
              classes,
              classes.tableEditCellStyle,
              this.getBodyCell,
              disabled,
            ),
          );
        }
      }

      content = (
        <Table
          className={classes.table}
          onContextMenu={this.handleTableRightClick}
        >
          <TableHead>{tableHeaderRow}</TableHead>
          <TableBody>{tableBodyRows}</TableBody>
        </Table>
      );
    } else if (displayMode === 'COPY') {
      content = (
        <div className={classes.copyPasteDiv}>
          <TextField
            autoFocus
            fullWidth
            multiline
            rows={20}
            variant="filled"
            label="Copy with CTRL+C"
            defaultValue={dataArray.asText}
            InputProps={{
              readOnly: true,
              className: classes.copyPasteInput,
            }}
            inputRef={input => {
              this.copyField = input;
            }}
            onClick={this.handleTableCellClick}
          />
          <Tooltip
            className={classes.tooltip}
            title="Go back to the stratified view"
            classes={{
              tooltip: classes.tooltipInner,
            }}
          >
            <Button
              size="small"
              color="primary"
              onClick={this.handleViewSwitch('STRATIFIED')}
            >
              Back
            </Button>
          </Tooltip>
        </div>
      );
    } else if (displayMode === 'PASTE') {
      content = (
        <div className={classes.copyPasteDiv}>
          <TextField
            autoFocus
            fullWidth
            multiline
            rows={20}
            variant="filled"
            label="Paste with CTRL+V"
            value="Not yet implemented"
            InputProps={{
              className: classes.copyPasteInput,
            }}
            inputRef={input => {
              this.pasteField = input;
            }}
          />
          <Tooltip
            className={classes.tooltip}
            title="Apply pasted content in the stratified view - UNDER DEVELOPMENT"
            classes={{
              tooltip: classes.tooltipInner,
            }}
          >
            <Button
              size="small"
              color="primary"
              onClick={this.handleViewSwitch('STRATIFIED')}
            >
              Apply
            </Button>
          </Tooltip>
        </div>
      );
    }

    return (
      <div className={classNames(classes.root, disabled && classes.disabled)}>
        {content}
        <Menu
          anchorReference="anchorPosition"
          anchorPosition={contextMenuPosition}
          open={contextMenuAnchor != null}
          onClose={this.handleCloseContextMenu}
          onContextMenu={this.handleCloseContextMenu}
        >
          {fillMenuOption}
          <MenuItem onClick={this.handleViewSwitch('COPY')}>
            Copy table
          </MenuItem>
          <MenuItem onClick={this.handleViewSwitch('PASTE')}>
            Paste table
          </MenuItem>
        </Menu>
      </div>
    );
  }
}

export default DataTable;
