import React from 'react';
import PropTypes from 'prop-types';
import { observer, PropTypes as MobxPropTypes } from 'mobx-react';
import { withStyles } from '@material-ui/core/styles';
import Button from '@material-ui/core/Button';
import Typography from '@material-ui/core/Typography';
import Dialog from '@material-ui/core/Dialog';
import Divider from '@material-ui/core/Divider';
import FormControl from '@material-ui/core/FormControl';
import InputLabel from '@material-ui/core/InputLabel';
import FormHelperText from '@material-ui/core/FormHelperText';
import Grid from '@material-ui/core/Grid';
import Select from '@material-ui/core/Select';
import MenuItem from '@material-ui/core/MenuItem';
import FilledInput from '@material-ui/core/FilledInput';
import SwipeableViews from 'react-swipeable-views';
import Tabs from '@material-ui/core/Tabs';
import Tab from '@material-ui/core/Tab';
import TextField from '@material-ui/core/TextField';
import ValidationTextField from './ValidationTextField';
import DataTable from './DataTable';
import TabContainer from './TabContainer';
import ValidateName from '../utilities/ValidateName';
import { HealthOutcomeType } from '../settings';
import GetHealthOutcomeTypeEnableStatus from '../utilities/GetHealthOutcomeTypeEnableStatus';

const validTxt = 'Type a new name';

const styles = () => ({
  root: {
    padding: 10,
    width: 600,
  },
  divider: {
    marginTop: 10,
    marginBottom: 10,
  },
});

@withStyles(styles)
@observer
class HealthOutcomeAddWidget extends React.Component {
  static propTypes = {
    classes: PropTypes.object.isRequired,
    outcomeTree: MobxPropTypes.observableObject.isRequired,
  };

  state = {
    outcomeTree: null,
    tabIdx: 0,
    dialogOpen: false,
    name: '',
    nameValid: false,
    type: '',
    typeValid: false,
    top: 0,
    left: 0,
    severityArray: null,
    durationArray: null,
  };

  static getDerivedStateFromProps(props, state) {
    const { outcomeTree } = props;

    let newState = null;
    if (outcomeTree && outcomeTree !== state.outcomeTree) {
      newState = {
        outcomeTree,
        tabIdx: 0,
        dialogOpen: false,
        name: '',
        nameValid: false,
        type: '',
        typeValid: false,
        top: 0,
        left: 0,
        severityArray: outcomeTree.model.createDataArray(),
        durationArray: outcomeTree.model.createDataArray(),
      };
    }
    return newState;
  }

  handleOpenBtnClick = event => {
    event.preventDefault();
    this.setState({
      dialogOpen: true,
    });
  };

  handleCloseNewWidgetRequest = event => {
    event.preventDefault();
    this.setState({
      outcomeTree: null,
      tabIdx: 0,
      dialogOpen: false,
    });
  };

  handleNameChange = (name, nameValid) => {
    this.setState({
      name,
      nameValid,
    });
  };

  handleTypeChange = (event, child) => {
    event.preventDefault();
    const { value: type } = child.props;

    const typeValid = Object.prototype.hasOwnProperty.call(
      HealthOutcomeType,
      type,
    );

    this.setState({
      type,
      typeValid,
    });
  };

  handlePositionLeftChange = event => {
    event.preventDefault();
    const left = event.currentTarget.value;
    this.setState({
      left,
    });
  };

  handlePositionTopChange = event => {
    event.preventDefault();
    const top = event.currentTarget.value;
    this.setState({
      top,
    });
  };

  handleSeverityArrayChange = severityArray => {
    this.setState({
      severityArray,
    });
  };

  handleDurationArrayChange = durationArray => {
    this.setState({
      durationArray,
    });
  };

  handleCreateNewHealthOutcome = event => {
    event.preventDefault();
    const {
      name,
      nameValid,
      type,
      typeValid,
      top,
      left,
      severityArray,
      durationArray,
    } = this.state;

    this.handleCloseNewWidgetRequest(event);
    if (nameValid && typeValid) {
      const { outcomeTree } = this.props;
      outcomeTree.addHealthOutcome(
        name,
        type,
        top,
        left,
        severityArray,
        durationArray,
      );
    }
  };

  handleTabChange = (event, tabIdx) => {
    event.preventDefault();
    this.setState({ tabIdx });
  };

  render() {
    const { classes, outcomeTree } = this.props;

    const {
      dialogOpen,
      tabIdx,
      nameValid,
      type,
      typeValid,
      top,
      left,
      severityArray,
      durationArray,
    } = this.state;
    const restrictedNames = outcomeTree.healthOutcomeNames;
    const addBtnDisabled = !(nameValid && typeValid);
    const typeEnableStatus = GetHealthOutcomeTypeEnableStatus(
      outcomeTree.healthOutcomesAsArray,
    );

    const tabs = [<Tab key='generalSettings' label="General settings" />];
    if (type == 'ACUTE' || type == 'SEQUELA') {
      tabs.push(<Tab key='tables' label="Tables" />);
    }

    return (
      <div>
        <Button
          onClick={this.handleOpenBtnClick}
          color="primary"
          aria-haspopup="true"
          aria-owns={null}
        >
          Add
        </Button>
        <Dialog
          open={dialogOpen}
          onClose={this.handleCloseNewWidgetRequest}
          maxWidth={false}
        >
          <form className={classes.root}>
            <Typography variant="h6">Add a new health outcome</Typography>
            <Tabs
              value={tabIdx}
              indicatorColor="primary"
              textColor="primary"
              onChange={this.handleTabChange}
            >
              {tabs}
            </Tabs>
            <SwipeableViews axis="x" index={tabIdx}>
              <TabContainer>
                <Grid container>
                  <Grid item xs={6}>
                    <ValidationTextField
                      label="Item name"
                      helperText={validTxt}
                      onChange={this.handleNameChange}
                      validationFunc={ValidateName(restrictedNames, validTxt)}
                      autoFocus
                      fullWidth
                      required
                      variant="filled"
                    />
                    <FormControl
                      required
                      style={{ width: '100%', marginTop: 25, marginBottom: 20 }}
                      variant="filled"
                    >
                      <InputLabel htmlFor="type-helper">Item type</InputLabel>
                      <Select
                        value={type}
                        onChange={this.handleTypeChange}
                        input={<FilledInput name="type" id="type-helper" />}
                      >
                        <MenuItem
                          value={HealthOutcomeType.INFECTION}
                          disabled={!typeEnableStatus.INFECTION}
                        >
                          {HealthOutcomeType.INFECTION}
                        </MenuItem>
                        <MenuItem
                          value={HealthOutcomeType.ACUTE}
                          disabled={!typeEnableStatus.ACUTE}
                        >
                          {HealthOutcomeType.ACUTE}
                        </MenuItem>
                        <MenuItem
                          value={HealthOutcomeType.ASYMPTOMATIC}
                          disabled={!typeEnableStatus.ASYMPTOMATIC}
                        >
                          {HealthOutcomeType.ASYMPTOMATIC}
                        </MenuItem>
                        <MenuItem
                          value={HealthOutcomeType.SEQUELA}
                          disabled={!typeEnableStatus.SEQUELA}
                        >
                          {HealthOutcomeType.SEQUELA}
                        </MenuItem>
                        <MenuItem
                          value={HealthOutcomeType.RECOVERY}
                          disabled={!typeEnableStatus.RECOVERY}
                        >
                          {HealthOutcomeType.RECOVERY}
                        </MenuItem>
                        <MenuItem
                          value={HealthOutcomeType.DEATH}
                          disabled={!typeEnableStatus.DEATH}
                        >
                          {HealthOutcomeType.DEATH}
                        </MenuItem>
                      </Select>
                      <FormHelperText>Select type</FormHelperText>
                    </FormControl>
                    <TextField
                      label="Position: left"
                      value={left}
                      type="number"
                      variant="filled"
                      margin="normal"
                      onChange={this.handlePositionLeftChange}
                      fullWidth
                    />
                    <TextField
                      label="Position: top"
                      value={top}
                      type="number"
                      variant="filled"
                      onChange={this.handlePositionTopChange}
                      fullWidth
                    />
                  </Grid>
                </Grid>
              </TabContainer>
              <TabContainer>
                <Grid container>
                  <Grid item xs={6}>
                    <Typography variant="overline">
                      Disability weights
                    </Typography>
                    <DataTable
                      dataArray={severityArray}
                      onChange={this.handleSeverityArrayChange}
                    />
                  </Grid>
                  <Grid item xs={6}>
                    <Typography variant="overline">Duration</Typography>
                    <DataTable
                      dataArray={durationArray}
                      onChange={this.handleDurationArrayChange}
                    />
                  </Grid>
                </Grid>
              </TabContainer>
            </SwipeableViews>
            <Divider className={classes.divider} />
            <div style={{ textAlign: 'right' }}>
              <Button
                color="secondary"
                onClick={this.handleCloseNewWidgetRequest}
              >
                Cancel
              </Button>
              <Button
                color="primary"
                disabled={addBtnDisabled}
                onClick={this.handleCreateNewHealthOutcome}
                type="submit"
              >
                OK
              </Button>
            </div>
          </form>
        </Dialog>
      </div>
    );
  }
}

export default HealthOutcomeAddWidget;
