import React from 'react';
import PropTypes from 'prop-types';
import { observer, PropTypes as MobxPropTypes } from 'mobx-react';
import { withStyles } from '@material-ui/core/styles';
import Button from '@material-ui/core/Button';
import Typography from '@material-ui/core/Typography';
import Dialog from '@material-ui/core/Dialog';
import Divider from '@material-ui/core/Divider';
import FormControl from '@material-ui/core/FormControl';
import InputLabel from '@material-ui/core/InputLabel';
import FormHelperText from '@material-ui/core/FormHelperText';
import Select from '@material-ui/core/Select';
import SwipeableViews from 'react-swipeable-views';
import Tabs from '@material-ui/core/Tabs';
import Tab from '@material-ui/core/Tab';
import Grid from '@material-ui/core/Grid';
import MenuItem from '@material-ui/core/MenuItem';
import FilledInput from '@material-ui/core/FilledInput';
import TextField from '@material-ui/core/TextField';
import TabContainer from './TabContainer';
import ValidationTextField from './ValidationTextField';
import DataTable from './DataTable';
import ValidateName from '../utilities/ValidateName';
import { HealthOutcomeType } from '../settings';
import GetHealthOutcomeTypeEnableStatus from '../utilities/GetHealthOutcomeTypeEnableStatus';

const validTxt = 'Type a new name';

const styles = () => ({
  root: {
    padding: 10,
    width: 600,
  },
  divider: {
    marginTop: 10,
    marginBottom: 10,
  },
});

@withStyles(styles)
@observer
class HealthOutcomeModifyWidget extends React.Component {
  static propTypes = {
    classes: PropTypes.object.isRequired,
    open: PropTypes.bool,
    healthOutcome: MobxPropTypes.observableObject,
    onClose: PropTypes.func,
  };

  static defaultProps = {
    open: false,
    healthOutcome: null,
    onClose: null,
  };

  state = {
    healthOutcome: null,
    tabIdx: 0,
    name: '',
    nameValid: false,
    nameValidFunc: null,
    type: '',
    typeValid: false,
    top: 0,
    left: 0,
    typeValidFunc: null,
    severityArray: null,
    durationArray: null,
  };

  static getDerivedStateFromProps(props, state) {
    const { open, healthOutcome } = props;

    if (open && healthOutcome && healthOutcome != state.healthOutcome) {
      const { healthOutcome: {outcomeTree} } = props;

      const nameValidFunc = ValidateName(
        outcomeTree.nonSelectedHealthOutcomeNames,
        validTxt,
      );

      const typeValidFunc = type => {
        const result = Object.prototype.hasOwnProperty.call(
          HealthOutcomeType,
          type,
        );
        return result;
      };

      return {
        healthOutcome,
        tabIdx: 0,
        name: healthOutcome.name,
        nameValid: nameValidFunc(healthOutcome.name) === validTxt,
        nameValidFunc,
        type: healthOutcome.type,
        typeValid: typeValidFunc(healthOutcome.type),
        typeValidFunc,
        top: healthOutcome.top,
        left: healthOutcome.left,
        severityArray: healthOutcome.severityArray,
        durationArray: healthOutcome.durationArray,
      };
    }
    return null;
  }

  handleCloseWidgetRequest = (event) => {
    this.setState({
      healthOutcome: null,
      tabIdx: 0,
      name: '',
      nameValid: false,
      nameValidFunc: null,
      type: '',
      typeValid: false,
      typeValidFunc: null,
      top: 0,
      left: 0,
      severityArray: null,
      durationArray: null,
    });
    const { onClose, healthOutcome } = this.props;
    if (onClose) {
      onClose(event, healthOutcome);
    }
  };

  handleNameChange = (name, nameValid) => {
    this.setState({
      name,
      nameValid,
    });
  };

  handleTypeChange = (event, child) => {
    event.preventDefault();
    const { value: type } = child.props;
    const { typeValidFunc } = this.state;

    const typeValid = typeValidFunc(type);

    this.setState({
      type,
      typeValid,
    });
  };

  handlePositionLeftChange = event => {
    event.preventDefault();
    const left = event.currentTarget.value;
    this.setState({
      left,
    });
  };

  handlePositionTopChange = event => {
    event.preventDefault();
    const top = event.currentTarget.value;
    this.setState({
      top,
    });
  };

  handleSeverityArrayChange = severityArray => {
    this.setState({
      severityArray,
    });
  };

  handleDurationArrayChange = durationArray => {
    this.setState({
      durationArray,
    });
  };

  handleModifyHealthOutcome = event => {
    event.preventDefault();
    const {
      name,
      type,
      top,
      left,
      severityArray,
      durationArray,
    } = this.state;
    const { healthOutcome } = this.props;
    const { outcomeTree } = healthOutcome;
    this.handleCloseWidgetRequest(event);
    outcomeTree.modifyHealthOutcome(
      healthOutcome.id,
      name,
      type,
      top,
      left,
      severityArray,
      durationArray,
    );
  };

  handleTabChange = (event, tabIdx) => {
    event.preventDefault();
    this.setState({ tabIdx });
  };

  render() {
    const { classes, healthOutcome, open } = this.props;
    if (!healthOutcome) return null;

    const {outcomeTree} = healthOutcome;
    if (!outcomeTree) return null;

    const {
      tabIdx,
      name,
      nameValid,
      nameValidFunc,
      type,
      typeValid,
      top,
      left,
      severityArray,
      durationArray,
    } = this.state;

    const addBtnDisabled = !(nameValid && typeValid);
    const typeEnableStatus = GetHealthOutcomeTypeEnableStatus(
      outcomeTree.healthOutcomesAsArray,
    );
    typeEnableStatus[healthOutcome.type] = true;

    const tabs = [<Tab key='generalSettings' label="General settings" />];
    if (type == 'ACUTE' || type == 'SEQUELA') {
      tabs.push(<Tab key='tables' label="Tables" />);
    }

    return (
      <Dialog
        open={open}
        onClose={this.handleCloseWidgetRequest}
        BackdropProps={{ className: classes.backdrop }}
      >
        <form className={classes.root}>
          <Typography variant="h6">Modify health outcome</Typography>
          <Tabs
            value={tabIdx}
            indicatorColor="primary"
            textColor="primary"
            onChange={this.handleTabChange}
          >
            {tabs}
          </Tabs>
          <SwipeableViews axis="x" index={tabIdx}>
            <TabContainer>
              <Grid container>
                <Grid item xs={6}>
                  <ValidationTextField
                    label="Item name"
                    helperText={validTxt}
                    value={name}
                    onChange={this.handleNameChange}
                    validationFunc={nameValidFunc}
                    autoFocus
                    fullWidth
                    required
                    variant="filled"
                  />
                  <FormControl
                    required
                    style={{ width: '100%', marginTop: 25, marginBottom: 20 }}
                    variant="filled"
                  >
                    <InputLabel htmlFor="type-helper">Item type</InputLabel>
                    <Select
                      value={type}
                      onChange={this.handleTypeChange}
                      input={<FilledInput name="type" id="type-helper" />}
                    >
                      <MenuItem
                        value={HealthOutcomeType.INFECTION}
                        disabled={!typeEnableStatus.INFECTION}
                      >
                        {HealthOutcomeType.INFECTION}
                      </MenuItem>
                      <MenuItem
                        value={HealthOutcomeType.ACUTE}
                        disabled={!typeEnableStatus.ACUTE}
                      >
                        {HealthOutcomeType.ACUTE}
                      </MenuItem>
                      <MenuItem
                        value={HealthOutcomeType.ASYMPTOMATIC}
                        disabled={!typeEnableStatus.ASYMPTOMATIC}
                      >
                        {HealthOutcomeType.ASYMPTOMATIC}
                      </MenuItem>
                      <MenuItem
                        value={HealthOutcomeType.SEQUELA}
                        disabled={!typeEnableStatus.SEQUELA}
                      >
                        {HealthOutcomeType.SEQUELA}
                      </MenuItem>
                      <MenuItem
                        value={HealthOutcomeType.RECOVERY}
                        disabled={!typeEnableStatus.RECOVERY}
                      >
                        {HealthOutcomeType.RECOVERY}
                      </MenuItem>
                      <MenuItem
                        value={HealthOutcomeType.DEATH}
                        disabled={!typeEnableStatus.DEATH}
                      >
                        {HealthOutcomeType.DEATH}
                      </MenuItem>
                    </Select>
                    <FormHelperText>Select type</FormHelperText>
                  </FormControl>
                  <TextField
                    label="Position: left"
                    value={left}
                    type="number"
                    variant="filled"
                    margin="normal"
                    onChange={this.handlePositionLeftChange}
                    fullWidth
                  />
                  <TextField
                    label="Position: top"
                    value={top}
                    type="number"
                    variant="filled"
                    onChange={this.handlePositionTopChange}
                    fullWidth
                  />
                </Grid>
              </Grid>
            </TabContainer>
            <TabContainer>
              <Grid container>
                <Grid item xs={6}>
                  <Typography variant="overline">Disability weights</Typography>
                  <DataTable
                    dataArray={severityArray}
                    onChange={this.handleSeverityArrayChange}
                  />
                </Grid>
                <Grid item xs={6}>
                  <Typography variant="overline">Duration</Typography>
                  <DataTable
                    dataArray={durationArray}
                    onChange={this.handleDurationArrayChange}
                  />
                </Grid>
              </Grid>
            </TabContainer>
          </SwipeableViews>
          <Divider className={classes.divider} />
          <div style={{ textAlign: 'right' }}>
            <Button color="secondary" onClick={this.handleCloseWidgetRequest}>
              Cancel
            </Button>
            <Button
              color="primary"
              disabled={addBtnDisabled}
              onClick={this.handleModifyHealthOutcome}
              type="submit"
            >
              OK
            </Button>
          </div>
        </form>
      </Dialog>
    );
  }
}

export default HealthOutcomeModifyWidget;
