import React from 'react';
import PropTypes from 'prop-types';
import { observer, PropTypes as MobxPropTypes } from 'mobx-react';
import { withStyles } from '@material-ui/core/styles';
import Button from '@material-ui/core/Button';
import Typography from '@material-ui/core/Typography';
import Dialog from '@material-ui/core/Dialog';
import Divider from '@material-ui/core/Divider';
import SwipeableViews from 'react-swipeable-views';
import Tabs from '@material-ui/core/Tabs';
import Tab from '@material-ui/core/Tab';
import Grid from '@material-ui/core/Grid';
import FormControlLabel from '@material-ui/core/FormControlLabel';
import Switch from '@material-ui/core/Switch';
import TabContainer from './TabContainer';
import ValidationTextField from './ValidationTextField';
import DataTable from './DataTable';
import ValidateName from '../utilities/ValidateName';

const validTxt = 'Type a new name';

const styles = () => ({
  root: {
    padding: 10,
    width: 900,
  },
  divider: {
    marginTop: 10,
    marginBottom: 10,
  },
});

@withStyles(styles)
@observer
class HealthStateAddWidget extends React.Component {
  static propTypes = {
    classes: PropTypes.object.isRequired,
    open: PropTypes.bool.isRequired,
    healthOutcome: MobxPropTypes.observableObject, // eslint-disable-line react/no-unused-prop-types
    onClose: PropTypes.func,
  };

  static defaultProps = {
    onClose: null,
  };

  state = {
    healthOutcome: null,
    tabIdx: 0,
    name: '',
    nameValid: false,
    nameValidFunc: null,
    severityArray: null,
    durationArray: null,
    proportionArray: null,
    remainingCases: false
  };

  static getDerivedStateFromProps(props, state) {
    const { open, healthOutcome } = props;
    if (open && healthOutcome && healthOutcome !== state.healthOutcome) {
      const nameValidFunc = ValidateName(
        healthOutcome.healthStateNames,
        validTxt,
      );
      return {
        healthOutcome,
        tabIdx: 0,
        name: '',
        nameValid: false,
        nameValidFunc,
        severityArray: healthOutcome.outcomeTree.model.createDataArray(),
        durationArray: healthOutcome.outcomeTree.model.createDataArray(),
        proportionArray: healthOutcome.outcomeTree.model.createDataArray(),
        remainingCases: false
      };
    }
    return null;
  }

  handleCloseWidgetRequest = event => {
    event.preventDefault();
    this.setState({
      healthOutcome: null,
      tabIdx: 0,
      name: '',
      nameValid: false,
      nameValidFunc: null,
      severityArray: null,
      durationArray: null,
      proportionArray: null,
      remainingCases: false
    });
    const { onClose } = this.props;
    if (onClose) {
      onClose();
    }
  };

  handleNameChange = (name, nameValid) => {
    this.setState({
      name,
      nameValid,
    });
  };

  handleSeverityArrayChange = severityArray => {
    this.setState({
      severityArray,
    });
  };

  handleDurationArrayChange = durationArray => {
    this.setState({
      durationArray,
    });
  };

  handleProportionArrayChange = proportionArray => {
    this.setState({
      proportionArray,
    });
  };

  handleRemainingCasesChange = event => {
    this.setState({ remainingCases: event.target.checked });
  };

  handleAddHealthState = event => {
    event.preventDefault();
    const {
      healthOutcome,
      name,
      severityArray,
      durationArray,
      proportionArray,
      remainingCases
    } = this.state;
    this.handleCloseWidgetRequest(event);
    healthOutcome.addHealthState(
      name,
      severityArray,
      durationArray,
      proportionArray,
      remainingCases
    );
  };

  handleTabChange = (event, tabIdx) => {
    this.setState({ tabIdx });
  };

  render() {
    const { classes, open, healthOutcome } = this.props;
    const {
      tabIdx,
      name,
      nameValid,
      nameValidFunc,
      severityArray,
      durationArray,
      proportionArray,
      remainingCases
    } = this.state;

    if (!(open && healthOutcome)) return null;
    const addBtnDisabled = !nameValid;

    const remainingCasesSwitch = (
      <Switch
        checked={remainingCases}
        value="remainingCases"
        color="primary"
      />
    );

    return (
      <Dialog
        open={open}
        onClose={this.handleCloseWidgetRequest}
        BackdropProps={{ className: classes.backdrop }}
        maxWidth={false}
      >
        <form className={classes.root}>
          <Typography variant="h6">Modify health state</Typography>
          <Tabs
            value={tabIdx}
            indicatorColor="primary"
            textColor="primary"
            onChange={this.handleTabChange}
          >
            <Tab label="General settings" />
            <Tab label="Tables" />
          </Tabs>
          <SwipeableViews axis="x" index={tabIdx}>
            <TabContainer>
              <Grid container>
                <Grid item xs={4}>
                  <ValidationTextField
                    label="Item name"
                    helperText={validTxt}
                    value={name}
                    onChange={this.handleNameChange}
                    validationFunc={nameValidFunc}
                    autoFocus
                    fullWidth
                    required
                    variant="filled"
                  />
                </Grid>
              </Grid>
              <Grid item container xs={12}>
                <Grid item xs={4}>
                  <FormControlLabel
                    control={remainingCasesSwitch}
                    onChange={this.handleRemainingCasesChange}
                    label="Remaining cases"
                  />
                </Grid>
                <Grid item xs={4}>
                  <Typography
                    variant="body2"
                  >
                    <b>Remaining cases</b> indicates that this health state covers cases
                    not covered by other health states.
                    <br />
                    There can be only one health state within a health outcome covering remaining cases.
                    <br />
                    If none of the health states is selected to cover remaining cases, then
                    the toolkit will use the heatlh state instantiated last as the one to cover
                    the remaining cases. This can be arbitrary.
                  </Typography>
                </Grid>
              </Grid>
            </TabContainer>
            <TabContainer>
              <Grid container>
                <Grid item xs={4}>
                  <Typography variant="overline">Disability weights</Typography>
                  <DataTable
                    dataArray={severityArray}
                    onChange={this.handleSeverityArrayChange}
                  />
                </Grid>
                <Grid item xs={4}>
                  <Typography variant="overline">Duration</Typography>
                  <DataTable
                    dataArray={durationArray}
                    onChange={this.handleDurationArrayChange}
                  />
                </Grid>
                <Grid item xs={4}>
                  <Typography variant="overline">Proportion</Typography>
                  <DataTable
                    dataArray={proportionArray}
                    onChange={this.handleProportionArrayChange}
                    disabled={remainingCases}
                  />
                </Grid>
              </Grid>
            </TabContainer>
          </SwipeableViews>
          <Divider className={classes.divider} />
          <div style={{ textAlign: 'right' }}>
            <Button color="secondary" onClick={this.handleCloseWidgetRequest}>
              Cancel
            </Button>
            <Button
              color="primary"
              disabled={addBtnDisabled}
              onClick={this.handleAddHealthState}
              type="submit"
            >
              OK
            </Button>
          </div>
        </form>
      </Dialog>
    );
  }
}

export default HealthStateAddWidget;
