import React from 'react';
import PropTypes from 'prop-types';
import classNames from 'classnames';
import { observer, PropTypes as MobxPropTypes } from 'mobx-react';
import { withStyles } from '@material-ui/core/styles';
import Table from '@material-ui/core/Table';
import TableBody from '@material-ui/core/TableBody';
import TableCell from '@material-ui/core/TableCell';
import TableHead from '@material-ui/core/TableHead';
import TableRow from '@material-ui/core/TableRow';
import Menu from '@material-ui/core/Menu';
import MenuItem from '@material-ui/core/MenuItem';
import TextField from '@material-ui/core/TextField';
import Button from '@material-ui/core/Button';
import Tooltip from '@material-ui/core/Tooltip';
import Select from '@material-ui/core/Select';
import FocusInput from '../utilities/FocusInput';

const styles = () => ({
  root: {
    fontSize: 11,
    width: 275,
  },
  disabled: {
    opacity: 0.4,
    pointerEvents: 'none',
  },
  table: {
    tableLayout: 'fixed',
    '& tr:nth-child(even)': {
      backgroundColor: '#fafafa',
    },
  },
  tableRowStyle: {
    height: 'auto',
  },
  tableHeaderCellStyle: {
    fontWeight: 'bold',
    height: 16,
    lineHeight: '16px',
    fontSize: 11,
    padding: 4,
    margin: 0,
    color: 'rgb(26, 50, 115)',
    backgroundColor: '#e7f0d8',
    textTransform: 'capitalize',
    width: 118,
    '&:first-child': {
      width: 40,
      textAlign: 'center',
    },
    '&:last-child': {
      paddingRight: 4,
    },
  },
  tableEditCellStyle: {
    height: 22,
    padding: 0,
    margin: 0,
    color: 'rgb(26, 50, 115)',
    '&:first-child': {
      width: 40,
      textAlign: 'center',
    },
    '&:last-child': {
      paddingRight: 0,
    },
  },
  editField: {
    fontSize: 11,
    height: 22,
    lineHeight: '22px',
    margin: 0,
    padding: 0,
    width: '100%',
    '&:before': {
      borderBottom: 'none',
    },
    '&:after': {
      borderBottom: '1px solid rgb(73, 123, 24)',
    },
    '&:hover': {
      '&:before': {
        transition: 'none',
        borderBottom: '1px solid rgb(73, 123, 24) !important',
      },
    },
  },
  select: {
    padding: '0',
    borderBottomColor: 'transparent',
  },
  selectMenu: {
    paddingRight: '16px !important',
  },
  icon: {
    fontSize: 16,
    marginTop: 5,
  },
  tableCellRoot: {
    borderBottom: 'none',
  },
  copyPasteDiv: {
    position: 'relative',
  },
  copyPasteInput: {
    fontSize: 11,
    lineHeight: '18px',
    '& ::selection': {
      backgroundColor: 'rgb(0, 0, 0, 0.14)',
    },
  },
  tooltip: {
    position: 'absolute',
    top: 0,
    right: 0,
  },
  tooltipInner: {
    fontSize: 11,
  },
});

@withStyles(styles)
@observer
class MappingTable extends React.Component {
  static propTypes = {
    classes: PropTypes.object.isRequired,
    model: MobxPropTypes.observableObject.isRequired,
    disabled: PropTypes.bool,
  };

  static defaultProps = {
    disabled: false,
  };

  state = {
    contextMenuEvent: null,
    displayMode: 'STRATIFIED',
  };

  componentDidUpdate = (prevProps, prevState) => {
    const { displayMode } = this.state;
    if (
      prevState.displayMode !== 'COPY' &&
      displayMode === 'COPY' &&
      this.copyField
    ) {
      FocusInput(this.copyField, 1);
    } else if (
      prevState.displayMode !== 'PASTE' &&
      displayMode === 'PASTE' &&
      this.pasteField
    ) {
      FocusInput(this.pasteField, 1);
    }
  };

  handleTableRightClick = event => {
    event.preventDefault();
    event.persist();
    this.setState({
      contextMenuEvent: event,
    });
  };

  handleCloseContextMenu = event => {
    event.preventDefault();
    this.setState({
      contextMenuEvent: null,
    });
  };

  handleTableCellChange = event => {
    event.preventDefault();

    const [row, col] = event.target.name.split(',').map(x => parseInt(x, 10));
    const val = event.target.value;
    const { model } = this.props;

    model.outcomeTreeMapping.setValue(row, col, val);
  };

  handleDataArrayFill = () => {
    const { contextMenuEvent } = this.state;

    if (contextMenuEvent) {
      const fillValue = contextMenuEvent.target.textContent;
      const { model } = this.props;
      model.outcomeTreeMapping.fillWithValue(fillValue);
    }
    this.handleCloseContextMenu(contextMenuEvent);
  };

  handleViewSwitch = view => () => {
    const { contextMenuEvent } = this.state;
    if (contextMenuEvent) {
      this.handleCloseContextMenu(contextMenuEvent);
    }

    this.setState({
      displayMode: view,
    });
  };

  getHeaderCell = val => val;

  getBodyCell = (val, selectOptions, i, j, classes) => (
    <Select
      inputProps={{ name: `${i},${j}` }}
      value={val}
      margin="dense"
      className={classes.editField}
      classes={{
        select: classes.select,
        selectMenu: classes.selectMenu,
        icon: classes.icon,
      }}
      onChange={this.handleTableCellChange}
    >
      {selectOptions}
    </Select>
  );

  getTableRow = (
    i,
    dataRow,
    selectOptions,
    classes,
    cellClassName,
    getCell,
  ) => {
    const cells = dataRow.rowVals.map((val, j) => {
      const key = j;
      return (
        <TableCell
          align="right"
          key={key}
          className={cellClassName}
          classes={{
            root: classes.tableCellRoot,
          }}
        >
          {getCell(val, selectOptions, i, j, classes)}
        </TableCell>
      );
    });

    return (
      <TableRow key={dataRow.rowName} className={classes.tableRowStyle}>
        <TableCell
          className={cellClassName}
          classes={{
            root: classes.tableCellRoot,
          }}
        >
          {dataRow.rowName}
        </TableCell>
        {cells}
      </TableRow>
    );
  };

  render() {
    const { contextMenuEvent, displayMode } = this.state;

    const {
      classes,
      model,
      model: { outcomeTreeMapping: dataArray },
      disabled,
    } = this.props;

    const idNameMapping = model.outcomeTreesAsArray.reduce(
      (map, obj) => {
        map[obj.id] = obj.name; // eslint-disable-line no-param-reassign
        return map;
      },
      { '': 'None' },
    );

    const selectOptions = Object.keys(idNameMapping).map(id => {
      return (
        <MenuItem key={id} value={id}>
          {idNameMapping[id]}
        </MenuItem>
      );
    });

    let contextMenuPosition;
    let contextMenuAnchor;
    let fillMenuOption = null;
    if (contextMenuEvent !== null) {
      contextMenuAnchor = contextMenuEvent.target;
      contextMenuPosition = {
        left: contextMenuEvent.clientX,
        top: contextMenuEvent.clientY,
      };
      if (contextMenuEvent.target.localName === 'div') {
        fillMenuOption = (
          <MenuItem onClick={this.handleDataArrayFill}>
            Fill table with this value
          </MenuItem>
        );
      }
    } else {
      contextMenuAnchor = null;
      contextMenuPosition = {
        left: 0,
        top: 0,
      };
    }

    let content;
    if (displayMode === 'STRATIFIED') {
      let tableHeaderRow = [];
      let tableBodyRows = [];
      if (dataArray) {
        const dataObject = dataArray.asObject;

        if (dataObject) {
          const dimNames = Object.keys(dataArray.dimOrder);
          if (dimNames.length === 2) {
            tableHeaderRow = this.getTableRow(
              0,
              dataObject.headRow,
              selectOptions,
              classes,
              classes.tableHeaderCellStyle,
              this.getHeaderCell,
              disabled,
            );
            tableBodyRows = dataObject.bodyRows.map((bodyRow, i) =>
              this.getTableRow(
                i,
                bodyRow,
                selectOptions,
                classes,
                classes.tableEditCellStyle,
                this.getBodyCell,
                disabled,
              ),
            );
          }
        } else {
          tableHeaderRow = (
            <TableRow className={classes.tableRowStyle}>
              <TableCell
                className={classes.tableHeaderCellStyle}
                classes={{
                  root: classes.tableCellRoot,
                }}
              >
                No content
              </TableCell>
            </TableRow>
          );
        }
      }

      content = (
        <Table
          className={classes.table}
          onContextMenu={this.handleTableRightClick}
        >
          <TableHead>{tableHeaderRow}</TableHead>
          <TableBody>{tableBodyRows}</TableBody>
        </Table>
      );
    } else if (displayMode === 'COPY') {
      content = (
        <div className={classes.copyPasteDiv}>
          <TextField
            autoFocus
            fullWidth
            multiline
            rows={20}
            variant="filled"
            label="Copy with CTRL+C"
            defaultValue={dataArray.asText}
            InputProps={{
              readOnly: true,
              className: classes.copyPasteInput,
            }}
            inputRef={input => {
              this.copyField = input;
            }}
            onClick={this.handleTableCellClick}
          />
          <Tooltip
            className={classes.tooltip}
            title="Go back to the stratified view"
            classes={{
              tooltip: classes.tooltipInner,
            }}
          >
            <Button
              size="small"
              color="primary"
              onClick={this.handleViewSwitch('STRATIFIED')}
            >
              Back
            </Button>
          </Tooltip>
        </div>
      );
    } else if (displayMode === 'PASTE') {
      content = (
        <div className={classes.copyPasteDiv}>
          <TextField
            autoFocus
            fullWidth
            multiline
            rows={20}
            variant="filled"
            label="Paste with CTRL+V"
            value="Not yet implemented"
            InputProps={{
              className: classes.copyPasteInput,
            }}
            inputRef={input => {
              this.pasteField = input;
            }}
          />
          <Tooltip
            className={classes.tooltip}
            title="Apply pasted content in the stratified view - UNDER DEVELOPMENT"
            classes={{
              tooltip: classes.tooltipInner,
            }}
          >
            <Button
              size="small"
              color="primary"
              onClick={this.handleViewSwitch('STRATIFIED')}
            >
              Apply
            </Button>
          </Tooltip>
        </div>
      );
    }

    return (
      <div className={classNames(classes.root, disabled && classes.disabled)}>
        {content}
        <Menu
          anchorReference="anchorPosition"
          anchorPosition={contextMenuPosition}
          open={contextMenuAnchor != null}
          onClose={this.handleCloseContextMenu}
          onContextMenu={this.handleCloseContextMenu}
        >
          {fillMenuOption}
          <MenuItem onClick={this.handleViewSwitch('COPY')}>
            Copy table
          </MenuItem>
          <MenuItem onClick={this.handleViewSwitch('PASTE')}>
            Paste table
          </MenuItem>
        </Menu>
      </div>
    );
  }
}

export default MappingTable;
