import React from 'react';
import PropTypes from 'prop-types';
import { observer, PropTypes as MobxPropTypes } from 'mobx-react';
import Grid from '@material-ui/core/Grid';
import TextField from '@material-ui/core/TextField';
import Switch from '@material-ui/core/Switch';
import Button from '@material-ui/core/Button';
import FormControlLabel from '@material-ui/core/FormControlLabel';
import { withStyles } from '@material-ui/core/styles';
import ValidationTextField from './ValidationTextField';
import GetTimeDateString from '../utilities/GetTimeDateString';
import ValidateName from '../utilities/ValidateName';

const validTxt = 'Type name';

const styles = () => ({
  labelPlacementTop: {
    flexDirection: 'column-reverse',
    marginLeft: 5,
    marginBottom: 5,
  },
});

@withStyles(styles)
@observer
class ModelTabMeta extends React.Component {
  static propTypes = {
    classes: PropTypes.object.isRequired,
    model: MobxPropTypes.observableObject.isRequired,
    nonSelectedModelsNames: PropTypes.arrayOf(PropTypes.string).isRequired, // eslint-disable-line react/no-unused-prop-types
  };

  state = {
    model: null,
    modelName: '',
    modelNameValid: false,
    modelNameValidFunc: () => validTxt, // eslint-disable-line react/no-unused-state
    modelDescription: '',
    modelVersion: '',
    modelAuthor: '',
    modelIsCongenitalType: false,
    modelCongenitalCategoryName: '',
    modelCongenitalCategoryNameValid: false,
  };

  static getDerivedStateFromProps(props, state) {
    if (props.model !== state.model) {
      const modelNameValidFunc = ValidateName(
        props.nonSelectedModelsNames,
        validTxt,
      );

      return {
        model: props.model,
        modelName: props.model.name,
        modelNameValid: modelNameValidFunc(props.model.name) === validTxt,
        modelNameValidFunc,
        modelDescription: props.model.description,
        modelVersion: props.model.version,
        modelAuthor: props.model.author,
        modelIsCongenitalType: props.model.isCongenitalType,
        modelCongenitalCategoryName: props.model.congenitalCategoryName,
        modelCongenitalCategoryNameValid:
          ValidateName()(props.model.congenitalCategoryName) === '',
      };
    }
    return null;
  }

  handleModelNameChange = (modelName, modelNameValid) => {
    this.setState({
      modelName,
      modelNameValid,
    });
  };

  handleModelDescriptionChange = event => {
    this.setState({
      modelDescription: event.target.value,
    });
  };

  handleModelVersionChange = event => {
    this.setState({
      modelVersion: event.target.value,
    });
  };

  handleModelAuthorChange = event => {
    this.setState({
      modelAuthor: event.target.value,
    });
  };

  handleCongenitalTypeChange = event => {
    this.setState({
      modelIsCongenitalType: event.target.checked,
    });
  };

  handleModelCongenitalCategoryNameChange = (
    modelCongenitalCategoryName,
    modelCongenitalCategoryNameValid,
  ) => {
    this.setState({
      modelCongenitalCategoryName,
      modelCongenitalCategoryNameValid,
    });
  };

  handleApplyChanges = event => {
    event.preventDefault();
    const { model } = this.state;
    const {
      modelName,
      modelNameValid,
      modelDescription,
      modelVersion,
      modelAuthor,
      modelIsCongenitalType,
      modelCongenitalCategoryName,
    } = this.state;

    if (modelNameValid) {
      model.updateMetaInfo(
        modelName,
        modelDescription,
        modelVersion,
        modelAuthor,
        modelIsCongenitalType,
        modelCongenitalCategoryName,
      );
    }
  };

  render() {
    const { model, classes } = this.props;
    const {
      modelName,
      modelNameValid,
      modelNameValidFunc,
      modelDescription,
      modelVersion,
      modelAuthor,
      modelIsCongenitalType,
      modelCongenitalCategoryName,
      modelCongenitalCategoryNameValid,
    } = this.state;
    const applyDisabled =
      !modelNameValid ||
      (modelIsCongenitalType && !modelCongenitalCategoryNameValid);

    let congenitalCategoryNameWidget = null;
    if (modelIsCongenitalType) {
      congenitalCategoryNameWidget = (
        <ValidationTextField
          label="Congenital age category name"
          value={modelCongenitalCategoryName}
          validationFunc={ValidateName()}
          onChange={this.handleModelCongenitalCategoryNameChange}
          helperText=" "
          variant="filled"
          fullWidth
        />
      );
    }

    return (
      <form noValidate autoComplete="off">
        <Grid
          container
          direction="column"
          justify="center"
          alignItems="center"
          spacing={1}
        >
          <Grid item xs={12} container>
            <Grid item xs={4} />
            <Grid item xs={4}>
              <ValidationTextField
                label="Name"
                value={modelName}
                onChange={this.handleModelNameChange}
                validationFunc={modelNameValidFunc}
                helperText={validTxt}
                variant="filled"
                fullWidth
                required
              />
            </Grid>
          </Grid>
          <Grid item xs={12} container>
            <Grid item xs={4} />
            <Grid item xs={4}>
              <TextField
                label="Description"
                value={modelDescription}
                onChange={this.handleModelDescriptionChange}
                helperText=" "
                variant="filled"
                fullWidth
              />
            </Grid>
          </Grid>
          <Grid item xs={12} container>
            <Grid item xs={4} />
            <Grid item xs={4}>
              <TextField
                label="Version"
                value={modelVersion}
                onChange={this.handleModelVersionChange}
                helperText=" "
                variant="filled"
                fullWidth
              />
            </Grid>
          </Grid>
          <Grid item xs={12} container>
            <Grid item xs={4} />
            <Grid item xs={4}>
              <TextField
                label="Author"
                value={modelAuthor}
                onChange={this.handleModelAuthorChange}
                helperText=" "
                variant="filled"
                fullWidth
              />
            </Grid>
          </Grid>
          <Grid item xs={12} container>
            <Grid item xs={4} />
            <Grid item xs={1}>
              <FormControlLabel
                control={
                  // eslint-disable-next-line react/jsx-wrap-multilines
                  <Switch
                    checked={modelIsCongenitalType}
                    onChange={this.handleCongenitalTypeChange}
                  />
                }
                label="Congenital type"
                labelPlacement="top"
                classes={{
                  labelPlacementTop: classes.labelPlacementTop,
                }}
              />
            </Grid>
            <Grid item xs={3}>
              {congenitalCategoryNameWidget}
            </Grid>
          </Grid>
          <Grid item xs={12} container>
            <Grid item xs={4} />
            <Grid item xs={4}>
              <TextField
                label="Create time"
                value={GetTimeDateString(model.createTime)}
                fullWidth
                InputProps={{ readOnly: true }}
              />
            </Grid>
          </Grid>
          <Grid item xs={12} container>
            <Grid item xs={4} />
            <Grid item xs={4}>
              <TextField
                label="Last update time"
                value={GetTimeDateString(model.lastUpdateTime)}
                fullWidth
                InputProps={{ readOnly: true }}
              />
            </Grid>
          </Grid>
          <Grid item xs={12} container>
            <Grid item xs={4} />
            <Grid item xs={4}>
              <div style={{ textAlign: 'right' }}>
                <Button
                  color="primary"
                  type="submit"
                  disabled={applyDisabled}
                  onClick={this.handleApplyChanges}
                >
                  Apply changes
                </Button>
              </div>
            </Grid>
          </Grid>
        </Grid>
      </form>
    );
  }
}

export default ModelTabMeta;
