/* global jsPlumb */

import React from 'react';
import PropTypes from 'prop-types';
import { observer, PropTypes as MobxPropTypes } from 'mobx-react';
import { withStyles } from '@material-ui/core/styles';
import Paper from '@material-ui/core/Paper';
import '../external/jsplumb';
import HealthOutcomeWidget from './HealthOutcomeWidget';

jsPlumb.importDefaults({
  RenderMode: 'svg',
  Connector: ['Straight', { gap: 3 }],
  PaintStyle: {
    strokeWidth: 1,
    stroke: '#888',
    outlineStroke: 'transparent',
    outlineWidth: 0,
  },
  HoverPaintStyle: { strokeWidth: 2 },
  Endpoint: ['Dot', { radius: 2 }],
  Anchor: ['Left', 'Right'],
  ConnectionsDetachable: false,
  ConnectionOverlays: [
    ['Arrow', { length: 8, width: 8, location: 1, fallback: 0.8 }],
  ],
});

const styles = () => ({
  root: {
    minHeight: 400,
    width: '100%',
    resize: 'vertical',
    overflow: 'auto',
    position: 'relative',
    flex: '1 1 0%',
    color: 'white',
    backgroundColor: 'white',
    backgroundImage: `linear-gradient(#eeeeee 1px, transparent 1px),
      linear-gradient(90deg, #eeeeee 1px, transparent 1px),
      linear-gradient(rgba(235, 235, 235, .6) 1px, transparent 1px),
      linear-gradient(90deg, rgba(235, 235, 235, .6) 1px, transparent 1px)`,
    backgroundSize: '100px 100px, 100px 100px, 10px 10px, 10px 10px',
    backgroundPosition: '0px -1px, -1px 0px, 0px 0px, 0px 0px',
    backgroundClip: 'border-box',
    backgroundOrigin: 'padding-box',
  },
  connectionLabel: {
    fontSize: 12,
    color: 'black',
    backgroundColor: 'white',
  },
});

@withStyles(styles)
@observer
class OutcomeTreeRenderWidget extends React.Component {
  static propTypes = {
    classes: PropTypes.object.isRequired,
    outcomeTree: MobxPropTypes.observableObject.isRequired,
  };

  componentDidMount = () => {
    jsPlumb.ready(() => {
      this.renderOutcomeTree();
    });
  };

  componentDidUpdate = () => {
    jsPlumb.ready(() => {
      this.renderOutcomeTree();
    });
  };

  componentWillUnmount = () => {
    jsPlumb.ready(() => {
      jsPlumb.reset();
    });
  };

  makeGroups = healthOutcomes => {
    healthOutcomes.forEach(healthOutcome => {
      const healthOutcomeId = healthOutcome.id;
      const healthOutcomeNode = document.getElementById(healthOutcomeId);
      if (healthOutcomeNode) {
        jsPlumb.addGroup({
          el: healthOutcomeNode,
          id: healthOutcomeId,
          anchor: 'AutoDefault',
          endpoint: ['Dot', { radius: 2 }],
          draggable: true,
          dragOptions: {
            containment: true,
            grid: [10, 10],
            stop(event) {
              const top = event.pos[1];
              const left = event.pos[0];
              healthOutcome.setPosition(top, left);
            },
          },
          droppable: false,
          constrain: true,
          dropOverride: true,
        });
      }
    });
  };

  makeDraggable = healthOutcomes => {
    healthOutcomes.forEach(healthOutcome => {
      const healthOutcomeId = healthOutcome.id;
      const healthOutcomeNode = document.getElementById(healthOutcomeId);
      if (healthOutcomeNode) {
        jsPlumb.draggable(healthOutcomeNode, {
          containment: true,
          grid: [10, 10],
          stop(event) {
            const top = event.pos[1];
            const left = event.pos[0];
            healthOutcome.setPosition(top, left);
          },
        });
      }
    });
  };

  makeConnections = transitions => {
    const { classes } = this.props;
    transitions.forEach(transition => {
      const conn = jsPlumb.connect({
        source: transition.source.id,
        target: transition.target.id,
        anchors: [['Right'], ['Left']],
        overlays: [
          [
            'Label',
            {
              label: transition.name,
              cssClass: classes.connectionLabel,
            },
          ],
        ],
      });
      if (conn) {
        conn.transitionId = transition.id;
      }
    });
  };

  getHealthOutcomeNodes = healthOutcomes => {
    const nodes = healthOutcomes.map(healthOutcome => {
      return (
        <HealthOutcomeWidget
          key={healthOutcome.id}
          healthOutcome={healthOutcome}
          onCollapseExpandBtnClick={this.handleCollapseExpandBtnClick}
        />
      );
    });
    return nodes;
  };

  renderOutcomeTree = () => {
    const widget = this;
    const { outcomeTree } = this.props;
    const healthOutcomes = outcomeTree.healthOutcomesAsArray;
    const transitions = outcomeTree.transitionsAsArray;

    jsPlumb.ready(() => {
      jsPlumb.setSuspendDrawing(true);
      jsPlumb.setContainer('graphContainer');
      jsPlumb.reset();
      widget.makeDraggable(healthOutcomes);
      widget.makeConnections(transitions);
      jsPlumb.setSuspendDrawing(false, true);
    });
  };

  render() {
    const {
      classes,
      outcomeTree: {
        healthOutcomesAsArray: healthOutcomes,
        transitionsAsArray: transitions,
      },
    } = this.props;

    // Force re-render on transition modification
    transitions.forEach(transition => {
      const { name, source, target } = transition; // eslint-disable-line no-unused-vars
    });

    const healthOutcomeNodes = this.getHealthOutcomeNodes(healthOutcomes);

    return (
      <Paper id="graphContainer" className={classes.root}>
        {healthOutcomeNodes}
      </Paper>
    );
  }
}

export default OutcomeTreeRenderWidget;
