import { observable, action, computed, values, reaction, toJS } from 'mobx';
import { serialize, deserialize } from 'serializr';
import GenerateId from '../utilities/GenerateId';
import SaveDataLocally from '../utilities/SaveDataLocally';
import GetTopologicalOrder from '../utilities/GetTopologicalOrder';
import { ElementType, Sex, Age, SupportedSpecVersions } from '../settings';

export class Dims {
  id = null;

  @observable
  values = new Map();

  @observable
  order = [];

  static default = () =>
    new Dims(new Map(Object.entries({ age: Age, sex: Sex })), ['age', 'sex']);

  constructor(valuesMap, order) {
    this.id = GenerateId('dims');
    this.values = valuesMap;
    this.order = order;
  }

  @action
  setValues = (dimName, dimCategories) => {
    this.values.set(dimName, dimCategories);
  };

  @computed
  get valid() {
    return (
      this.values &&
      this.order &&
      this.values.size === 2 &&
      this.order.length === 2
    );
  }
}

export class DataArray {
  defVal = null;

  @observable
  dims = null;

  @observable
  vals = [];

  static createDataArray = (defVal = 0, dims = Dims.default) =>
    new DataArray(defVal, dims);

  static GetDimsValuesCompatibility = (dims, vals) => {
    if (!dims) {
      return false;
    }
    const dimValues = dims.values;
    const dimOrder = dims.order;

    const dimsRowNum =
      (dimValues && dimOrder && dimValues.get(dimOrder[0]).length) || 0;
    const dimsColNum =
      (dimValues && dimOrder && dimValues.get(dimOrder[1]).length) || 0;

    const valsRowNum = (vals && vals.length) || 0;
    const valsColNum = (vals && vals.length > 0 && vals[0].length) || 0;

    return (
      dimsRowNum > 0 &&
      dimsColNum > 0 &&
      dimsRowNum === valsRowNum &&
      dimsColNum === valsColNum
    );
  };

  constructor(defVal = '', dims = Dims.default, vals = null) {
    this.defVal = defVal;
    this.dims = dims;
    if (DataArray.GetDimsValuesCompatibility(dims)) {
      this.vals = vals;
    } else if (this.dims) {
      let rowNum = 0;
      let colNum = 0;
      if (dims && dims.valid) {
        rowNum = dims.values.get(dims.order[0]).length;
        colNum = dims.values.get(dims.order[1]).length;
      }

      this.vals = new Array(rowNum)
        .fill(defVal)
        .map(() => new Array(colNum).fill(defVal));
    } else {
      this.vals = [];
    }

    reaction(() => toJS(this.dims), () => this.adjustValuesToDims());
  }

  @computed
  get dimOrder() {
    return this.dims.order;
  }

  @computed
  get asObject() {
    const { dims, vals, defVal } = this;

    if (!dims) {
      return null;
    }

    const defRowVals = dims.order.map(() => defVal);

    const headRow = {
      rowName: dims.order[0],
      rowVals: dims.values.get(dims.order[1]),
    };

    const bodyRows = dims.values.get(dims.order[0]).map((val, i) => {
      const rowVals = vals[i] || defRowVals;
      return { rowName: val, rowVals };
    });

    return { headRow, bodyRows };
  }

  @computed
  get asText() {
    const { dims, vals, defVal } = this;
    const defRowVals = dims.order.map(() => defVal);

    const headText = `${dims.order[0]}\t ${dims.values
      .get(dims.order[1])
      .join('\t ')}\n`;

    const bodyText = dims.values.get(dims.order[0]).map((val, i) => {
      const rowVals = vals[i] || defRowVals;
      return `'${val}\t ${rowVals.join('\t ')}`;
    });

    return headText + bodyText.join('\n');
  }

  @action
  adjustValuesToDims = () => {
    if (!this.dims) {
      return;
    }

    const valuesLength = this.vals.length;
    const dimsLength = this.dims.values.get(this.dims.order[0]).length;

    const lengthDiff = dimsLength - valuesLength;
    if (lengthDiff > 0) {
      this.vals = new Array(lengthDiff)
        .fill([this.defVal, this.defVal])
        .concat(this.vals);
    } else if (lengthDiff < 0) {
      this.vals.splice(0, -lengthDiff);
    }
  };

  @action
  fillWithValue = value => {
    this.vals.map(row => row.fill(value));
  };

  @action
  setValue = (row, col, value) => {
    this.vals[row][col] = value;
  };

  @action
  replaceValue = (oldValue, newValue = '') => {
    for (let i = 0; i < this.vals.length; i += 1) {
      for (let j = 0; j < this.vals[i].length; j += 1) {
        if (this.vals[i][j] === oldValue) {
          this.vals[i][j] = newValue;
        }
      }
    }
  };
}

export class HealthState {
  elementType = ElementType.HEALTH_STATE;

  healthOutcome = null;

  id = null;

  constructor(
    healthOutcome,
    name,
    severityArray = null,
    durationArray = null,
    proportionArray = null,
    remainingCases
  ) {
    this.healthOutcome = healthOutcome;
    this.id = GenerateId('healthState');
    this.name = name;
    this.severityArray = severityArray;
    this.durationArray = durationArray;
    this.proportionArray = proportionArray;
    this.remainingCases = remainingCases;
  }

  @observable
  name = '';

  @observable
  transitionsFromIds = [];

  @observable
  transitionsToIds = [];

  @observable
  severityArray = null;

  @observable
  durationArray = null;

  @observable
  proportionArray = null;

  @observable
  remainingCases = false;

  @computed
  get healthOutcomeId() {
    return this.healthOutcome.id;
  }

  @action
  addTransitionFromId = transitionId => {
    if (!this.transitionsFromIds.includes(transitionId)) {
      this.transitionsFromIds.push(transitionId);
    }
  };

  @action
  removeTansitionFromId = transitionId => {
    if (this.transitionsFromIds.includes(transitionId)) {
      this.transitionsFromIds.pop(transitionId);
    }
  };

  @action
  addTransitionToId = transitionId => {
    if (!this.transitionsToIds.includes(transitionId)) {
      this.transitionsToIds.push(transitionId);
    }
  };

  @action
  removeTansitionToId = transitionId => {
    if (this.transitionsToIds.includes(transitionId)) {
      this.transitionsToIds.pop(transitionId);
    }
  };

  @action
  modify = (
    name,
    severityArray = this.healthOutcome.outcomeTree.model.createDataArray(),
    durationArray = this.healthOutcome.outcomeTree.model.createDataArray(),
    proportionArray = this.healthOutcome.outcomeTree.model.createDataArray(),
    remainingCases,
  ) => {
    this.name = name;
    this.severityArray = severityArray;
    this.durationArray = durationArray;
    this.proportionArray = proportionArray;
    this.remainingCases = remainingCases;
    return this;
  };
}

export class HealthOutcome {
  elementType = ElementType.HEALTH_OUTCOME;

  outcomeTree = null;

  id = null;

  constructor(
    outcomeTree,
    name,
    type,
    top = 0,
    left = 0,
    severityArray = null,
    durationArray = null,
  ) {
    this.outcomeTree = outcomeTree;
    this.id = GenerateId('healthOutcome');
    this.name = name;
    this.type = type;
    this.top = parseInt(top, 10);
    this.left = parseInt(left, 10);
    this.severityArray = severityArray;
    this.durationArray = durationArray;
  }

  @observable
  name = '';

  @observable
  type = '';

  @observable
  top = 0;

  @observable
  left = 0;

  @observable
  healthStates = new Map();

  @observable
  selectedHealthState = null;

  @observable
  collapsed = false;

  @observable
  transitionsFromIds = [];

  @observable
  transitionsToIds = [];

  @observable
  severityArray = null;

  @observable
  durationArray = null;

  @computed
  get healthStatesAsArray() {
    const healthStates = Array.from(this.healthStates.values());
    return healthStates;
  }

  @computed
  get healthStateNames() {
    const names = Array.from(this.healthStates.values()).map(
      healthState => healthState.name,
    );
    return names;
  }

  @computed
  get deepTransitionsToIds() {
    let { transitionsToIds } = this;
    this.healthStatesAsArray.forEach(healthState => {
      transitionsToIds = [
        ...new Set([...transitionsToIds, ...healthState.transitionsToIds]),
      ];
    });
    return transitionsToIds;
  }

  @computed
  get deepTransitionsFromIds() {
    let { transitionsFromIds } = this;
    this.healthStatesAsArray.forEach(healthState => {
      transitionsFromIds = [
        ...new Set([...transitionsFromIds, ...healthState.transitionsFromIds]),
      ];
    });
    return transitionsFromIds;
  }

  @computed
  get healthOutcomeId() {
    return this.id;
  }

  @action
  addHealthState = (
    name,
    severityArray = this.outcomeTree.model.createDataArray(),
    durationArray = this.outcomeTree.model.createDataArray(),
    proportionArray = this.outcomeTree.model.createDataArray(),
    remainingCases
  ) => {
    const healthState = new HealthState(
      this,
      name,
      severityArray,
      durationArray,
      proportionArray,
      remainingCases
    );
    if (healthState) {
      this.healthStates.set(healthState.id, healthState);
      this.outcomeTree.model.store.sentMsg(
        `Health state with name ${healthState.name} created`,
      );
    } else {
      this.outcomeTree.model.store.sentMsg('Health state could not be created');
    }

    return healthState;
  };

  @action
  deleteHealthState = (id, name = null) => {
    if (this.healthStates.has(id)) {
      const healthState = this.healthStates.get(id);
      values(healthState.transitionsFromIds).forEach(transitionId =>
        this.outcomeTree.removeTransition(transitionId),
      );
      values(healthState.transitionsToIds).forEach(transitionId =>
        this.outcomeTree.removeTransition(transitionId),
      );

      this.healthStates.delete(id);
      if (name) {
        this.outcomeTree.model.store.sentMsg(
          `Health state with name "${name}" deleted`,
        );
      }
    } else if (name) {
      this.outcomeTree.model.store.sentMsg(
        `Health state with name "${name}" does not exist and cannot be deleted`,
      );
    }
  };

  @action
  deleteAllHealthStates = () => {
    this.healthStates.forEach((value, key) => this.deleteHealthState(key));
  };

  @action
  modify = (
    name,
    type,
    top,
    left,
    severityArray = this.outcomeTree.model.createDataArray(),
    durationArray = this.outcomeTree.model.createDataArray(),
  ) => {
    this.name = name;
    this.type = type;
    this.top = parseInt(top, 10);
    this.left = parseInt(left, 10);
    this.severityArray = severityArray;
    this.durationArray = durationArray;
    return this;
  };

  @action
  setPosition = (top, left) => {
    this.top = top;
    this.left = left;
    return this;
  };

  @action
  setCollapsed = collapsed => {
    this.collapsed = collapsed;
    return this;
  };

  @action
  addTransitionFromId = transitionId => {
    if (!this.transitionsFromIds.includes(transitionId)) {
      this.transitionsFromIds.push(transitionId);
    }
  };

  @action
  removeTansitionFromId = transitionId => {
    if (this.transitionsFromIds.includes(transitionId)) {
      this.transitionsFromIds.pop(transitionId);
    }
  };

  @action
  addTransitionToId = transitionId => {
    if (!this.transitionsToIds.includes(transitionId)) {
      this.transitionsToIds.push(transitionId);
    }
  };

  @action
  removeTansitionToId = transitionId => {
    if (this.transitionsToIds.includes(transitionId)) {
      this.transitionsToIds.pop(transitionId);
    }
  };
}

export class Transition {
  elementType = ElementType.TRANSITION;

  outcomeTree = null;

  id = null;

  constructor(
    outcomeTree,
    name,
    source = null,
    target = null,
    annualType = false,
    annualDuration = -1.0,
    redistributeType = false,
    probabilityArray = null,
    redistributeArray = null,
    complementaryType = false,
    complementaryTransition = null,
  ) {
    this.outcomeTree = outcomeTree;
    this.id = GenerateId('transition');
    this.name = name;
    this.setSource(source);
    this.setTarget(target);
    this.annualType = annualType;
    this.annualDuration = annualDuration;
    this.redistributeType = redistributeType;
    this.probabilityArray = probabilityArray;
    this.redistributeArray = redistributeArray;
    this.complementaryType = complementaryType;
    this.complementaryTransition = complementaryTransition;
  }

  @observable
  name = '';

  @observable
  annualType = false;

  @observable
  annualDuration = -1.0;

  @observable
  redistributeType = false;

  @observable
  source = null;

  @observable
  target = null;

  @observable
  complementaryType = false;

  @observable
  complementaryTransition = null;

  @observable
  probabilityArray = null;

  @observable
  redistributeArray = null;

  @action
  setSource = source => {
    if (source !== this.source) {
      if (this.source) {
        this.source.removeTansitionFromId(this.id);
      }
      if (source) {
        source.addTransitionFromId(this.id);
      }
      this.source = source;
    }
  };

  @action
  setTarget = target => {
    if (target !== this.target) {
      if (this.target) {
        this.target.removeTansitionToId(this.id);
      }
      if (target) {
        target.addTransitionToId(this.id);
      }
      this.target = target;
    }
  };

  @action
  modify = (
    name,
    source,
    target,
    annualType,
    annualDuration,
    redistributeType,
    probabilityArray = this.outcomeTree.model.createDataArray(),
    redistributeArray = this.outcomeTree.model.createDataArray(),
    complementaryType,
    complementaryTransition,
  ) => {
    this.name = name;
    this.setSource(source);
    this.setTarget(target);
    this.annualType = annualType;
    this.annualDuration = annualDuration;
    this.redistributeType = redistributeType;
    this.probabilityArray = probabilityArray;
    this.redistributeArray = redistributeArray;
    this.complementaryType = complementaryType;
    this.complementaryTransition = complementaryTransition;
    return this;
  };
}

Transition.prototype.isValid = () => {
  const outcomeTreeIdValid = !!this.outcomeTree;

  return outcomeTreeIdValid;
};

export class OutcomeTree {
  model = null;

  id = null;

  constructor(model, name) {
    this.model = model;
    this.id = GenerateId('outcomeTree');
    this.name = name;
  }

  @observable
  name = '';

  @observable
  healthOutcomes = new Map();

  @observable
  transitions = new Map();

  @observable
  selectedHealthOutcome = null;

  @observable
  selectedTransition = null;

  @observable
  healthOutcomesRunOrder = [];

  @computed
  get healthOutcomesAsArray() {
    const healthOutcomes = Array.from(this.healthOutcomes.values());
    return healthOutcomes;
  }

  @computed
  get runOrderedHealthOutcomesAsArray() {
    const healthOutcomes = this.healthOutcomesRunOrder.map(id =>
      this.healthOutcomes.get(id),
    );
    return healthOutcomes;
  }

  @computed
  get healthOutcomeNames() {
    const names = Array.from(this.healthOutcomes.values()).map(
      healthOutcome => healthOutcome.name,
    );
    return names;
  }

  @computed
  get nonSelectedHealthOutcomeNames() {
    let names;
    if (this.selectedHealthOutcome) {
      names = this.healthOutcomeNames.filter(
        name => name !== this.selectedHealthOutcome.name,
      );
    } else {
      names = this.healthOutcomeNames;
    }

    return names;
  }

  @computed
  get nonSelectedTransitionNames() {
    let names;
    if (this.selectedTransition) {
      names = this.transitionNames.filter(
        name => name !== this.selectedTransition.name,
      );
    } else {
      names = this.transitionNames;
    }

    return names;
  }

  @computed
  get transitionsAsArray() {
    const transitions = Array.from(this.transitions.values());
    return transitions;
  }

  @computed
  get complTransCandidatesAsArray() {
    const transitions = this.transitionsAsArray.filter(
      transition =>
        !transition.complementaryType &&
        ((this.selectedTransition !== null &&
          transition.id !== this.selectedTransition.id) ||
          this.selectedTransition == null),
    );
    return transitions;
  }

  @computed
  get transitionNames() {
    const names = Array.from(this.transitions.values()).map(
      transition => transition.name,
    );
    return names;
  }

  @computed
  get graph() {
    const nodes = this.healthOutcomesAsArray.reduce((map, elem) => {
      // eslint-disable-next-line no-param-reassign
      map[elem.id] = {
        id: elem.id,
        type: elem.type,
      };
      return map;
    }, {});

    const edges = this.transitionsAsArray
      // Get array of edge specifications
      .map(transition => ({
        sourceId: transition.source.healthOutcomeId,
        targetId: transition.target.healthOutcomeId,
      }))
      // Remove duplicates
      .filter(
        (elem, index, self) =>
          self.findIndex(
            t => t.sourceId === elem.sourceId && t.targetId === elem.targetId,
          ) === index,
      );

    return { nodes, edges };
  }

  @action
  addHealthOutcome = (
    name,
    type,
    top = 0,
    left = 0,
    severityArray = this.model.createDataArray(),
    durationArray = this.model.createDataArray(),
  ) => {
    const healthOutcome = new HealthOutcome(
      this,
      name,
      type,
      top,
      left,
      severityArray,
      durationArray,
    );
    if (healthOutcome) {
      this.healthOutcomes.set(healthOutcome.id, healthOutcome);
      const orderResult = GetTopologicalOrder(this.graph);
      if (orderResult.success) {
        this.healthOutcomesRunOrder = orderResult.order;
      }
      this.model.store.sentMsg(
        `Health outcome with name "${healthOutcome.name}" created`,
      );
    } else {
      this.model.store.sentMsg(
        `Health outcome with name ${name} could not be created`,
      );
    }
    return healthOutcome;
  };

  @action
  modifyHealthOutcome = (
    id,
    name,
    type,
    top,
    left,
    severityArray = this.model.createDataArray(),
    durationArray = this.model.createDataArray(),
  ) => {
    const healthOutcome = this.healthOutcomes.get(id);
    if (healthOutcome) {
      healthOutcome.modify(name, type, top, left, severityArray, durationArray);
      this.model.store.sentMsg(
        `Health outcome with name "${healthOutcome.name}" modified`,
      );
    } else {
      this.model.store.sentMsg(
        `Health outcome with name ${name} could not be modified`,
      );
    }

    return healthOutcome;
  };

  @action
  addTransition = (
    name,
    source = null,
    target = null,
    annualType = false,
    annualDuration = -1.0,
    redistributeType = false,
    probabilityArray = this.model.createDataArray(),
    redistributeArray = this.model.createDataArray(),
    complementaryType = false,
    complementaryTransition = null,
  ) => {
    let transition = null;
    const { graph } = this;
    graph.edges.push({
      sourceId: source.healthOutcomeId,
      targetId: target.healthOutcomeId,
    });
    const orderResult = GetTopologicalOrder(this.graph);

    if (orderResult.success) {
      transition = new Transition(
        this,
        name,
        source,
        target,
        annualType,
        annualDuration,
        redistributeType,
        probabilityArray,
        redistributeArray,
        complementaryType,
        complementaryTransition,
      );
      this.transitions.set(transition.id, transition);
      this.healthOutcomesRunOrder = orderResult.order;
      this.model.store.sentMsg(
        `Transition with name "${transition.name}" created`,
      );
    } else {
      this.model.store.sentMsg(
        'Transition could not be added because it would create a cycle',
      );
    }
    return transition;
  };

  @action
  modifyTransition = (
    id,
    name,
    source,
    target,
    annualType,
    annualDuration,
    redistributeType,
    probabilityArray = this.model.createDataArray(),
    redistributeArray = this.model.createDataArray(),
    complementaryType,
    complementaryTransition,
  ) => {
    const transition = this.transitions.get(id);
    if (transition) {
      transition.modify(
        name,
        source,
        target,
        annualType,
        annualDuration,
        redistributeType,
        probabilityArray,
        redistributeArray,
        complementaryType,
        complementaryTransition,
      );
      const orderResult = GetTopologicalOrder(this.graph);
      if (orderResult.success) {
        this.healthOutcomesRunOrder = orderResult.order;
      }

      this.model.store.sentMsg(
        `Transition with name "${transition.name}" modified`,
      );
    } else {
      this.model.store.sentMsg(
        `Transition with name ${name} could not be modified`,
      );
    }

    return transition;
  };

  @action
  deleteHealthOutcome = (id, name) => {
    if (this.healthOutcomes.has(id)) {
      const healthOutcome = this.healthOutcomes.get(id);
      // Delete transitions
      values(healthOutcome.transitionsFromIds).forEach(transitionId =>
        this.removeTransition(transitionId),
      );
      values(healthOutcome.transitionsToIds).forEach(transitionId =>
        this.removeTransition(transitionId),
      );
      // Delete health states
      healthOutcome.deleteAllHealthStates();
      // Delete the health outcome itself
      this.healthOutcomes.delete(id);
      // Update the run order
      const orderResult = GetTopologicalOrder(this.graph);
      if (orderResult.success) {
        this.healthOutcomesRunOrder = orderResult.order;
      }

      this.model.store.sentMsg(`Health outcome with name "${name}" deleted`);
    } else {
      this.model.store.sentMsg(
        `Health outcome with name "${name}" does not exist and cannot be deleted`,
      );
    }
  };

  removeTransition = id => {
    if (this.transitions.has(id)) {
      const transition = this.transitions.get(id);
      transition.source.removeTansitionFromId(id);
      transition.target.removeTansitionToId(id);
      this.transitions.delete(id);
    }
  };

  @action
  deleteTransition = (id, name = null) => {
    if (this.transitions.has(id)) {
      this.removeTransition(id);
      this.transitions.delete(id);
      const orderResult = GetTopologicalOrder(this.graph);
      if (orderResult.success) {
        this.healthOutcomesRunOrder = orderResult.order;
      }

      if (name) {
        this.model.store.sentMsg(`Transition with name "${name}" deleted`);
      }
    } else if (name) {
      this.model.store.sentMsg(
        `Transition with name "${name}" does not exist and cannot be deleted`,
      );
    }
  };

  @action
  selectHealthOutcome = id => {
    if (this.healthOutcomes.has(id)) {
      this.selectedHealthOutcome = this.healthOutcomes.get(id);
    } else {
      this.deselectHealthOutcome();
    }
  };

  @action
  deselectHealthOutcome = () => {
    this.selectedHealthOutcome = null;
  };

  @action
  selectTransition = id => {
    if (this.transitions.has(id)) {
      this.selectedTransition = this.transitions.get(id);
    } else {
      this.deselectTransition();
    }
  };

  @action
  deselectTransition = () => {
    this.selectedTransition = null;
  };
}

export class Model {
  specVersion = 1;

  store = null;

  id = null;

  @observable
  dims = null;

  constructor(store = null, name = '') {
    this.store = store;
    this.id = GenerateId('model');
    this.dims = Dims.default();
    this.outcomeTreeMapping = this.createDataArray('');
    this.rawCasesCountArray = this.createDataArray();
    this.name = name.trim();
    this.createTime = new Date();
    this.lastUpdateTime = this.createTime;
  }

  @observable
  name = '';

  @observable
  description = '';

  @observable
  version = '';

  @observable
  author = '';

  @observable
  createTime = null;

  @observable
  lastUpdateTime = null;

  @observable
  isCongenitalType = false;

  @observable
  congenitalCategoryName = '';

  @observable
  outcomeTrees = new Map();

  @observable
  selectedOutcomeTree = null;

  @observable
  outcomeTreeMapping = null;

  @observable
  rawCasesCountArray = null;

  @action
  setStore = store => {
    this.store = store;
  };

  @action
  updateMetaInfo = (
    name = null,
    description = null,
    version = null,
    author = null,
    isCongenitalType = null,
    congenitalCategoryName = null,
  ) => {
    if (name != null) this.name = name.trim();
    if (description != null) this.description = description.trim();
    if (version != null) this.version = version.trim();
    if (author != null) this.author = author.trim();

    this.setCongenitalType(isCongenitalType, congenitalCategoryName);

    this.lastUpdateTime = new Date();

    this.store.sentMsg(`Details of model "${name}" updated`);
  };

  @action
  updateRawCasesCountArray = rawCasesCountArray => {
    this.rawCasesCountArray = rawCasesCountArray;

    this.lastUpdateTime = new Date();

    this.store.sentMsg(`Average number of cases for model "${name}" updated`);
  };

  @computed
  get outcomeTreesAsArray() {
    const outcomeTrees = Array.from(this.outcomeTrees.values());
    return outcomeTrees;
  }

  @computed
  get outcomeTreeNames() {
    const names = Array.from(this.outcomeTrees.values()).map(
      outcomeTree => outcomeTree.name,
    );
    return names;
  }

  @action
  setCongenitalType = (
    isCongenitalType = null,
    congenitalCategoryName = null,
  ) => {
    if (isCongenitalType != null) {
      if (!isCongenitalType) {
        this.isCongenitalType = isCongenitalType;
        this.dims.setValues('age', Age);
      } else if (congenitalCategoryName !== '') {
        this.isCongenitalType = isCongenitalType;
        this.congenitalCategoryName = congenitalCategoryName.trim();
        const ageCategories = Age.slice();
        ageCategories.unshift(this.congenitalCategoryName);
        this.dims.setValues('age', ageCategories);
      }
    }
  };

  @action
  addOutcomeTree = name => {
    const outcomeTree = new OutcomeTree(this, name);
    if (outcomeTree) {
      this.outcomeTrees.set(outcomeTree.id, outcomeTree);
      if (this.outcomeTrees.size === 1) {
        this.outcomeTreeMapping.fillWithValue(outcomeTree.id);
      }
      this.selectOutcomeTree(outcomeTree.id, outcomeTree.name);
    } else {
      this.store.sentMsg(
        `Outcome tree with name "${name}" could not be created`,
      );
    }

    return outcomeTree;
  };

  @action
  selectOutcomeTree = (id, name) => {
    if (this.selectedOutcomeTree) {
      this.selectedOutcomeTree.deselectHealthOutcome();
      this.selectedOutcomeTree.deselectTransition();
    }
    if (this.outcomeTrees.has(id)) {
      this.selectedOutcomeTree = this.outcomeTrees.get(id);
      this.selectedOutcomeTree.deselectHealthOutcome();
      this.selectedOutcomeTree.deselectTransition();
      this.store.sentMsg(
        `Outcome tree with name "${this.selectedOutcomeTree.name}" selected`,
      );
    } else {
      this.store.sentMsg(
        `Outcome tree with name "${name}" does not exist and cannot be selected`,
      );
    }
  };

  @action
  deleteOutcomeTree = (id, name) => {
    if (this.outcomeTrees.has(id)) {
      if (this.selectedOutcomeTree === this.outcomeTrees.get(id)) {
        this.selectedOutcomeTree = null;
      }
      this.outcomeTrees.delete(id);

      const lastOutcomeTreeId = Array.from(this.outcomeTrees.keys()).pop();
      if (lastOutcomeTreeId) {
        const lastOutcomeTree = this.outcomeTrees.get(lastOutcomeTreeId);
        this.selectOutcomeTree(lastOutcomeTree.id, lastOutcomeTree.name);
        this.outcomeTreeMapping.replaceValue(id, lastOutcomeTree.id);
      } else {
        this.outcomeTreeMapping.replaceValue(id, '');
      }

      this.store.sentMsg(`Outcome tree with name "${name}" deleted`);
    } else {
      this.store.sentMsg(
        `Outcome tree with name "${name}" does not exist and cannot be deleted`,
      );
    }
  };

  @action
  createDataArray = (defVal = 0, dims = this.dims) =>
    DataArray.createDataArray(defVal, dims);
}

export default class ModelsStore {
  rootStore = null;

  constructor(store) {
    this.rootStore = store;
  }

  sentMsg = msg => this.rootStore.notificationStore.setMsg(msg);

  @observable
  models = new Map();

  @observable
  selectedModel = null;

  @computed
  get modelsAsArray() {
    const models = Array.from(this.models.values());
    return models;
  }

  @computed
  get modelsNames() {
    const names = Array.from(this.models.values()).map(model => model.name);
    return names;
  }

  @computed
  get nonSelectedModelsNames() {
    const names = this.modelsNames.filter(
      name => name !== this.selectedModel.name,
    );
    return names;
  }

  @action
  createModel = name => {
    const model = new Model(this, name);
    if (model) {
      this.models.set(model.id, model);
      this.selectModel(model.id, model.name);
    } else {
      this.sentMsg(`Model with name "${model.name}" could not be created`);
    }

    return model;
  };

  @action
  addModelFromSpec = modelSpec => {
    let model = null;
    const specVersion = parseInt(modelSpec.specVersion, 10) || 0;
    if (SupportedSpecVersions.includes(specVersion)) {
      console.log(modelSpec);
      model = deserialize(Model, modelSpec);
      if (model) {
        if (!this.models.has(model.id)) {
          model.setStore(this);
          this.models.set(model.id, model);
          this.selectModel(model.id, model.name);
        } else {
          this.sentMsg(
            `Model with id "${
              model.id
            }" is already loaded. New model will not be added.`,
          );
        }
      } else {
        this.sentMsg(`Model with name "${model.name}" could not be added`);
      }
    } else {
      this.sentMsg(`Specification version "${specVersion}" is not supported.`);
    }

    return model;
  };

  @action
  selectModel = (id, name) => {
    if (this.models.has(id)) {
      this.selectedModel = this.models.get(id);
      this.sentMsg(`Model with name "${this.selectedModel.name}" selected`);
    } else {
      this.sentMsg(
        `Model with name "${name}" does not exist and cannot be selected`,
      );
    }
  };

  @action
  selectLastModel = () => {
    if (this.models.size > 0) {
      this.selectedModel = this.modelsAsArray[this.models.size - 1];
      this.sentMsg(`Model with name "${this.selectedModel.name}" selected`);
    } else {
      this.selectedModel = null;
      this.sentMsg('No model selected');
    }
  };

  @action
  deleteModel = (id, name) => {
    if (this.models.has(id)) {
      if (this.models.delete(id)) {
        this.sentMsg(`Model with name "${name}" deleted`);
        this.selectLastModel();
      }
    } else {
      this.sentMsg(
        `Model with name "${name}" does not exist and cannot be deleted`,
      );
    }
  };

  @action
  saveModel = (id, name) => {
    if (this.models.has(id)) {
      const model = this.models.get(id);
      const fileName = model.name.replace(/\s+/g, '-');
      const jsModel = serialize(model);
      const jsonModel = JSON.stringify(jsModel, function(k, v) {
        if (v instanceof Array) {
          return JSON.stringify(v);
        }
        return v;
      }, 2)
      .replace(/\\/g, '')
      .replace(/\"\[/g, '[')
      .replace(/\]\"/g, ']')
      .replace(/\"\{/g, '{')
      .replace(/\}\"/g, '}');
      if (window.modelCreateWebGUI) {
        window.modelCreateWebGUI.saveModel(jsonModel);
      } else {
        const mime = 'application/json';
        const fileNameSuffix = '.json';
        SaveDataLocally(jsonModel, fileName + fileNameSuffix, mime);
        this.sentMsg(`Model with name "${name}" saved`);
      }
    } else {
      this.sentMsg(
        `Model with name "${name}" does not exist and cannot be saved`,
      );
    }
  };

  @action
  emailModel = (id, name) => {
    if (this.models.has(id)) {
      if (window.modelCreateWebGUI) {
        const model = this.models.get(id);
        const jsModel = serialize(model);
        window.modelCreateWebGUI.emailModel(jsModel);
        this.sentMsg(`Model "${name}" has been emailed`);
      } else {
        this.sentMsg(
          'Sending models via email is only supported when integrated with BCoDE toolkit.',
        );
      }
    } else {
      this.sentMsg(
        `Model with name "${name}" does not exist and cannot be emailed`,
      );
    }
  };
}
