import {
  createModelSchema,
  identifier,
  primitive,
  date,
  reference,
  map,
  list,
  object,
  custom,
} from 'serializr';
import {
  Model,
  OutcomeTree,
  Transition,
  HealthOutcome,
  HealthState,
  DataArray,
  Dims,
} from './Models';

createModelSchema(
  Dims,
  {
    id: identifier(),
    values: map(list(primitive())),
    order: list(primitive()),
  },
  context => {
    const jsonObject = context.json;
    const dims = new Dims(
      new Map(Object.entries(jsonObject.values)),
      jsonObject.order,
    );
    dims.id = jsonObject.id;
    return dims;
  },
);

createModelSchema(DataArray, {
  defVal: primitive(),
  dims: reference(Dims),
  vals: list(list(primitive())),
});

createModelSchema(HealthState, {
  elementType: primitive(),
  healthOutcome: reference(HealthOutcome),
  id: identifier(),
  name: primitive(),
  transitionsFromIds: list(primitive()),
  transitionsToIds: list(primitive()),
  severityArray: object(DataArray),
  durationArray: object(DataArray),
  proportionArray: object(DataArray),
  remainingCases: primitive()
});

createModelSchema(HealthOutcome, {
  elementType: primitive(),
  outcomeTree: reference(OutcomeTree),
  id: identifier(),
  selectedHealthState: reference(HealthState),
  name: primitive(),
  type: primitive(),
  top: primitive(),
  left: primitive(),
  healthStates: map(object(HealthState)),
  transitionsFromIds: list(primitive()),
  transitionsToIds: list(primitive()),
  severityArray: object(DataArray),
  durationArray: object(DataArray),
});

createModelSchema(Transition, {
  elementType: primitive(),
  outcomeTree: reference(OutcomeTree),
  id: identifier(),
  name: primitive(),
  annualType: primitive(),
  annualDuration: primitive(),
  redistributeType: primitive(),
  complementaryType: primitive(),
  complementaryTransition: reference(Transition),
  source: custom(
    (value, key, obj) => ({
      id: obj.source.id,
      elementType: obj.source.elementType,
      parentId:
        obj.source.elementType === 'HEALTH_STATE'
          ? obj.source.healthOutcome.id
          : null,
    }),
    (value, context) => {
      const { id, elementType, parentId } = value;
      if (elementType === 'HEALTH_OUTCOME') {
        return context.parentContext.target.healthOutcomes.get(id);
      }
      return context.parentContext.target.healthOutcomes
        .get(parentId)
        .healthStates.get(id);
    },
  ),
  target: custom(
    (value, key, obj) => ({
      id: obj.target.id,
      elementType: obj.target.elementType,
      parentId:
        obj.target.elementType === 'HEALTH_STATE'
          ? obj.target.healthOutcome.id
          : null,
    }),
    (value, context) => {
      const { id, elementType, parentId } = value;
      if (elementType === 'HEALTH_OUTCOME') {
        return context.parentContext.target.healthOutcomes.get(id);
      }
      return context.parentContext.target.healthOutcomes
        .get(parentId)
        .healthStates.get(id);
    },
  ),
  probabilityArray: object(DataArray),
  redistributeArray: object(DataArray),
});

createModelSchema(OutcomeTree, {
  model: reference(Model),
  id: identifier(),
  name: primitive(),
  healthOutcomesRunOrder: list(primitive()),
  healthOutcomes: map(object(HealthOutcome)),
  transitions: map(object(Transition)),
  selectedHealthOutcome: reference(HealthOutcome),
  selectedTransition: reference(Transition),
});

export const ModelSchema = createModelSchema(Model, {
  specVersion: primitive(),
  id: identifier(),
  name: primitive(),
  description: primitive(),
  dims: object(Dims),
  version: primitive(),
  author: primitive(),
  createTime: date(),
  lastUpdateTime: date(),
  isCongenitalType: primitive(),
  congenitalCategoryName: primitive(),
  selectedOutcomeTree: reference(OutcomeTree),
  outcomeTrees: map(object(OutcomeTree)),
  outcomeTreeMapping: object(DataArray),
  rawCasesCountArray: object(DataArray),
});
