#' Run BCoDE models
#'
#' Executes BCoDE models using the deployed Windows version of the toolkit.
#'
#' Results are returned to R as list object. Optionally, results can be written
#' to output file and/or clipboard.
#'
#' @param inputFile String with path to the Excel or JSON file with BCoDE
#'   model(s). Required input.
#' @param outputFile String with path to the output JSON file. If NULL then no
#'   output file is created. Default = NULL.
#' @param outputClipboard Logical. If TRUE then results will be written to
#'   clipboard. Default = FALSE.
#' @param sampleSize Integer number of iterations for Monte Carlo.
#'   Default = 1000L
#' @param randomSeed Integer number to be used as seed for the random number
#'   generator. If 0 then the seed effectively used will be time based, thus
#'   every run will result in a new sample generated. Default = 0L
#' @param timeDiscounting Time discounting. If 0, then no discounting is
#'   applied. Default = 0.0.
#' @param countryID ID of country to be used for population data. Required if
#'   input file is a JSON file. Default = 'EUEEA'.
#' @param bcodeFilePath String with path to the BCoDE executable. Required.
#'   Default = NULL.
#'
#' @return
#' @importFrom jsonlite read_json
#' @export
#'
#' @examples
#' \dontrun{
#' results <- RunBCoDE(
#'   inputFile = 'D:/_REPOSITORIES/Test.xls',
#'   outputFile = 'D:/_REPOSITORIES/Results.json',
#'   outputClipboard = FALSE,
#'   bcodeFilePath = 'D:/_DEPLOYMENT/BCoDE/BCoDE.exe'
#' )
#' results <- RunBCoDE(
#'   inputFile = './testModels/Listeriosis.json',
#'   outputFile = 'D:/_REPOSITORIES/Results.json',
#'   outputClipboard = FALSE,
#'   sampleSize = 1000L,
#'   countryID = 'Poland',
#'   bcodeFilePath = 'D:/_DEPLOYMENT/BCoDE/BCoDE.exe'
#' )
#' }
RunBCoDE <- function(
  inputFile,
  outputFile = NULL,
  outputClipboard = FALSE,
  sampleSize = 1000L,
  randomSeed = 0L,
  timeDiscounting = 0.0,
  countryID = 'EUEEA',
  bcodeFilePath = NULL
) {
  if (is.null(bcodeFilePath) || !file.exists(bcodeFilePath)) {
    stop('Argument "bcodeFilePath" does not contain path to valid BCoDE toolkit')
  }

  if (missing(inputFile)) {
    stop('Argument "inputFile" must be specified')
  }

  if (
    !file.exists(inputFile) &&
    !file.exists(file.path(dirname(bcodeFilePath), inputFile))
  ) {
    stop('Input file "', inputFile, '" does not exist')
  }

  if (
    tolower(tools::file_ext(inputFile)) == 'json' &&
    (!is.atomic(countryID) || !is.character(countryID) || nchar(countryID) == 0)
  ) {
      stop('Argument "country" must be a valid country name')
  }

  if (is.null(outputFile) || !is.character(outputFile)) {
    outputFile <- tempfile()
    on.exit({
      unlink(outputFile)
    })
  }

  args <- sprintf(
    '-i "%s" -o "%s" -s %d -r %d -t %6.9f -p %s',
    inputFile,
    outputFile,
    as.integer(sampleSize),
    as.integer(randomSeed),
    as.numeric(timeDiscounting),
    countryID
  )
  if (outputClipboard) {
    args <- paste(args, "-c")
  }

  system2(command = bcodeFilePath, args = args)

  isOutputPathAbsolute <- regexpr("^.:(/|\\\\)", outputFile) != -1L
  if (!isOutputPathAbsolute) {
    outputFile <- file.path(dirname(bcodeFilePath), outputFile)
  }
  print(outputFile)
  results <- jsonlite::read_json(outputFile)

  return(results)
}
