//==========================================================================================================================
//  Implementation file for "ArrayHelpers.h"
//  Written by Daniel Lewandowski
//  Copyright @ 2011 NextPage Software
//==========================================================================================================================

#include "fn/ArrayHelpers.h"

#include <QVariant>
#include "fn/DataArray.h"

namespace fn {

void AppendMetaDataToVariantMap(QVariantMap& variantMap, const fn::StringToStringMap& metaData)
{

    QVariantList list;

    QStringList order = QStringList() << "description" << "weblink" << "remaining cases" << "redistribute" << "direct type" << "duration";
    QMap<QString, QString> names = QMap<QString, QString>();
    names.insert("description", "Description");
    names.insert("weblink", "Weblink");
    names.insert("remaining cases", "Remaining cases");
    names.insert("redistribute", "Age-group redistribution");
    names.insert("direct type", "Lifetime transition probability");
    names.insert("duration", "Duration");

    int counter = 0;
    for (QStringList::const_iterator it = order.begin(); it != order.end(); ++it) {
        QString key = *it;
        if (metaData.contains(key)) {
            QVariantMap item;
            QString value = metaData[key];
            item.insert("id", counter);
            item.insert("ParameterName", key);
            item.insert("DisplayString", names[key]);
            item.insert("Value", value);
            list.append(item);
            counter++;
        }
    }

    variantMap.insert("meta", list);
}

QVariantMap FormulaArrayToVariantMap(const fn::StringBaseArray& formulaArray, const fn::ConstInputFactorPtrList& inputFactorList)
{
    QVariantMap result;

    QVariantList factorNames;
    QVariantList array;

    if ((inputFactorList.length() == 2) & (formulaArray.length() > 0) & ((formulaArray.length() % 2) == 0)) {
        factorNames.append(inputFactorList.at(0)->getName());
        factorNames.append(inputFactorList.at(1)->getName());

        const QStringList& rowNames = inputFactorList.at(0)->getCategoryList();
        const QStringList& colNames = inputFactorList.at(1)->getCategoryList();

        for (int i = 0; i < rowNames.length(); ++i) {
            QVariantMap map;
            map["id"] = i;
            map["Age"] = rowNames[i];
            for (int j = 0; j < colNames.length(); ++j) {
                map[colNames[j]] = formulaArray.at(i * 2 + j);
            }
            array.append(map);
        }
    }
    result.insert("factorNames", factorNames);
    result.insert("stratified", array);

    return result;
}

QVariantMap NumericArrayToVariantMap(const fn::NumericBaseArray& valueArray, const fn::ConstInputFactorPtrList& inputFactorList)
{
    QVariantMap result;

    QVariantList factorNames;
    QVariantList array;

    if ((inputFactorList.length() == 2) & (valueArray.size() > 0) & ((valueArray.size() % 2) == 0)) {
        factorNames.append(inputFactorList.at(0)->getName());
        factorNames.append(inputFactorList.at(1)->getName());

        const QStringList& rowNames = inputFactorList.at(0)->getCategoryList();
        const QStringList& colNames = inputFactorList.at(1)->getCategoryList();

        for (int i = 0; i < rowNames.length(); ++i) {
            QVariantMap map;
            map["id"] = i;
            map["Age"] = rowNames[i];
            for (int j = 0; j < colNames.length(); ++j) {
                map[colNames[j]] = valueArray(i * 2 + j);
            }
            array.append(map);
        }
    }
    result.insert("factorNames", factorNames);
    result.insert("stratified", array);

    return result;
}

QVariantMap BoolArrayToVariantMap(const fn::BoolBaseArray& boolArray, const fn::ConstInputFactorPtrList& inputFactorList)
{
    QVariantMap result;

    QVariantList factorNames;
    QVariantList array;

    if ((inputFactorList.length() == 2) & (boolArray.length() > 0) & ((boolArray.length() % 2) == 0)) {
        factorNames.append(inputFactorList.at(0)->getName());
        factorNames.append(inputFactorList.at(1)->getName());

        const QStringList& rowNames = inputFactorList.at(0)->getCategoryList();
        const QStringList& colNames = inputFactorList.at(1)->getCategoryList();

        for (int i = 0; i < rowNames.length(); ++i) {
            QVariantMap map;
            map["id"] = i;
            map["Age"] = rowNames[i];
            for (int j = 0; j < colNames.length(); ++j) {
                map[colNames[j]] = boolArray.at(i * 2 + j);
            }
            array.append(map);
        }
    }
    result.insert("factorNames", factorNames);
    result.insert("stratified", array);

    return result;
}

fn::StringBaseArray VariantMapToFormulaArray(const QVariantMap& variantArray, const fn::ConstInputFactorPtrList& inputFactorList)
{
    QVariantList variantFormulaArray = variantArray["stratified"].toList();

    fn::StringBaseArray result;
    const QStringList& colNames = inputFactorList[1]->getCategoryList();
    for (int i = 0; i < variantFormulaArray.length(); ++i) {
        QVariantMap row = variantFormulaArray[i].toMap();
        for (int j = 0; j < colNames.length(); ++j) {
            const QString name = colNames[j];
            const QString val = row[name].toString();
            result.append(val);
        }
    }

    return result;
}

fn::StringToStringMap VariantMapToMetaData(const QVariantMap& variantArray)
{
    QVariantList variantMetaData = variantArray["meta"].toList();

    fn::StringToStringMap result;
    for (int i = 0; i < variantMetaData.length(); ++i) {
        QVariantMap row = variantMetaData[i].toMap();
        result.insert(row.value("ParameterName").toString(), row.value("Value").toString());
    }

    return result;
}

QVariantList DataArrayToVariantList(const fn::DataArray& array)
{
    QVariantList result;

    const fn::ConstInputFactorPtrList& inputFactorList = array.getInputFactorList();
    QVariantList variantArray = fn::FormulaArrayToVariantList(array.getInputFormulaArray(), inputFactorList);
    QVariantList variantMeta = fn::MetaDataToVariantList(array.getMetaData());
    QVariant variantEmptyRow = QVariant(QVariantList() << QVariant() << QVariant() << QVariant());

    result.append(QVariant(QVariantList() << array.getName() << QString("") << QString("")));
    result.append(variantEmptyRow);
    if (variantMeta.length() > 0) {
        result.append(variantMeta);
        result.append(variantEmptyRow);
    }
    result.append(variantArray);

    return result;
}

QVariantList MetaDataToVariantList(const fn::StringToStringMap& metaData)
{
    QVariantList result;

    if (metaData.size() > 0) {
        for (fn::StringToStringMap::const_iterator it = metaData.begin(); it != metaData.end(); ++it) {
            const QString& metaInfoName = it.key().toLower();
            const QString& metaInfoValue = it.value();
            QVariantList list = QVariantList() << metaInfoName << metaInfoValue << QString("");
            result.append(QVariant(list));
        }
    }

    return result;
}

QVariantList FormulaArrayToVariantList(const fn::StringBaseArray& formulaArray, const fn::ConstInputFactorPtrList& inputFactorList)
{
    QVariantList result;

    if (inputFactorList.length() == 2) {

        const QStringList& rowNames = inputFactorList.at(0)->getCategoryList();
        const QStringList& colNames = inputFactorList.at(1)->getCategoryList();

        QVariantList firstRow;
        firstRow.append(inputFactorList[1]->getName());
        firstRow.append(QVariant(colNames).toList());
        result.append(QVariant(firstRow));
        QVariantList secondRow = QVariantList() << inputFactorList[0]->getName() << QString("") << QString("");
        result.append(QVariant(secondRow));

        for (int i = 0; i < rowNames.length(); ++i) {
            QVariantList list;
            list.append(rowNames[i]);
            for (int j = 0; j < colNames.length(); ++j) {
                list.append(formulaArray.at(i * 2 + j));
            }
            result.append(QVariant(list));
        }
    }

    return result;
}

QVariantMap DataArrayToVariantMap(const fn::DataArray& dataArray, const fn::DataArray::ArrayType& dataArrayType, const fn::DataArray::FormulaType& formulaType)
{
    QVariantMap result;

    switch (dataArrayType) {
        case fn::DataArray::Formula: {
                switch (formulaType) {
                    case fn::DataArray::Input: {
                            const fn::ConstInputFactorPtrList& inputFactorList = dataArray.getInputFactorList();
                            const fn::StringBaseArray& formulaArray = dataArray.getInputFormulaArray();
                            result = fn::FormulaArrayToVariantMap(formulaArray, inputFactorList);
                        }
                        break;
                    case (fn::DataArray::Run) : {
                            const fn::ConstInputFactorPtrList& runFactorList = dataArray.getRunFactorList();
                            const fn::StringBaseArray& formulaArray = dataArray.getRunFormulaArray();
                            result = fn::FormulaArrayToVariantMap(formulaArray, runFactorList);
                        }
                        break;
                }
            }
            break;
        case fn::DataArray::Numeric: {
                const fn::ConstInputFactorPtrList& inputFactorList = dataArray.getInputFactorList();
                const fn::NumericBaseArray& numericArray = dataArray.getValueArray();
                result = fn::NumericArrayToVariantMap(numericArray, inputFactorList);
            }
            break;
    }

    const fn::StringToStringMap& metaData = dataArray.getMetaData();
    fn::AppendMetaDataToVariantMap(result, metaData);
    result["name"] = dataArray.getName();

    return result;
}

DataArray VariantMapToDataArray(const QVariantMap& variantArray, const fn::ConstInputFactorPtrList& inputFactorList)
{
    fn::InputFactorList inputFactors;
    for (fn::ConstInputFactorPtrList::const_iterator it = inputFactorList.begin(); it != inputFactorList.end(); ++it) {
        inputFactors.append((**it));
    }

    QString name = variantArray["name"].toString();
    fn::StringToStringMap metaData = VariantMapToMetaData(variantArray);
    fn::StringBaseArray formulaArray = VariantMapToFormulaArray(variantArray, inputFactorList);

    fn::DataArray result = fn::DataArray(inputFactors, name, metaData, formulaArray);

    return result;
}

}  // namespace fn

