//==========================================================================================================================
//  Implementation file for "DataArray.h"
//  Written by Daniel Lewandowski
//  Copyright @ 2011 NextPage Software
//==========================================================================================================================

#include "DataArray.h"

#include "fn/formulaParserHelper.h"

namespace fn {

DataArray::DataArray(const fn::InputFactorList& inputFactorList, const QString& name, const fn::StringToStringMap& metaData, const fn::StringBaseArray& inputFormulaArray) :
    m_name(name),
    m_parser(),
    m_metaData(metaData),
    m_inputFactorList(inputFactorList),
    m_inputDims(),
    m_inputFormulaArray(inputFormulaArray),
    m_runFactorList(),
    m_runDims(),
    m_runFormulaArray(inputFormulaArray),
    m_valueArray()
{
    for (int i = 0; i < inputFactorList.length(); ++i) {
        const fn::InputFactor& inputFactor = inputFactorList.at(i);
        m_inputDims.append(inputFactor.getNumCategories());
        m_runFactorList.append(&inputFactor);
    }

    m_runDims = m_inputDims;

    m_valueArray = fn::NumericBaseArray(marray::SkipInitialization, m_runDims.begin(), m_runDims.end(), marray::FirstMajorOrder);
}

DataArray::DataArray(const fn::InputFactorList& inputFactorList, const QString& name, const fn::StringToStringMap& metaData, const QString& initFormula) :
    m_name(name),
    m_parser(),
    m_metaData(metaData),
    m_inputFactorList(inputFactorList),
    m_inputDims(),
    m_inputFormulaArray(),
    m_runFactorList(),
    m_runDims(),
    m_runFormulaArray(),
    m_valueArray()
{
    for (int i = 0; i < inputFactorList.length(); ++i) {
        const fn::InputFactor& inputFactor = inputFactorList.at(i);
        m_inputDims.append(inputFactor.getNumCategories());
        m_runFactorList.append(&inputFactor);
    }

    m_runDims = m_inputDims;

    int numFormulas = std::accumulate(m_inputDims.begin(), m_inputDims.end(), 1, std::multiplies<short int>());
    for (int i = 0; i < numFormulas; ++i) {
        m_inputFormulaArray.append(initFormula);
    }
    m_runFormulaArray = m_inputFormulaArray;

    m_valueArray = fn::NumericBaseArray(marray::SkipInitialization, m_runDims.begin(), m_runDims.end(), marray::FirstMajorOrder);
}

DataArray::DataArray(const fn::ConstInputFactorPtrList& inputFactorList, const QString& initFormula) :
    m_name(),
    m_parser(),
    m_metaData(),
    m_inputFactorList(),
    m_inputDims(),
    m_inputFormulaArray(),
    m_runFactorList(inputFactorList),
    m_runDims(),
    m_runFormulaArray(),
    m_valueArray()
{
    for (int i = 0; i < inputFactorList.length(); ++i) {
        const fn::InputFactor& inputFactor = *inputFactorList.at(i);
        m_inputDims.append(inputFactor.getNumCategories());
        m_inputFactorList.append(inputFactor);
    }

    m_runDims = m_inputDims;

    int numFormulas = std::accumulate(m_inputDims.begin(), m_inputDims.end(), 1, std::multiplies<short int>());
    for (int i = 0; i < numFormulas; ++i) {
        m_inputFormulaArray.append(initFormula);
    }
    m_runFormulaArray = m_inputFormulaArray;

    m_valueArray = fn::NumericBaseArray(marray::SkipInitialization, m_runDims.begin(), m_runDims.end(), marray::FirstMajorOrder);
}

DataArray::DataArray() :
    m_name(""),
    m_parser(),
    m_metaData(),
    m_inputFactorList(),
    m_inputDims(),
    m_inputFormulaArray(),
    m_runFactorList(),
    m_runDims(),
    m_runFormulaArray(),
    m_valueArray()
{

}

DataArray::~DataArray()
{

}

void DataArray::reshape(const fn::Dims& dims)
{
    m_valueArray.reshape(dims.begin(), dims.end());
    m_inputDims = dims;
}

const fn::Dims& DataArray::getDims() const
{
    return m_inputDims;
}

const QString DataArray::asString()
{
    return QString::fromStdString(m_valueArray.asString());
}

const QString DataArray::getMetaInfo(const QString& infoType) const
{
    QString result = "";

    if (m_metaData.contains(infoType)) {
        result = m_metaData.value(infoType);
    }

    return result;
}

void DataArray::setMetaInfo(const QString& infoType, const QString& value)
{
    m_metaData[infoType] = value;
}

const QString& DataArray::getName() const
{
    return m_name;
}

void DataArray::setName(const QString& name)
{
    m_name = name;
}

const fn::StringToStringMap& DataArray::getMetaData() const
{
    return m_metaData;
}

void DataArray::setMetaData(const fn::StringToStringMap& metaData)
{
    m_metaData = metaData;
}

const fn::StringBaseArray& DataArray::getInputFormulaArray() const
{
    return m_inputFormulaArray;
}

void DataArray::setInputFormulaArray(const fn::StringBaseArray& formulaArray)
{
    m_inputFormulaArray = formulaArray;
    m_runFormulaArray = formulaArray;
    adjustDimensions(m_runFactorList);
}

const fn::StringBaseArray& DataArray::getRunFormulaArray() const
{
    return m_runFormulaArray;
}

void DataArray::setRunFormulaArray(const fn::StringBaseArray& formulaArray)
{
    m_runFormulaArray = formulaArray;
}

const fn::NumericBaseArray& DataArray::getValueArray() const
{
    return m_valueArray;
}

void DataArray::setValueArray(const fn::NumericBaseArray& valueArray)
{
    m_valueArray = valueArray;
}

fn::ConstInputFactorPtrList DataArray::getInputFactorList() const
{
    fn::ConstInputFactorPtrList inputFactors;
    QListIterator<fn::InputFactor> inputFactorIter(m_inputFactorList);
    inputFactorIter.toFront();
    while (inputFactorIter.hasNext()) {
        inputFactors.append(&(inputFactorIter.next()));
    }
    return inputFactors;
}

const fn::ConstInputFactorPtrList& DataArray::getRunFactorList() const
{
    return m_runFactorList;
}

const fn::InputFactorList& DataArray::getInputFactors() const
{
    return m_inputFactorList;
}

void DataArray::setInputFactors(const fn::InputFactorList& inputFactorList)
{
    m_inputFactorList = inputFactorList;
}

void DataArray::adjustDimensions(const fn::ConstInputFactorPtrList& runFactorList)
{
    //+ Hard coded condition, but this is a very specific adjustment
    //  and will have to be updated for another specification of dimensions anyway.

    // Available action types
    enum ActionType { NONE, EXPAND_INPUT, RENAME_CONGENITAL_AGE_CAT, LIMIT_INPUT };

    // Initialize run dimensions
    m_runFactorList = runFactorList;
    m_runDims.clear();
    for (int i = 0; i < m_runFactorList.size(); ++i)
        m_runDims.append(m_runFactorList.at(i)->getNumCategories());

    // Determine action to be taken
    ActionType action = ActionType::NONE;
    if ((m_runDims[0] == 20) && (m_inputDims[0] == 19) && (m_runDims[1] == m_inputDims[1]) && (m_runDims[1] == 2)) {

        // CASE 1: input has 19 age categories, run needs 20 age categories
        action = ActionType::EXPAND_INPUT;

    } else if ((m_runDims[0] == 20) && (m_inputDims[0] == 20) && (m_runDims[1] == m_inputDims[1]) && (m_runDims[1] == 2)) {

        // CASE 2: input has 20 age categories, run needs 20 age categories
        action = ActionType::RENAME_CONGENITAL_AGE_CAT;

    } else if ((m_runDims[0] == 19) && (m_inputDims[0] == 20) && (m_runDims[1] == m_inputDims[1]) && (m_runDims[1] == 2)) {

        // CASE 3: input has 20 age categories, run needs 19 age categories
        action = ActionType::LIMIT_INPUT;
    }

    // Take action
    switch (action) {
    case ActionType::EXPAND_INPUT: {
        // Expand input array to run array
        m_inputFormulaArray.prepend(m_inputFormulaArray.at(1));
        m_inputFormulaArray.prepend(m_inputFormulaArray.at(2));
        m_runFormulaArray = m_inputFormulaArray;

        // Expand input dimensions to run dimensions
        m_inputDims = m_runDims;

        // Copy factors
        m_inputFactorList.clear();
        fn::ConstInputFactorPtrList::const_iterator iter = runFactorList.begin();
        for (; iter != runFactorList.end(); ++iter) {
            fn::InputFactor inputFactor = **iter;
            m_inputFactorList.append(inputFactor);
        }

        break;
    }
    case ActionType::RENAME_CONGENITAL_AGE_CAT: {
        // Copy factors
        m_inputFactorList.clear();
        fn::ConstInputFactorPtrList::const_iterator iter = runFactorList.begin();
        for (; iter != runFactorList.end(); ++iter) {
            fn::InputFactor inputFactor = **iter;
            m_inputFactorList.append(inputFactor);
        }

        break;
    }
    case ActionType::LIMIT_INPUT: {
        // Limit input array to run array
        m_inputFormulaArray.pop_front();
        m_inputFormulaArray.pop_front();
        m_runFormulaArray = m_inputFormulaArray;

        // Limit input dimensions to run dimensions
        m_inputDims = m_runDims;

        // Copy factors
        m_inputFactorList.clear();
        fn::ConstInputFactorPtrList::const_iterator iter = runFactorList.begin();
        for (; iter != runFactorList.end(); ++iter) {
            fn::InputFactor inputFactor = **iter;
            m_inputFactorList.append(inputFactor);
        }
        break;
    }
    default:
        break;
    }

    m_valueArray = fn::NumericBaseArray(marray::SkipInitialization, m_runDims.begin(), m_runDims.end(), marray::FirstMajorOrder);
}

void DataArray::initializeParser()
{
    try {
        m_parser.SetArgSep(',');
        m_parser.SetDecSep('.');
        fn::defineAdditionalFunctions(&m_parser);
        const mu::string_type formula = m_runFormulaArray.join(',').toStdWString();
        m_parser.SetExpr(formula);
    } catch (mu::Parser::exception_type&) {
        m_runFormulaArray.replaceInStrings(QRegExp("*"), "0");
        m_parser.SetExpr(m_runFormulaArray.join(',').toStdWString());
        UT_THROW(QString("One or more formulas for parameter %1 cannot be parsed.\nInstead all formulas have been initialized to evaluate to 0.").arg(m_name))
    }
}

void DataArray::deinitializeParser()
{

}

void DataArray::evaluateFormulas()
{
    Q_ASSERT(static_cast<size_t>(m_runFormulaArray.length()) == m_valueArray.size());

    int numFormulas;
    mu::value_type* values = m_parser.Eval(numFormulas);
    for (int i = 0; i < numFormulas; ++i) {
        m_valueArray(i) = values[i];
    }
}

}  // namespace fn
