//==========================================================================================================================
//  'DataArray' DataArray
//			This class implements n-dimensional array
//
//  Written by Daniel Lewandowski
//  Copyright @ 2011 NextPage Software
//==========================================================================================================================

#ifndef _fn_DataArray_
#define _fn_DataArray_

#include "fn/globals.h"
#include "fn/CustomDataTypes.h"
#include "fn/InputFactor.h"

namespace fn {

class DataArray
{
public:

    enum ArrayType {Formula = 0, Numeric};

    enum FormulaType {Input = 0, Run};

    // Constructor.
    //
    // 'inputFactorList' - [in] List of input factors
    // 'name' - [in] Name
    // 'metaData' - [in] Meta information
    // 'formulaArray' - [in] Array of formulas
    //
    DataArray(const fn::InputFactorList& inputFactorList, const QString& name, const fn::StringToStringMap& metaData, const fn::StringBaseArray& inputFormulaArray);
    DataArray(const fn::InputFactorList& inputFactorList, const QString& name, const fn::StringToStringMap& metaData, const QString& initFormula);
    DataArray(const fn::ConstInputFactorPtrList& inputFactorList, const QString& initFormula);

    // Default constructor
    //
    DataArray();

    // Destructor.
    //
    virtual ~DataArray();

    // Reshape the array to new set of dimensions
    //
    // 'dims' - [in] Vector of extents of dimensions
    //
    void reshape(const fn::Dims& dims);

    // Get dimensions
    //
    // Return - dimensions
    const fn::Dims& getDims() const;

    // Return table as string
    //
    // Return - string
    const QString asString();

    // Get meta information
    //
    // 'infoType' - [in] Information type
    //
    // Return - information type as QVariant
    const QString getMetaInfo(const QString& infoType) const;

    // Set meta information
    //
    // 'infoType' - [in] Information type
    // 'value' - [in] Value
    //
    // Return - information type as QVariant
    void setMetaInfo(const QString& infoType, const QString& value);

    // Get name
    //
    // Return - Name as string
    const QString& getName() const;

    // Set name
    //
    // Return - Name as string
    void setName(const QString& name);

    // Get whole meta data object
    //
    // Return - Meta data
    const fn::StringToStringMap& getMetaData() const;

    // Set whole meta data object
    //
    // 'metaData' - [in] Meta data
    //
    void setMetaData(const fn::StringToStringMap& metaData);

    // Get array of formulae
    //
    const fn::StringBaseArray& getInputFormulaArray() const;

    // Set formula array (for aiding serialization)
    //
    // 'formulaArray' - [in] Array of formulae
    //
    void setInputFormulaArray(const fn::StringBaseArray& formulaArray);

    // Get array of formulae
    //
    const fn::StringBaseArray& getRunFormulaArray() const;

    // Set formula array (for aiding serialization)
    //
    // 'formulaArray' - [in] Array of formulae
    //
    void setRunFormulaArray(const fn::StringBaseArray& formulaArray);

    // Get array of values
    //
    const fn::NumericBaseArray& getValueArray() const;

    // Set array of values
    //
    void setValueArray(const fn::NumericBaseArray& valueArray);

    // Get input factor list
    fn::ConstInputFactorPtrList getInputFactorList() const;

    // Get run factor list
    const fn::ConstInputFactorPtrList& getRunFactorList() const;

    // Get input factor list (for serialization)
    const fn::InputFactorList& getInputFactors() const;

    // Set input factor list (for serialization)
    void setInputFactors(const fn::InputFactorList& inputFactorList);

    // Adjust run formula array if necessary
    void adjustDimensions(const fn::ConstInputFactorPtrList& runFactorList);

    // Initialize parser
    //
    void initializeParser();

    // Deinitialize parser
    //
    void deinitializeParser();

    // Evaluate formulas
    //
    void evaluateFormulas();

private:

    // Name
    QString m_name;

    // Formula parser
    fn::FormulaParser m_parser;

    // Meta data (name, description, weblink, etc)
    fn::StringToStringMap m_metaData;

    // Input factor list
    fn::InputFactorList m_inputFactorList;

    // Input data dimensions
    fn::Dims m_inputDims;

    // Input formula array
    fn::StringBaseArray m_inputFormulaArray;

    // Input factor list
    fn::ConstInputFactorPtrList m_runFactorList;

    // Run dimensions
    fn::Dims m_runDims;

    // Run formula array
    fn::StringBaseArray m_runFormulaArray;

    // Value array
    fn::NumericBaseArray m_valueArray;
};

}  // namespace fn

#endif  // _fn_DataArray_
