//==========================================================================================================================
//  Implementation file for "DiseaseModel.h"
//  Written by Daniel Lewandowski
//  Copyright @ 2011 NextPage Software
//==========================================================================================================================

#include "DiseaseModel.h"

#include "ut/VariantUtils.h"
#include "fn/Infection.h"
#include "fn/Overloads.h"
#include "fn/HealthOutcome.h"
#include "fn/HealthState.h"
#include "fn/TransitionParamsMgr.h"
#include "fn/TransitionProbabilityComplement.h"
#include "fn/ArrayHelpers.h"
#include "ui/Chief.h"

namespace fn {

DiseaseModel::DiseaseModel(ui::Chief* chief, const ID& id,
                           const fn::InputType::Country& country, const QString& countryName,
                           const fn::DiseaseTemplate* diseaseTemplate,
                           const fn::ConstInputFactorPtrList& runFactorList) :
    m_id(id),
    m_chief(chief),
    m_diseaseTemplate(diseaseTemplate),
    m_jsonDiseaseTemplate(nullptr),
    m_country(country),
    m_countryName(countryName),
    m_disease(),
    m_runFactorList(runFactorList),
    m_outcomeTreeMap(),
    m_outcomeTreeList(),
    m_rawCasesCount(),
    m_ageDistr(),
    m_lifeExp(),
    m_outcomeTreeMapping(),
    m_totalPopulation(0),
    m_dims(),
    m_underEstimType(0),
    m_coverage(),
    m_underestimation(),
    m_underreporting(),
    m_underascertainment(),
    m_modelSampled(false),
    m_numSamples(0),
    m_sampleAcuteYLD(),
    m_sampleAcuteYLL(),
    m_sampleSequelaYLD(),
    m_sampleSequelaYLL(),
    m_samplePopulationCasesCount(),
    m_sampleDeathCasesCount(),
    m_sampleNumSymptomaticCasesCount(),
    m_categorizedResults(),
    m_sequelaTypeResults(),
    m_aggregatedPageResults()
{
    m_diseaseTemplateType = fn::DiseaseModel::DiseaseTemplateType::Original;

    fn::ConstInputFactorPtrList::ConstIterator iter = m_runFactorList.constBegin();
    for (; iter != m_runFactorList.constEnd(); ++iter) {
        m_dims.append((**iter).getNumCategories());
    }

    m_disease = diseaseTemplate->getDisease();
}

DiseaseModel::DiseaseModel(ui::Chief* chief, const DiseaseModel::ID& id,
                           const fn::InputType::Country& country, const QString& countryName,
                           const fn::JSONDiseaseTemplate* jsonDiseaseTemplate) :
    m_id(id),
    m_chief(chief),
    m_diseaseTemplate(nullptr),
    m_jsonDiseaseTemplate(jsonDiseaseTemplate),
    m_country(country),
    m_countryName(countryName),
    m_disease(),
    m_runFactorList(),
    m_outcomeTreeMap(),
    m_outcomeTreeList(),
    m_rawCasesCount(),
    m_ageDistr(),
    m_lifeExp(),
    m_outcomeTreeMapping(),
    m_totalPopulation(0),
    m_dims(),
    m_underEstimType(0),
    m_coverage(),
    m_underestimation(),
    m_underreporting(),
    m_underascertainment(),
    m_modelSampled(false),
    m_numSamples(0),
    m_sampleAcuteYLD(),
    m_sampleAcuteYLL(),
    m_sampleSequelaYLD(),
    m_sampleSequelaYLL(),
    m_samplePopulationCasesCount(),
    m_sampleDeathCasesCount(),
    m_sampleNumSymptomaticCasesCount(),
    m_categorizedResults(),
    m_sequelaTypeResults(),
    m_aggregatedPageResults()
{
    m_runFactorList = jsonDiseaseTemplate->getConstInputFactorPtrList();

    m_diseaseTemplateType = fn::DiseaseModel::DiseaseTemplateType::JSON;

    fn::ConstInputFactorPtrList::ConstIterator iter = m_runFactorList.constBegin();
    for (; iter != m_runFactorList.constEnd(); ++iter) {
        m_dims.append((**iter).getNumCategories());
    }

    m_disease = jsonDiseaseTemplate->getDisease();
}

DiseaseModel::~DiseaseModel()
{

}

void DiseaseModel::initialize()
{
    if (m_diseaseTemplateType == fn::DiseaseModel::DiseaseTemplateType::Original) {
        // Initialize outcome tree mapping
        m_outcomeTreeMapping = m_chief->getInputDataMgr().getOutcomeTreeMappingData(m_country, m_disease);

        // Initialize correction factors
        m_coverage = m_chief->getInputDataMgr().getCorrectionWithName(m_country, m_disease, "coverage");
        m_underestimation = m_chief->getInputDataMgr().getCorrectionWithName(m_country, m_disease, "underestimation");
        m_underreporting = m_chief->getInputDataMgr().getCorrectionWithName(m_country, m_disease, "underreporting");
        m_underascertainment = m_chief->getInputDataMgr().getCorrectionWithName(m_country, m_disease, "underascertainment");
        setUnderestimationType(m_chief->getInputDataMgr().getCorrectionType(m_country, m_disease));

        // Initialize raw cases count
        m_rawCasesCount = m_chief->getInputDataMgr().getNotificationData(m_country, m_disease);

        // Population data
        m_ageDistr = m_chief->getInputDataMgr().getAgeDistributionData(m_country);
        m_lifeExp = m_chief->getInputDataMgr().getLifeExpectancyData(m_country);

        createInternalStructure(m_diseaseTemplate);

    } else if (m_diseaseTemplateType == fn::DiseaseModel::DiseaseTemplateType::JSON) {
        // Initialize outcome tree mapping
        m_outcomeTreeMapping = m_jsonDiseaseTemplate->getOutcomeTreeMapping();

        fn::StringToStringMap metaData;
        const fn::InputFactorList& inputFactorList = m_jsonDiseaseTemplate->getInputFactorList();

        // Initialize default correction factors
        m_coverage = fn::DataArray(inputFactorList, "Coverage", metaData, "1");
        m_underestimation = fn::DataArray(inputFactorList, "Underestimation", metaData, "1");
        m_underreporting = fn::DataArray(inputFactorList, "Underreporting", metaData, "1");
        m_underascertainment = fn::DataArray(inputFactorList, "Underascertainment", metaData, "1");
        setUnderestimationType(0);

        // Initialize default raw cases count
        m_rawCasesCount = m_jsonDiseaseTemplate->getRawCasesCount();

        // Population data
        m_ageDistr = m_chief->getInputDataMgr().getAgeDistributionData(m_country);
        m_lifeExp = m_chief->getInputDataMgr().getLifeExpectancyData(m_country);

        createInternalStructure(m_jsonDiseaseTemplate);
    }

    // Pass population data to Infection
    propagateChangesInModelParameters();
}

DiseaseModel::ID DiseaseModel::getId() const
{
    return m_id;
}

QString DiseaseModel::getCaption() const
{
    QString result = "";

    if (m_diseaseTemplateType == fn::DiseaseModel::DiseaseTemplateType::Original) {
        result = getCountryName() + " - " + getDiseaseName() + " - " + m_id;
    }

    return result;
}

const InputType::Country& DiseaseModel::getCountry() const
{
    return m_country;
}

const QString& DiseaseModel::getCountryName() const
{
    return m_countryName;
}

const InputType::Disease& DiseaseModel::getDisease() const
{
    return m_disease;
}

QString DiseaseModel::getDiseaseName() const
{
    QString result = QStringLiteral("");
    if (m_diseaseTemplateType == fn::DiseaseModel::DiseaseTemplateType::Original) {
        result = m_diseaseTemplate->getCaption();
    } else {
        result = m_jsonDiseaseTemplate->getCaption();
    }
    return result;
}

bool DiseaseModel::getIsCongenital() const
{
    bool result = false;
    if (m_diseaseTemplateType == fn::DiseaseModel::DiseaseTemplateType::Original) {
        result = m_diseaseTemplate->getIsCongenital();
    } else {
        result = m_jsonDiseaseTemplate->getIsCongenital();
    }
    return result;
}

const fn::Dims& DiseaseModel::getDims() const
{
    return m_dims;
}

const fn::ConstInputFactorPtrList& DiseaseModel::getInputFactorsList() const
{
    return m_runFactorList;
}

const QVariantMap DiseaseModel::getVariantModelInfo() const
{
    QVariantMap result;
    result["countryID"] = m_country;
    result["diseaseID"] = m_disease;
    result["modelID"] = m_id;
    result["countryName"] = getCountryName();;
    result["diseaseName"] = getDiseaseName();
    result["isCongenital"] = getIsCongenital();

    QJsonObject jsonDiseaseSpec;
    if (m_diseaseTemplateType == fn::DiseaseModel::JSON && m_jsonDiseaseTemplate != nullptr) {
        jsonDiseaseSpec = m_jsonDiseaseTemplate->getJsonObject();
    }
    result["jsonDiseaseSpec"] = jsonDiseaseSpec;

    return result;
}

DiseaseModel::DiseaseTemplateType DiseaseModel::getDiseaseTemplateType() const
{
    return m_diseaseTemplateType;
}

const fn::JSONDiseaseTemplate* DiseaseModel::getJSONDiseaseTemplate() const
{
    return m_jsonDiseaseTemplate;
}

void DiseaseModel::runModel(const int numSamples, const double timeDiscountRate, int& currentOutcomeTree)
{
    m_modelSampled = false;

    m_numSamples = numSamples;
    initializeSample();
    initializeAllFormulaParsers();
    clearResults();

    // Evaluate Population Data
    m_ageDistr.evaluateFormulas();
    m_lifeExp.evaluateFormulas();

    // Population is assumed NOT random, therefore it is calculated only once
    // and used through out the simulation.
    const fn::NumericBaseArray totalPopulation = m_ageDistr.getValueArray();
    m_totalPopulation = fn::getSum2D(totalPopulation);

    /* Removed per request from Alessandro Cassini
    // Deduct duplicated population
    if (getIsCongenital())
    {
        m_totalPopulation -= totalPopulation(0);
        m_totalPopulation -= totalPopulation(1);
    }
    */

    for (int i = 0; i < m_outcomeTreeList.size(); ++i) {
        currentOutcomeTree = i;
        m_outcomeTreeList[i].m_totalPopulation =
                fn::getSum2D(fn::copyFilteredNumericBaseArray(totalPopulation, m_outcomeTreeMapping.getValueArray(), i + 1));
        m_outcomeTreeList[i].runModel(numSamples, timeDiscountRate);
    }

    deinitializeAllFormulaParsers();

    computeResults();

    m_modelSampled = true;
}

bool DiseaseModel::isModelSampled() const
{
    return m_modelSampled;
}

int DiseaseModel::getUnderestimationType() const
{
    return m_underEstimType;
}

void DiseaseModel::setUnderestimationType(const int underEstimType)
{
    Q_ASSERT((underEstimType >= 0) & (underEstimType <= 1));
    m_underEstimType = underEstimType;
}

const fn::DataArray& DiseaseModel::getParameter(const fn::DiseaseModel::ParamType& paramType) const
{
    switch (paramType) {
    case OutcomeTreeMapping:
        return m_outcomeTreeMapping;
    case RawCasesCount:
        return m_rawCasesCount;
    case AgeDistribution:
        return m_ageDistr;
    case LifeExpectancy:
        return m_lifeExp;
    case Coverage:
        return m_coverage;
    case Underestimation:
        return m_underestimation;
    case Underreporting:
        return m_underreporting;
    case Underascertainment:
        return m_underascertainment;
    }

    UT_THROW(QString("Unknown parameter type requested from disease model %1").arg(m_id));
}

fn::DataArray& DiseaseModel::getParameter(const fn::DiseaseModel::ParamType& paramType)
{
    switch (paramType) {
    case OutcomeTreeMapping:
        return m_outcomeTreeMapping;
    case RawCasesCount:
        return m_rawCasesCount;
    case AgeDistribution:
        return m_ageDistr;
    case LifeExpectancy:
        return m_lifeExp;
    case Coverage:
        return m_coverage;
    case Underestimation:
        return m_underestimation;
    case Underreporting:
        return m_underreporting;
    case Underascertainment:
        return m_underascertainment;
    }

    UT_THROW(QString("Unknown parameter type requested from disease model %1").arg(m_id));
}

void DiseaseModel::setParameter(const fn::DiseaseModel::ParamType& paramType, const fn::DataArray& parameter)
{
    switch (paramType) {
    case OutcomeTreeMapping:
        m_outcomeTreeMapping = parameter;
        break;
    case RawCasesCount:
        m_rawCasesCount = parameter;
        break;
    case AgeDistribution:
        m_ageDistr = parameter;
        break;
    case LifeExpectancy:
        m_lifeExp = parameter;
        break;
    case Coverage:
        m_coverage = parameter;
        break;
    case Underestimation:
        m_underestimation = parameter;
        break;
    case Underreporting:
        m_underreporting = parameter;
        break;
    case Underascertainment:
        m_underascertainment = parameter;
        break;
    }
}

void DiseaseModel::propagateChangesInModelParameters()
{
    adjustDimensionsInAllObjects();

    m_outcomeTreeMapping.initializeParser();
    m_outcomeTreeMapping.evaluateFormulas();

    for (int i = 0; i < m_outcomeTreeList.size(); ++i) {
        m_outcomeTreeList[i].m_infection->setRawCasesCount(fn::copyFilteredDataArray(m_rawCasesCount, m_outcomeTreeMapping, i + 1));
        m_outcomeTreeList[i].m_infection->setLifeExpectancy(fn::copyFilteredDataArray(m_lifeExp, m_outcomeTreeMapping, i + 1));
    }

    m_modelSampled = false;
}

int DiseaseModel::getOutcomeTreeCount()
{
    return m_outcomeTreeList.size();
}

const fn::OutcomeTree& DiseaseModel::getOutcomeTree(const int index) const
{
    if (index < 0 || index > m_outcomeTreeList.count()) {
        UT_THROW(QString("Disease model %1 has no outcome tree with index %1").arg(m_id).arg(index));
    };
    return m_outcomeTreeList[index];
}

fn::OutcomeTree& DiseaseModel::getOutcomeTree(const int index)
{
    if (index < 0 || index > m_outcomeTreeList.count()) {
        UT_THROW(QString("Disease model %1 has no outcome tree with index %1").arg(m_id).arg(index));
    };
    return m_outcomeTreeList[index];
}

const OutcomeTree& DiseaseModel::getOutcomeTree(const OutcomeTree::ID outcomeTreeId) const
{
    if (m_outcomeTreeMap.contains(outcomeTreeId)) {
        int index = m_outcomeTreeMap.value(outcomeTreeId);
        return m_outcomeTreeList[index];
    } else {
        UT_THROW(QString("Disease model %1 has no outcome tree with id %2").arg(m_id).arg(outcomeTreeId));
    }
}

fn::OutcomeTree& DiseaseModel::getOutcomeTree(const fn::OutcomeTree::ID outcomeTreeId)
{
    if (m_outcomeTreeMap.contains(outcomeTreeId)) {
        int index = m_outcomeTreeMap.value(outcomeTreeId);
        return m_outcomeTreeList[index];
    } else {
        UT_THROW(QString("Disease model %1 has no outcome tree with id %2").arg(m_id).arg(outcomeTreeId));
    }
}

const QVariantMap& DiseaseModel::getAggregatedPageResults() const
{
    return m_aggregatedPageResults;
}

const QVariantMap DiseaseModel::getDetailedPageResults() const
{
    QVariantMap result;
    QVariantList detailedResults;
    for (int i = 0; i < m_outcomeTreeList.size(); ++i) {
        const fn::OutcomeTree& outcomeTree = m_outcomeTreeList[i];
        QVariantMap outcomeTreeResults;
        outcomeTreeResults.insert("outcomeTreeID", outcomeTree.getId());
        outcomeTreeResults.insert("outcomeTreeName", outcomeTree.getName());
        outcomeTreeResults.insert("sequelaTypeAggregatedResults", outcomeTree.getSequelaTypeAggregatedResults());
        outcomeTreeResults.insert("healthOutcomeAggregatedResults", outcomeTree.getHealthOutcomeAggregatedResults());
        detailedResults.append(outcomeTreeResults);
    }
    result.insert("detailedResults", detailedResults);
    result.insert("totalCategorizedResults", m_categorizedResults);
    result.insert("totalSequelaTypeResults", m_sequelaTypeResults);

    return result;
}

fn::NumericBaseArray DiseaseModel::getSampleYLD()
{
    return m_sampleAcuteYLD + m_sampleSequelaYLD;
}

fn::NumericBaseArray DiseaseModel::getSampleYLL()
{
    return m_sampleAcuteYLL + m_sampleSequelaYLL;
}

fn::NumericBaseArray DiseaseModel::getSampleAgeDistr()
{
    return fn::replicate(m_ageDistr.getValueArray(), m_numSamples);
}

fn::NumericBaseArray DiseaseModel::getSampleNumSymptomaticCasesCount()
{
    return m_sampleNumSymptomaticCasesCount;
}

fn::NumericBaseArray DiseaseModel::getSampleDeathCasesCount()
{
    return m_sampleDeathCasesCount;
}

const fn::SampleVal& DiseaseModel::getTotalPopulation() const
{
    return m_totalPopulation;
}

void DiseaseModel::createInternalStructure(const fn::DiseaseTemplate* diseaseTemplate)
{
    for (int i = 0; i < diseaseTemplate->getNumOutcomeTreeTemplates(); ++i) {
        const fn::OutcomeTree::ID outcomeTreeID = diseaseTemplate->getOutcomeTreeTemplateID(i);
        const fn::OutcomeTreeTemplate& outcomeTreeTemplate = diseaseTemplate->getOutcomeTreeTemplate(outcomeTreeID);
        m_outcomeTreeList.append(fn::OutcomeTree(outcomeTreeID, i + 1, *this, outcomeTreeTemplate));
        m_outcomeTreeMap.insert(outcomeTreeID, i);
    }
}

void DiseaseModel::createInternalStructure(const fn::JSONDiseaseTemplate* diseaseTemplate)
{
    for (int i = 0; i < diseaseTemplate->getNumOutcomeTreeTemplates(); ++i) {
        const fn::OutcomeTree::ID outcomeTreeID = diseaseTemplate->getOutcomeTreeTemplateID(i);
        const fn::OutcomeTreeTemplate& outcomeTreeTemplate = diseaseTemplate->getOutcomeTreeTemplate(outcomeTreeID);
        m_outcomeTreeList.append(fn::OutcomeTree(outcomeTreeID, i + 1, *this, outcomeTreeTemplate));
        m_outcomeTreeMap.insert(outcomeTreeID, i);
    }
}

void DiseaseModel::adjustDimensionsInAllObjects()
{
    m_ageDistr.adjustDimensions(m_runFactorList);

    m_outcomeTreeMapping.adjustDimensions(m_runFactorList);
    m_coverage.adjustDimensions(m_runFactorList);
    m_underestimation.adjustDimensions(m_runFactorList);
    m_underreporting.adjustDimensions(m_runFactorList);
    m_underascertainment.adjustDimensions(m_runFactorList);
    m_rawCasesCount.adjustDimensions(m_runFactorList);
    m_lifeExp.adjustDimensions(m_runFactorList);

    for (int i = 0; i < m_outcomeTreeList.size(); ++i) {
        m_outcomeTreeList[i].adjustDimensionsInAllObjects(m_runFactorList);
    }
}

fn::NumericBaseArray DiseaseModel::getUnderestimationSample()
{
    fn::NumericBaseArray result;

    if (m_underEstimType == 0) {
        result = m_underestimation.getValueArray();
    } else {
        result = m_underreporting.getValueArray() * m_underascertainment.getValueArray();
    }

    return result;
}

void DiseaseModel::initializeAllFormulaParsers()
{
    m_rawCasesCount.initializeParser();
    m_ageDistr.initializeParser();
    m_lifeExp.initializeParser();
    m_coverage.initializeParser();
    m_underestimation.initializeParser();
    m_underreporting.initializeParser();
    m_underascertainment.initializeParser();
}

void DiseaseModel::deinitializeAllFormulaParsers()
{
    m_rawCasesCount.deinitializeParser();
    m_ageDistr.deinitializeParser();
    m_lifeExp.deinitializeParser();
    m_coverage.deinitializeParser();
    m_underestimation.deinitializeParser();
    m_underreporting.deinitializeParser();
    m_underascertainment.deinitializeParser();
}

void DiseaseModel::evaluateAllFormulas()
{
    m_rawCasesCount.evaluateFormulas();
    m_ageDistr.evaluateFormulas();
    m_lifeExp.evaluateFormulas();
    m_coverage.evaluateFormulas();
    m_underestimation.evaluateFormulas();
    m_underreporting.evaluateFormulas();
    m_underascertainment.evaluateFormulas();
}

void DiseaseModel::initializeSample()
{
    const fn::Dims sampleDims = fn::Dims() << m_dims << static_cast<size_t>(m_numSamples);

    m_samplePopulationCasesCount = fn::NumericBaseArray(sampleDims.begin(), sampleDims.end(), 0);
    m_sampleDeathCasesCount = fn::NumericBaseArray(sampleDims.begin(), sampleDims.end(), 0);
    m_sampleNumSymptomaticCasesCount = fn::NumericBaseArray(sampleDims.begin(), sampleDims.end(), 0);
    m_sampleAcuteYLD = fn::NumericBaseArray(sampleDims.begin(), sampleDims.end(), 0);
    m_sampleAcuteYLL = fn::NumericBaseArray(sampleDims.begin(), sampleDims.end(), 0);
    m_sampleSequelaYLD = fn::NumericBaseArray(sampleDims.begin(), sampleDims.end(), 0);
    m_sampleSequelaYLL = fn::NumericBaseArray(sampleDims.begin(), sampleDims.end(), 0);
}

void DiseaseModel::clearResults()
{
    m_categorizedResults.clear();
    m_sequelaTypeResults.clear();
    m_aggregatedPageResults.clear();
}

void DiseaseModel::clearSample()
{
    m_samplePopulationCasesCount = fn::NumericBaseArray();
    m_sampleDeathCasesCount = fn::NumericBaseArray();
    m_sampleNumSymptomaticCasesCount = fn::NumericBaseArray();
    m_sampleAcuteYLD = fn::NumericBaseArray();
    m_sampleAcuteYLL = fn::NumericBaseArray();
    m_sampleSequelaYLD = fn::NumericBaseArray();
    m_sampleSequelaYLL = fn::NumericBaseArray();
}

void DiseaseModel::computeResults()
{
    const fn::SampleVal numPer = fn::SampleVal(100000);
    const fn::SampleVal unit = numPer / m_totalPopulation;
    const fn::NumericBaseArray sampleAgeDistr = fn::replicate(m_ageDistr.getValueArray(), m_numSamples);
    const fn::NumericBaseArray sampleUnit = numPer / sampleAgeDistr;
    const fn::NumericBaseArray aggrNumSymptomaticCasesCount = fn::getAggregatedSample(m_sampleNumSymptomaticCasesCount);
    const fn::NumericBaseArray aggrDeathCasesCount = fn::getAggregatedSample(m_sampleDeathCasesCount);

    // A. Detailed page results.
    // A1a. Compute DALY, DALY per case, DALY per 100,000 per sex/age category.
    const fn::NumericBaseArray sampleYLD = m_sampleAcuteYLD + m_sampleSequelaYLD;
    const fn::NumericBaseArray sampleYLDPerCase = sampleYLD / m_sampleNumSymptomaticCasesCount;
    const fn::NumericBaseArray sampleYLDPer100k = sampleYLD * sampleUnit;
    const fn::NumericBaseArray sampleYLDPer100kTotalPopulation = sampleYLD * unit;
    const fn::NumericBaseArray sampleYLL = m_sampleAcuteYLL + m_sampleSequelaYLL;
    const fn::NumericBaseArray sampleYLLPerCase = sampleYLL / m_sampleNumSymptomaticCasesCount;
    const fn::NumericBaseArray sampleYLLPer100k = sampleYLL * sampleUnit;
    const fn::NumericBaseArray sampleYLLPer100kTotalPopulation = sampleYLL * unit;
    const fn::NumericBaseArray sampleDALY = sampleYLD + sampleYLL;
    const fn::NumericBaseArray sampleDALYPerCase = sampleDALY / m_sampleNumSymptomaticCasesCount;
    const fn::NumericBaseArray sampleDALYPer100k = sampleDALY * sampleUnit;
    const fn::NumericBaseArray sampleDALYPer100kTotalPopulation = sampleDALY * unit;

    // A1b. Categorized results. Store all statistics (YLD, YLL, DALY).
    // DALY are displayed in the charts. YLD and YLL are displayed in the tooltips.
    m_categorizedResults.insert("yld", fn::getStatistics(sampleYLD, m_runFactorList));
    m_categorizedResults.insert("yldPerCase", fn::getStatistics(sampleYLDPerCase, m_runFactorList));
    m_categorizedResults.insert("yldPer100k", fn::getStatistics(sampleYLDPer100k, m_runFactorList));
    m_categorizedResults.insert("yldPer100kTotalPopulation", fn::getStatistics(sampleYLDPer100kTotalPopulation, m_runFactorList));
    m_categorizedResults.insert("yll", fn::getStatistics(sampleYLL, m_runFactorList));
    m_categorizedResults.insert("yllPerCase", fn::getStatistics(sampleYLLPerCase, m_runFactorList));
    m_categorizedResults.insert("yllPer100k", fn::getStatistics(sampleYLLPer100k, m_runFactorList));
    m_categorizedResults.insert("yllPer100kTotalPopulation", fn::getStatistics(sampleYLLPer100kTotalPopulation, m_runFactorList));
    m_categorizedResults.insert("daly", fn::getStatistics(sampleDALY, m_runFactorList));
    m_categorizedResults.insert("dalyPerCase", fn::getStatistics(sampleDALYPerCase, m_runFactorList));
    m_categorizedResults.insert("dalyPer100k", fn::getStatistics(sampleDALYPer100k, m_runFactorList));
    m_categorizedResults.insert("dalyPer100kTotalPopulation", fn::getStatistics(sampleDALYPer100kTotalPopulation, m_runFactorList));

    // A2. Acute and sequela aggregated results from all outcome trees
    const fn::NumericBaseArray sampleAcuteDALY = m_sampleAcuteYLD + m_sampleAcuteYLL;
    const fn::NumericBaseArray sampleSequelaDALY = m_sampleSequelaYLD + m_sampleSequelaYLL;

    // Aggregate
    const fn::NumericBaseArray aggrAcuteYLD = fn::getAggregatedSample(m_sampleAcuteYLD);
    const fn::NumericBaseArray aggrAcuteYLL = fn::getAggregatedSample(m_sampleAcuteYLL);
    const fn::NumericBaseArray aggrAcuteDALY = fn::getAggregatedSample(sampleAcuteDALY);

    const fn::NumericBaseArray aggrSequelaYLD = fn::getAggregatedSample(m_sampleSequelaYLD);
    const fn::NumericBaseArray aggrSequelaYLL = fn::getAggregatedSample(m_sampleSequelaYLL);
    const fn::NumericBaseArray aggrSequelaDALY = fn::getAggregatedSample(sampleSequelaDALY);

    const fn::NumericBaseArray aggrYLD = fn::getAggregatedSample(sampleYLD);
    const fn::NumericBaseArray aggrYLL = fn::getAggregatedSample(sampleYLL);
    const fn::NumericBaseArray aggrDALY = fn::getAggregatedSample(sampleDALY);

    // Store
    QVariantMap acuteResults;
    acuteResults.insert("yld", fn::getAggregatedStatistics(aggrAcuteYLD));
    acuteResults.insert("yldPer100k", fn::getAggregatedStatistics(aggrAcuteYLD * unit));
    acuteResults.insert("yll", fn::getAggregatedStatistics(aggrAcuteYLL));
    acuteResults.insert("yllPer100k", fn::getAggregatedStatistics(aggrAcuteYLL * unit));
    acuteResults.insert("daly", fn::getAggregatedStatistics(aggrAcuteDALY));
    acuteResults.insert("dalyPerCase", fn::getAggregatedStatistics(aggrAcuteDALY / aggrNumSymptomaticCasesCount));
    acuteResults.insert("dalyPer100k", fn::getAggregatedStatistics(aggrAcuteDALY * unit));

    QVariantMap sequelaResults;
    sequelaResults.insert("yld", fn::getAggregatedStatistics(aggrSequelaYLD));
    sequelaResults.insert("yldPer100k", fn::getAggregatedStatistics(aggrSequelaYLD * unit));
    sequelaResults.insert("yll", fn::getAggregatedStatistics(aggrSequelaYLL));
    sequelaResults.insert("yllPer100k", fn::getAggregatedStatistics(aggrSequelaYLL * unit));
    sequelaResults.insert("daly", fn::getAggregatedStatistics(aggrSequelaDALY));
    sequelaResults.insert("dalyPerCase", fn::getAggregatedStatistics(aggrSequelaDALY / aggrNumSymptomaticCasesCount));
    sequelaResults.insert("dalyPer100k", fn::getAggregatedStatistics(aggrSequelaDALY * unit));

    QVariantMap totalResults;
    totalResults.insert("yld", fn::getAggregatedStatistics(aggrYLD));
    totalResults.insert("yldPer100k", fn::getAggregatedStatistics(aggrYLD * unit));
    totalResults.insert("yll", fn::getAggregatedStatistics(aggrYLL));
    totalResults.insert("yllPer100k", fn::getAggregatedStatistics(aggrYLL * unit));
    totalResults.insert("daly", fn::getAggregatedStatistics(aggrDALY));
    totalResults.insert("dalyPerCase", fn::getAggregatedStatistics(aggrDALY / aggrNumSymptomaticCasesCount));
    totalResults.insert("dalyPer100k", fn::getAggregatedStatistics(aggrDALY * unit));

    m_sequelaTypeResults.insert("acute", acuteResults);
    m_sequelaTypeResults.insert("sequela", sequelaResults);
    m_sequelaTypeResults.insert("total", totalResults);

    // B. Aggregated page results
    m_aggregatedPageResults.insert("yld", totalResults["yld"]);
    m_aggregatedPageResults.insert("yldPer100k", totalResults["yldPer100k"]);
    m_aggregatedPageResults.insert("yll", totalResults["yll"]);
    m_aggregatedPageResults.insert("yllPer100k", totalResults["yllPer100k"]);
    m_aggregatedPageResults.insert("daly", totalResults["daly"]);
    m_aggregatedPageResults.insert("dalyPerCase", totalResults["dalyPerCase"]);
    m_aggregatedPageResults.insert("dalyPer100k", totalResults["dalyPer100k"]);

    m_aggregatedPageResults.insert("cases", fn::getAggregatedStatistics(aggrNumSymptomaticCasesCount));
    m_aggregatedPageResults.insert("incidencePer100k", fn::getAggregatedStatistics(aggrNumSymptomaticCasesCount * unit));
    m_aggregatedPageResults.insert("deaths", fn::getAggregatedStatistics(aggrDeathCasesCount));
    m_aggregatedPageResults.insert("mortalityPer100k", fn::getAggregatedStatistics(aggrDeathCasesCount  * unit));
}

} // namespace fn
