//==========================================================================================================================
//  'DiseaseModel' class
//      Represents a single disease model, which can consist of multiple sub-model blocks and transitions.
//
//  Written by Daniel Lewandowski
//  Copyright @ 2010 NextPage Software
//==========================================================================================================================

#ifndef _fn_DiseaseModel_
#define _fn_DiseaseModel_

#include "fn/OutcomeTree.h"
#include "fn/JSONDiseaseTemplate.h"

namespace ui {
    class Chief;
}

namespace fn {

class DiseaseTemplate;

class DiseaseModel
{

public:

    enum ParamType {OutcomeTreeMapping = 0, RawCasesCount, AgeDistribution, LifeExpectancy, Coverage, Underestimation,
                   Underreporting, Underascertainment};

    enum DiseaseTemplateType {Original = 0, JSON = 1};

    // Disease model ID data type.
    typedef QString ID;

    // Constructor.
    //
    DiseaseModel(ui::Chief* chief, const ID& id,
                 const fn::InputType::Country& country, const QString& countryName,
                 const fn::DiseaseTemplate* diseaseTemplate,
                 const fn::ConstInputFactorPtrList& runFactorList);
    DiseaseModel(ui::Chief* chief, const ID& id,
                 const fn::InputType::Country& country, const QString& countryName,
                 const fn::JSONDiseaseTemplate* jsonDiseaseTemplate);

    // Destructor.
    //
    virtual ~DiseaseModel();

    // Initialize the model. Setup all internals
    //
    void initialize();

    // ATTRIBUTES ----------------------------------------------------------------------------------------------------------

    // Get model ID.
    ID getId() const;

    // Get caption.
    QString getCaption() const;

    // Get country
    const fn::InputType::Country& getCountry() const;
    const QString& getCountryName() const;

    // Get disease
    const fn::InputType::Disease& getDisease() const;
    QString getDiseaseName() const;

    // Get indicator that the disease model is congenital
    bool getIsCongenital() const;

    // Get dimensions
    //
    // Return - dimensions
    const fn::Dims& getDims() const;

    // Get list of input factors
    //
    // Return - input factors pointers list
    const fn::ConstInputFactorPtrList& getInputFactorsList() const;

    // Get model info
    //
    // Return - Model info as Variant
    const QVariantMap getVariantModelInfo() const;

    // Get disease template type
    //
    // Return - disease template type
    DiseaseTemplateType getDiseaseTemplateType() const;

    // Get pointer to JSON disease template
    //
    // Return - pointer to JSON disease template
    const fn::JSONDiseaseTemplate* getJSONDiseaseTemplate() const;

    // RUN -----------------------------------------------------------------------------------------------------------------

    // Perform model run
    //
    // 'numSamples' - [in] Number of samples
    // 'timeDiscountRate' - [in] Time discount rate
    // 'currentOutcomeTree' - [out] Currently processed outcome tree
    //
    void runModel(const int numSamples, const double timeDiscountRate, int& currentOutcomeTree);

    // Get indicator whether the model has been sampled
    //
    // Return - TRUE if model has been sampled
    bool isModelSampled() const;

    // PARAMETERS ----------------------------------------------------------------------------------------------------------

    // Get underestimation type
    //
    // Return - underestimation type
    int getUnderestimationType() const;

    // Set underestimation type
    //
    // 'underEstimType' - [in] Underestimation type (0 - underestimation, 1 - underreporting + underascertainment)
    void setUnderestimationType(const int underEstimType);

    // Get data array type parameter
    //
    // Return - data array for parameter
    const fn::DataArray& getParameter(const fn::DiseaseModel::ParamType& paramType) const;
          fn::DataArray& getParameter(const fn::DiseaseModel::ParamType& paramType);

    // Set parameter
    //
    // 'parameter' - [in] New parameter
    //
    void setParameter(const fn::DiseaseModel::ParamType& paramType, const fn::DataArray& parameter);

    // Propagate changes to infections. This propagates the initial raw cases count and age distribution to underlying
    // outcome trees.
    //
    void propagateChangesInModelParameters();

    // OUTCOME TREE --------------------------------------------------------------------------------------------------------

    // Get the number of outcome trees
    //
    // Return - count of outcome trees
    int getOutcomeTreeCount();

    // Get outcome tree
    //
    // 'index' - [in] Index of outcome tree
    //
    // Return - Reference to outcome tree. Exception is thrown if index is out of bounds
    const fn::OutcomeTree& getOutcomeTree(const int index) const;
          fn::OutcomeTree& getOutcomeTree(const int index);

    // Get outcome tree
    //
    // 'outcomeTreeID' - [in] ID of outcome tree
    //
    // Return - Reference to outcome tree. Exception is thrown if index is out of bounds
    const fn::OutcomeTree& getOutcomeTree(const fn::OutcomeTree::ID outcomeTreeId) const;
          fn::OutcomeTree& getOutcomeTree(const fn::OutcomeTree::ID outcomeTreeId);

    // RESULTS -------------------------------------------------------------------------------------------------------------

    // Get aggregated page results
    //
    // Return - aggregated page results as Variant
    const QVariantMap& getAggregatedPageResults() const;

    // Get detailed page results
    //
    // Return - detailed page results as Variant
    const QVariantMap getDetailedPageResults() const;

    // Get YLD sample
    //
    fn::NumericBaseArray getSampleYLD();

    // Get YLL sample
    //
    fn::NumericBaseArray getSampleYLL();

    // Get age distribution sample
    //
    fn::NumericBaseArray getSampleAgeDistr();

    // Get number of symptomatic cases count sample
    //
    fn::NumericBaseArray getSampleNumSymptomaticCasesCount();

    // Get number of death cases count sample
    //
    fn::NumericBaseArray getSampleDeathCasesCount();

    // Get total population
    //
    const fn::SampleVal& getTotalPopulation() const;

    // Clear the sample tables
    //
    void clearSample();

private:

    friend class OutcomeTree;

    typedef QList<fn::OutcomeTree> OutcomeTreeList;
    typedef QMap<fn::OutcomeTree::ID, int> OutcomeTreeIDToIndexMap;

    // INITIALIZATION ------------------------------------------------------------------------------------------------------

    // Creates all outcome trees objects
    //
    void createInternalStructure(const fn::DiseaseTemplate* diseaseTemplate);
    void createInternalStructure(const fn::JSONDiseaseTemplate* diseaseTemplate);

    // Adjust run formula arrays
    void adjustDimensionsInAllObjects();

    // SAMPLING ------------------------------------------------------------------------------------------------------------

    // Get sample of the underestimation (one iteration)
    //
    // Return - Numeric array containing sample of the underestimation parameter
    fn::NumericBaseArray getUnderestimationSample();

    // Initialize all formula parsers.
    //
    void initializeAllFormulaParsers();

    // Deinitialize all formula parsers
    //
    void deinitializeAllFormulaParsers();

    // Evaluate all formula parsers. This basically creates a new sample of the inputs
    //
    void evaluateAllFormulas();

    // RESULTS -------------------------------------------------------------------------------------------------------------

    // Initialize the results tables - mainly dimensions
    //
    void initializeSample();

    // Compute total categorized results (disease model level)
    //
    void computeResults();

    // Clear the results tables
    //
    void clearResults();

    // PRIVATE MEMBERS -----------------------------------------------------------------------------------------------------

    // Internal ID
    ID m_id;

    // Disease template type
    DiseaseTemplateType m_diseaseTemplateType;

    // Chief
    ui::Chief* m_chief;

    // Disease template
    const fn::DiseaseTemplate* m_diseaseTemplate;
    const fn::JSONDiseaseTemplate* m_jsonDiseaseTemplate;

    // Country
    fn::InputType::Country m_country;
    QString m_countryName;

    // Disease
    fn::InputType::Disease m_disease;

    // (Master) input factor list (used in calculations)
    fn::ConstInputFactorPtrList m_runFactorList;

    // Map of outcome trees IDs to age/sex categories mapping indices
    OutcomeTreeIDToIndexMap m_outcomeTreeMap;

    // Outcome tree list
    OutcomeTreeList m_outcomeTreeList;

    // Raw cases count
    fn::DataArray m_rawCasesCount;

    // Age distribution
    fn::DataArray m_ageDistr;

    // Life expectation
    fn::DataArray m_lifeExp;

    // Outcome tree mapping
    fn::DataArray m_outcomeTreeMapping;

    // Total population
    fn::SampleVal m_totalPopulation;

    // Vector of dimensions
    fn::Dims m_dims;

    // Estimation type
    int m_underEstimType; // 0 - underestimation, 1 - underreporting and underascertainment
    // Population coverage
    fn::DataArray m_coverage;
    // Underestimation
    fn::DataArray m_underestimation;
    // Underreporting
    fn::DataArray m_underreporting;
    // Underascertainment
    fn::DataArray m_underascertainment;

    // Indicate that the model has been sampled
    bool m_modelSampled;

    // Number of samples per variable
    int m_numSamples;

    fn::NumericBaseArray m_sampleAcuteYLD;
    fn::NumericBaseArray m_sampleAcuteYLL;
    fn::NumericBaseArray m_sampleSequelaYLD;
    fn::NumericBaseArray m_sampleSequelaYLL;
    fn::NumericBaseArray m_samplePopulationCasesCount;
    fn::NumericBaseArray m_sampleDeathCasesCount;
    fn::NumericBaseArray m_sampleNumSymptomaticCasesCount;

    // DETAILED PAGE RESULTS

    // CATEGORIZED RESULTS (PER SEX AND AGE)
    QVariantMap m_categorizedResults;

    // SEQUELA TYPE RESULTS
    QVariantMap m_sequelaTypeResults;

    // AGGREGATED PAGE RESULTS
    QVariantMap m_aggregatedPageResults;

};

} // namespace fn

#endif // _fn_DiseaseModel_
