//==========================================================================================================================
//  Implementation file for "HealthOutcome.h"
//  Written by Daniel Lewandowski
//  Copyright @ 2011 NextPage Software
//==========================================================================================================================

#include "HealthOutcome.h"

#include "Overloads.h"

namespace fn {

HealthOutcome::HealthOutcome(fn::OutcomeTreeNode::ID id, const QString& caption, const fn::Dims& dims, const bool acuteType, const bool visible) :
    fn::OutcomeTreeNode(id, caption, fn::OutcomeTreeNode::HealthOutcome, dims),
    m_acuteType(acuteType),
    m_visible(visible),
    m_healthStateList(),
    m_remainingCasesHealthState(nullptr)
{

}

HealthOutcome::~HealthOutcome()
{

}

void HealthOutcome::addInternalHealthState(fn::HealthState* healthState)
{
    m_healthStateList.append(healthState);
}

void HealthOutcome::invalidateRemainingCasesHealthState()
{
    m_remainingCasesHealthState = nullptr;
    for (fn::HealthStatePtrList::const_iterator it = m_healthStateList.begin(); it != m_healthStateList.end(); ++it) {
        fn::HealthState* healthState = *it;
        if (healthState->isRemainingCases()) {
            m_remainingCasesHealthState = healthState;
            break;
        }
    }
}

void HealthOutcome::runCalculation(const double timeDiscountRate)
{
    // Initialize proportions in case of multiple health states
    // This is done, only if there is exactly one parent transition (and at least one health state obviously)
    if ((m_healthStateList.length() > 0) && (m_parentTransProbs.size() > 0)) {
        fn::NumericBaseArray totalProportion = m_healthStateList.at(0)->m_proportion.getValueArray();
        for (int i = 1; i < m_healthStateList.length(); ++i) {
            totalProportion += m_healthStateList.at(i)->m_proportion.getValueArray();
        }

        // Set proportion of the health state covering remaining cases
        if (m_remainingCasesHealthState != nullptr) {
            m_remainingCasesHealthState->m_proportion.setValueArray(fn::max(1.0 - totalProportion, 0));
            totalProportion += m_remainingCasesHealthState->m_proportion.getValueArray();
        }

        // Set normalized proportion in all health states
        for (int i = 0; i < m_healthStateList.length(); ++i) {
            fn::HealthState* healthState = m_healthStateList.at(i);
            const fn::NumericBaseArray& unNormProp = healthState->m_proportion.getValueArray();
            healthState->setNormalizedProportion(unNormProp / totalProportion);
        }
    }

    // Run simulation in all health states
    for (int i = 0; i < m_healthStateList.length(); ++i) {
        m_healthStateList.at(i)->runCalculation(timeDiscountRate);
    }

    // Pass calculations to children transition probabilities if any
    for (int i = 0; i < m_childTransProbs.length(); ++i) {
        m_childTransProbs.at(i)->runCalculation();
    }
}

const fn::NumericBaseArray HealthOutcome::getOutputRemLifeExp() const
{
    return m_healthStateList.at(0)->getOutputRemLifeExp();
}

const fn::NumericBaseArray HealthOutcome::getOutputCasesCount() const
{
    fn::NumericBaseArray result = m_healthStateList[0]->getOutputCasesCount();
    for (int i = 1; i < m_healthStateList.length(); ++i) {
        result += m_healthStateList[i]->getOutputCasesCount();
    }
    return result;
}

const NumericBaseArray HealthOutcome::getOutputCumulativeDuration() const
{
    return m_healthStateList.at(0)->getOutputCumulativeDuration();
}

HealthState* HealthOutcome::getHealthState(const size_t idx) const
{
    return m_healthStateList.at(static_cast<int>(idx));
}

bool HealthOutcome::getAcuteType() const
{
    return m_acuteType;
}

bool HealthOutcome::getVisible() const
{
    return m_visible;
}

int HealthOutcome::getNumHealthStates() const
{
    return m_healthStateList.length();
}

InputType::HealthState HealthOutcome::getHealthStatesType() const
{
    //+ Assumption: all health states must have the same type.
    if (m_healthStateList.size() == 0) {
        return fn::InputType::INFECTION;
    } else {
        return m_healthStateList[0]->getType();
    }
}

fn::NumericBaseArray HealthOutcome::getTotalSampleMeanBurden(const fn::Dims& dims)
{
    fn::NumericBaseArray result(dims.begin(), dims.end(), 0);
    // Process all health states
    for (int i = 0; i < m_healthStateList.length(); ++i) {
        result += m_healthStateList[i]->getSampleMeanBurden();
    }
    return result;
}

fn::NumericBaseArray HealthOutcome::getSampleYLD(const fn::Dims& dims)
{
    fn::NumericBaseArray result(dims.begin(), dims.end(), 0);
    // Process all health states
    for (int i = 0; i < m_healthStateList.length(); ++i) {
        result += m_healthStateList[i]->getSampleBurden();
    }
    return result;
}

NumericBaseArray HealthOutcome::getSampleYLL(const fn::Dims& dims)
{
    fn::NumericBaseArray result(dims.begin(), dims.end(), 0);

    // Health outcome is connected directly
    // Process all transitions
    for (int i = 0; i < m_childTransProbs.length(); ++i) {
        const fn::TransitionProbabilityBase& transProb = *m_childTransProbs[i];
        const fn::OutcomeTreeNode& outcomeTreeNode = transProb.getChild();
        if (outcomeTreeNode.getNodeType() == fn::OutcomeTreeNode::HealthState) {
            // Connected to health state
            const fn::HealthState& childHealthState = static_cast<const fn::HealthState&>(outcomeTreeNode);
            if (childHealthState.getType() == fn::InputType::YLL) {
                result += childHealthState.getSampleBurden();
            }
        } else if (outcomeTreeNode.getNodeType() == fn::OutcomeTreeNode::HealthOutcome) {
            // Connected to health outcome
            const fn::HealthOutcome& childHealthOutcome = static_cast<const fn::HealthOutcome&>(outcomeTreeNode);
            for (int i = 0; i < childHealthOutcome.getNumHealthStates(); ++i) {
                fn::HealthState* childHealthState = childHealthOutcome.getHealthState(i);
                if (childHealthState->getType() == fn::InputType::YLL) {
                    result += childHealthState->getSampleBurden();
                }
            }
        }
    }

    // Health states are connected to YLL health outcome/state
    // Process all internal health states
    for (int i = 0; i < m_healthStateList.length(); ++i) {
        fn::HealthState& healthState = *m_healthStateList[i];
        if (healthState.getType() == fn::InputType::YLD) {
            result += healthState.getSampleChildrenYLLBurden(dims);
        }
    }

    return result;
}

fn::NumericBaseArray HealthOutcome::getAggregatedYLD(const fn::Dims& dims)
{
    fn::NumericBaseArray result(dims.begin(), dims.end(), 0);
    // Process all health states
    for (int i = 0; i < m_healthStateList.length(); ++i) {
        const fn::HealthState& healthState = *m_healthStateList[i];
        if (healthState.getType() == fn::InputType::YLD) {
            result += healthState.getAggregatedBurden();
        }
    }
    return result;
}

fn::NumericBaseArray HealthOutcome::getAggregatedYLL(const fn::Dims& dims)
{

    fn::NumericBaseArray result(dims.begin(), dims.end(), 0);

    // Health outcome is connected directly to YLL health outcome/state
    // Process all transitions
    for (int i = 0; i < m_childTransProbs.length(); ++i) {
        const fn::TransitionProbabilityBase& transProb = *m_childTransProbs[i];
        const fn::OutcomeTreeNode& outcomeTreeNode = transProb.getChild();
        if (outcomeTreeNode.getNodeType() == fn::OutcomeTreeNode::HealthState) {
            // Connected to health state
            const fn::HealthState& childHealthState = static_cast<const fn::HealthState&>(outcomeTreeNode);
            if (childHealthState.getType() == fn::InputType::YLL) {
                result += childHealthState.getAggregatedBurden();
            }
        } else if (outcomeTreeNode.getNodeType() == fn::OutcomeTreeNode::HealthOutcome) {
            // Connected to health outcome
            const fn::HealthOutcome& childHealthOutcome = static_cast<const fn::HealthOutcome&>(outcomeTreeNode);
            for (int i = 0; i < childHealthOutcome.getNumHealthStates(); ++i) {
                fn::HealthState* childHealthState = childHealthOutcome.getHealthState(i);
                if (childHealthState->getType() == fn::InputType::YLL) {
                    result += childHealthState->getAggregatedBurden();
                }
            }
        }
    }

    // Health states are connected to YLL health outcome/state
    // Process all internal health states
    for (int i = 0; i < m_healthStateList.length(); ++i) {
        fn::HealthState& healthState = *m_healthStateList[i];
        if (healthState.getType() == fn::InputType::YLD) {
            result += healthState.getAggregatedChildrenYLLBurden(dims);
        }
    }

    return result;
}

NumericBaseArray HealthOutcome::getSamplePopulationCasesCount(const fn::Dims& dims)
{
    fn::NumericBaseArray result(dims.begin(), dims.end(), 0);
    // Process all health states
    for (int i = 0; i < m_healthStateList.length(); ++i) {
        const fn::HealthState& healthState = *m_healthStateList[i];
        if (healthState.getType() == fn::InputType::YLD) {
            result += healthState.getSampleCasesCount();
        }
    }
    return result;
}

NumericBaseArray HealthOutcome::getSampleDeathCasesCount(const fn::Dims& dims)
{
    fn::NumericBaseArray result(dims.begin(), dims.end(), 0);

    // Health outcome is connected directly to YLL health outcome/state
    // Process all transitions
    for (int i = 0; i < m_childTransProbs.length(); ++i) {
        const fn::TransitionProbabilityBase& transProb = *m_childTransProbs[i];
        const fn::OutcomeTreeNode& outcomeTreeNode = transProb.getChild();
        if (outcomeTreeNode.getNodeType() == fn::OutcomeTreeNode::HealthState) {
            // Connected to health state
            const fn::HealthState& childHealthState = static_cast<const fn::HealthState&>(outcomeTreeNode);
            if (childHealthState.getType() == fn::InputType::YLL) {
                result += childHealthState.getSampleCasesCount();
            }
        } else if (outcomeTreeNode.getNodeType() == fn::OutcomeTreeNode::HealthOutcome) {
            // Connected to health outcome
            const fn::HealthOutcome& childHealthOutcome = static_cast<const fn::HealthOutcome&>(outcomeTreeNode);
            for (int i = 0; i < childHealthOutcome.getNumHealthStates(); ++i) {
                fn::HealthState* childHealthState = childHealthOutcome.getHealthState(i);
                if (childHealthState->getType() == fn::InputType::YLL) {
                    result += childHealthState->getSampleCasesCount();
                }
            }
        }
    }

    // Health states are connected to YLL health outcome/state
    // Process all internal health states
    for (int i = 0; i < m_healthStateList.length(); ++i) {
        fn::HealthState& healthState = *m_healthStateList[i];
        if (healthState.getType() == fn::InputType::YLD) {
            result += healthState.getSampleChildrenYLLCasesCount(dims);
        }
    }

    return result;
}

fn::NumericBaseArray HealthOutcome::getAggregatedPopulationCasesCount(const fn::Dims& dims)
{
    fn::NumericBaseArray result(dims.begin(), dims.end(), 0);
    // Process all health states
    for (int i = 0; i < m_healthStateList.length(); ++i) {
        const fn::HealthState& healthState = *m_healthStateList[i];
        if (healthState.getType() == fn::InputType::YLD) {
            result += healthState.getAggregatedCasesCount();
        }
    }
    return result;
}

fn::NumericBaseArray HealthOutcome::getAggregatedDeathCasesCount(const fn::Dims& dims)
{

    fn::NumericBaseArray result(dims.begin(), dims.end(), 0);

    // Health outcome is connected directly to YLL health outcome/state
    // Process all transitions
    for (int i = 0; i < m_childTransProbs.length(); ++i) {
        const fn::TransitionProbabilityBase& transProb = *m_childTransProbs[i];
        const fn::OutcomeTreeNode& outcomeTreeNode = transProb.getChild();
        if (outcomeTreeNode.getNodeType() == fn::OutcomeTreeNode::HealthState) {
            // Connected to health state
            const fn::HealthState& childHealthState = static_cast<const fn::HealthState&>(outcomeTreeNode);
            if (childHealthState.getType() == fn::InputType::YLL) {
                result += childHealthState.getAggregatedCasesCount();
            }
        } else if (outcomeTreeNode.getNodeType() == fn::OutcomeTreeNode::HealthOutcome) {
            // Connected to health outcome
            const fn::HealthOutcome& childHealthOutcome = static_cast<const fn::HealthOutcome&>(outcomeTreeNode);
            for (int i = 0; i < childHealthOutcome.getNumHealthStates(); ++i) {
                fn::HealthState* childHealthState = childHealthOutcome.getHealthState(i);
                if (childHealthState->getType() == fn::InputType::YLL) {
                    result += childHealthState->getAggregatedCasesCount();
                }
            }
        }
    }

    // Health states are connected to YLL health outcome/state
    // Process all internal health states
    for (int i = 0; i < m_healthStateList.length(); ++i) {
        fn::HealthState& healthState = *m_healthStateList[i];
        if (healthState.getType() == fn::InputType::YLD) {
            result += healthState.getAggregatedChildrenYLLCasesCount(dims);
        }
    }

    return result;

}

}  // namespace fn
