//==========================================================================================================================
//  'HealthOutcome' HealthOutcome
//			This class implements a health outcome.
//
//  Written by Daniel Lewandowski
//  Copyright @ 2011 NextPage Software
//==========================================================================================================================

#ifndef _fn_HealthOutcome_
#define _fn_HealthOutcome_

#include "fn/globals.h"
#include "fn/OutcomeTreeNode.h"
#include "fn/HealthState.h"

namespace fn {

class HealthOutcome : public fn::OutcomeTreeNode
{
public:

    // Constructor.
    //
    // 'in' - [in] Unique ID
    // 'caption' - [in] Caption
    // 'acuteType' - [in] TRUE if acute, FALSE if sequela
    // 'visible' - [in] TRUE if visible, FALSE if not
    //
    HealthOutcome(fn::OutcomeTreeNode::ID id, const QString& caption, const fn::Dims& dims, const bool acuteType, const bool visible);

    // Destructor.
    //
    virtual ~HealthOutcome();

    // Add new internal health state
    //
    // 'healthState' - [in] Health state
    //
    void addInternalHealthState(fn::HealthState* healthState);

    // Reinitialize local reference to the health state holding remaining cases
    //
    void invalidateRemainingCasesHealthState();

    // Run calculation
    //
    void runCalculation(const double timeDiscountRate);

    // Get output remaining life expectancy
    // Warning! This function will return remaining life expectancy from the first internal health state
    //
    // Return - output remaining life expectancy numeric array
    virtual const fn::NumericBaseArray getOutputRemLifeExp() const;

    // Get output cases count table
    //
    // Return - output cases count numeric array
    virtual const fn::NumericBaseArray getOutputCasesCount() const;

    // Get output cumulative duration
    // Warning! This function will return output cumulative duration from the first transition from the parent
    //
    // Return - output cumulative duration numeric array
    virtual const fn::NumericBaseArray getOutputCumulativeDuration() const;

    // Get health state
    //
    // 'id' - [in] ID of the health state
    //
    // Return - list of results
    fn::HealthState* getHealthState(const size_t idx) const;

    // Get acute type
    //
    bool getAcuteType() const;

    // Get visibility
    //
    bool getVisible() const;

    // Get the number of health states
    //
    int getNumHealthStates() const;

    // Get type of underlying states
    //
    fn::InputType::HealthState getHealthStatesType() const;

    // Get mean total burden from all health states of this health outcome
    //
    fn::NumericBaseArray getTotalSampleMeanBurden(const fn::Dims& dims);

    // Get total burden from all health states of this health outcome
    //
    fn::NumericBaseArray getSampleYLD(const fn::Dims& dims);
    fn::NumericBaseArray getSampleYLL(const fn::Dims& dims);

    // Get total aggregated burden from all health states of this health outcome
    //
    fn::NumericBaseArray getAggregatedYLD(const fn::Dims& dims);
    fn::NumericBaseArray getAggregatedYLL(const fn::Dims& dims);

    // Get total cases count from all health states of this health outcome
    //
    fn::NumericBaseArray getSamplePopulationCasesCount(const fn::Dims& dims);
    fn::NumericBaseArray getSampleDeathCasesCount(const fn::Dims& dims);

    // Get total aggregated cases count from all health states of this health outcome
    //
    fn::NumericBaseArray getAggregatedPopulationCasesCount(const fn::Dims& dims);
    fn::NumericBaseArray getAggregatedDeathCasesCount(const fn::Dims& dims);

private:

    // Acute or sequela?
    bool m_acuteType;

    // Visible?
    bool m_visible;

    // List of contained health states
    fn::HealthStatePtrList m_healthStateList;

    // Pointer to the health state covering remaining cases
    fn::HealthState* m_remainingCasesHealthState;

};

}  // namespace fn

#endif  // _fn_HealthOutcome_

