//==========================================================================================================================
//  Implementation file for "HealthState.h"
//  Written by Daniel Lewandowski
//  Copyright @ 2011 NextPage Software
//==========================================================================================================================

#include "HealthState.h"

#include "fn/Overloads.h"

namespace fn {

HealthState::HealthState(const fn::OutcomeTreeNode::ID id, const QString& caption, const fn::DataArray& proportion,
                         const fn::DataArray& severity, const fn::DataArray& duration,
                         const fn::InputType::HealthState& type, const fn::Dims& dims, const fn::HealthOutcome& parent) :
    fn::SequelaBase(id, caption, type, dims, parent),
    m_proportion(proportion),
    m_severity(severity),
    m_duration(duration),
    m_proportionRemainingCases(false),
    m_normProportion(),
    m_burden(),
    m_sampleBurden(),
    m_sampleMeanBurden(),
    m_sampleAggregatedBurden()
{
    completeMetaData(m_severity);
    completeMetaData(m_duration);
    completeMetaData(m_proportion);

    m_proportionRemainingCases = m_proportion.getMetaInfo("remaining cases").toUpper() == "YES";
}

HealthState::~HealthState()
{

}

void HealthState::runCalculation(const double timeDiscountRate)
{

    m_currentIteration++;

    // Initialize the case count to that coming from transitions directly entering this health state.
    m_casesCount = getInputCasesCount();
    // Initialize the remaining life expectancy to that coming from the first transition directly entering this health state
    fn::NumericBaseArray inputRemLifeExp = getInputRemLifeExp();
    // Initialize the cumulative duration to that coming from the first transition directly entering this health state
    fn::NumericBaseArray inputCumulativeDuration = getInputCumulativeDuration();

    // If transitions are coming to the parent health outcome, rather than directly to the health state, then take its inputs.
    if (m_parent.getCountTransitionFromParent() > 0) {
        // Accumulate cases from the encapsulating health outcome
        m_casesCount = m_parent.getInputCasesCount() * m_normProportion;

        // Overwrite remaining life expectancy
        inputRemLifeExp = m_parent.getInputRemLifeExp();

        // Overwrite cumulative duration
        inputCumulativeDuration = m_parent.getInputCumulativeDuration();
    }

    // Process duration
    fn::NumericBaseArray duration = m_duration.getValueArray();
    // TODO: remaining life expectancy as functions, for instance "RLE"
    // Take remaining life expectancy if "-1" inserted in the duration table (only the first cell is checked!!!)
    if (static_cast<int>(duration(0)) == -1) {
        // Take input remaining life expectancy
        // No time discounting applied as it is already discounted
        duration = inputRemLifeExp;
    } else {
        fn::applyTimeDiscountRate(timeDiscountRate, inputCumulativeDuration, duration);
    }
    // Duration cannot be longer than the input remaining life expectancy
    duration = fn::min(duration, inputRemLifeExp);
    m_cumulativeDuration = inputCumulativeDuration + duration;
    m_duration.setValueArray(duration);

    // Compute the number of life years left
    m_remLifeExp = inputRemLifeExp - m_duration.getValueArray();

    switch (m_type) {
        case fn::InputType::ASYMPTOMATIC:
        case fn::InputType::YLD: // Life Years lost due to disability
            {
            fn::NumericBaseArray duration = min(m_duration.getValueArray(), inputRemLifeExp);
            m_burden = m_casesCount * m_severity.getValueArray() * duration;
            }
            break;
        case fn::InputType::YLL: // Years Life Lost
            m_burden = m_casesCount * m_remLifeExp;
            break;
        case fn::InputType::INFECTION:
            m_burden = fn::NumericBaseArray(0);
            break;
    }

    // Store samples
    storeSample(m_casesCount, m_sampleCasesCount, static_cast<PosInt>(m_currentIteration));
    storeSample(m_burden, m_sampleBurden, static_cast<PosInt>(m_currentIteration));

    // Pass calculations to children transition probabilities
    for (int i = 0; i < m_childTransProbs.length(); ++i) {
        m_childTransProbs.at(i)->runCalculation();
    }
}

const fn::DataArray& HealthState::getParameter(const fn::HealthState::ParamType& paramType) const
{
    switch (paramType) {
    case fn::HealthState::Severity:
        return m_severity;
    case fn::HealthState::Duration:
        return m_duration;
    case fn::HealthState::Proportion:
        return m_proportion;
    }

    Q_ASSERT(false);
    abort();
}

fn::DataArray& HealthState::getParameter(const fn::HealthState::ParamType& paramType)
{
    switch (paramType) {
    case fn::HealthState::Severity:
        return m_severity;
    case fn::HealthState::Duration:
        return m_duration;
    case fn::HealthState::Proportion:
        return m_proportion;
    }

    Q_ASSERT(false);
    abort();
}

void HealthState::setParameter(const fn::HealthState::ParamType& paramType, const fn::DataArray& parameter)
{
    switch (paramType)
    {
    case fn::HealthState::Severity:
        m_severity = parameter;
        completeMetaData(m_severity);
        break;
    case fn::HealthState::Duration:
        m_duration = parameter;
        completeMetaData(m_duration);
         break;
    case fn::HealthState::Proportion:
        m_proportion = parameter;
        completeMetaData(m_proportion);
        m_proportionRemainingCases = m_proportion.getMetaInfo("remaining cases").toLower() == "yes";
        break;
    }
}

bool HealthState::isRemainingCases() const
{
    return m_proportionRemainingCases;
}

void HealthState::initializeSampling(const fn::Dims& sampleDims)
{
    fn::SequelaBase::initializeSampling(sampleDims);
    m_sampleBurden = fn::NumericBaseArray(sampleDims.begin(), sampleDims.end(), 0);
}

void HealthState::clearSample()
{
    fn::SequelaBase::clearSample();
    m_sampleBurden = fn::NumericBaseArray();
}

void HealthState::setNormalizedProportion(const NumericBaseArray& normProportion)
{
    m_normProportion = normProportion;
}

const fn::NumericBaseArray& HealthState::getNormalizedProportion() const
{
    return m_normProportion;
}

void HealthState::computeSampleMeanResults()
{
    fn::SequelaBase::computeSampleMeanResults();
    m_sampleMeanBurden = fn::getSampleMean(m_sampleBurden);
}

const fn::NumericBaseArray& HealthState::getSampleMeanBurden() const
{
    return m_sampleMeanBurden;
}

const fn::NumericBaseArray& HealthState::getSampleBurden() const
{
    return m_sampleBurden;
}

const fn::InputType::HealthState& HealthState::getType() const
{
    return m_type;
}

void HealthState::computeAggregatedResults()
{
    fn::SequelaBase::computeAggregatedResults();
    m_sampleAggregatedBurden = fn::getAggregatedSample(m_sampleBurden);
}

const fn::NumericBaseArray& HealthState::getAggregatedBurden() const
{
    return m_sampleAggregatedBurden;
}

NumericBaseArray HealthState::getSampleChildrenYLLBurden(const fn::Dims& dims)
{
    fn::NumericBaseArray result(dims.begin(), dims.end(), 0);
    // Process all health states
    for (int i = 0; i < m_childTransProbs.length(); ++i) {
        const fn::TransitionProbabilityBase& transProb = *m_childTransProbs[i];
        const fn::OutcomeTreeNode& outcomeTreeNode = transProb.getChild();
        if (outcomeTreeNode.getNodeType() == fn::OutcomeTreeNode::HealthState) {
            const fn::HealthState& childHealthState = static_cast<const fn::HealthState&>(outcomeTreeNode);
            if (childHealthState.getType() == fn::InputType::YLL) {
                result += childHealthState.getSampleBurden();
            }
        } else if (outcomeTreeNode.getNodeType() == fn::OutcomeTreeNode::HealthOutcome) {
            // Connected to health outcome
            const fn::HealthOutcome& childHealthOutcome = static_cast<const fn::HealthOutcome&>(outcomeTreeNode);
            for (int i = 0; i < childHealthOutcome.getNumHealthStates(); ++i) {
                fn::HealthState* childHealthState = childHealthOutcome.getHealthState(i);
                if (childHealthState->getType() == fn::InputType::YLL) {
                    result += childHealthState->getSampleBurden();
                }
            }
        }
    }
    return result;
}

fn::NumericBaseArray HealthState::getAggregatedChildrenYLLBurden(const fn::Dims& dims)
{
    fn::NumericBaseArray result(dims.begin(), dims.end(), 0);
    // Process all health states
    for (int i = 0; i < m_childTransProbs.length(); ++i) {
        const fn::TransitionProbabilityBase& transProb = *m_childTransProbs[i];
        const fn::OutcomeTreeNode& outcomeTreeNode = transProb.getChild();
        if (outcomeTreeNode.getNodeType() == fn::OutcomeTreeNode::HealthState) {
            const fn::HealthState& childHealthState = static_cast<const fn::HealthState&>(outcomeTreeNode);
            if (childHealthState.getType() == fn::InputType::YLL) {
                result += childHealthState.getAggregatedBurden();
            }
        } else if (outcomeTreeNode.getNodeType() == fn::OutcomeTreeNode::HealthOutcome) {
            // Connected to health outcome
            const fn::HealthOutcome& childHealthOutcome = static_cast<const fn::HealthOutcome&>(outcomeTreeNode);
            for (int i = 0; i < childHealthOutcome.getNumHealthStates(); ++i) {
                fn::HealthState* childHealthState = childHealthOutcome.getHealthState(i);
                if (childHealthState->getType() == fn::InputType::YLL) {
                    result += childHealthState->getAggregatedBurden();
                }
            }
        }
    }
    return result;
}

NumericBaseArray HealthState::getSampleChildrenYLLCasesCount(const Dims &dims)
{
    fn::NumericBaseArray result(dims.begin(), dims.end(), 0);
    // Process all health states
    for (int i = 0; i < m_childTransProbs.length(); ++i) {
        const fn::TransitionProbabilityBase& transProb = *m_childTransProbs[i];
        const fn::OutcomeTreeNode& outcomeTreeNode = transProb.getChild();
        if (outcomeTreeNode.getNodeType() == fn::OutcomeTreeNode::HealthState) {
            const fn::HealthState& childHealthState = static_cast<const fn::HealthState&>(outcomeTreeNode);
            if (childHealthState.getType() == fn::InputType::YLL) {
                result += childHealthState.getSampleCasesCount();
            }
        } else if (outcomeTreeNode.getNodeType() == fn::OutcomeTreeNode::HealthOutcome) {
            // Connected to health outcome
            const fn::HealthOutcome& childHealthOutcome = static_cast<const fn::HealthOutcome&>(outcomeTreeNode);
            for (int i = 0; i < childHealthOutcome.getNumHealthStates(); ++i) {
                fn::HealthState* childHealthState = childHealthOutcome.getHealthState(i);
                if (childHealthState->getType() == fn::InputType::YLL) {
                    result += childHealthState->getSampleCasesCount();
                }
            }
        }
    }
    return result;
}

fn::NumericBaseArray HealthState::getAggregatedChildrenYLLCasesCount(const Dims& dims)
{
    fn::NumericBaseArray result(dims.begin(), dims.end(), 0);
    // Process all health states
    for (int i = 0; i < m_childTransProbs.length(); ++i) {
        const fn::TransitionProbabilityBase& transProb = *m_childTransProbs[i];
        const fn::OutcomeTreeNode& outcomeTreeNode = transProb.getChild();
        if (outcomeTreeNode.getNodeType() == fn::OutcomeTreeNode::HealthState) {
            const fn::HealthState& childHealthState = static_cast<const fn::HealthState&>(outcomeTreeNode);
            if (childHealthState.getType() == fn::InputType::YLL) {
                result += childHealthState.getAggregatedCasesCount();
            }
        } else if (outcomeTreeNode.getNodeType() == fn::OutcomeTreeNode::HealthOutcome) {
            // Connected to health outcome
            const fn::HealthOutcome& childHealthOutcome = static_cast<const fn::HealthOutcome&>(outcomeTreeNode);
            for (int i = 0; i < childHealthOutcome.getNumHealthStates(); ++i) {
                fn::HealthState* childHealthState = childHealthOutcome.getHealthState(i);
                if (childHealthState->getType() == fn::InputType::YLL) {
                    result += childHealthState->getAggregatedCasesCount();
                }
            }
        }
    }
    return result;
}

void HealthState::initializeAllFormulaParsers()
{
    m_severity.initializeParser();
    m_duration.initializeParser();
    m_proportion.initializeParser();
}

void HealthState::deinitializeAllFormulaParsers()
{
    m_severity.deinitializeParser();
    m_duration.deinitializeParser();
    m_proportion.deinitializeParser();
}

void HealthState::evaluateAllFormulas()
{
    m_severity.evaluateFormulas();
    m_duration.evaluateFormulas();
    m_proportion.evaluateFormulas();
}

void HealthState::completeMetaData(fn::DataArray& array)
{
    fn::StringToStringMap metaData = array.getMetaData();
    QString arrayName = array.getName();

    if (!metaData.contains("description")) {
        metaData["description"] = "";
    }
    if (!metaData.contains("weblink")) {
        metaData["weblink"] = "";
    }
    if (arrayName.toLower() == "proportion" && !metaData.contains("remaining cases")) {
        metaData["remaining cases"] = "No";
    }

    array.setMetaData(metaData);
}

}  // namespace fn
