//==========================================================================================================================
//  'HealthState' HealthState
//			This class implements a health state, a state within a health outcome.
//
//  Written by Daniel Lewandowski
//  Copyright @ 2011 NextPage Software
//==========================================================================================================================

#ifndef _fn_HealthState_
#define _fn_HealthState_

#include "fn/SequelaBase.h"

namespace fn {

class HealthOutcome;

class HealthState : public fn::SequelaBase
{
public:

    enum ParamType {Severity = 0, Duration, Proportion};

    // Constructor.
    //
    // 'id' - [in] Unique ID
    // 'caption' - [in] Caption
    // 'proportion' - [in] Proportion of cases covered by this health state (scalar)
    // 'severity' - [in] Severity
    // 'duration' - [in] Duration
    // 'type' - [in] Type
    // 'parent' - [in] Parent health outcome
    //
    HealthState(const fn::OutcomeTreeNode::ID id, const QString& caption, const fn::DataArray& proportion,
                const fn::DataArray& severity, const fn::DataArray& duration,
                const fn::InputType::HealthState& type, const fn::Dims& dims, const fn::HealthOutcome& parent);

    // Destructor.
    //
    virtual ~HealthState();

    // Run calculation
    //
    void runCalculation(const double timeDiscountRate);

    // Get data array type parameter
    //
    // Return - data array for parameter
    const fn::DataArray& getParameter(const fn::HealthState::ParamType& paramType) const;
          fn::DataArray& getParameter(const fn::HealthState::ParamType& paramType);

    // Set parameter
    //
    // 'parameter' - [in] New parameter
    //
    void setParameter(const fn::HealthState::ParamType& paramType, const fn::DataArray& parameter);

    // Flag indicating whether this health state covers the remaining cases over all health states of the parent health outcome
    //
    // Return - TRUE is remaining cases
    bool isRemainingCases() const;

    // Initialize sample containers to the specified size
    //
    // 'dims' - [in] Dimensions
    //
    void initializeSampling(const fn::Dims& sampleDims);

    // Clear sample
    //
    void clearSample();

    // Set normalized proportion
    //
    void setNormalizedProportion(const fn::NumericBaseArray& normProportion);

    // Get normalized proportion
    //
    const fn::NumericBaseArray& getNormalizedProportion() const;

    // Compute mean sample results
    //
    void computeSampleMeanResults();

    // Get sample mean burden
    //
    const fn::NumericBaseArray& getSampleMeanBurden() const;

    // Get sample burden
    //
    const fn::NumericBaseArray& getSampleBurden() const;

    // Get type
    //
    const fn::InputType::HealthState& getType() const;

    // Get aggregated burden
    //
    const fn::NumericBaseArray& getAggregatedBurden() const;

    // Get sample burden from children YLL
    //
    fn::NumericBaseArray getSampleChildrenYLLBurden(const fn::Dims& dims);

    // Get aggregated burden from children YLL
    //
    fn::NumericBaseArray getAggregatedChildrenYLLBurden(const fn::Dims& dims);

    // Get aggregated cases count from children YLL
    //
    fn::NumericBaseArray getAggregatedChildrenYLLCasesCount(const fn::Dims& dims);

    // Get cases count from children YLL
    //
    fn::NumericBaseArray getSampleChildrenYLLCasesCount(const fn::Dims& dims);

    // Compute aggregated burden
    //
    void computeAggregatedResults();

    // Formula parser handlers
    //
    void initializeAllFormulaParsers();
    void deinitializeAllFormulaParsers();
    void evaluateAllFormulas();

private:

    friend class fn::HealthOutcome;

    // Make sure meta data is complete
    //
    // 'array' - [in] array to be check and completed
    //
    void completeMetaData(fn::DataArray& array);

    // ----------------------------------------------
    // Permanent parameters

    // Proportion of cases covered by this health state
    fn::DataArray m_proportion;

    // Severity parameter
    fn::DataArray m_severity;

    // Duration parameter
    fn::DataArray m_duration;

    // Is it remaining proportion?
    bool m_proportionRemainingCases;

    // ----------------------------------------------
    // Iteration input parameters

    // Normalized proportion
    fn::NumericBaseArray m_normProportion;

    // ----------------------------------------------
    // Iteration output results

    // Burden of sequela
    fn::NumericBaseArray m_burden;

    // ----------------------------------------------
    // Sample storage
    fn::NumericBaseArray m_sampleBurden;

    // ----------------------------------------------
    // Sample mean output
    fn::NumericBaseArray m_sampleMeanBurden;

    // ----------------------------------------------
    // Total burden sample
    fn::NumericBaseArray m_sampleAggregatedBurden;

};

// List of constant pointers to health states.
typedef QList<fn::HealthState*> HealthStatePtrList;

}  // namespace fn

#endif  // _fn_HealthState_
