//==========================================================================================================================
//  Implementation file for "Infection.h"
//  Written by Daniel Lewandowski
//  Copyright @ 2011 NextPage Software
//==========================================================================================================================

#include "Infection.h"

#include "fn/DiseaseModel.h"
#include "fn/Overloads.h"

namespace fn {

Infection::Infection(const fn::OutcomeTreeNode::ID id, const QString& caption, const fn::Dims& dims, const fn::HealthOutcome& parent) :
    fn::SequelaBase(id, caption, fn::InputType::INFECTION, dims, parent),
    m_lifeExp(),
    m_rawCasesCount(),
    m_sampleCoverage(),
    m_sampleUnderestimation(),
    m_alpha(nullptr)
{

}

Infection::~Infection()
{

}

void Infection::runCalculation(const fn::NumericBaseArray& coverage, const double timeDiscountRate, const fn::NumericBaseArray& underestimation)
{
    m_currentIteration++;

    const fn::NumericBaseArray& alpha = m_alpha->getIterProbability();

    // Derive total incidence from the first input (Acute illness)
    m_casesCount = computeCorrectedCasesCount(alpha, coverage, underestimation);

    // Store sample output incidence
    storeSample(m_casesCount, m_sampleCasesCount, static_cast<PosInt>(m_currentIteration));

    // Apply time discounting to remaining life expectancy
    // This is the starting life expectancy of this outcome tree since this is health outcome "Infection".
    fn::applyTimeDiscountRate(timeDiscountRate, m_remLifeExp);

    // Pass calculations to transition probabilities
    for (int i = 0; i < m_childTransProbs.length(); ++i) {
        m_childTransProbs.at(i)->runCalculation();
    }
}

const fn::DataArray& Infection::getRawCasesCount() const
{
    return m_rawCasesCount;
}

void Infection::setRawCasesCount(const fn::DataArray& rawCasesCount)
{
    m_rawCasesCount = rawCasesCount;
}

const fn::DataArray& Infection::getLifeExpectancy() const
{
    return m_lifeExp;
}

void Infection::setLifeExpectancy(const fn::DataArray& lifeExp)
{
    m_lifeExp = lifeExp;
}

void Infection::initializeSampling(const fn::Dims& sampleDims)
{
    fn::SequelaBase::initializeSampling(sampleDims);
    m_sampleCoverage = fn::NumericBaseArray(sampleDims.begin(), sampleDims.end(), 0);
    m_sampleUnderestimation = fn::NumericBaseArray(sampleDims.begin(), sampleDims.end(), 0);
}

fn::NumericBaseArray Infection::computeCorrectedCasesCount(const fn::NumericBaseArray& alpha, const fn::NumericBaseArray& coverage,
                                                    const fn::NumericBaseArray& underestimation)
{
    return m_rawCasesCount.getValueArray() / alpha / coverage * underestimation;
}

void Infection::determineAlpha()
{
    m_alpha = nullptr;

    // Pass calculations to transition probabilities
    for (int i = 0; i < m_childTransProbs.length(); ++i) {
        fn::TransitionProbabilityBase* transProb = m_childTransProbs.at(i);
        const fn::OutcomeTreeNode& child = transProb->getChild();
        if (child.getNodeType() == fn::OutcomeTreeNode::NodeType::HealthOutcome) {
            const fn::HealthOutcome& ho = static_cast<const fn::HealthOutcome&>(child);
            if (ho.getAcuteType()) {
                m_alpha = transProb;
                break;
            }
        } else if (child.getNodeType() == fn::OutcomeTreeNode::NodeType::HealthState) {
            const fn::SequelaBase& sb = static_cast<const fn::SequelaBase&>(child);
            const fn::HealthOutcome& ho = sb.getParentHealthOutcome();
            if (ho.getAcuteType() && ho.getNumHealthStates() == 1) {
                m_alpha = transProb;
                break;
            }
        }
    }
}

const fn::TransitionProbabilityBase* Infection::getAlpha() const
{
    return m_alpha;
}

void Infection::clearSample()
{
    fn::SequelaBase::clearSample();
    m_sampleCoverage = fn::NumericBaseArray();
    m_sampleUnderestimation = fn::NumericBaseArray();
}

void Infection::createAllFormulaParsers()
{
    m_lifeExp.initializeParser();
    m_rawCasesCount.initializeParser();
}

void Infection::destroyAllFormulaParsers()
{
    m_lifeExp.deinitializeParser();
    m_rawCasesCount.deinitializeParser();
}

void Infection::evaluateAllFormulas()
{
    m_lifeExp.evaluateFormulas();
    m_remLifeExp = m_lifeExp.getValueArray();
    m_rawCasesCount.evaluateFormulas();
}

}  // namespace fn
