//==========================================================================================================================
//  Implementation file for "InputDataMgr.h"
//  Written by Daniel Lewandowski
//  Copyright @ 2011 NextPage Software
//==========================================================================================================================

#include "InputDataMgr.h"

#include <QStringList>
#include <QDir>
#include "fn/Overloads.h"
#include "fn/DataArray.h"
#include "fn/ImportExportHelpers.h"
#include "ut/Excel.h"

namespace fn {

#if (defined(FINAL))
    const QString kReferenceDataPath = "./referenceData/";
#else
    const QString kReferenceDataPath = "./../../../../../_referenceData/";
#endif

const QString kDiseaseDataRefFileName = "diseaseData.xls";
const QString kCountryDataRefFileName = "countryData.xls";
const QString kModelDataRefFileNamePrefix = "modelData-";


InputDataMgr::InputDataMgr():
    m_appPath(""),
    m_diseaseFileInfo(""),
    m_countryFileInfo(""),
    m_modelFileInfos(),
    m_diseaseTemplateMgr(),
    m_inputFactorMgr(),
    m_countryNames(),
    m_lifeExpectancyData(),
    m_ageDistributionData(),
    m_paramsMgr(),
    m_notificationData(),
    m_outcomeTreeMappingData(),
    m_transProbInfoMgr(),
    m_correctionData(),
    m_correctionType(),
    m_availCountryDiseasePairs(),
    m_diseaseIds()
{

}

InputDataMgr::~InputDataMgr()
{

}

const DataArray& InputDataMgr::getNotificationData(const InputType::Country& country, const fn::InputType::Disease& disease)
{
    fn::InputType::Country finalCountry = country;
    if (!m_notificationData.contains(country)) {
        finalCountry = "Generic";
    }

    return m_notificationData[finalCountry][disease];
}

const DataArray& InputDataMgr::getOutcomeTreeMappingData(const InputType::Country& country, const InputType::Disease& disease)
{
    fn::InputType::Country finalCountry = country;
    if (!m_outcomeTreeMappingData.contains(country)) {
        finalCountry = "Generic";
    }

    return m_outcomeTreeMappingData[finalCountry][disease];
}

DataArray& InputDataMgr::getLifeExpectancyData(const InputType::Country& country)
{
    return m_lifeExpectancyData[country];
}

void InputDataMgr::setLifeExpectancyData(const InputType::Country& country, const DataArray& leData)
{
    m_lifeExpectancyData[country] = leData;
}

DataArray& InputDataMgr::getAgeDistributionData(const InputType::Country& country)
{
    return m_ageDistributionData[country];
}

void InputDataMgr::setAgeDistributionData(const InputType::Country& country, const DataArray& ageData)
{
    m_ageDistributionData[country] = ageData;
}

void InputDataMgr::importReferenceData()
{
    if (m_diseaseFileInfo.exists() && m_countryFileInfo.exists() && m_modelFileInfos.length() > 0) {
        emit excelSheetImporting(fn::FileOperationType::Import, "Reference data", "importStarted", "");
        // Clear previous data
        m_lifeExpectancyData.clear();
        m_ageDistributionData.clear();
        m_notificationData.clear();
        m_transProbInfoMgr.clearData();
        m_correctionData.clear();
        m_correctionType.clear();
        m_paramsMgr.clearData();
        m_inputFactorMgr.clearData();
        m_diseaseTemplateMgr.clearData();
        m_availCountryDiseasePairs.clear();
        m_diseaseIds.clear();
        // Read new data
        readData();
        // Save in a binary format
        saveSerializedData();
        emit excelSheetImporting(fn::FileOperationType::Import, "Reference data", "importFinished", "");
    } else {
        emit showMessage("Reference data", "Reference data Excel files could not be found", "error");
    }
}

bool InputDataMgr::countryDataExist(const InputType::Country& country) const
{
    return (m_lifeExpectancyData.contains(country) && m_ageDistributionData.contains(country));
}

QString InputDataMgr::getCountryName(const InputType::Country& country) const
{
    QString result = QStringLiteral("");
    if (countryDataExist(country)) {
        result = m_countryNames[country];
    }

    return result;
}

bool InputDataMgr::diseaseDataExist(const InputType::Disease& disease) const
{
    return m_diseaseTemplateMgr.diseaseExists(disease);
}

bool InputDataMgr::countryDiseaseCombinationAvailable(const InputType::Country& country, const InputType::Disease& disease) const
{
    fn::InputType::Country finalCountry = country;
    if (!m_availCountryDiseasePairs.contains(country)) {
        finalCountry = "Generic";
    }

    return m_availCountryDiseasePairs.contains(finalCountry, disease);
}

const DiseaseTemplateMgr& InputDataMgr::getDiseaseTemplateMgr() const
{
    return m_diseaseTemplateMgr;
}

DiseaseTemplateMgr& InputDataMgr::getDiseaseTemplateMgr()
{
    return m_diseaseTemplateMgr;
}

const QList<fn::InputType::Country> InputDataMgr::getAvailableCountries() const
{
    return m_lifeExpectancyData.keys();
}

const QList<fn::InputType::Disease> InputDataMgr::getAvailableDiseases() const
{
    return m_diseaseIds;
}

const ParamsMgr& InputDataMgr::getParametersMgr() const
{
    return m_paramsMgr;
}

ParamsMgr& InputDataMgr::getParametersMgr()
{
    return m_paramsMgr;
}

const TransitionParamsMgr& InputDataMgr::getTransProbInfoMgr() const
{
    return m_transProbInfoMgr;
}

TransitionParamsMgr& InputDataMgr::getTransProbInfoMgr()
{
    return m_transProbInfoMgr;
}

fn::DataArray InputDataMgr::getCorrectionWithName(const fn::InputType::Country& country,
                                                  const fn::InputType::Disease& disease,
                                                  const QString& corrName)
{
    fn::InputType::Country finalCountry = country;
    if (!m_correctionData.contains(country)) {
        finalCountry = "Generic";
    }

    fn::DataArray result = m_correctionData[finalCountry][disease].value(corrName);
    return result;
}

int InputDataMgr::getCorrectionType(const InputType::Country& country, const InputType::Disease& disease)
{
    fn::InputType::Country finalCountry = country;
    if (!m_correctionType.contains(country)) {
        finalCountry = "Generic";
    }

    int result = 0;
    if (m_correctionType[finalCountry].contains(disease)) {
        if (m_correctionType[finalCountry][disease].contains("type")) {
            result = m_correctionType[finalCountry][disease]["type"].toInt();
        }
    }

    return result;
}

void InputDataMgr::setPath(const QString& appPath)
{
    m_appPath = appPath;
    loadSerializedData();

    m_diseaseFileInfo.setFile(m_appPath + kReferenceDataPath + kDiseaseDataRefFileName);
    m_countryFileInfo.setFile(m_appPath + kReferenceDataPath + kCountryDataRefFileName);
    QDir dir(m_appPath + kReferenceDataPath);
    dir.setNameFilters(QStringList() << kModelDataRefFileNamePrefix + "*");
    dir.setFilter(QDir::Files | QDir::Readable | QDir::NoSymLinks);
    dir.setSorting(QDir::Name);
    m_modelFileInfos = dir.entryInfoList();
}

void InputDataMgr::saveSerializedData()
{
    QFile file(m_appPath + "/referenceData.dat");

    file.open(QIODevice::WriteOnly);
    QDataStream out(&file);
    out.setVersion(QDataStream::Qt_5_14);

    // Write the data
    out << m_inputFactorMgr;

    out << m_countryNames;
    out << m_lifeExpectancyData;
    out << m_ageDistributionData;

    out << m_outcomeTreeMappingData;

    out << m_notificationData;

    out << m_paramsMgr;

    out << m_diseaseTemplateMgr;

    out << m_transProbInfoMgr;
    out << m_correctionData;
    out << m_correctionType;

    out << m_availCountryDiseasePairs;

    out << m_diseaseIds;

    file.flush();
    file.close();
}

void InputDataMgr::loadSerializedData()
{
    QFile file(m_appPath + "/referenceData.dat");
    if (file.exists()) {
        file.open(QIODevice::ReadOnly);
        QDataStream in(&file);
        in.setVersion(QDataStream::Qt_5_14);

        // Write the data
        in >> m_inputFactorMgr;

        in >>  m_countryNames;
        in >> m_lifeExpectancyData;
        in >> m_ageDistributionData;

        in >> m_outcomeTreeMappingData;

        in >> m_notificationData;

        in >> m_paramsMgr;

        in >> m_diseaseTemplateMgr;

        in >> m_transProbInfoMgr;
        in >> m_correctionData;
        in >> m_correctionType;

        in >> m_availCountryDiseasePairs;

        in >> m_diseaseIds;

        file.close();
    }
}

void InputDataMgr::readData()
{
    const QString kRunFactorsDefString = "Run factors";
    const QString kLifeExpectancyString = "Life expectancy";
    const QString kAgeDistributionString = "Age distribution";
    const QString kOutcomeTreeMappingDataString = "Outcome tree mapping";
    const QString kNotificationDataString = "Notification data";
    const QString kCoverageDataString = "coverage";
    const QString kUnderestimationDataString = "underestimation";
    const QString kUnderreportingDataString = "underreporting";
    const QString kUnderascertainmentDataString = "underascertainment";
    const QString kUnderEstimationType = "underestimation type";
    const QString kCaptionDefString = "Caption";

    // Run Excel application
    ut::Excel excelApp;

    // 1. Load disease templates
    if (m_diseaseFileInfo.exists()) {

        try {
            excelApp.openWorkbook(m_diseaseFileInfo.canonicalFilePath());

            int sheetCount = excelApp.getWorksheetCount();
            if (sheetCount > 0) {
                // Extract data from all sheets
                for (int sheetIdx = 0; sheetIdx < sheetCount; ++sheetIdx) {

                    excelApp.setCurrentSheet(sheetIdx + 1);

                    QString diseaseId = excelApp.getCurrentSheetName();
                    emit excelSheetImporting(FileOperationType::Import, "Reference data", m_diseaseFileInfo.fileName(), diseaseId);
                    readDiseaseTemplate(excelApp, diseaseId);
                }
            }
        } catch (...) {
            // Close country data workbook
            excelApp.closeWorkbook();
            QMessageBox::critical(nullptr, "Error", "An error occured");
        }
    }

    // 2. Load country data
    if (m_countryFileInfo.exists()) {

        try {
            excelApp.openWorkbook(m_countryFileInfo.canonicalFilePath());

            int sheetCount = excelApp.getWorksheetCount();
            if (sheetCount > 0) {
                // Extract data from all sheets
                for (int sheetIdx = 0; sheetIdx < sheetCount; ++sheetIdx) {

                    excelApp.setCurrentSheet(sheetIdx + 1);

                    QString countryId = excelApp.getCurrentSheetName();
                    if (fn::InputType::CountryIds.contains(countryId)) {
                        emit excelSheetImporting(FileOperationType::Import, "Reference data", m_countryFileInfo.fileName(), countryId);

                        if (countryId != "Generic") {
                            int lastRow = 0;

                            QAxObject* headerCell = excelApp.getCellWithText(kCaptionDefString, "A");
                            QString countryName = excelApp.getStringCellValue(ut::Excel::getCellRow(headerCell) + 1, 1);
                            m_countryNames.insert(countryId, countryName);

                            // Read input data tables ----------------------------------------------------------------------------------
                            headerCell = excelApp.getCellWithText(kLifeExpectancyString, "A");
                            m_lifeExpectancyData.insert(countryId, readInputDataFromExcel(headerCell, excelApp, lastRow));

                            headerCell = excelApp.getCellWithText(kAgeDistributionString, "A");
                            m_ageDistributionData.insert(countryId, readInputDataFromExcel(headerCell, excelApp, lastRow));
                        }
                    }
                }
            }
        } catch (...) {
            // Close country data workbook
            excelApp.closeWorkbook();
            QMessageBox::critical(nullptr, "Error", "An error occured");
        }
    }

    // 3. Load model data
    for (int i = 0; i < m_modelFileInfos.length(); ++i) {
        const QFileInfo& modelDataFileInfo = m_modelFileInfos.at(i);
        QString countryId = modelDataFileInfo
                .baseName()
                .remove(kModelDataRefFileNamePrefix)
                .trimmed();
        // Check if country is recognized in the internal list of countries and diseases
        if (!fn::InputType::CountryIds.contains(countryId)) {
            continue;
        }

        if (modelDataFileInfo.exists()) {

            try {

                excelApp.openWorkbook(modelDataFileInfo.canonicalFilePath());

                int sheetCount = excelApp.getWorksheetCount();
                if (sheetCount > 0) {
                    // Extract data from all sheets
                    for (int sheetIdx = 0; sheetIdx < sheetCount; ++sheetIdx) {

                        excelApp.setCurrentSheet(sheetIdx + 1);

                        QString diseaseId = excelApp
                                .getCurrentSheetName()
                                .trimmed();
                        emit excelSheetImporting(FileOperationType::Import, "Reference data", modelDataFileInfo.fileName(), diseaseId);

                        m_availCountryDiseasePairs.insert(countryId, diseaseId);
                        // Check if country and disease data have been imported already
                        if (((countryId != "Generic") && !countryDataExist(countryId)) || !diseaseDataExist(diseaseId)) {
                            continue;
                        }

                        int lastRow = 0;

                        // Read input factors ----------------------------------------------------------------------------------
                        readFactors(excelApp, kRunFactorsDefString, diseaseId);

                        // Read outcome tree mapping data ----------------------------------------------------------------------
                        QAxObject* headerCell = excelApp.getCellWithText(kOutcomeTreeMappingDataString);
                        m_outcomeTreeMappingData[countryId][diseaseId] = readInputDataFromExcel(headerCell, excelApp, lastRow);

                        // Read infection data -------------------------------------------------------------------------------------
                        headerCell = excelApp.getCellWithText(kNotificationDataString);
                        m_notificationData[countryId][diseaseId] = readInputDataFromExcel(headerCell, excelApp, lastRow);

                        // Read corrections data -------------------------------------------------------------------------------
                        headerCell = excelApp.getCellWithText(kCoverageDataString);
                        m_correctionData[countryId][diseaseId]["coverage"] = readInputDataFromExcel(headerCell, excelApp, lastRow);

                        headerCell = excelApp.getCellWithText(kUnderestimationDataString);
                        m_correctionData[countryId][diseaseId]["underestimation"] = readInputDataFromExcel(headerCell, excelApp, lastRow);

                        headerCell = excelApp.getCellWithText(kUnderreportingDataString);
                        m_correctionData[countryId][diseaseId]["underreporting"] = readInputDataFromExcel(headerCell, excelApp, lastRow);

                        headerCell = excelApp.getCellWithText(kUnderascertainmentDataString);
                        m_correctionData[countryId][diseaseId]["underascertainment"] = readInputDataFromExcel(headerCell, excelApp, lastRow);

                        headerCell = excelApp.getCellWithText(kUnderEstimationType);
                        int startCell = ut::Excel::getCellRow(headerCell) + 2;
                        m_correctionType[countryId][diseaseId] = readStringMapData(excelApp, startCell, lastRow);

                        // Read health states parameters -----------------------------------------------------------------------
                        fn::DiseaseTemplate& diseaseTemplate = m_diseaseTemplateMgr.getDiseaseTemplate(diseaseId);
                        for (int i = 0; i < diseaseTemplate.getNumOutcomeTreeTemplates(); ++i) {

                            QString outcomeTreeID = diseaseTemplate.getOutcomeTreeTemplateID(i);
                            const fn::OutcomeTreeTemplate outcomeTreeTemplate = diseaseTemplate.getOutcomeTreeTemplate(outcomeTreeID);

                            // Read parameters ---------------------------------------------------------------------------------
                            const fn::StringToStringListMap& healthStateMap = outcomeTreeTemplate.getHealthStateMap();
                            fn::StringToStringListMap::const_iterator healthStateIter = healthStateMap.constBegin();
                            for (; healthStateIter != healthStateMap.constEnd(); ++healthStateIter) {
                                const QString healthStateName = healthStateIter.key();
                                headerCell = excelApp.getCellWithText(healthStateName);
                                if (headerCell != nullptr) {

                                    // Read severity
                                    QAxObject* severityHeaderCell = excelApp.getCell(ut::Excel::getCellRow(headerCell) + 2, 1);
                                    m_paramsMgr.addParameter(countryId, diseaseId, outcomeTreeID, healthStateName, fn::ParamsMgr::Severity, readInputDataFromExcel(severityHeaderCell, excelApp, lastRow));

                                    // Read duration
                                    QAxObject* durationHeaderCell = excelApp.getCell(lastRow + 2, 1);
                                    m_paramsMgr.addParameter(countryId, diseaseId, outcomeTreeID, healthStateName, fn::ParamsMgr::Duration, readInputDataFromExcel(durationHeaderCell, excelApp, lastRow));

                                    // Read proportion (specified optionally)
                                    if (excelApp.getCellValue(lastRow + 2, 1).toString().toLower() == "proportion") {
                                        QAxObject* proportionHeaderCell = excelApp.getCell(lastRow + 2, 1);
                                        m_paramsMgr.addParameter(countryId, diseaseId, outcomeTreeID, healthStateName, fn::ParamsMgr::Proportion, readInputDataFromExcel(proportionHeaderCell, excelApp, lastRow));
                                    }
                                }
                            }

                            // Read transition probabilities data --------------------------------------------------------------
                            const fn::StringToStringListMap& transProbMap = outcomeTreeTemplate.getTransProbMap();
                            fn::StringToStringListMap::const_iterator transProbIter = transProbMap.constBegin();
                            for (; transProbIter != transProbMap.constEnd(); ++transProbIter) {
                                const QString transProbName = transProbIter.key();
                                headerCell = excelApp.getCellWithText(transProbName);
                                if (headerCell != nullptr) {

                                    // Read probability
                                    QAxObject* probabilityHeaderCell = excelApp.getCell(ut::Excel::getCellRow(headerCell) + 2, 1);
                                    m_transProbInfoMgr.addParameter(countryId, diseaseId, outcomeTreeID, transProbName, fn::TransitionParamsMgr::Probability, readInputDataFromExcel(probabilityHeaderCell, excelApp, lastRow));

                                    // Read redistribution (specified optionally)
                                    if (excelApp.getCellValue(lastRow + 2, 1).toString().toLower() == "redistribution") {
                                        QAxObject* redistributionHeaderCell = excelApp.getCell(lastRow + 2, 1);
                                        m_transProbInfoMgr.addParameter(countryId, diseaseId, outcomeTreeID, transProbName, fn::TransitionParamsMgr::Redistribution, readInputDataFromExcel(redistributionHeaderCell, excelApp, lastRow));
                                    }

                                }
                            }
                        }
                    }
                }
            } catch (...) {
                // Close country data workbook
                excelApp.closeWorkbook();
                QMessageBox::critical(nullptr, "Error", "An error occured");
            }

        }
    }

}

void InputDataMgr::readDiseaseTemplate(ut::Excel& excelApp, const InputType::Disease& disease)
{
    const QString kCaptionDefString = "Caption";

    // Initialize disease template object
    fn::DiseaseTemplate diseaseTemplate;

    // Find "Caption" header cell and go the caption cell
    QAxObject* cellCaptionHeader = excelApp.getCellWithText(kCaptionDefString, "A");
    // Set row presumably holding the caption text
    int captionRow = ut::Excel::getCellRow(cellCaptionHeader) + 1;
    QString caption = excelApp.getStringCellValue(captionRow, 1);

    // Set row presumably holding the first outcome tree ID
    int outcomeTreeIDRow = captionRow + 5;
    QString outcomeTreeID = excelApp.getStringCellValue(outcomeTreeIDRow, 1);
    bool isCongenital = false;
    int numOutcomeTrees = 0;
    if (!outcomeTreeID.isEmpty()) {
        // There is at least one outcome tree. Enter and read in all outcome tree IDs.
        while (!outcomeTreeID.isEmpty()) {
            if ((outcomeTreeID.toLower() == "congenital") || (outcomeTreeID.toLower() == "perinatal")) {
                isCongenital = true;
            }
            diseaseTemplate.addOutcomeTreeTemplateID(outcomeTreeID);
            // Move to next health outcome row
            outcomeTreeID = excelApp.getStringCellValue(++outcomeTreeIDRow, 1);
            numOutcomeTrees++;
        }
    }
    diseaseTemplate.setDisease(disease);
    diseaseTemplate.setCaption(caption);
    diseaseTemplate.setIsCongenital(isCongenital);

    int row = outcomeTreeIDRow;
    // Go over all outcome trees and collect the templates
    for (int i = 0; i < numOutcomeTrees; ++i) {

        // Initialize outcome tree template object
        fn::OutcomeTreeTemplate outcomeTreeTemplate;

        // Read health outcomes
        int healthOutcomeRow = row + 6;
        QString healthOutcomeID = excelApp.getStringCellValue(healthOutcomeRow, 1);
        if (!healthOutcomeID.isEmpty()) {
            // There is at least one health outcome. Enter and read in all health outcomes
            while (!healthOutcomeID.isEmpty()) {
                QString healthOutcomeCaption = excelApp.getStringCellValue(healthOutcomeRow, 2);
                QString healthOutcomeSequelaType = excelApp.getStringCellValue(healthOutcomeRow, 3).toUpper();
                QString healthOutcomeVisible = excelApp.getStringCellValue(healthOutcomeRow, 4).toUpper();
                outcomeTreeTemplate.addHealthOutcome(healthOutcomeID, healthOutcomeCaption, healthOutcomeSequelaType, healthOutcomeVisible);
                // Move to next health outcome row
                healthOutcomeID = excelApp.getStringCellValue(++healthOutcomeRow, 1);
            }
        }

        // Read health states
        int healthStateRow = healthOutcomeRow + 4;
        QString healthStateID = excelApp.getStringCellValue(healthStateRow, 1);
        if (!healthStateID.isEmpty()) {
            // There is at least one health state. Enter and read in all health states
            while (!healthStateID.isEmpty()) {
                QString healthStateCaption = excelApp.getStringCellValue(healthStateRow, 2);
                QString healthStateParentHSID = excelApp.getStringCellValue(healthStateRow, 3);
                QString healthStateType = excelApp.getStringCellValue(healthStateRow, 4).toUpper();
                outcomeTreeTemplate.addHealthState(healthStateID, healthStateCaption, healthStateParentHSID, healthStateType);
                // Move to next health state row
                healthStateID = excelApp.getStringCellValue(++healthStateRow, 1);
            }
        }

        // Read transition probabilities
        int transProbRow = healthStateRow + 4;
        QString transProbID = excelApp.getStringCellValue(transProbRow, 1);
        if (!transProbID.isEmpty()) {
            // There is at least one trans prob. Enter and read in all trans probs
            while (!transProbID.isEmpty()) {
                QString transProbCaption = excelApp.getStringCellValue(transProbRow, 2);
                QString transProbParentID = excelApp.getStringCellValue(transProbRow, 3);
                QString transProbChildID = excelApp.getStringCellValue(transProbRow, 4);
                QString sourceTransProbID = excelApp.getStringCellValue(transProbRow, 5);
                outcomeTreeTemplate.addTransProb(transProbID, transProbCaption, transProbParentID, transProbChildID, sourceTransProbID);
                // Move to next transition probability row
                transProbID = excelApp.getStringCellValue(++transProbRow, 1);
            }
        }

        outcomeTreeID = diseaseTemplate.getOutcomeTreeTemplateID(i);
        outcomeTreeTemplate.setId(outcomeTreeID);
        outcomeTreeTemplate.setName(outcomeTreeID);
        diseaseTemplate.addOutcomeTreeTemplate(outcomeTreeID, outcomeTreeTemplate);
        row = transProbRow;
    }

    m_diseaseTemplateMgr.addDiseaseTemplate(disease, diseaseTemplate);
    m_diseaseIds.append(disease);

}

void InputDataMgr::readFactors(ut::Excel& excelApp, const QString& inputFactorLabel, const fn::InputType::Disease& disease)
{
    // Find "Input factors" header cell and go the input factors list
    QAxObject* cellInputFactorHeader = excelApp.getCellWithText(inputFactorLabel, "A");

    // Set row presumably holding the first input factor
    int inputFactorRow = ut::Excel::getCellRow(cellInputFactorHeader) + 3;

    QString inputFactorName = excelApp.getStringCellValue(inputFactorRow, 1).trimmed();

    // If there is at least one input factor name. Enter and read in all input factors
    while (!inputFactorName.isEmpty()) {
        int inputFactorCatCol = 2;
        QString inputFactorCatName = excelApp.getStringCellValue(inputFactorRow, inputFactorCatCol);
        if (!inputFactorCatName.isEmpty()) {
            // There is at least one category for this input factor. Enter and read in all input factor categories
            QStringList inputFactorCatNameList;
            while (!inputFactorCatName.isEmpty()) {
                inputFactorCatNameList.append(inputFactorCatName);
                inputFactorCatName = excelApp.getStringCellValue(inputFactorRow, ++inputFactorCatCol);
            }
            m_inputFactorMgr.addInputFactor(disease, inputFactorName, inputFactorCatNameList);
        }

        // Move to next input factor row
        inputFactorName = excelApp.getStringCellValue(++inputFactorRow, 1);
    }
}

const InputFactorMgr& InputDataMgr::getInputFactorMgr() const
{
    return m_inputFactorMgr;
}

InputFactorMgr& InputDataMgr::getInputFactorMgr()
{
    return m_inputFactorMgr;
}

}  // namespace fn
