//==========================================================================================================================
//  'InputDataMgr' class
//			This class manages all input data
//
//  Written by Daniel Lewandowski
//  Copyright @ 2011 NextPage Software
//==========================================================================================================================

#ifndef _fn_InputDataMgr_
#define _fn_InputDataMgr_

#include <QFileInfo>
#include "fn/InputFactorMgr.h"
#include "fn/ParamsMgr.h"
#include "fn/DiseaseTemplateMgr.h"
#include "fn/TransitionParamsMgr.h"

namespace ut {

class Excel;

}

namespace fn {

class InputDataMgr : public QObject
{

    Q_OBJECT

public:

    // Constructor.
    //
    // 'filename' - [in] Full path to the reference data file
    //
    explicit InputDataMgr();

    // Destructor.
    //
    virtual ~InputDataMgr();

    // Get notification data for a given country and disease
    //
    // 'country' - [in] Country
    // 'disease' - [in] Disease
    //
    // Return - Reference to the notification data.
    const fn::DataArray& getNotificationData(const fn::InputType::Country& country, const fn::InputType::Disease& disease);

    // Get outcome tree mapping for a given country and disease
    //
    // 'country' - [in] Country
    // 'disease' - [in] Disease
    //
    // Return - Reference to the outcome tree mapping data.
    const fn::DataArray& getOutcomeTreeMappingData(const fn::InputType::Country& country, const fn::InputType::Disease& disease);

    // Get life expectancy data for a given country
    //
    // 'country' - [in] Country
    //
    // Return - Reference to the life expectancy data.
    fn::DataArray& getLifeExpectancyData(const fn::InputType::Country& country);

    // Set life expectancy data for a given country
    //
    // 'country' - [in] Country
    // 'leData' - [in] Life expectancy data
    void setLifeExpectancyData(const fn::InputType::Country& country, const fn::DataArray& leData);

    // Get age distribution data for a given country
    //
    // 'country' - [in] Country
    //
    // Return - reference to the age distribution data.
    fn::DataArray& getAgeDistributionData(const fn::InputType::Country& country);

    // Set age distribution data for a given country
    //
    // 'country' - [in] Country
    // 'ageData' - [in] Age distribution data
    void setAgeDistributionData(const fn::InputType::Country& country, const fn::DataArray& ageData);

    // Get input factor manager
    //
    // Return - Reference to the input factor manager.
    const InputFactorMgr& getInputFactorMgr() const;
          InputFactorMgr& getInputFactorMgr();

    // Import reference data from external file
    //
    void importReferenceData();

    // Check if country data exist in the database
    //
    // 'country' - [in] Country
    //
    // Return - TRUE only if exists
    bool countryDataExist(const fn::InputType::Country& country) const;

    // Get country name
    //
    // 'country' - [in] Country
    //
    // Return - Country name
    QString getCountryName(const fn::InputType::Country& country) const;

    // Check if disease data exist in the database
    //
    // 'disease' - [in] Disease
    //
    // Return - TRUE only if exists
    bool diseaseDataExist(const InputType::Disease& disease) const;

    // Check if model data for country and disease exist in the database
    //
    // 'disease' - [in] Disease
    //
    // Return - TRUE only if exists
    bool countryDiseaseCombinationAvailable(const InputType::Country& country, const InputType::Disease& disease) const;

    // Get disease template manager
    //
    // Return - Reference to the disease template manager
    const fn::DiseaseTemplateMgr& getDiseaseTemplateMgr() const;
          fn::DiseaseTemplateMgr& getDiseaseTemplateMgr();

    // Get list of country data available in the library
    //
    // Return - List of countries
    const QList<fn::InputType::Country> getAvailableCountries() const;

    // Get list of diseases data available in the library
    //
    // Return - List of diseases
    const QList<fn::InputType::Disease> getAvailableDiseases() const;

    // Get disability weights manager
    //
    // Return - Reference to the disability weights manager
    const fn::ParamsMgr& getParametersMgr() const;
          fn::ParamsMgr& getParametersMgr();

    // Get transition probability info manager
    //
    // Return - Reference to the transition probability info manager
    const fn::TransitionParamsMgr& getTransProbInfoMgr() const;
          fn::TransitionParamsMgr& getTransProbInfoMgr();

    // Get correction info of a given name (coverage, underestimation, ...)
    fn::DataArray getCorrectionWithName(const fn::InputType::Country& country,
                                                    const fn::InputType::Disease& disease,
                                                    const QString& corrName);

    // Get correction type (underestimation or underreporting with underascertainment)
    int getCorrectionType(const fn::InputType::Country& country, const fn::InputType::Disease& disease);

public slots:

    // Set path to reference data file
    //
    void setPath(const QString& appPath);

signals:

    // Signal import of Excel sheet from reference data
    void excelSheetImporting(const fn::FileOperationType&, const QString&, const QString&, const QString&);
    void showMessage(QString, QString, QString);

private:

    // Save data in binary file
    //
    void saveSerializedData();

    // Load data from binary file
    //
    void loadSerializedData();

    // Main work horse of data loading
    //
    void readData();

    // Reads disease template data
    //
    // 'excelApp' - [in] Instance of Excel holding the data
    // 'disease' - [in] Disease for which the input table is read.
    //
    void readDiseaseTemplate(ut::Excel& excelApp, const fn::InputType::Disease& disease);

    // Reads factor data.
    //
    // 'excelApp' - [in] Instance of Excel holding the data
    // 'inputFactorLabel' - [in] Label marking start of the input factor section
    // 'disease' - [in] Disease for which input factors will be stored
    //
    void readFactors(ut::Excel& excelApp, const QString& inputFactorLabel, const fn::InputType::Disease& disease);

    // Directory of current executable
    QString m_appPath;

    // Disease reference data input file info
    QFileInfo m_diseaseFileInfo;

    // Country reference data input file info
    QFileInfo m_countryFileInfo;

    // Models reference data input file infos
    QFileInfoList m_modelFileInfos;

    // DISEASE TEMPLATE DATA

    // Disease template manager
    fn::DiseaseTemplateMgr m_diseaseTemplateMgr;

    // Input factor manager
    fn::InputFactorMgr m_inputFactorMgr;

    // COUNTRY DATA

    // Country names
    fn::StringToStringMap m_countryNames;

    // Life expectancy data keyed on country
    fn::CountryToInputDataMap m_lifeExpectancyData;

    // Age distribution data keyed on country
    fn::CountryToInputDataMap m_ageDistributionData;

    // MODEL DATA
    // Parameters manager
    fn::ParamsMgr m_paramsMgr;

    // Notification data keyed on country, disease, outcome tree ID
    fn::CountryToDiseaseToInputDataMap m_notificationData;

    // Outcome tree mapping data keyed on country and disease
    fn::CountryToDiseaseToInputDataMap m_outcomeTreeMappingData;

    // Transition probability info manager
    fn::TransitionParamsMgr m_transProbInfoMgr;

    // Corrections
    fn::TableType m_correctionData;

    // Corrections type
    fn::CountryToDiseaseToStringToStringMap m_correctionType;

    // Available combinations of countries and diseases
    QMultiMap<fn::InputType::Country, fn::InputType::Disease> m_availCountryDiseasePairs;

    // List of disease Ids in the order as provided in the reference data
    QList<fn::InputType::Disease> m_diseaseIds;
};

}  // namespace fn

#endif  // _fn_InputDataMgr_
