//==========================================================================================================================
//  Implementation file for "JSONDiseaseTemplate.h"
//  Written by Daniel Lewandowski
//  Copyright @ 2019 NextPage Software
//==========================================================================================================================

#include "JSONDiseaseTemplate.h"

#include <QJsonArray>

#include "fn/ImportExportHelpers.h"

namespace fn {

JSONDiseaseTemplate::JSONDiseaseTemplate() :
    m_jsonObj(),
    m_outcomeTreeTemplateMap(),
    m_outcomeTreeIdList(),
    m_severityParameters(),
    m_durationParameters(),
    m_proportionParameters(),
    m_probabilityParameters(),
    m_redistributionParameters(),
    m_inputFactorList(),
    m_outcomeTreeMapping(),
    m_rawCasesCount()
{
}

JSONDiseaseTemplate::~JSONDiseaseTemplate()
{

}

void JSONDiseaseTemplate::setJsonObject(const QJsonObject& jsonObj)
{
    m_jsonObj = jsonObj;

    parseJsonObject();
}

const QJsonObject& JSONDiseaseTemplate::getJsonObject() const
{
    return m_jsonObj;
}

fn::InputType::Disease JSONDiseaseTemplate::getDisease() const
{
    QString diseaseStr = m_jsonObj.value(QStringLiteral("id")).toString();
    fn::InputType::Disease disease = static_cast<fn::InputType::Disease>(diseaseStr);

    return(disease);
}

QString JSONDiseaseTemplate::getCaption() const
{
    return m_jsonObj.value(QStringLiteral("name")).toString();
}

bool JSONDiseaseTemplate::getIsCongenital() const
{
    bool result = false;
    QJsonValue val = m_jsonObj.value(QStringLiteral("isCongenitalType"));
    if (!val.isUndefined()) {
        result = val.toBool();
    }

    return result;
}

int JSONDiseaseTemplate::getNumOutcomeTreeTemplates() const
{
    int result = m_outcomeTreeIdList.length();

    return result;
}

const QString JSONDiseaseTemplate::getOutcomeTreeTemplateID(const int index) const
{
    return m_outcomeTreeIdList.at(index);
}

const OutcomeTreeTemplate JSONDiseaseTemplate::getOutcomeTreeTemplate(const QString& id) const
{
    return m_outcomeTreeTemplateMap[id];
}

void JSONDiseaseTemplate::parseJsonObject()
{
    // Read input factors
    QJsonObject dimsJson = m_jsonObj.value(QStringLiteral("dims")).toObject();
    QJsonArray dimsOrderJson = dimsJson.value(QStringLiteral("order")).toArray();
    QJsonObject dimsValuesJson = dimsJson.value(QStringLiteral("values")).toObject();
    for (int i = 0; i < dimsOrderJson.size(); ++i) {
        QString inputFactorName = dimsOrderJson[i].toString();
        QStringList inputFactorCatNameList = dimsValuesJson[inputFactorName].toVariant().toStringList();
        // Capitalize
        inputFactorName.replace(0, 1, inputFactorName[0].toUpper());
        fn::InputFactor inputFactor(inputFactorName, inputFactorCatNameList);
        m_inputFactorList.append(inputFactor);
    }

    // Read average number of cases
    m_rawCasesCount = fn::readInputDataFromJSON(m_inputFactorList, m_jsonObj["rawCasesCountArray"].toObject(), "Notification data");

    // Read outcome trees
    QJsonObject otsJson = m_jsonObj.value(QStringLiteral("outcomeTrees")).toObject();

    QJsonObject::const_iterator it = otsJson.constBegin();
    for (; it != otsJson.constEnd(); ++it) {
        QJsonObject otJson = it->toObject();

        QString id = otJson["id"].toString();
        QString name = otJson["name"].toString();

        QJsonArray hosRunOrderJson = otJson["healthOutcomesRunOrder"].toArray();
        QJsonObject hosJson = otJson["healthOutcomes"].toObject();
        QJsonObject tsJson = otJson["transitions"].toObject();
        if (hosRunOrderJson.size() == 0 ||
                hosJson.length() == 0 ||
                hosRunOrderJson.size() != hosJson.length()) {
            continue;
        }

        // Initialize outcome tree template object
        fn::OutcomeTreeTemplate outcomeTreeTemplate;
        outcomeTreeTemplate.setId(id);
        outcomeTreeTemplate.setName(name);

        fn::StringTableType severityParameters;
        fn::StringTableType durationParameters;
        fn::StringTableType proportionParameters;

        // Read health outcomes
        QString hoInfectionId = "";
        for (int i = 0; i < hosRunOrderJson.size(); ++i) {
            QString hoId = hosRunOrderJson[i].toString();

            QJsonObject hoJson = hosJson[hoId].toObject();
            QString hoCaption = hoJson["name"].toString();
            QString hoSequalaType = hoJson["type"].toString().toUpper();
            QString hoVisible = QStringLiteral("TRUE");

            if (hoSequalaType.toUpper() != "RECOVERY") {
                outcomeTreeTemplate.addHealthOutcome(hoId, hoCaption, hoSequalaType, hoVisible);

                // Initialize health state type from health outcome type
                QString hsType = "";
                if (hoSequalaType == "INFECTION") {
                    hsType = "INFECTION";
                    hoInfectionId = hoId;
                } else if (hoSequalaType == "ASYMPTOMATIC") {
                    hsType = "ASYMPTOMATIC";
                } else if (hoSequalaType == "ACUTE") {
                    hsType = "YLD";
                } else if (hoSequalaType == "SEQUELA") {
                    hsType = "YLD";
                } else if (hoSequalaType == "DEATH") {
                    hsType = "YLL";
                }

                // Read health states
                QJsonObject hssJson = hoJson["healthStates"].toObject();
                QJsonObject::const_iterator hsIt = hssJson.constBegin();
                for (; hsIt != hssJson.constEnd(); ++hsIt) {
                    QJsonObject hsJson = hsIt->toObject();

                    QString hsId = hsJson["id"].toString();
                    QString hsCaption = hsJson["name"].toString();
                    bool jsRemainingCases = hsJson["remainingCases"].toBool();

                    outcomeTreeTemplate.addHealthState(hsId, hsCaption, hoId, hsType);

                    severityParameters.insert(hsId, fn::readInputDataFromJSON(m_inputFactorList, hsJson["severityArray"].toObject(), "Severity"));
                    durationParameters.insert(hsId, fn::readInputDataFromJSON(m_inputFactorList, hsJson["durationArray"].toObject(), "Duration"));
                    fn::DataArray proportionArray = fn::readInputDataFromJSON(m_inputFactorList, hsJson["proportionArray"].toObject(), "Proportion");
                    proportionArray.setMetaInfo("remaining cases", jsRemainingCases ? "yes" : "no");
                    proportionParameters.insert(hsId, proportionArray);
                }

                if (hssJson.length() == 0) {
                    QString hsId = hoId + "_hs";
                    outcomeTreeTemplate.addHealthState(hsId, hoCaption, hoId, hsType);

                    severityParameters.insert(hsId, fn::readInputDataFromJSON(m_inputFactorList, hoJson["severityArray"].toObject(), "Severity"));
                    durationParameters.insert(hsId, fn::readInputDataFromJSON(m_inputFactorList, hoJson["durationArray"].toObject(), "Duration"));
                }
            }
        }

        m_severityParameters.insert(id, severityParameters);
        m_durationParameters.insert(id, durationParameters);
        m_proportionParameters.insert(id, proportionParameters);

        // Read transitions
        fn::StringTableType probabilityParameters;
        fn::StringTableType redistributionParameters;

        QJsonObject::const_iterator tsIt = tsJson.constBegin();
        for (; tsIt != tsJson.constEnd(); ++tsIt) {
            QJsonObject tJson = tsIt->toObject();

            QString tId = tJson["id"].toString();
            QString tCaption = tJson["name"].toString();

            QString tParentId = tJson["source"].toObject()["id"].toString();
            QString tChildId = tJson["target"].toObject()["id"].toString();

            if (tParentId == hoInfectionId) {
                tParentId = tParentId + "_hs";
            }

            fn::DataArray probability = fn::readInputDataFromJSON(m_inputFactorList, tJson["probabilityArray"].toObject(), "Probability");
            fn::StringToStringMap metaData;
            metaData["redistribute"] = tJson["redistributeType"].toBool() ? "yes" : "no";
            metaData["direct type"] = tJson["annualType"].toBool() ? "no" : "yes";
            metaData["duration"] = tJson["annualDuration"].toVariant().toString();
            probability.setMetaData(metaData);
            probabilityParameters.insert(tId, probability);
            redistributionParameters.insert(tId, fn::readInputDataFromJSON(m_inputFactorList, tJson["redistributeArray"].toObject(), "Redistribution"));

            QString tSourceId = QStringLiteral("");
            bool tComplementaryType = tJson["complementaryType"].toBool();
            if (tComplementaryType) {
                tSourceId = tJson["complementaryTransition"].toString();
            }

            bool parentExists = outcomeTreeTemplate.hasHealthOutcome(tParentId) || outcomeTreeTemplate.hasHealthState(tParentId);
            bool childExists = outcomeTreeTemplate.hasHealthOutcome(tChildId) || outcomeTreeTemplate.hasHealthState(tChildId);

            if (parentExists && childExists) {
                outcomeTreeTemplate.addTransProb(tId, tCaption, tParentId, tChildId, tSourceId);
            }
        }
        m_probabilityParameters.insert(id, probabilityParameters);
        m_redistributionParameters.insert(id, redistributionParameters);

        m_outcomeTreeIdList.append(id);
        m_outcomeTreeTemplateMap.insert(id, outcomeTreeTemplate);
    }

    // Read outcome tree mapping
    m_outcomeTreeMapping = fn::readInputDataFromJSON(m_inputFactorList, m_jsonObj["outcomeTreeMapping"].toObject(), "Outcome tree mapping");
    // Remap from IDs to indices of outcome trees
    fn::StringBaseArray otMappingFormulas = m_outcomeTreeMapping.getInputFormulaArray();
    for (int i = 0; i < otMappingFormulas.size(); ++i) {
        QString otId = otMappingFormulas.at(i);
        int otIndex = m_outcomeTreeIdList.indexOf(otId);
        otMappingFormulas[i] = QString::number(otIndex + 1);
    }
    m_outcomeTreeMapping.setInputFormulaArray(otMappingFormulas);
}

fn::DataArray JSONDiseaseTemplate::getParameter(const QString& outcomeTreeId, const QString& elementId, const JSONDiseaseTemplate::ParamType& paramType) const
{
    const fn::StringToStringTableType* map;
    QString name;
    QString initFormula;
    fn::StringToStringMap metaData;
    switch (paramType)
    {
        case Severity:
            map = &m_severityParameters;
            name = "Severity";
            initFormula = "0";
            break;
        case Duration:
            map = &m_durationParameters;
            name = "Duration";
            initFormula = "0";
            break;
        case Proportion:
            map = &m_proportionParameters;
            name = "Proportion";
            initFormula = "1";
            break;
        case Probability:
            map = &m_probabilityParameters;
            name = "Probability";
            initFormula = "1";
            break;
        case Redistribution:
            map = &m_redistributionParameters;
            name = "Redistribution";
            initFormula = "0";
            break;
    }

    if (map->contains(outcomeTreeId)) {
        const fn::StringTableType& elementMap = (*map)[outcomeTreeId];
        if (elementMap.contains(elementId)) {
            return(elementMap.value(elementId));
        }
    }

    fn::DataArray defaultResult = fn::DataArray(m_inputFactorList, name, metaData, initFormula);

    return defaultResult;
}


const fn::ConstInputFactorPtrList JSONDiseaseTemplate::getConstInputFactorPtrList() const
{
    fn::ConstInputFactorPtrList inputFactors;
    QListIterator<fn::InputFactor> inputFactorIter(m_inputFactorList);
    inputFactorIter.toFront();
    while (inputFactorIter.hasNext()) {
        inputFactors.append(&(inputFactorIter.next()));
    }
    return inputFactors;
}

const fn::InputFactorList& JSONDiseaseTemplate::getInputFactorList() const
{
    return m_inputFactorList;
}

const fn::DataArray& JSONDiseaseTemplate::getOutcomeTreeMapping() const
{
    return m_outcomeTreeMapping;
}

const fn::DataArray& JSONDiseaseTemplate::getRawCasesCount() const
{
    return m_rawCasesCount;
}

}  // namespace fn

