//==========================================================================================================================
//  Implementation file for "OutcomeTree.h"
//  Written by Daniel Lewandowski
//  Copyright @ 2011 NextPage Software
//==========================================================================================================================

#include "OutcomeTree.h"

#include "ut/VariantUtils.h"

#include "fn/ArrayHelpers.h"
#include "fn/Infection.h"
#include "fn/Overloads.h"
#include "fn/HealthOutcome.h"
#include "fn/HealthState.h"
#include "fn/TransitionParamsMgr.h"
#include "fn/TransitionProbabilityComplement.h"
#include "fn/JSONDiseaseTemplate.h"
#include "ui/Chief.h"

namespace fn {

OutcomeTree::OutcomeTree(const ID& id, const int index, fn::DiseaseModel& model, const fn::OutcomeTreeTemplate& outcomeTreeTemplate) :
    m_model(model),
    m_id(id),
    m_name(),
    m_index(index),
    m_infection(nullptr),
    m_totalPopulation(0),
    m_healthOutcomeList(),
    m_healthStateMap(),
    m_healthStateList(),
    m_transProbList(),
    m_transProbComplList(),
    m_transProbMap(),
    m_healthOutcomeResults(),
    m_sequelaTypeResults(),
    m_numSamples(1)
{
    createInternalStructure(outcomeTreeTemplate);
}

OutcomeTree::~OutcomeTree()
{

}

const QString& OutcomeTree::getName() const
{
    return m_name;
}

// Get model ID.
const fn::OutcomeTree::ID& OutcomeTree::getId() const
{
    return m_id;
}

const fn::Infection& OutcomeTree::getInfection() const
{
    return *m_infection.data();
}

fn::Infection& OutcomeTree::getInfection()
{
    return *m_infection.data();
}

const fn::HealthState& OutcomeTree::getHealthState(const int index) const
{
    if (index < 0 || index > m_healthStateList.count()) {
        UT_THROW(QString("Outcome tree %1 has no health state with index %1").arg(m_id).arg(index));
    };
    return m_healthStateList[index];
}

fn::HealthState& OutcomeTree::getHealthState(const int index)
{
    if (index < 0 || index > m_healthStateList.count()) {
        UT_THROW(QString("Outcome tree %1 has no health state with index %1").arg(m_id).arg(index));
    };
    return m_healthStateList[index];
}

const fn::HealthState& OutcomeTree::getHealthState(const OutcomeTreeNode::ID healthStateId) const
{
    if (m_healthStateMap.contains(healthStateId)) {
        return static_cast<const fn::HealthState&>(*m_healthStateMap.value(healthStateId));
    } else {
        UT_THROW(QString("Outcome tree %1 has no health state with id %1").arg(m_id).arg(healthStateId));
    }
}

fn::HealthState& OutcomeTree::getHealthState(const OutcomeTreeNode::ID healthStateId)
{
    if (m_healthStateMap.contains(healthStateId)) {
        return static_cast<fn::HealthState&>(*m_healthStateMap.value(healthStateId));
    } else {
        UT_THROW(QString("Outcome tree %1 has no health state with id %1").arg(m_id).arg(healthStateId));
    }
}

int OutcomeTree::getHealthStateCount() const
{
    return m_healthStateList.count();
}

fn::TransitionProbability& OutcomeTree::getTransitionProbability(const int index)
{
    if (index < 0 || index > m_transProbList.count()) {
        UT_THROW(QString("Outcome tree %1 has no transition probability with index %1").arg(m_id).arg(index));
    };
    return m_transProbList[index];
}

const fn::TransitionProbability& OutcomeTree::getTransitionProbability(const int index) const
{
    if (index < 0 || index > m_transProbList.count()) {
        UT_THROW(QString("Outcome tree %1 has no transition probability with index %1").arg(m_id).arg(index));
    };
    return m_transProbList[index];
}

const fn::TransitionProbabilityBase& OutcomeTree::getTransitionProbability(const TransitionProbabilityBase::ID transProdId) const
{
    if (m_transProbMap.contains(transProdId)) {
        return *m_transProbMap[transProdId];
    } else {
        UT_THROW(QString("Outcome tree %1 has no transition probability with id %1").arg(m_id).arg(transProdId));
    }
}

fn::TransitionProbabilityBase& OutcomeTree::getTransitionProbability(const fn::TransitionProbabilityBase::ID transProdId)
{
    if (m_transProbMap.contains(transProdId)) {
        return *m_transProbMap[transProdId];
    } else {
        UT_THROW(QString("Outcome tree %1 has no transition probability with id %1").arg(m_id).arg(transProdId));
    }
}

int OutcomeTree::getTransProbCount() const
{
    return m_transProbList.count();
}

void OutcomeTree::createInternalStructure(const fn::OutcomeTreeTemplate& outcomeTreeTemplate)
{
    m_name = outcomeTreeTemplate.getName();

    QString otId = outcomeTreeTemplate.getId();

    QMap<QString, fn::OutcomeTreeNode*> nodes;

    // 1. Create health outcomes
    QMap<QString, fn::HealthOutcome*> healthOutcomes;

    const fn::StringToStringListMap& healthOutcomesMap = outcomeTreeTemplate.getHealthOutcomeMap();
    const QStringList& healthOutcomeIds = outcomeTreeTemplate.getHealthOutcomeRunOrder();
    QStringList::ConstIterator hoIdsIter = healthOutcomeIds.constBegin();
    for (; hoIdsIter != healthOutcomeIds.constEnd(); ++hoIdsIter) {
        QString id = *hoIdsIter;
        const QStringList& params = healthOutcomesMap[id];
        const QString& caption = params[1];
        const QString& sequelaType = params[2];
        const QString& visible = params[3];
        healthOutcomes[id] = &addHealthOutcome(
                    id,
                    caption,
                    m_model.m_dims,
                    sequelaType.toUpper() == "ACUTE",
                    visible.toUpper() == "TRUE");
        nodes[id] = static_cast<fn::OutcomeTreeNode*>(healthOutcomes.value(id));
    }

    // 2. Create health states
    QMap<QString, fn::SequelaBase*> healthStates;
    const fn::StringToStringListMap& healthStatesIds = outcomeTreeTemplate.getHealthStateMap();
    fn::StringToStringListMap::ConstIterator hsIdsIter = healthStatesIds.constBegin();
    for (; hsIdsIter != healthStatesIds.constEnd(); ++hsIdsIter) {
        QString id = hsIdsIter.key();
        const QStringList& params = hsIdsIter.value();
        const QString& caption = params[1];
        const QString& parentHOID = params[2];
        fn::HealthOutcome& parentHealthOutcome = *static_cast<fn::HealthOutcome*>(*healthOutcomes.find(parentHOID));
        const QString& type = params[3];
        if (type.toUpper() == "INFECTION") {
            m_infection = QSharedPointer<fn::Infection>(new fn::Infection(id, caption, m_model.m_dims, parentHealthOutcome));
            healthStates[id] = m_infection.data();
            m_healthStateMap.insert(id, m_infection.data());
        } else {
            if (m_model.getDiseaseTemplateType() == fn::DiseaseModel::Original) {
                fn::ParamsMgr& paramsMgr = m_model.m_chief->getInputDataMgr().getParametersMgr();
                fn::DataArray severityArray = paramsMgr.getParameter(m_model.m_country, m_model.m_disease, m_id, id, fn::ParamsMgr::Severity, m_model.getInputFactorsList());
                fn::DataArray durationArray = paramsMgr.getParameter(m_model.m_country, m_model.m_disease, m_id, id, fn::ParamsMgr::Duration, m_model.getInputFactorsList());
                fn::DataArray proportionArray = paramsMgr.getParameter(m_model.m_country, m_model.m_disease, m_id, id, fn::ParamsMgr::Proportion, m_model.getInputFactorsList());
                healthStates[id] = &addHealthState(id, caption, parentHealthOutcome, proportionArray, severityArray, durationArray, fn::getHealthStateTypeByName(type), m_model.m_dims);
            } else if (m_model.getDiseaseTemplateType() == fn::DiseaseModel::JSON) {
                fn::DataArray severityArray = m_model.getJSONDiseaseTemplate()->getParameter(otId, id, fn::JSONDiseaseTemplate::Severity);
                fn::DataArray durationArray = m_model.getJSONDiseaseTemplate()->getParameter(otId, id, fn::JSONDiseaseTemplate::Duration);
                fn::DataArray proportionArray = m_model.getJSONDiseaseTemplate()->getParameter(otId, id, fn::JSONDiseaseTemplate::Proportion);
                healthStates[id] = &addHealthState(id, caption, parentHealthOutcome, proportionArray, severityArray, durationArray, fn::getHealthStateTypeByName(type), m_model.m_dims);
            }
        }
        nodes[id] = static_cast<fn::OutcomeTreeNode*>(healthStates.value(id));
    }

    // 3. Create transition probabilities
    const fn::StringToStringListMap& transProbs = outcomeTreeTemplate.getTransProbMap();

    // First process non-complementary transitions
    fn::StringToStringListMap::ConstIterator transProbIter = transProbs.constBegin();
    for (; transProbIter != transProbs.constEnd(); ++transProbIter) {
        const QStringList& params = transProbIter.value();
        const QString& sourceTransProbID = params[4];
        if (sourceTransProbID.isEmpty()) {
            const QString& id = params[0];
            const QString& caption = params[1];
            const QString& parentID = params[2];
            const QString& childID = params[3];

            OutcomeTreeNode& parent = *nodes[parentID];
            OutcomeTreeNode& child = *nodes[childID];
            const fn::TransitionProbabilityBase* transProb = nullptr;
            if (m_model.getDiseaseTemplateType() == fn::DiseaseModel::Original) {
                fn::TransitionParamsMgr& transProbInfoMgr = m_model.m_chief->getInputDataMgr().getTransProbInfoMgr();
                const fn::DataArray probabilityArray = transProbInfoMgr.getParameter(m_model.m_country, m_model.m_disease, m_id, id, fn::TransitionParamsMgr::Probability, m_model.getInputFactorsList());
                const fn::DataArray redistributionArray = transProbInfoMgr.getParameter(m_model.m_country, m_model.m_disease, m_id, id, fn::TransitionParamsMgr::Redistribution, m_model.getInputFactorsList());
                transProb = &addTransitionProbability(id, caption, parent, child, probabilityArray, redistributionArray);
            } else if (m_model.getDiseaseTemplateType() == fn::DiseaseModel::JSON) {
                fn::DataArray probabilityArray = m_model.getJSONDiseaseTemplate()->getParameter(otId, id, fn::JSONDiseaseTemplate::Probability);
                fn::DataArray redistributionArray = m_model.getJSONDiseaseTemplate()->getParameter(otId, id, fn::JSONDiseaseTemplate::Redistribution);
                transProb = &addTransitionProbability(id, caption, parent, child, probabilityArray, redistributionArray);
            }
        }
    }

    // Determine alpha transition (from sequela type INFECTION to ACUTE)
    m_infection->determineAlpha();
    if (m_infection->getAlpha() == nullptr) {
        UT_THROW("Non-complementary transition between health outcomes of sequela types INFECTION and ACUTE does not exist");
    }

    // No process complementary transitions
    transProbIter = transProbs.constBegin();
    for (; transProbIter != transProbs.constEnd(); ++transProbIter) {
        const QStringList& params = transProbIter.value();
        const QString& sourceTransProbID = params[4];
        if (!sourceTransProbID.isEmpty()) {
            const QString& id = params[0];
            const QString& caption = params[1];
            const QString& parentID = params[2];
            const QString& childID = params[3];

            if (m_transProbMap.contains(sourceTransProbID)) {
                fn::TransitionProbability* sourceTransProb = static_cast<fn::TransitionProbability*>(m_transProbMap.value(sourceTransProbID));
                addTransitionProbabilityComplement(id, caption, *nodes[parentID], *nodes[childID], *sourceTransProb);
            } else {
                UT_THROW("Source transition probability is missing");
            }
        }
    }
}

void OutcomeTree::adjustDimensionsInAllObjects(const fn::ConstInputFactorPtrList& runFactorList)
{
    for (int i = 0; i < m_healthStateList.size(); ++i) {
        fn::HealthState& healthState = m_healthStateList[i];
        healthState.getParameter(fn::HealthState::Severity).adjustDimensions(runFactorList);
        healthState.getParameter(fn::HealthState::Duration).adjustDimensions(runFactorList);
        healthState.getParameter(fn::HealthState::Proportion).adjustDimensions(runFactorList);
    }

    for (int i = 0; i < m_transProbList.size(); ++i) {
        fn::TransitionProbability& transProb = m_transProbList[i];
        transProb.getProbability().adjustDimensions(runFactorList);
        transProb.getRedistribution().adjustDimensions(runFactorList);
    }
}

fn::HealthOutcome& OutcomeTree::addHealthOutcome(const fn::HealthOutcome::ID id, const QString& caption, const fn::Dims& dims,
                                                 const bool acuteType, const bool visible)
{
    fn::HealthOutcome healthOutcome(id, caption, dims, acuteType, visible);
    m_healthOutcomeList.append(healthOutcome);
    fn::HealthOutcome& newHealthOutcome = m_healthOutcomeList.last();
    return newHealthOutcome;
}

fn::HealthState& OutcomeTree::addHealthState(const fn::SequelaBase::ID id, const QString& caption,
                                              fn::HealthOutcome& parent, const fn::DataArray& proportion,
                                              const fn::DataArray& severity, const fn::DataArray& duration,
                                              const fn::InputType::HealthState& type, const fn::Dims& dims)
{
    // Create a new health state
    m_healthStateList.append(fn::HealthState(id, caption, proportion, severity, duration, type, dims, parent));
    // Get reference to it
    fn::HealthState& newHealthState = m_healthStateList.last();
    m_healthStateMap.insert(id, &newHealthState);
    // Add to the list of internal health states of its parent health outcome
    parent.addInternalHealthState(&newHealthState);

    return newHealthState;
}

const fn::TransitionProbabilityBase& OutcomeTree::addTransitionProbability(
        const fn::TransitionProbability::ID id,
        const QString& caption,
        OutcomeTreeNode& parent, OutcomeTreeNode& child,
        const fn::DataArray& probability, const fn::DataArray& redistribution)
{
    fn::TransitionProbability transProb(id, caption, parent, child, probability, redistribution);
    m_transProbList.append(transProb);
    fn::TransitionProbability* addedTrans = &(m_transProbList.last());
    m_transProbMap.insert(id, addedTrans);
    parent.addTransitionToChild(addedTrans);
    child.addTransitionFromParent(addedTrans);
    return *addedTrans;
}

const TransitionProbabilityBase& OutcomeTree::addTransitionProbabilityComplement(
        const TransitionProbabilityBase::ID id,
        const QString& caption,
        OutcomeTreeNode& parent,
        OutcomeTreeNode& child,
        TransitionProbability& sourceTransProb)
{
    fn::TransitionProbabilityComplement transProb(id, caption, parent, child, sourceTransProb);
    m_transProbComplList.append(transProb);
    fn::TransitionProbabilityComplement* addedTrans = &(m_transProbComplList.last());
    m_transProbMap.insert(id, addedTrans);
    parent.addTransitionToChild(addedTrans);
    child.addTransitionFromParent(addedTrans);
    return *addedTrans;
}

void OutcomeTree::runModel(const int numSamples, const double timeDiscountRate)
{
    // 1. INITIALIZATION
    m_numSamples = static_cast<PosInt>(numSamples);

    fn::Dims sampleDims = m_model.m_dims;
    sampleDims.append(m_numSamples);

    // Initialize formula parsers
    initializeAllFormulaParsers();

    // Initialize sample containers
    initializeSampling(sampleDims);

    // Invalidate health outcomes
    invalidateRemainingCasesHealthStates();

    // 2. SAMPLING
    for (PosInt i = 0; i < m_numSamples; ++i) {

        // Evaluate underestimation parameters (coverage, underestimation, underreporting, underascertainment)
        m_model.evaluateAllFormulas();

        // Evaluate the remaining parameters
        evaluateAllFormulas();

        // Get final underestimation (underestimation or underreporting * underascertainment)
        fn::NumericBaseArray underestimation = m_model.getUnderestimationSample();

        // Compute sums of child transition probabilities for all health outcomes
        for (int i = 0; i < m_healthOutcomeList.length(); ++i) {
            m_healthOutcomeList[i].computeSumProbsToChildren(m_model.m_dims);
        }

        // Compute sums of child transition probabilities for all health states
        for (int i = 0; i < m_healthStateList.length(); ++i) {
            m_healthStateList[i].computeSumProbsToChildren(m_model.m_dims);
        }

        // Compute sums of child transition probabilities in infection
        m_infection->computeSumProbsToChildren(m_model.m_dims);

        // Run calculations in the infection
        m_infection->runCalculation(m_model.m_coverage.getValueArray(), timeDiscountRate, underestimation);

        // Run calculations in the remaining health outcomes
        for (int i = 0; i < m_healthOutcomeList.length(); ++i) {
            m_healthOutcomeList[i].runCalculation(timeDiscountRate);
        }
    }

    // 3. FINALIZING

    // These results are presented in partial results views (outcome tree)
    m_infection->computeSampleMeanResults();
    for (int i = 0; i < m_transProbList.length(); ++i) {
        m_transProbList[i].computeSampleMeanResults();
    }
    for (int i = 0; i < m_transProbComplList.length(); ++i) {
        m_transProbComplList[i].computeSampleMeanResults();
    }
    for (int i = 0; i < m_healthStateList.length(); ++i) {
        m_healthStateList[i].computeSampleMeanResults();
    }

    // These are used to get aggregated results per health state
    m_infection->computeAggregatedResults();
    for (int i = 0; i < m_healthStateList.length(); ++i) {
        m_healthStateList[i].computeAggregatedResults();
    }

    // 4. COMPUTE OUTPUT RESULTS
    collectAllResults();

    // 5. RELEASE SAMPLE
    clearSample();

    // 6. Destroy formula parsers
    m_infection->destroyAllFormulaParsers();
    for (int i = 0; i < m_transProbList.length(); ++i) {
        m_transProbList[i].deinitializeAllFormulaParsers();
    }
    for (int i = 0; i < m_healthStateList.length(); ++i) {
        m_healthStateList[i].deinitializeAllFormulaParsers();
    }
}

void OutcomeTree::initializeSampling(const fn::Dims& sampleDims)
{
    m_infection->initializeSampling(sampleDims);
    for (int i = 0; i < m_transProbList.length(); ++i) {
        m_transProbList[i].initializeSampling(sampleDims);
    }
    for (int i = 0; i < m_transProbComplList.length(); ++i) {
        m_transProbComplList[i].initializeSampling(sampleDims);
    }
    for (int i = 0; i < m_healthStateList.length(); ++i) {
        m_healthStateList[i].initializeSampling(sampleDims);
    }
}

void OutcomeTree::invalidateRemainingCasesHealthStates()
{
    for (fn::HealthOutcomeList::iterator it = m_healthOutcomeList.begin(); it != m_healthOutcomeList.end(); ++it) {
        (*it).invalidateRemainingCasesHealthState();
    }
}

void OutcomeTree::clearSample()
{
    m_infection->clearSample();
    for (int i = 0; i < m_healthStateList.length(); ++i) {
        m_healthStateList[i].clearSample();
    }
    for (int i = 0; i < m_transProbList.length(); ++i) {
        m_transProbList[i].clearSample();
    }
}

void OutcomeTree::initializeAllFormulaParsers()
{
    m_infection->createAllFormulaParsers();
    for (int i = 0; i < m_healthStateList.length(); ++i) {
        m_healthStateList[i].initializeAllFormulaParsers();
    }
    for (int i = 0; i < m_transProbList.length(); ++i) {
        m_transProbList[i].initializeAllFormulaParsers();
    }
}

void OutcomeTree::deinitializeAllFormulaParsers()
{
    m_infection->destroyAllFormulaParsers();
    for (int i = 0; i < m_healthStateList.length(); ++i) {
        m_healthStateList[i].deinitializeAllFormulaParsers();
    }
    for (int i = 0; i < m_transProbList.length(); ++i) {
        m_transProbList[i].deinitializeAllFormulaParsers();
    }
}

void OutcomeTree::evaluateAllFormulas()
{
    m_infection->evaluateAllFormulas();
    for (int i = 0; i < m_healthStateList.length(); ++i) {
        m_healthStateList[i].evaluateAllFormulas();
    }
    for (int i = 0; i < m_transProbList.length(); ++i) {
        m_transProbList[i].evaluateAllFormulas();
    }
    for (int i = 0; i < m_transProbComplList.length(); ++i) {
        m_transProbComplList[i].evaluateAllFormulas();
    }
}

void OutcomeTree::collectAllResults()
{
    // Clear previous results
    m_healthOutcomeResults.clear();
    m_sequelaTypeResults.clear();

    // Initialize dimensions for arrays
    //const fn::Dims aggregatedDims = fn::Dims() << m_numSamples;
    const fn::Dims sampleDims = fn::Dims() << m_model.m_dims << m_numSamples;

    computeHealthOutcomeResults(sampleDims);
    computeSequelaResults(sampleDims);
}

void OutcomeTree::computeHealthOutcomeResults(const fn::Dims& sampleDims)
{
    const fn::SampleVal numPer = fn::SampleVal(100000);
    const fn::SampleVal unit = numPer / m_totalPopulation;
    const fn::SampleVal unitTotalPopulation = numPer / m_model.m_totalPopulation;
    const fn::NumericBaseArray& sampleNumSymptomaticCases = m_infection->getAlpha()->getSampleOutputCasesCount();
    const fn::NumericBaseArray aggrNumSymptomaticCases = fn::getAggregatedSample(sampleNumSymptomaticCases);

    // Fetch YLD and YLL and aggregate appropriately
    for (int i = 0; i < m_healthOutcomeList.length(); ++i) {
        fn::HealthOutcome& healthOutcome = m_healthOutcomeList[i];
        if ((healthOutcome.getHealthStatesType() == fn::InputType::YLD) && healthOutcome.getVisible()) {
            QVariantMap result;

            // HEALTH OUTCOME AGGREGATION
            const QString name = healthOutcome.getCaption();
            const QString acuteType = healthOutcome.getAcuteType() ? "acute" : "sequela";
            const fn::NumericBaseArray& sampleYLD = healthOutcome.getSampleYLD(sampleDims);
            const fn::NumericBaseArray& sampleYLL = healthOutcome.getSampleYLL(sampleDims);
            const fn::NumericBaseArray sampleDALY = sampleYLD + sampleYLL;
            const fn::NumericBaseArray& samplePopulationCasesCount = healthOutcome.getSamplePopulationCasesCount(sampleDims);
            const fn::NumericBaseArray& sampleDeathCasesCount = healthOutcome.getSampleDeathCasesCount(sampleDims);
            const fn::NumericBaseArray sampleAllCasesCount = samplePopulationCasesCount + sampleDeathCasesCount;

            const fn::NumericBaseArray aggrYLD = fn::getAggregatedSample(sampleYLD);
            const fn::NumericBaseArray aggrYLL = fn::getAggregatedSample(sampleYLL);
            const fn::NumericBaseArray aggrDALY = fn::getAggregatedSample(sampleDALY);
            const fn::NumericBaseArray aggrPopulationCasesCount = fn::getAggregatedSample(samplePopulationCasesCount);
            const fn::NumericBaseArray aggrDeathCasesCount = fn::getAggregatedSample(sampleDeathCasesCount);
            const fn::NumericBaseArray aggrAllCasesCount = fn::getAggregatedSample(sampleAllCasesCount);

            result.insert("name", name);
            result.insert("acuteType", acuteType);
            result.insert("yld", fn::getAggregatedStatistics(aggrYLD));
            result.insert("yldPer100k", fn::getAggregatedStatistics(aggrYLD * unit));
            result.insert("yldPer100kTotalPopulation", fn::getAggregatedStatistics(aggrYLD * unitTotalPopulation));
            result.insert("yll", fn::getAggregatedStatistics(aggrYLL));
            result.insert("yllPer100k", fn::getAggregatedStatistics(aggrYLL * unit));
            result.insert("yllPer100kTotalPopulation", fn::getAggregatedStatistics(aggrYLL * unitTotalPopulation));
            result.insert("daly", fn::getAggregatedStatistics(aggrDALY));
            result.insert("dalyPerCase", fn::getAggregatedStatistics(aggrDALY / aggrNumSymptomaticCases));
            result.insert("dalyPer100k", fn::getAggregatedStatistics(aggrDALY * unit));
            result.insert("dalyPer100kTotalPopulation", fn::getAggregatedStatistics(aggrDALY * unitTotalPopulation));
            result.insert("populationCasesCount", fn::getAggregatedStatistics(aggrPopulationCasesCount));
            result.insert("deathCasesCount", fn::getAggregatedStatistics(aggrDeathCasesCount));
            result.insert("allCasesCount", fn::getAggregatedStatistics(aggrAllCasesCount));
            result.insert("incidencePer100k", fn::getAggregatedStatistics(aggrPopulationCasesCount * unit));
            result.insert("mortalityPer100k", fn::getAggregatedStatistics(aggrDeathCasesCount * unitTotalPopulation));

            m_healthOutcomeResults.append(result);

            m_model.m_samplePopulationCasesCount += samplePopulationCasesCount;
            m_model.m_sampleDeathCasesCount += sampleDeathCasesCount;
        }
    }
}

void OutcomeTree::computeSequelaResults(const fn::Dims& sampleDims)
{
    const fn::SampleVal numPer = fn::SampleVal(100000);
    const fn::SampleVal unit = numPer / m_totalPopulation;
    const fn::SampleVal unitTotalPopulation = numPer / m_model.m_totalPopulation;
    const fn::NumericBaseArray& sampleNumSymptomaticCases = m_infection->getAlpha()->getSampleOutputCasesCount();
    const fn::NumericBaseArray aggrNumSymptomaticCases = fn::getAggregatedSample(sampleNumSymptomaticCases);

    fn::NumericBaseArray sampleAcuteYLD = fn::NumericBaseArray(sampleDims.begin(), sampleDims.end(), 0);
    fn::NumericBaseArray sampleAcuteYLL = fn::NumericBaseArray(sampleDims.begin(), sampleDims.end(), 0);
    fn::NumericBaseArray sampleSequelaYLD = fn::NumericBaseArray(sampleDims.begin(), sampleDims.end(), 0);
    fn::NumericBaseArray sampleSequelaYLL = fn::NumericBaseArray(sampleDims.begin(), sampleDims.end(), 0);

    // Fetch YLD and YLL and aggregate appropriately
    for (int i = 0; i < m_healthOutcomeList.length(); ++i) {
        fn::HealthOutcome& healthOutcome = m_healthOutcomeList[i];
        if (healthOutcome.getHealthStatesType() == fn::InputType::YLD) {
            const fn::NumericBaseArray sampleYLD = healthOutcome.getSampleYLD(sampleDims);
            const fn::NumericBaseArray sampleYLL = healthOutcome.getSampleYLL(sampleDims);

            // SEQUELA TYPE AGGREGATION
            if (healthOutcome.getAcuteType()) {
                sampleAcuteYLD += sampleYLD;
                sampleAcuteYLL += sampleYLL;
            } else {
                sampleSequelaYLD += sampleYLD;
                sampleSequelaYLL += sampleYLL;
            }
        }
    }

    const fn::NumericBaseArray sampleAcuteDALY = sampleAcuteYLD + sampleAcuteYLL;
    const fn::NumericBaseArray sampleSequelaDALY = sampleSequelaYLD + sampleSequelaYLL;

    const fn::NumericBaseArray sampleTotalYLD = sampleAcuteYLD + sampleSequelaYLD;
    const fn::NumericBaseArray sampleTotalYLL = sampleAcuteYLL + sampleSequelaYLL;
    const fn::NumericBaseArray sampleTotalDALY = sampleAcuteDALY + sampleSequelaDALY;

    m_model.m_sampleAcuteYLD += sampleAcuteYLD;
    m_model.m_sampleAcuteYLL += sampleAcuteYLL;
    m_model.m_sampleSequelaYLD += sampleSequelaYLD;
    m_model.m_sampleSequelaYLL += sampleSequelaYLL;
    m_model.m_sampleNumSymptomaticCasesCount += sampleNumSymptomaticCases;

    // Aggregate
    const fn::NumericBaseArray aggrAcuteYLD = fn::getAggregatedSample(sampleAcuteYLD);
    const fn::NumericBaseArray aggrAcuteYLL = fn::getAggregatedSample(sampleAcuteYLL);
    const fn::NumericBaseArray aggrAcuteDALY = fn::getAggregatedSample(sampleAcuteDALY);

    const fn::NumericBaseArray aggrSequelaYLD = fn::getAggregatedSample(sampleSequelaYLD);
    const fn::NumericBaseArray aggrSequelaYLL = fn::getAggregatedSample(sampleSequelaYLL);
    const fn::NumericBaseArray aggrSequelaDALY = fn::getAggregatedSample(sampleSequelaDALY);

    const fn::NumericBaseArray aggrYLD = fn::getAggregatedSample(sampleTotalYLD);
    const fn::NumericBaseArray aggrYLL = fn::getAggregatedSample(sampleTotalYLL);
    const fn::NumericBaseArray aggrDALY = fn::getAggregatedSample(sampleTotalDALY);

    // Store
    QVariantMap acuteResults;
    acuteResults.insert("yld", fn::getAggregatedStatistics(aggrAcuteYLD));
    acuteResults.insert("yldPer100k", fn::getAggregatedStatistics(aggrAcuteYLD * unit));
    acuteResults.insert("yldPer100kTotalPopulation", fn::getAggregatedStatistics(aggrAcuteYLD * unitTotalPopulation));
    acuteResults.insert("yll", fn::getAggregatedStatistics(aggrAcuteYLL));
    acuteResults.insert("yllPer100k", fn::getAggregatedStatistics(aggrAcuteYLL * unit));
    acuteResults.insert("yllPer100kTotalPopulation", fn::getAggregatedStatistics(aggrAcuteYLL * unitTotalPopulation));
    acuteResults.insert("daly", fn::getAggregatedStatistics(aggrAcuteDALY));
    acuteResults.insert("dalyPerCase", fn::getAggregatedStatistics(aggrAcuteDALY / aggrNumSymptomaticCases));
    acuteResults.insert("dalyPer100k", fn::getAggregatedStatistics(aggrAcuteDALY * unit));
    acuteResults.insert("dalyPer100kTotalPopulation", fn::getAggregatedStatistics(aggrAcuteDALY * unitTotalPopulation));

    QVariantMap sequelaResults;
    sequelaResults.insert("yld", fn::getAggregatedStatistics(aggrSequelaYLD));
    sequelaResults.insert("yldPer100k", fn::getAggregatedStatistics(aggrSequelaYLD * unit));
    sequelaResults.insert("yldPer100kTotalPopulation", fn::getAggregatedStatistics(aggrSequelaYLD * unitTotalPopulation));
    sequelaResults.insert("yll", fn::getAggregatedStatistics(aggrSequelaYLL));
    sequelaResults.insert("yllPer100k", fn::getAggregatedStatistics(aggrSequelaYLL * unit));
    sequelaResults.insert("yllPer100kTotalPopulation", fn::getAggregatedStatistics(aggrSequelaYLL * unitTotalPopulation));
    sequelaResults.insert("daly", fn::getAggregatedStatistics(aggrSequelaDALY));
    sequelaResults.insert("dalyPerCase", fn::getAggregatedStatistics(aggrSequelaDALY / aggrNumSymptomaticCases));
    sequelaResults.insert("dalyPer100k", fn::getAggregatedStatistics(aggrSequelaDALY * unit));
    sequelaResults.insert("dalyPer100kTotalPopulation", fn::getAggregatedStatistics(aggrSequelaDALY * unitTotalPopulation));

    QVariantMap totalResults;
    totalResults.insert("yld", fn::getAggregatedStatistics(aggrYLD));
    totalResults.insert("yldPer100k", fn::getAggregatedStatistics(aggrYLD * unit));
    totalResults.insert("yldPer100kTotalPopulation", fn::getAggregatedStatistics(aggrYLD * unitTotalPopulation));
    totalResults.insert("yll", fn::getAggregatedStatistics(aggrYLL));
    totalResults.insert("yllPer100k", fn::getAggregatedStatistics(aggrYLL * unit));
    totalResults.insert("yllPer100kTotalPopulation", fn::getAggregatedStatistics(aggrYLL * unitTotalPopulation));
    totalResults.insert("daly", fn::getAggregatedStatistics(aggrDALY));
    totalResults.insert("dalyPerCase", fn::getAggregatedStatistics(aggrDALY / aggrNumSymptomaticCases));
    totalResults.insert("dalyPer100k", fn::getAggregatedStatistics(aggrDALY * unit));
    totalResults.insert("dalyPer100kTotalPopulation", fn::getAggregatedStatistics(aggrDALY * unitTotalPopulation));

    m_sequelaTypeResults.insert("acute", acuteResults);
    m_sequelaTypeResults.insert("sequela", sequelaResults);
    m_sequelaTypeResults.insert("total", totalResults);
}

const QVariantList& OutcomeTree::getHealthOutcomeAggregatedResults() const
{
    return m_healthOutcomeResults;
}

const QVariantMap& OutcomeTree::getSequelaTypeAggregatedResults() const
{
    return m_sequelaTypeResults;
}

} // namespace fn
