//==========================================================================================================================
//  'OutcomeTree' class
//      Represents a single disease model, which can consist of multiple sub-model blocks and transitions.
//
//  Written by Daniel Lewandowski
//  Copyright @ 2010 NextPage Software
//==========================================================================================================================

#ifndef _fn_OutcomeTree_
#define _fn_OutcomeTree_

#include <QSharedPointer>
#include "fn/globals.h"
#include "fn/CustomDataTypes.h"
#include "fn/InputFactor.h"
#include "fn/Infection.h"
#include "fn/HealthOutcome.h"
#include "fn/HealthState.h"


namespace fn {

class DiseaseModel;
class OutcomeTreeTemplate;
class TransitionProbabilityBase;

class OutcomeTree
{

public:

    // Disease model ID data type.
    typedef QString ID;

    // Constructor.
    //
    OutcomeTree(const ID& id, const int index, fn::DiseaseModel& model, const fn::OutcomeTreeTemplate& outcomeTreeTemplate);

    // Destructor.
    //
    virtual ~OutcomeTree();

    // Get outcome tree ID.
    const ID& getId() const;

    // Get outcome tree name.
    const QString& getName() const;

    // HEALTH STATE --------------------------------------------------------------------------------------------------------

    // Get infection
    //
    // Return - Reference to infection.
    const fn::Infection& getInfection() const;
          fn::Infection& getInfection();

    // Get health state with given index
    //
    // 'index' - [in] Index of health state
    //
    // Return - Reference to health state. Exception is thrown if index is out of bounds
    const fn::HealthState& getHealthState(const int index) const;
          fn::HealthState& getHealthState(const int index);

    // Get health state
    //
    // 'healthStateId' - [in] Id of the health state
    //
    // Return - Reference to health state. Exception is thrown if Id is not recognized
    const fn::HealthState& getHealthState(const fn::OutcomeTreeNode::ID healthStateId) const;
          fn::HealthState& getHealthState(const fn::OutcomeTreeNode::ID healthStateId);

    // Get count of health states
    //
    // Return - Count
    int getHealthStateCount() const;

    // TRANSITION PROBABILITY ----------------------------------------------------------------------------------------------

    // Get transition probability. Only non-complementary transition probabilities are returned
    //
    // 'transProbId' - [in] Transition probability Id
    //
    // Return - Reference to transition probability. Exception is thrown if Id is not recognized
    const fn::TransitionProbability& getTransitionProbability(const int index) const;
          fn::TransitionProbability& getTransitionProbability(const int index);

    // Get transition probability
    //
    // 'transProbId' - [in] Transition probability Id
    //
    // Return - Reference to transition probability. Exception is thrown if Id is not recognized
    const fn::TransitionProbabilityBase& getTransitionProbability(const fn::TransitionProbabilityBase::ID transProdId) const;
          fn::TransitionProbabilityBase& getTransitionProbability(const fn::TransitionProbabilityBase::ID transProdId);

    // Get count of transition probabilities
    //
    // Return - Count
    int getTransProbCount() const;

private:

    friend class DiseaseModel;

    typedef QMap<fn::SequelaBase::ID, fn::SequelaBase*> IDtoHealthStateMap;
    typedef QMap<fn::TransitionProbabilityBase::ID, fn::TransitionProbabilityBase*> IDtoTransProbMap;

    // INTERNAL STRUCTURE --------------------------------------------------------------------------------------------------

    // Creates all sequela and transition probability objects
    //
    void createInternalStructure(const fn::OutcomeTreeTemplate& outcomeTreeTemplate);

    // Adjust run formula arrays
    void adjustDimensionsInAllObjects(const fn::ConstInputFactorPtrList& runFactorList);

    // Create health outcome
    //
    fn::HealthOutcome& addHealthOutcome(const fn::HealthOutcome::ID id, const QString& caption, const fn::Dims& dims,
                                        const bool acuteType, const bool visible);

    // Create health state
    //
    fn::HealthState& addHealthState(const fn::SequelaBase::ID id, const QString& caption,
                                    fn::HealthOutcome& parent, const fn::DataArray& proportion,
                                    const fn::DataArray& severity, const fn::DataArray& duration,
                                    const fn::InputType::HealthState& type, const fn::Dims& dims);

    // Add transition probability
    //
    const fn::TransitionProbabilityBase& addTransitionProbability(const fn::TransitionProbabilityBase::ID id,
                                                              const QString& caption, OutcomeTreeNode& parent, OutcomeTreeNode& child,
                                                              const DataArray& probability, const DataArray& redistribution);

    // Add complement transition probability
    //
    const fn::TransitionProbabilityBase& addTransitionProbabilityComplement(const fn::TransitionProbabilityBase::ID id,
                                                              const QString& caption, OutcomeTreeNode& parent, OutcomeTreeNode& child,
                                                              fn::TransitionProbability& sourceTransProb);

    // RUN -----------------------------------------------------------------------------------------------------------------

    // Perform model run
    //
    // 'numSamples' - [in] Number of samples
    //
    void runModel(const int numSamples, const double timeDiscountRate);

    // Initialize sampling
    //
    // 'sampleDims'- [in] Dimensions of the sample storage
    //
    void initializeSampling(const fn::Dims& sampleDims);

    // For all health outcomes reinitialize local references to health states holding remaining cases
    //
    void invalidateRemainingCasesHealthStates();

    // Clear samples after model run.
    //
    void clearSample();

    // Initialize all formula parsers.
    //
    void initializeAllFormulaParsers();

    // Deinitialize all formula parsers
    //
    void deinitializeAllFormulaParsers();

    // Evaluate all formula parsers. This basically creates a new sample of the inputs
    //
    void evaluateAllFormulas();

    // RESULTS -------------------------------------------------------------------------------------------------------------

    // Collect all results
    //
    void collectAllResults();

    // Process health outcome results
    //
    void computeHealthOutcomeResults(const fn::Dims& sampleDims);

    // Process sequela aggregated results
    //
    void computeSequelaResults(const fn::Dims& sampleDims);

    // Get results per health outcome
    //
    // Return - Health outcome aggregated results as Variant
    const QVariantList& getHealthOutcomeAggregatedResults() const;

    // Get results per sequela type
    //
    // Return - Sequela type aggregated results as Variant
    const QVariantMap& getSequelaTypeAggregatedResults() const;

    // PRIVATE MEMBERS -----------------------------------------------------------------------------------------------------

    // Disease model
    fn::DiseaseModel& m_model;

    // Internal ID
    ID m_id;

    // Name
    QString m_name;

    // Index among all outcome trees
    int m_index;

    // Infection
    QSharedPointer<fn::Infection> m_infection;

    // Total population
    fn::SampleVal m_totalPopulation;

    // List of health outcomes
    fn::HealthOutcomeList m_healthOutcomeList;
    IDtoHealthStateMap m_healthStateMap;

    // List of health states
    fn::HealthStateList m_healthStateList;

    // List of transition probabilities
    fn::TransitionProbabilityList m_transProbList;

    // List of complement transition probabilities
    fn::TransitionProbabilityComplementList m_transProbComplList;

    // Map of all transition probabilities
    IDtoTransProbMap m_transProbMap;

    // Health outcome results
    QVariantList m_healthOutcomeResults;

    // Sequela type results
    QVariantMap m_sequelaTypeResults;

    // Number of samples per variable
    PosInt m_numSamples;

};

} // namespace fn

#endif // _fn_OutcomeTree_
