//==========================================================================================================================
//  'OutcomeTreeNode' class
//			This class implements basic outcome tree node
//
//  Written by Daniel Lewandowski
//  Copyright @ 2012 NextPage Software
//==========================================================================================================================

#ifndef _fn_OutcomeTreeNode_
#define _fn_OutcomeTreeNode_

#include "fn/TransitionProbabilityBase.h"
#include "fn/DataArray.h"

namespace fn {

class OutcomeTreeNode 
{
public:

    // Node type
    enum NodeType {HealthOutcome = 0, HealthState};

    // Outcome tree node ID data type.
    typedef QString ID;

    // Constructor.
    //
    OutcomeTreeNode(ID id, const QString& caption, NodeType nodeType, const fn::Dims& dims);
    
    // Destructor.
    //
    virtual ~OutcomeTreeNode();

    // Get ID
    const ID getId() const;

    // Get caption
    //
    const QString& getCaption() const;

    // Get node type
    const NodeType getNodeType() const;

    // Add transition from parent
    //
    void addTransitionFromParent(fn::TransitionProbabilityBase* trans);

    // Add transition to child
    //
    void addTransitionToChild(fn::TransitionProbabilityBase* trans);

    // Get count of transitions from parent
    //
    // Return - count of parent transitions
    const int getCountTransitionFromParent() const;

    // Get count of transitions to children
    //
    // Return - count of child transitions
    const int getCountTransitionToChild() const;

    // Compute sums of all probs
    //
    void computeSumProbsToChildren(const fn::Dims& runDims);

    // Get sum of all probs
    //
    const fn::NumericBaseArray& getSumProbsToChildren() const;

    // Get input remaining life expectancy
    // Warning! This function will return output remaining life expectancy from the parent outcome tree node
    //
    // Return - input remaining life expectancy numeric array
    const fn::NumericBaseArray getInputRemLifeExp() const;

    // Get input cumulative duration
    // Warning! This function will return output cumulative duration from the parent outcome tree node
    //
    // Return - input cumulative duration numeric array
    const fn::NumericBaseArray getInputCumulativeDuration() const;

    // Get input cases count table
    //
    // Return - input cases count numeric array
    const fn::NumericBaseArray getInputCasesCount() const;

    // Get cases count
    //
    virtual const fn::NumericBaseArray getOutputCasesCount() const = 0;

    // Get remaining life expectancy
    //
    virtual const fn::NumericBaseArray getOutputRemLifeExp() const = 0;

    // Get cumulative duration
    //
    virtual const fn::NumericBaseArray getOutputCumulativeDuration() const = 0;

protected:

    // Node type
    NodeType m_nodeType;

    // ID
    ID m_id;

    // Name
    QString m_caption;

    // List of transitions from parents
    fn::TransProbPtrList m_parentTransProbs;

    // List of transitions to children
    fn::TransProbPtrList m_childTransProbs;

    // Sum of all transition probability tables leaving this health state
    fn::NumericBaseArray m_sumProbs;

    // Dimensions
    fn::Dims m_dims;

};

}  // namespace fn

#endif  // _fn_OutcomeTreeNode_
