//==========================================================================================================================
//  Implementation file for "Overloads.h"
//  Written by Daniel Lewandowski
//  Copyright @ 2011 NextPage Software
//==========================================================================================================================

#include "Overloads.h"

#include <boost/bind.hpp>
#include <boost/ref.hpp>
#include <boost/array.hpp>
#include <boost/accumulators/accumulators.hpp>
#include <boost/accumulators/statistics/mean.hpp>
#include <boost/accumulators/statistics/extended_p_square.hpp>

#include "external/marray.hxx"
#include "fn/DiseaseTemplateMgr.h"

namespace fn {

using namespace boost::accumulators;

// Disability weight manager
QDataStream& operator << (QDataStream& out, const ParamsMgr& paramsMgr)
{
    out << paramsMgr.getSeverityParameters();
    out << paramsMgr.getDurationParameters();
    out << paramsMgr.getProportionParameters();
    out << paramsMgr.getTimeToDeathParameters();
    return out;
}

QDataStream& operator >> (QDataStream& in, ParamsMgr& paramsMgr)
{
    fn::CountryToDiseaseToStringToStringTableType severityMap;
    fn::CountryToDiseaseToStringToStringTableType durationMap;
    fn::CountryToDiseaseToStringToStringTableType proportionMap;
    fn::CountryToDiseaseToStringToStringTableType timeToDeathMap;

    in >> severityMap;
    in >> durationMap;
    in >> proportionMap;
    in >> timeToDeathMap;
    paramsMgr.setSeverityParameters(severityMap);
    paramsMgr.setDurationParameters(durationMap);
    paramsMgr.setProportionParameters(proportionMap);
    paramsMgr.setTimeToDeathParameters(timeToDeathMap);

    return in;
}

// Transition probability info manager
QDataStream& operator << (QDataStream& out, const TransitionParamsMgr& transProbInfoMgr)
{
    out << transProbInfoMgr.getProbabilityParameters();
    out << transProbInfoMgr.getRedistributionParameters();
    return out;
}

QDataStream& operator >> (QDataStream& in, TransitionParamsMgr& transProbInfoMgr)
{
    fn::CountryToDiseaseToStringToStringTableType probabilityParameters;
    fn::CountryToDiseaseToStringToStringTableType redistributionParameters;

    in >> probabilityParameters;
    in >> redistributionParameters;

    transProbInfoMgr.setProbabilityParameters(probabilityParameters);
    transProbInfoMgr.setRedistributionParameters(redistributionParameters);

    return in;
}

// Input factor manager
QDataStream& operator << (QDataStream& out, const InputFactorMgr& inputFactorMgr)
{
    out << inputFactorMgr.getInputFactors();
    return out;
}

QDataStream& operator >> (QDataStream& in, InputFactorMgr& inputFactorMgr)
{
    fn::DiseaseToInputFactorListMap inputFactorMap;
    in >> inputFactorMap;
    inputFactorMgr.setInputFactors(inputFactorMap);
    return in;
}

// Input factor
QDataStream& operator << (QDataStream& out, const InputFactor& inputFactor)
{
    out << inputFactor.getName() << inputFactor.getCategoryList();
    return out;
}

QDataStream& operator >> (QDataStream& in, InputFactor& inputFactor)
{
    QString name;
    QStringList categories;
    in >> name >> categories;
    inputFactor = InputFactor(name, categories);
    return in;
}

// Outcome tree template
QDataStream& operator << (QDataStream& out, const OutcomeTreeTemplate& outcomeTreeTemplate)
{
    out << outcomeTreeTemplate.getId();
    out << outcomeTreeTemplate.getHealthOutcomeMap();
    out << outcomeTreeTemplate.getHealthStateMap();
    out << outcomeTreeTemplate.getTransProbMap();
    out << outcomeTreeTemplate.getHealthOutcomeRunOrder();
    return out;
}

QDataStream& operator >> (QDataStream& in, OutcomeTreeTemplate& outcomeTreeTemplate)
{
    QString id;
    fn::StringToStringListMap healthOutcomeMap;
    fn::StringToStringListMap healthStateMap;
    fn::StringToStringListMap transProbMap;
    QStringList healthOutcomeRunOrder;

    in >> id;
    in >> healthOutcomeMap;
    in >> healthStateMap;
    in >> transProbMap;
    in >> healthOutcomeRunOrder;

    outcomeTreeTemplate.setHealthOutcomeMap(healthOutcomeMap);
    outcomeTreeTemplate.setHealthStateMap(healthStateMap);
    outcomeTreeTemplate.setTransProbMap(transProbMap);
    outcomeTreeTemplate.setHealthOutcomeRunOrder(healthOutcomeRunOrder);

    return in;
}

// Disease template
QDataStream& operator << (QDataStream& out, const DiseaseTemplate& diseaseTemplate)
{
    out << diseaseTemplate.getDisease();
    out << diseaseTemplate.getCaption();
    out << diseaseTemplate.getOutcomeTreeIDList();
    out << diseaseTemplate.getOutcomeTreeTemplateMap();
    out << diseaseTemplate.getIsCongenital();
    return out;
}

QDataStream& operator >> (QDataStream& in, DiseaseTemplate& diseaseTemplate)
{
    fn::InputType::Disease disease;
    QString caption;
    QStringList outcomeTreeIDList;
    StringToOutcomeTreeTemplateMap outcomeTreeTemplateMap;
    bool isCongenital;

    in >> disease;
    in >> caption;
    in >> outcomeTreeIDList;
    in >> outcomeTreeTemplateMap;
    in >> isCongenital;

    diseaseTemplate.setDisease(disease);
    diseaseTemplate.setCaption(caption);
    diseaseTemplate.setOutcomeTreeIDList(outcomeTreeIDList);
    diseaseTemplate.setOutcomeTreeTemplateMap(outcomeTreeTemplateMap);
    diseaseTemplate.setIsCongenital(isCongenital);

    return in;
}

// Disease template manager
QDataStream& operator << (QDataStream& out, const DiseaseTemplateMgr& diseaseTemplateMgr)
{
    out << diseaseTemplateMgr.getDiseaseTemplates();
    return out;
}

QDataStream& operator >> (QDataStream& in, DiseaseTemplateMgr& diseaseTemplateMgr)
{
    fn::DiseaseToDiseaseTemplateMap diseaseTemplatesMap;

    in >> diseaseTemplatesMap;
    diseaseTemplateMgr.setDiseaseTemplates(diseaseTemplatesMap);

    return in;
}

// Data array
QDataStream& operator << (QDataStream& out, const fn::DataArray& dataArray)
{
    out << dataArray.getInputFactors();
    out << dataArray.getName();
    out << dataArray.getMetaData();
    out << dataArray.getInputFormulaArray();

    return out;
}

QDataStream& operator >> (QDataStream& in, fn::DataArray& dataArray)
{
    fn::InputFactorList inputFactors;
    QString name;
    fn::StringToStringMap metaData;
    QStringList formulaArray;

    in >> inputFactors;
    in >> name;
    in >> metaData;
    in >> formulaArray;

    dataArray = fn::DataArray(inputFactors, name, metaData, formulaArray);

    return in;
}

fn::NumericBaseArray operator - (const fn::SampleVal& lhs, const fn::NumericBaseArray& rhs)
{
    fn::NumericBaseArray result = rhs;
    size_t j = 0;
    fn::NumericBaseArray::const_iterator itRhs = rhs.begin();
    for (; itRhs != rhs.end(); ++itRhs) {
        result(j) = lhs - *itRhs;
        ++j;
    }
    return result;
}


fn::NumericBaseArray pow(const fn::NumericBaseArray& lhs, const fn::NumericBaseArray& rhs)
{
    fn::NumericBaseArray result = lhs;
    size_t j = 0;
    fn::NumericBaseArray::const_iterator itLhs = lhs.begin();
    fn::NumericBaseArray::const_iterator itRhs = rhs.begin();
    for (; itLhs != lhs.end(); ++itLhs) {
        result(j) = std::pow(*itLhs, *itRhs);
        ++j;
        ++itRhs;
    }
    return result;
}

fn::NumericBaseArray max(const fn::NumericBaseArray& lhs, const fn::NumericBaseArray& rhs)
{
    fn::NumericBaseArray result = lhs;
    size_t j = 0;
    fn::NumericBaseArray::const_iterator itLhs = lhs.begin();
    fn::NumericBaseArray::const_iterator itRhs = rhs.begin();
    for (; itLhs != lhs.end(); ++itLhs) {
        result(j) = std::max(*itLhs, *itRhs);
        ++j;
        ++itRhs;
    }
    return result;
}

fn::NumericBaseArray max(const fn::NumericBaseArray& lhs, const fn::SampleVal& rhs)
{
    fn::NumericBaseArray result = lhs;
    size_t j = 0;
    fn::NumericBaseArray::const_iterator itLhs = lhs.begin();
    for (; itLhs != lhs.end(); ++itLhs) {
        result(j) = std::max(*itLhs, rhs);
        ++j;
    }
    return result;
}

fn::NumericBaseArray min(const fn::NumericBaseArray& lhs, const fn::NumericBaseArray& rhs)
{
    fn::NumericBaseArray result = lhs;
    size_t j = 0;
    fn::NumericBaseArray::const_iterator itLhs = lhs.begin();
    fn::NumericBaseArray::const_iterator itRhs = rhs.begin();
    for (; itLhs != lhs.end(); ++itLhs) {
        result(j) = std::min(*itLhs, *itRhs);
        ++j;
        ++itRhs;
    }
    return result;
}

fn::NumericBaseArray min(const fn::NumericBaseArray& lhs, const fn::SampleVal& rhs)
{
    fn::NumericBaseArray result = lhs;
    size_t j = 0;
    fn::NumericBaseArray::const_iterator itLhs = lhs.begin();
    for (; itLhs != lhs.end(); ++itLhs) {
        result(j) = std::min(*itLhs, rhs);
        ++j;
    }
    return result;
}

fn::DataArray copyFilteredDataArray(const fn::DataArray& source, const fn::DataArray& filter, const fn::SampleVal condition)
{
    fn::DataArray result = source;

    const fn::NumericBaseArray& filterArray = filter.getValueArray();
    fn::NumericBaseArray::const_iterator itFilter = filterArray.begin();

    const fn::StringBaseArray& sourceArray = source.getRunFormulaArray();
    fn::StringBaseArray::const_iterator itSource = sourceArray.begin();

    fn::StringBaseArray resultFormulaArray;

    for (; itSource != sourceArray.end(); ++itSource) {
        if (*itFilter != condition) {
            resultFormulaArray.append("0");
        } else {
            resultFormulaArray.append(*itSource);
        }
        ++itFilter;
    }

    result.setRunFormulaArray(resultFormulaArray);
    return result;
}

fn::NumericBaseArray copyFilteredNumericBaseArray(const fn::NumericBaseArray& source, const fn::NumericBaseArray& filter, const fn::SampleVal condition)
{
    fn::NumericBaseArray result = source;
    size_t j = 0;

    fn::NumericBaseArray::const_iterator itFilter = filter.begin();
    fn::NumericBaseArray::const_iterator itSource = source.begin();

    for (; itSource != source.end(); ++itSource) {
        if (*itFilter != condition) {
            result(j) = 0;
        } else {
            result(j) = *itSource;
        }
        ++j;
        ++itFilter;
    }
    return result;
}

fn::NumericBaseArray getSampleMean(const fn::NumericBaseArray& data)
{
    //+ Generalize++
    fn::Dims dims = fn::getDims(data);
    size_t numSamples = dims.last();
    dims.pop_back();

    fn::NumericBaseArray result(dims.begin(), dims.end(), 0);
    if (numSamples > 0) {
        fn::Indices indices(2, 0);
        for (size_t i = 0; i < dims[0]; ++i) {
            indices[0] = i;
            for (size_t j = 0; j < dims[1]; ++j) {
                indices[1] = j;
                size_t base[] = {i, j, 0};
                size_t subShape[] = {1, 1, numSamples};
                marray::View<fn::SampleVal> v = data.view(base, subShape, marray::FirstMajorOrder);
                if (numSamples > 1) {v.squeeze();};
                fn::SampleVal sum = std::accumulate(v.begin(), v.end(), 0.0);
                result(&indices[0]) = sum/numSamples;
            }
        }
    }

    return result;
}

fn::NumericBaseArray getAggregatedSample(const fn::NumericBaseArray& data)
{
    //+ Generalize++
    fn::Dims dims = fn::getDims(data);
    size_t numSamples = dims.last();
    size_t numRows = dims[0];
    size_t numCols = dims[1];
    dims.remove(0, 2);

    fn::Indices indices(1, 0);
    fn::NumericBaseArray result(dims.begin(), dims.end(), 0);
    for (size_t i = 0; i < numSamples; ++i) {
        indices[0] = i;
        size_t base[] = {0, 0, i};
        size_t subShape[] = {numRows, numCols, 1};
        marray::View<fn::SampleVal> v = data.view(base, subShape, marray::FirstMajorOrder);
        if (numSamples > 1) {v.squeeze();};
        fn::SampleVal sum = std::accumulate(v.begin(),v.end(), 0.0);
        result(&indices[0]) = sum;
    }

    return result;
}

void storeSample(const fn::NumericBaseArray& sample, fn::NumericBaseArray& sampleContainer, size_t iteration)
{
    //+ Generalize!! 2 dimensions hard coded!!
    size_t base[] = {0, 0, iteration - 1};
    size_t subShape[] = {sample.shape(0), sample.shape(1), 1};
    marray::View<fn::SampleVal> v = sampleContainer.view(base, subShape, marray::FirstMajorOrder);
    v.squeeze();
    v = sample;
}

fn::NumericBaseArray replicate(const fn::NumericBaseArray& data, int numReplicates)
{
    //+ Generalize!!
    fn::Dims dims = fn::getDims(data);
    const PosInt n = static_cast<PosInt>(numReplicates);
    dims.push_back(n);
    fn::NumericBaseArray result(dims.begin(), dims.end(), 0);

    fn::Indices indices(3, 0);
    fn::Indices indices2(2, 0);
    for (PosInt i = 0; i < dims[0]; ++i) {
        indices[0] = i;
        indices2[0] = i;
        for (PosInt j = 0; j < dims[1]; ++j) {
            indices[1] = j;
            indices2[1] = j;
            for (PosInt k = 0; k < n; ++k) {
                indices[2] = k;
                result(&indices[0]) = data(&indices2[0]);
            }
        }
    }

    return result;
}

QVariantMap getStatistics(const fn::NumericBaseArray& data, const fn::ConstInputFactorPtrList& inputFactors)
{
    // Set up accumulators
    boost::array<fn::SampleVal, 3> probs = {0.025, 0.5, 0.975};
    typedef accumulator_set<fn::SampleVal, features<tag::mean, tag::extended_p_square> > Battery;
    Battery maleAcc(tag::extended_p_square::probabilities = probs);
    Battery femaleAcc(tag::extended_p_square::probabilities = probs);

    //+ Generalize!!
    QVariantMap result;

    QVariantList meta;
    result.insert("meta", meta);

    QVariantList arrayMale;
    QVariantList arrayFemale;

    fn::Dims dims = fn::getDims(data);
    PosInt numSamples = dims.last();

    if (inputFactors.length() == 2) {
        const QStringList& rowNames = inputFactors[0]->getCategoryList();

        // Compute and store
        if (numSamples > 1) {
            PosInt subShape[] = {1, 1, numSamples};
            PosInt rowCount = static_cast<PosInt>(rowNames.length());
            for (PosInt i = 0; i < rowCount; ++i) {
                // Female
                {
                    femaleAcc = Battery(tag::extended_p_square::probabilities = probs);

                    PosInt base[] = {i, 0, 0};
                    marray::View<fn::SampleVal> v = data.view(base, subShape, marray::FirstMajorOrder);
                    v.squeeze();
                    std::for_each(v.begin(), v.end(), boost::bind<void>(boost::ref(femaleAcc), _1));
                    QVariantMap map;
                    map["id"] = i;
                    map[inputFactors[0]->getName()] = rowNames[static_cast<int>(i)];
                    for (PosInt k = 0; k < probs.size(); ++k) {
                        map["Quantile" + QString::number(probs[k])] = extended_p_square(femaleAcc)[static_cast<int>(k)];
                    }
                    map["Mean"] = mean(femaleAcc);
                    arrayFemale.append(map);
                }

                // Male
                {
                    maleAcc = Battery(tag::extended_p_square::probabilities = probs);
                    PosInt base[] = {i, 1, 0};
                    marray::View<fn::SampleVal> v = data.view(base, subShape, marray::FirstMajorOrder);
                    v.squeeze();
                    std::for_each(v.begin(), v.end(), boost::bind<void>(boost::ref(maleAcc), _1));
                    QVariantMap map;
                    map["id"] = i;
                    map[inputFactors[0]->getName()] = rowNames[static_cast<int>(i)];
                    for (PosInt k = 0; k < probs.size(); ++k) {
                        map["Quantile" + QString::number(probs[k])] = extended_p_square(maleAcc)[static_cast<int>(k)];
                    }
                    map["Mean"] = mean(maleAcc);
                    arrayMale.append(map);
                }
            }
        }
    }

    result.insert("arrayFemale", arrayFemale);
    result.insert("arrayMale", arrayMale);
    return result;
}

QVariantMap getAggregatedStatistics(const fn::NumericBaseArray& data)
{
    //+ Generalize!!
    QVariantMap result;

    // Set up accumulators
    boost::array<fn::SampleVal, 3> probs = {0.025, 0.5, 0.975};
    accumulator_set<fn::SampleVal, features<tag::mean, tag::extended_p_square> >
            acc(tag::extended_p_square::probabilities = probs);

    PosInt numSamples = data.shape(data.dimension() - 1);

    if (numSamples > 0) {
        // Compute and store
        std::for_each(data.begin(), data.end(), boost::bind<void>(boost::ref(acc), _1));
        for (PosInt k = 0; k < probs.size(); ++k) {
            result["Quantile" + QString::number(probs[k])] = extended_p_square(acc)[static_cast<int>(k)];
        }
        result["Mean"] = mean(acc);
    } else {
        for (PosInt k = 0; k < probs.size(); ++k) {
            result["Quantile" + QString::number(probs[k])] = "-";
        }
        result["Mean"] = "-";
    }
    return result;
}

fn::SampleVal getSum2D(const fn::NumericBaseArray& data)
{
    fn::Dims dims = fn::getDims(data);

    size_t base[] = {0, 0};
    size_t shape2D[] = {dims[0], dims[1]};
    size_t shape1D[] = {dims[0] * dims[1]};
    marray::View<fn::SampleVal, true> v = data.view(base, shape2D, marray::FirstMajorOrder);
    v.reshape(shape1D, shape1D + 1);
    fn::SampleVal sum = std::accumulate(v.begin(), v.end(), 0.0);
    return sum;
}

fn::NumericBaseArray getColumnSum2D(const fn::NumericBaseArray& data)
{
    fn::NumericBaseArray result = data;

    fn::Dims dims = fn::getDims(data);

    for (size_t i = 0; i < 2; ++i) {
        size_t base[] = {0, i};
        size_t shape[] = {dims[0], 1};
        marray::View<fn::SampleVal, true> v = data.view(base, shape, marray::FirstMajorOrder);
        fn::SampleVal sum = std::accumulate(v.begin(), v.end(), 0.0);
        fn::Indices indices(2, 0);
        indices[1] = i;
        for (size_t j = 0; j < dims[0]; ++j) {
            indices[0] = j;
            result(&indices[0]) = sum;
        }
    }
    return result;
}

fn::Dims getDims(const fn::NumericBaseArray& array)
{

    size_t dimension = array.dimension();
    fn::Dims result;

    for (size_t i = 0; i < dimension; ++i) {
        result.append(array.shape(i));
    }

    return result;
}

double discountFunction(const double d, const double t)
{
    // Cashflow present value, file "Chap3_Present-value.pdf", page 49.
    if (d == 0) {
        return 0;
    } else {
        return std::pow(d + 1, 1 - t) / d;
    }
}

void applyTimeDiscountRate(const double d, fn::NumericBaseArray& duration)
{
    if (d > 0) {
        for (size_t i = 0; i < duration.size(); ++i) {
            duration(i) = discountFunction(d, 0) - discountFunction(d, duration(i));
        }
    }
}

void applyTimeDiscountRate(const double d, const fn::NumericBaseArray& start, fn::NumericBaseArray& duration)
{
    if (d > 0) {
        for (size_t i = 0; i < duration.size(); ++i) {
            duration(i) = discountFunction(d, start(i)) - discountFunction(d, start(i) + duration(i));
        }
    }
}

void filterNumericBaseArray(fn::NumericBaseArray& source, const fn::NumericBaseArray& filter, const SampleVal condition)
{
    //size_t j = 0;
    fn::NumericBaseArray::const_iterator itFilter = filter.begin();
    for (size_t j = 0; itFilter != filter.end(); ++itFilter, j++) {
        if (*itFilter != condition) {
            source(j) = 0;
        }
    }
}

}  // namespace fn

