//==========================================================================================================================
//  'ParamsMgr' class
//			This class manages input data - disability weights, duration, average time between infection and death
//
//  Written by Daniel Lewandowski
//  Copyright @ 2011 NextPage Software
//==========================================================================================================================

#ifndef _fn_ParamsMgr_
#define _fn_ParamsMgr_

#include "fn/DataArray.h"
#include "fn/CustomDataTypes.h"

namespace fn {

class ParamsMgr
{

public:

    enum ParamType {Severity, Duration, TimeToDeath, Proportion};

	// Constructor.
	//
    ParamsMgr();

	// Destructor.
	//
    virtual ~ParamsMgr();

    // Add parameter
    //
    // 'country' - [in] Country for which this disability weight is specified
    // 'disease' - [in] Disease for which this disability weight is specified
    // 'outcomeTreeID' - [in] ID of the outcome tree for which this disability weight is specified
    // 'healthStateID' - [in] ID of the health state for which this disability weight is specified
    // 'type' - [in] Disability weight type (Severity/Duration)
    // 'inputData' - [in] Disability weight input data.
    //
    void addParameter(const fn::InputType::Country& country, const fn::InputType::Disease& disease,
                             const QString& outcomeTreeID, const QString& healthStateID, const ParamType& paramType, const fn::DataArray& inputData);

    // Get parameter
    //
    // 'country' - [in] Country
    // 'disease' - [in] Disease
    // 'outcomeTreeID' - [in] ID of the outcome tree
    // 'healthStateID' - [in] ID of the health state
    // 'paramType' - [in] Parameter type (severity, duration, time to death, proportion)
    // 'defaultInputFactors' - [in] Input factors to be used to create a preinitialized table in case it does not exist
    //
    const fn::DataArray getParameter(const fn::InputType::Country& country, const fn::InputType::Disease& disease,
                                     const QString& outcomeTreeID, const QString& healthStateID, const ParamType& paramType,
                                     const fn::ConstInputFactorPtrList& defaultInputFactors);

    // Get all severity parameters
    //
    // Return - map of severity parameters
    const fn::CountryToDiseaseToStringToStringTableType& getSeverityParameters() const;
          fn::CountryToDiseaseToStringToStringTableType& getSeverityParameters();

    // Set all severity parameters
    //
    // 'severityParameters' - [in] Severity parameters
    //
    void setSeverityParameters(const fn::CountryToDiseaseToStringToStringTableType& severityParameters);

    // Get all duration parameters
    //
    // Return - map of duration parameters
    const fn::CountryToDiseaseToStringToStringTableType& getDurationParameters() const;
          fn::CountryToDiseaseToStringToStringTableType& getDurationParameters();

    // Set all duration parameters
    //
    // 'durationParameters' - [in] Duration parameters
    //
    void setDurationParameters(const fn::CountryToDiseaseToStringToStringTableType& durationParameters);

    // Get all time to death parameters
    //
    // Return - map of time to death parameters
    const fn::CountryToDiseaseToStringToStringTableType& getTimeToDeathParameters() const;
          fn::CountryToDiseaseToStringToStringTableType& getTimeToDeathParameters();

    // Set all time to death parameters
    //
    // 'timeToDeathParameters' - [in] Time to death parameters
    //
    void setTimeToDeathParameters(const fn::CountryToDiseaseToStringToStringTableType& timeToDeathParameters);

    // Get all proportion parameters
    //
    // Return - map of proportion parameters
    const fn::CountryToDiseaseToStringToStringTableType& getProportionParameters() const;
          fn::CountryToDiseaseToStringToStringTableType& getProportionParameters();

    // Set all proportion parameters
    //
    // 'proportionParameters' - [in] Proportion parameters
    //
    void setProportionParameters(const fn::CountryToDiseaseToStringToStringTableType& proportionParameters);

    // Clear internal data
    //
    void clearData();

private:	

    // Local map of Severity parameters
    fn::CountryToDiseaseToStringToStringTableType m_severityParameters;

    // Local map of Duration parameters
    fn::CountryToDiseaseToStringToStringTableType m_durationParameters;

    // Local map of Time Between Infection and Death parameters
    fn::CountryToDiseaseToStringToStringTableType m_timeToDeathParameters;

    // Local map of Proportion parameters
    fn::CountryToDiseaseToStringToStringTableType m_proportionParameters;

};

}  // namespace fn

#endif  // _fn_ParamsMgr_

