//==========================================================================================================================
//  Implementation file for "ResultsComputer.h"
//  Written by Daniel Lewandowski
//  Copyright @ 2015 NextPage Software
//==========================================================================================================================

#include "ResultsComputer.h"

#include "DiseaseModel.h"
#include "Overloads.h"

namespace fn {

ResultsComputer::ResultsComputer(const fn::DiseaseModelList& diseaseModelList) :
    m_sampleDims(),
    m_totalPopulation(0),
    m_numCongenitalModels(0),
    m_runFactorList(),
    m_sampleYLD(),
    m_sampleYLL(),
    m_sampleAgeDistr(),
    m_sampleNumSymptomaticCasesCount(),
    m_sampleDeathCasesCount(),
    m_diseaseModelList(diseaseModelList),
    m_categorizedResults(),
    m_aggregatedPageResults()
{

}

ResultsComputer::~ResultsComputer()
{

}

void ResultsComputer::initializeSample(const int numSamples)
{
    // Determine if the model set includes a congenital disease model
    QList<fn::DiseaseModel>::const_iterator diseaseModelIter = m_diseaseModelList.begin();
    m_numCongenitalModels = 0;
    fn::ConstInputFactorPtrList congRunFactorList;
    fn::ConstInputFactorPtrList nonCongRunFactorList;
    for (; diseaseModelIter != m_diseaseModelList.end(); ++diseaseModelIter)
    {
        const fn::DiseaseModel& diseaseModel = *diseaseModelIter;
        if (diseaseModel.getIsCongenital())
        {
            congRunFactorList = diseaseModel.getInputFactorsList();
            m_numCongenitalModels++;
        } else {
            nonCongRunFactorList = diseaseModel.getInputFactorsList();
        }
    }

    // Select factor list to use
    if (m_numCongenitalModels > 0) {
        m_runFactorList = congRunFactorList;
    } else {
        m_runFactorList = nonCongRunFactorList;
    }

    // Initialize sample dimensions
    m_sampleDims.clear();
    fn::ConstInputFactorPtrList::ConstIterator runFactorIter = m_runFactorList.constBegin();
    for (; runFactorIter != m_runFactorList.constEnd(); ++runFactorIter) {
        m_sampleDims.append((*runFactorIter)->getNumCategories());
    }
    m_sampleDims.append(static_cast<PosInt>(numSamples));

    // Initiliaze sample containers
    m_sampleYLD = fn::NumericBaseArray(m_sampleDims.begin(), m_sampleDims.end(), 0);
    m_sampleYLL = fn::NumericBaseArray(m_sampleDims.begin(), m_sampleDims.end(), 0);
    m_sampleAgeDistr = fn::NumericBaseArray(m_sampleDims.begin(), m_sampleDims.end(), 0);
    m_sampleNumSymptomaticCasesCount = fn::NumericBaseArray(m_sampleDims.begin(), m_sampleDims.end(), 0);
    m_sampleDeathCasesCount = fn::NumericBaseArray(m_sampleDims.begin(), m_sampleDims.end(), 0);
}

void ResultsComputer::clearResults()
{
    m_categorizedResults.clear();
    m_aggregatedPageResults.clear();
}

void ResultsComputer::clearSample()
{
    m_totalPopulation = 0;
    m_sampleYLD = fn::NumericBaseArray();
    m_sampleYLL = fn::NumericBaseArray();
    m_sampleAgeDistr = fn::NumericBaseArray();
    m_sampleNumSymptomaticCasesCount = fn::NumericBaseArray();
    m_sampleDeathCasesCount = fn::NumericBaseArray();
}

void ResultsComputer::computeAggregatedResults()
{
    clearResults();

    const fn::SampleVal numPer = fn::SampleVal(100000);
    const fn::SampleVal unit = numPer / m_totalPopulation;
    const fn::NumericBaseArray sampleUnit = numPer / m_sampleAgeDistr;
    const fn::NumericBaseArray aggrNumSymptomaticCasesCount = fn::getAggregatedSample(m_sampleNumSymptomaticCasesCount);
    const fn::NumericBaseArray aggrDeathCasesCount = fn::getAggregatedSample(m_sampleDeathCasesCount);

    // A. Categorized results for aggregated results page
    // A1a. Compute DALY, DALY per case, DALY per 100,000 per sex/age category.
    const fn::NumericBaseArray sampleYLDPerCase = m_sampleYLD / m_sampleNumSymptomaticCasesCount;
    const fn::NumericBaseArray sampleYLDPer100k = m_sampleYLD * sampleUnit;
    const fn::NumericBaseArray sampleYLDPer100kTotalPopulation = m_sampleYLD * unit;
    const fn::NumericBaseArray sampleYLLPerCase = m_sampleYLL / m_sampleNumSymptomaticCasesCount;
    const fn::NumericBaseArray sampleYLLPer100k = m_sampleYLL * sampleUnit;
    const fn::NumericBaseArray sampleYLLPer100kTotalPopulation = m_sampleYLL * unit;
    const fn::NumericBaseArray sampleDALY = m_sampleYLD + m_sampleYLL;
    const fn::NumericBaseArray sampleDALYPerCase = sampleDALY / m_sampleNumSymptomaticCasesCount;
    const fn::NumericBaseArray sampleDALYPer100k = sampleDALY * sampleUnit;
    const fn::NumericBaseArray sampleDALYPer100kTotalPopulation = sampleDALY * unit;

    // A1b. Store all statistics (YLD, YLL, DALY). YLD and YLL are displayed in the tooltips.
    m_categorizedResults.insert("yld", fn::getStatistics(m_sampleYLD, m_runFactorList));
    m_categorizedResults.insert("yldPerCase", fn::getStatistics(sampleYLDPerCase, m_runFactorList));
    m_categorizedResults.insert("yldPer100k", fn::getStatistics(sampleYLDPer100k, m_runFactorList));
    m_categorizedResults.insert("yldPer100kTotalPopulation", fn::getStatistics(sampleYLDPer100kTotalPopulation, m_runFactorList));
    m_categorizedResults.insert("yll", fn::getStatistics(m_sampleYLL, m_runFactorList));
    m_categorizedResults.insert("yllPerCase", fn::getStatistics(sampleYLLPerCase, m_runFactorList));
    m_categorizedResults.insert("yllPer100k", fn::getStatistics(sampleYLLPer100k, m_runFactorList));
    m_categorizedResults.insert("yllPer100kTotalPopulation", fn::getStatistics(sampleYLLPer100kTotalPopulation, m_runFactorList));
    m_categorizedResults.insert("daly", fn::getStatistics(sampleDALY, m_runFactorList));
    m_categorizedResults.insert("dalyPerCase", fn::getStatistics(sampleDALYPerCase, m_runFactorList));
    m_categorizedResults.insert("dalyPer100k", fn::getStatistics(sampleDALYPer100k, m_runFactorList));
    m_categorizedResults.insert("dalyPer100kTotalPopulation", fn::getStatistics(sampleDALYPer100kTotalPopulation, m_runFactorList));

    // B. Totals over all models for the aggregated resutls table
    const fn::NumericBaseArray aggrYLD = fn::getAggregatedSample(m_sampleYLD);
    const fn::NumericBaseArray aggrYLL = fn::getAggregatedSample(m_sampleYLL);
    const fn::NumericBaseArray aggrDALY = fn::getAggregatedSample(sampleDALY);
    m_aggregatedPageResults.insert("yld", fn::getAggregatedStatistics(aggrYLD));
    m_aggregatedPageResults.insert("yldPer100k", fn::getAggregatedStatistics(aggrYLD * unit));
    m_aggregatedPageResults.insert("yll", fn::getAggregatedStatistics(aggrYLL));
    m_aggregatedPageResults.insert("yllPer100k", fn::getAggregatedStatistics(aggrYLL * unit));
    m_aggregatedPageResults.insert("daly", fn::getAggregatedStatistics(aggrDALY));
    m_aggregatedPageResults.insert("dalyPerCase", fn::getAggregatedStatistics(aggrDALY / aggrNumSymptomaticCasesCount));
    m_aggregatedPageResults.insert("dalyPer100k", fn::getAggregatedStatistics(aggrDALY * unit));
    m_aggregatedPageResults.insert("cases", fn::getAggregatedStatistics(aggrNumSymptomaticCasesCount));
    m_aggregatedPageResults.insert("incidencePer100k", fn::getAggregatedStatistics(aggrNumSymptomaticCasesCount * unit));
    m_aggregatedPageResults.insert("deaths", fn::getAggregatedStatistics(aggrDeathCasesCount));
    m_aggregatedPageResults.insert("mortalityPer100k", fn::getAggregatedStatistics(aggrDeathCasesCount * unit));

    clearSample();
}

void ResultsComputer::addYLDSample(const fn::NumericBaseArray& sample)
{
    size_t base[] = {m_sampleYLD.shape(0) - sample.shape(0), 0, 0};
    size_t shape[] = {m_sampleYLD.shape(0) - base[0], m_sampleYLD.shape(1), m_sampleYLD.shape(2)};
    marray::View<fn::SampleVal> view = m_sampleYLD.view(base, shape);
    view += sample;
}

void ResultsComputer::addYLLSample(const fn::NumericBaseArray& sample)
{
    size_t base[] = {m_sampleYLL.shape(0) - sample.shape(0), 0, 0};
    size_t shape[] = {m_sampleYLL.shape(0) - base[0], m_sampleYLL.shape(1), m_sampleYLL.shape(2)};
    marray::View<fn::SampleVal> view = m_sampleYLL.view(base, shape);
    view += sample;
}

void ResultsComputer::addAgeDistrSample(const fn::NumericBaseArray& sample)
{
    size_t base[] = {m_sampleAgeDistr.shape(0) - sample.shape(0), 0, 0};
    size_t shape[] = {m_sampleAgeDistr.shape(0) - base[0], m_sampleAgeDistr.shape(1), m_sampleAgeDistr.shape(2)};
    marray::View<fn::SampleVal> view = m_sampleAgeDistr.view(base, shape);
    view += sample;
}

void ResultsComputer::addNumSymptomaticCasesCountSample(const fn::NumericBaseArray& sample)
{
    size_t base[] = {m_sampleNumSymptomaticCasesCount.shape(0) - sample.shape(0), 0, 0};
    size_t shape[] = {m_sampleNumSymptomaticCasesCount.shape(0) - base[0],
                      m_sampleNumSymptomaticCasesCount.shape(1),
                      m_sampleNumSymptomaticCasesCount.shape(2)};
    marray::View<fn::SampleVal> view = m_sampleNumSymptomaticCasesCount.view(base, shape);
    view += sample;
}

void ResultsComputer::addDeathCasesCountSample(const fn::NumericBaseArray& sample)
{
    size_t base[] = {m_sampleDeathCasesCount.shape(0) - sample.shape(0), 0, 0};
    size_t shape[] = {m_sampleDeathCasesCount.shape(0) - base[0],
                      m_sampleDeathCasesCount.shape(1),
                      m_sampleDeathCasesCount.shape(2)};
    marray::View<fn::SampleVal> view = m_sampleDeathCasesCount.view(base, shape);
    view += sample;
}

void ResultsComputer::addTotalPopulation(const fn::SampleVal& totalPopulation)
{
    m_totalPopulation += totalPopulation;
}

QVariantMap ResultsComputer::getAggregatedResults() const
{
    QVariantMap result;
    QVariantList modelResults;

    QList<fn::DiseaseModel>::const_iterator diseaseModelIter = m_diseaseModelList.begin();
    for (; diseaseModelIter != m_diseaseModelList.end(); ++diseaseModelIter)
    {
        const fn::DiseaseModel& model = *diseaseModelIter;
        QVariantMap modelResult;
        modelResult.insert("id", model.getId());
        modelResult.insert("country", model.getCountry());
        modelResult.insert("disease", model.getDisease());
        modelResult.insert("countryName", model.getCountryName());
        modelResult.insert("diseaseName", model.getDiseaseName());
        modelResult.insert("results", model.getAggregatedPageResults());
        modelResults.append(modelResult);
    }
    result.insert("modelResults", modelResults);
    result.insert("totalResults", m_aggregatedPageResults);
    result.insert("allResults", m_categorizedResults);

    return result;
}

}  // namespace fn
