//==========================================================================================================================
//  Implementation file for "TransitionProbability.h"
//  Written by Daniel Lewandowski
//  Copyright @ 2011 NextPage Software
//==========================================================================================================================
#define QT_USE_FAST_CONCATENATION
#define QT_USE_FAST_OPERATOR_PLUS

#include "fn/TransitionProbability.h"

#include "fn/globals.h"
#include "fn/Overloads.h"
#include "fn/FormulaChecker.h"

namespace fn {

TransitionProbability::TransitionProbability(const ID& id, const QString& caption,  const fn::OutcomeTreeNode& parent, const fn::OutcomeTreeNode& child,
                                             const fn::DataArray& probability, const fn::DataArray& redistribution) :
    fn::TransitionProbabilityBase(id, caption, probability.getDims(), parent, child, fn::TransitionProbabilityBase::Base),
    m_redistributeType(false),
    m_directType(true),
    m_probability(probability),
    m_redistribution(redistribution)
{
    completeMetaData();
    invalidateRedistributeType();
    invalidateDirectType();
    invalidateDuration();
}

TransitionProbability::~TransitionProbability()
{

}

bool TransitionProbability::isRedistributeType() const
{
    return m_redistributeType;
}

bool TransitionProbability::isDirectType() const
{
    return m_directType;
}

void TransitionProbability::setMetaData(const fn::StringToStringMap& metaData)
{
    m_probability.setMetaData(metaData);
    invalidateRedistributeType();
    invalidateDirectType();
    invalidateDuration();
}

const fn::DataArray& TransitionProbability::getProbability() const
{
    return m_probability;
}

fn::DataArray& TransitionProbability::getProbability()
{
    return m_probability;
}

void TransitionProbability::setProbability(const fn::DataArray& probability)
{
    m_probability = probability;
    invalidateRedistributeType();
    invalidateDirectType();
    invalidateDuration();
}

const fn::DataArray& TransitionProbability::getRedistribution() const
{
    return m_redistribution;
}

fn::DataArray& TransitionProbability::getRedistribution()
{
    return m_redistribution;
}

void TransitionProbability::setRedistribution(const fn::DataArray& redistribution)
{
    m_redistribution = redistribution;
}

void TransitionProbability::initializeAllFormulaParsers()
{
    m_probability.initializeParser();
    m_redistribution.initializeParser();
}

void TransitionProbability::deinitializeAllFormulaParsers()
{
    m_probability.deinitializeParser();
    m_redistribution.deinitializeParser();
}

void TransitionProbability::evaluateAllFormulas()
{
    m_probability.evaluateFormulas();
    m_redistribution.evaluateFormulas();
    m_iterProbability = m_probability.getValueArray();
    m_iterRedistribution = m_redistribution.getValueArray();
}

void TransitionProbability::completeMetaData()
{
    fn::StringToStringMap metaData = m_probability.getMetaData();

    if (!metaData.contains("description")) {
        metaData["description"] = "";
    }
    if (!metaData.contains("weblink")) {
        metaData["weblink"] = "";
    }
    if (!metaData.contains("redistribute")) {
        metaData["redistribute"] = "no";
    }
    if (!metaData.contains("direct type")) {
        metaData["direct type"] = "yes";
    }
    if (!metaData.contains("duration")) {
        metaData["duration"] = "-1";
    }

    m_probability.setMetaData(metaData);
}

void TransitionProbability::invalidateRedistributeType()
{
    m_redistributeType = m_probability.getMetaInfo("redistribute").toLower() == "yes";
}

void TransitionProbability::invalidateDirectType()
{
    m_directType = m_probability.getMetaInfo("direct type").toLower() == "yes";
}

void TransitionProbability::invalidateDuration()
{
    QString durationStr = m_probability.getMetaInfo("duration");
    if (fn::FormulaChecker::checkFormula(durationStr)) {
        m_duration.setFormula(durationStr);
    } else {
        m_duration.setFormula("-1");
    }
}

}  // namespace fn
