//==========================================================================================================================
//  Implementation file for "TransitionProbabilityBase.h"
//  Written by Daniel Lewandowski
//  Copyright @ 2012 NextPage Software
//==========================================================================================================================

#include "fn/TransitionProbabilityBase.h"

#include "fn/SequelaBase.h"
#include "fn/Overloads.h"

namespace fn {

TransitionProbabilityBase::TransitionProbabilityBase(const ID& id, const QString& caption, const fn::Dims& dims,
                                                     const fn::OutcomeTreeNode& parent, const fn::OutcomeTreeNode& child,
                                                     const TransProbSourceType& sourceType) :
    m_id(id),
    m_caption(caption),
    m_dims(dims),
    m_sourceType(sourceType),
    m_parent(parent),
    m_child(child),
    m_currentIteration(0),
    m_duration("-1"),
    m_iterProbability(),
    m_outputCasesCount(),
    m_outputRemLifeExp(),
    m_outputCumulativeDuration(),
    m_sampleProbability(),
    m_sampleRedistribution(),
    m_sampleOutputCasesCount(),
    m_sampleMeanProbability(),
    m_sampleMeanRedistribution(),
    m_sampleMeanOutputCasesCount()
{
    
}

TransitionProbabilityBase::~TransitionProbabilityBase()
{
    
}

const TransitionProbabilityBase::ID TransitionProbabilityBase::getId() const
{
    return m_id;
}

const QString& TransitionProbabilityBase::getCaption() const
{
    return m_caption;
}

const fn::Dims& TransitionProbabilityBase::getDims() const
{
    return m_dims;
}

const fn::TransitionProbabilityBase::TransProbSourceType& TransitionProbabilityBase::getTransProbSourceType() const
{
    return m_sourceType;
}

const fn::OutcomeTreeNode& TransitionProbabilityBase::getParent() const
{
    return m_parent;
}

const fn::OutcomeTreeNode& TransitionProbabilityBase::getChild() const
{
    return m_child;
}

void TransitionProbabilityBase::initializeSampling(const fn::Dims& sampleDims)
{
    m_currentIteration = 0;
    m_sampleProbability = fn::NumericBaseArray(sampleDims.begin(), sampleDims.end(), 0);
    m_sampleRedistribution = fn::NumericBaseArray(sampleDims.begin(), sampleDims.end(), 0);
    m_sampleOutputCasesCount = fn::NumericBaseArray(sampleDims.begin(), sampleDims.end(), 0);
}

void TransitionProbabilityBase::clearSample()
{
    m_sampleProbability = fn::NumericBaseArray();
    m_sampleRedistribution = fn::NumericBaseArray();
    m_sampleOutputCasesCount = fn::NumericBaseArray();
}

const fn::NumericBaseArray& TransitionProbabilityBase::getIterProbability() const
{
    return m_iterProbability;
}

const fn::NumericBaseArray& TransitionProbabilityBase::getIterRedistribution() const
{
    return m_iterRedistribution;
}

const fn::NumericBaseArray& TransitionProbabilityBase::getOutputCasesCount() const
{
    return m_outputCasesCount;
}

const fn::NumericBaseArray& TransitionProbabilityBase::getOutputRemLifeExp() const
{
    return m_outputRemLifeExp;
}

const fn::NumericBaseArray& TransitionProbabilityBase::getOutputCumulativeDuration() const
{
    return m_outputCumulativeDuration;
}

void TransitionProbabilityBase::runCalculation()
{
    m_currentIteration++;

    // Get data from the parent first
    m_outputRemLifeExp = m_parent.getOutputRemLifeExp();
    m_outputCumulativeDuration = m_parent.getOutputCumulativeDuration();
    fn::NumericBaseArray inputCasesCount = m_parent.getOutputCasesCount();

    // Deal with Direct/Cumulative type
    if (isDirectType()) {
        m_outputCasesCount = inputCasesCount * m_iterProbability;
    } else {
        fn::NumericBaseArray lifeExp = m_outputRemLifeExp;
        fn::SampleVal duration = m_duration.getValue();
        if (duration > -1) {
            lifeExp = fn::min(lifeExp, duration);
        }
        fn::NumericBaseArray sumTransProbs = m_parent.getSumProbsToChildren();
        fn::NumericBaseArray cumulativeRisk = (1.0 - fn::pow(1 - sumTransProbs, lifeExp)) * m_iterProbability / sumTransProbs;
        m_outputCasesCount = inputCasesCount * cumulativeRisk;
    }

    // Deal with redistribution
    if (isRedistributeType()) {
        m_outputCasesCount = fn::getColumnSum2D(m_outputCasesCount) * m_iterRedistribution;
    }

    // Store samples
    storeSample(m_iterProbability, m_sampleProbability, m_currentIteration);
    storeSample(m_iterRedistribution, m_sampleRedistribution, m_currentIteration);
    storeSample(m_outputCasesCount, m_sampleOutputCasesCount, m_currentIteration);
}

void TransitionProbabilityBase::computeSampleMeanResults()
{
    m_sampleMeanProbability = fn::getSampleMean(m_sampleProbability);
    m_sampleMeanRedistribution = fn::getSampleMean(m_sampleRedistribution);
    m_sampleMeanOutputCasesCount = fn::getSampleMean(m_sampleOutputCasesCount);
}

const fn::NumericBaseArray& TransitionProbabilityBase::getSampleOutputCasesCount() const
{
    return m_sampleOutputCasesCount;
}

const fn::NumericBaseArray& TransitionProbabilityBase::getSampleMeanOutputCasesCount() const
{
    return m_sampleMeanOutputCasesCount;
}

const fn::NumericBaseArray& TransitionProbabilityBase::getSampleMeanProbability() const
{
    return m_sampleMeanProbability;
}

const fn::NumericBaseArray& TransitionProbabilityBase::getSampleMeanRedistribution() const
{
    return m_sampleMeanRedistribution;
}

}  // namespace fn

