//==========================================================================================================================
//  'TransitionProbabilityBase' class
//			This class...
//
//  Written by Daniel Lewandowski
//  Copyright @ 2012 NextPage Software
//==========================================================================================================================

#ifndef _fn_TransitionProbabilityBase_
#define _fn_TransitionProbabilityBase_

#include "fn/globals.h"

#include "fn/DataValue.h"

namespace fn {

class DataArray;
class OutcomeTreeNode;

class TransitionProbabilityBase 
{
public:

    // Transition probability types
    enum TransProbSourceType {Base = 0, Complement};

    // Transition probability ID data type.
    typedef QString ID;

    // Constructor.
    //
    TransitionProbabilityBase(const ID& id, const QString& caption, const fn::Dims& dims,
                              const fn::OutcomeTreeNode& parent, const fn::OutcomeTreeNode& child,
                              const TransProbSourceType& sourceType);
    
    // Destructor.
    //
    virtual ~TransitionProbabilityBase();

    // Get Id
    //
    // Return - Id
    const ID getId() const;

    // Get caption
    //
    // Return - caption
    const QString& getCaption() const;

    // Get dimensions used
    //
    const fn::Dims& getDims() const;

    // Get transition probability source type
    //
    // Return - source type
    const TransProbSourceType& getTransProbSourceType() const;

    // Get parent node
    //
    // Return - parent node
    const fn::OutcomeTreeNode& getParent() const;

    // Get child node
    //
    // Return - child node
    const fn::OutcomeTreeNode& getChild() const;

    // Initialize sample containers to the specified size
    //
    // 'dims' - [in] Dimensions
    //
    void initializeSampling(const fn::Dims& sampleDims);

    // Clear sample
    //
    void clearSample();

    // Run calculation
    //
    void runCalculation();

    // Get current transition probability table
    //
    const fn::NumericBaseArray& getIterProbability() const;

    // Get current redistribution table
    //
    const fn::NumericBaseArray& getIterRedistribution() const;

    // Get output cases count table
    //
    // Return - output data array
    const fn::NumericBaseArray& getOutputCasesCount() const;

    // Get output remaining life expectancy table
    //
    // Return - output data array
    const fn::NumericBaseArray& getOutputRemLifeExp() const;

    // Get output cumulative duration
    //
    // Return - output data array
    const fn::NumericBaseArray& getOutputCumulativeDuration() const;

    // Get sample cases count
    //
    const fn::NumericBaseArray& getSampleOutputCasesCount() const;

    // Compute mean sample results
    //
    void computeSampleMeanResults();

    // Get sample mean cases count
    //
    const fn::NumericBaseArray& getSampleMeanOutputCasesCount() const;

    // Get sample mean probability
    //
    const fn::NumericBaseArray& getSampleMeanProbability() const;

    // Get sample mean redistribution
    //
    const fn::NumericBaseArray& getSampleMeanRedistribution() const;

    //----------------------------------------------------
    // Pure virtual methods

    // Generate a new sample for the transition probability
    //
    virtual void evaluateAllFormulas() = 0;

    // Return TRUE is transition probability is age dependent
    virtual bool isRedistributeType() const = 0;

    // Return type of transition probability
    virtual bool isDirectType() const = 0;

protected:

    // Transition probability source type
    TransProbSourceType m_sourceType;

    // ID
    ID m_id;

    // Name
    QString m_caption;

    // Dimensions
    fn::Dims m_dims;

    // Parent sequela
    const fn::OutcomeTreeNode& m_parent;

    // Child sequela
    const fn::OutcomeTreeNode& m_child;

    // Overwrite duration in cumulative transition probability if supplied
    fn::DataValue m_duration;

    // ----------------------------------------------
    // Single iteration data storage
    int m_currentIteration;                    // Current iteration
    fn::NumericBaseArray m_iterProbability;    // Transition probability table
    fn::NumericBaseArray m_iterRedistribution; // Redistribution table
    fn::NumericBaseArray m_outputCasesCount;   // Output cases count
    fn::NumericBaseArray m_outputRemLifeExp;   // Output remaining life expectancy
    fn::NumericBaseArray m_outputCumulativeDuration; // Output cumulative duration

    // ----------------------------------------------
    // Sample storage
    fn::NumericBaseArray m_sampleProbability;
    fn::NumericBaseArray m_sampleRedistribution;
    fn::NumericBaseArray m_sampleOutputCasesCount;

    // ----------------------------------------------
    // Sample mean output
    fn::NumericBaseArray m_sampleMeanProbability;
    fn::NumericBaseArray m_sampleMeanRedistribution;
    fn::NumericBaseArray m_sampleMeanOutputCasesCount;

};

// List of pointers to transition probabilities.
typedef QList<TransitionProbabilityBase*> TransProbPtrList;

}  // namespace fn

#endif  // _fn_TransitionProbabilityBase_


