/* This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this file,
 * You can obtain one at http://mozilla.org/MPL/2.0/. */

/*
 * flot-tickrotor: flot plugin to display angled X-axis tick labels.
 *
 * Requires flot 0.7 or higher and a browser supporting <canvas>.
 *
 * To activate, just set xaxis.rotateTicks to an angle in degrees.  Labels
 * are rotated clockwise, so if you want the labels to angle up and to the
 * right (/) you need to provide an angle > 90.  The text will be flipped so
 * that it is still right-side-up.
 * Angles greater than or equal to 180 are ignored.
 */
(function ($) {
  var options = {};

  function init(plot) {
    // Taken from flot-axislabels.
    // This is kind of a hack. There are no hooks in Flot between
    // the creation and measuring of the ticks (setTicks, measureTickLabels
    // in setupGrid() ) and the drawing of the ticks and plot box
    // (insertAxisLabels in setupGrid() ).
    //
    // Therefore, we use a trick where we run the draw routine twice:
    // the first time to get the tick measurements, so that we can change
    // them, and then have it draw it again.

    var ticks = [];  // preserve between draw() calls.
    var font;
    var secondPass = false;
    var rotateTicks, rotateTicksRads, radsAboveHoriz;
    var i;

    plot.hooks.draw.push(function (plot, ctx) {
      var xaxis;  // for convenience
      if (!secondPass) {
        var opts = plot.getAxes().xaxis.options;
        if (opts.rotateTicks === undefined) {
          return;
        }

        font = opts.rotateTicksFont;
        if (!font) {
          font = $('.tickLabel').css('font');
        }
        if (!font) {
          font = 'smaller sans-serif';
        }

        var elem, maxLabelWidth = 0, maxLabelHeight = 0, minX = 0, maxX = 0;

        // We have to clear the ticks option so that flot core
        // doesn't draw ticks superimposed with ours, but we preserve
        // the tick data as xaxis.rotatedTicks so that external code
        // can still get to it.

        // FIXME: It would obviously be better to just interrupt
        // the drawing of the ticks and preserve the 'ticks'
        // property.  That probably requires another hook.

        xaxis = plot.getAxes().xaxis;
        ticks = plot.getAxes().xaxis.ticks;
        xaxis.rotatedTicks = ticks;
        opts.ticks = [];  // we'll make our own

        var x;
        var body = $('body');
        for (i = 0; i < ticks.length; i++) {
          elem = $('<span style="font:' + font + '">' + ticks[i].label + '</span>');
          body.append(elem);
          ticks[i].height = elem.outerHeight(true);
          ticks[i].width = elem.outerWidth(true);
          elem.remove();
          if (ticks[i].width > maxLabelWidth) {
            maxLabelWidth = ticks[i].width;
          }
        }

        // Center such that the top of the label is at the center of the tick.
        opts.labelHeight = Math.ceil(maxLabelWidth);

        // re-draw with new label widths and heights
        secondPass = true;
        plot.setupGrid();
        plot.draw();
      } else {
        if (ticks.length === 0) {
          return;
        }
        xaxis = plot.getAxes().xaxis;
        var box = xaxis.box;
        var tick, label, xoffset, yoffset;
        for (i = 0; i < ticks.length; i++) {
          tick = ticks[i];
          if (!tick.label) {
            continue;
          }
          ctx.save();
          ctx.font = font;
          // Center such that the top of the label is at the center of the tick.
          xoffset = Math.floor(tick.height / 2);
          yoffset = Math.ceil(tick.width);
          ctx.translate(Math.round(plot.getPlotOffset().left + xaxis.p2c(tick.v)) + xoffset,
            box.top + box.padding + plot.getOptions().grid.labelMargin + yoffset);
          ctx.rotate(-Math.PI / 2);
          ctx.fillText(tick.label, 0, 0);
          ctx.restore();
        }
      }
    });
  }

  $.plot.plugins.push({
    init: init,
    options: options,
    name: 'tickRotor',
    version: '1.0'
  });
})(jQuery);
