/* global
    $, dispFloat, bcode, tooltipContentFormatter, capitalise
*/

/*
 * All charting functionality goes here
 */

var chartingModule = function () { // eslint-disable-line no-unused-vars

  // Default chart options
  let _defaultChartOptions = {
    canvas: true,
    points: {
      show: false
    },
    lines: {
      show: false
    },
    bars: {
      barWidth: 0.9,
      fill: true,
      lineWidth: 1,
      fillColor: {
        colors: [{
          opacity: 0.6
        }, {
          opacity: 0.8
        }]
      }
    },
    xaxis: {
      color: "#dddddd",
      autoscaleMargin: 0.1,
      axisLabelUseCanvas: true,
      axisLabelPadding: 20,
      axisLabelFontColor: "#555555",
      font: {
        size: 11,
        color: "#555555"
      }
    },
    yaxis: {
      color: "#dddddd",
      autoscaleMargin: 0.1,
      axisLabelUseCanvas: true,
      axisLabelPadding: 20,
      axisLabelFontColor: "#555555",
      font: {
        size: 11,
        color: "#555555"
      }
    },
    grid: {
      aboveData: false,
      borderWidth: 0,
      margin: 10,
      axisMargin: 10,
      minBorderMargin: 0,
      hoverable: true,
      clickable: false,
      autoHighlight: true
    },
    legend: {
      show: false,
      noColumns: 4
    }
  };

  // Standard tick formatter
  var _defaultTickFormatter = function (val) {
    return dispFloat(val);
  };

  // Nice upper bound
  var _niceUpperBound = function niceUpperBound(val, tickCount) {
    var tickCountLocal = tickCount;
    if (typeof (tickCount) == "undefined") {
      tickCountLocal = 10;
    }
    var unroundedTickSize = val / (tickCountLocal - 1);
    var pow10x = Math.pow(10, Math.ceil(Math.log(unroundedTickSize) / Math.log(10) - 1));
    var roundedTickSize = Math.ceil(unroundedTickSize / pow10x) * pow10x;
    return roundedTickSize * tickCountLocal;
  };

  var _plotAggrResults3Chart = function (data, sortItem, sortBy) {

    var chartContainer = $("#aggrResults3Chart");
    var dataLabel = bcode.maps.items[sortItem];

    var _chartTooltip = function (elem, event, coords, item, data) {
      // Grab the API reference
      var tooltip = elem.data("tooltip");

      // If we weren't passed the item object, hide the tooltip and remove cached point data
      if (!item) {
        tooltip.cachedPoint = null;
        tooltip.hide();
        return false;
      }

      // Proceed only if the data point has changed
      if (tooltip.cachedPoint !== item.dataIndex) {

        // Update the cached point data
        tooltip.cachedPoint = item.dataIndex;

        var localData = data[item.seriesIndex];

        var contentObject = {
          label: [localData.countryName, localData.diseaseName, localData.id].join(" - "),
          entries: [{
            label: dataLabel + " Mean",
            value: localData.results[sortItem].Mean
          },
          {
            label: dataLabel + " 2.5%",
            value: localData.results[sortItem]["Quantile0.025"]
          },
          {
            label: dataLabel + " 50%",
            value: localData.results[sortItem]["Quantile0.5"]
          },
          {
            label: dataLabel + " 97.5%",
            value: localData.results[sortItem]["Quantile0.975"]
          }
          ]
        };

        // Update the tooltip content and show
        tooltip.updateContent(tooltipContentFormatter(contentObject));
        tooltip.show({
          left: coords.pageX,
          top: coords.pageY
        });
      }
    };

    var locData = data.map(function (elem, index) {
      var newElem = elem;
      newElem.color = index;
      return newElem;
    });

    // Get data order
    var ranking = locData.rank(function (a, b) {
      return b.results[sortItem][sortBy] - a.results[sortItem][sortBy];
    });

    var errorData = locData.map(function (elem, index) {
      return [index + 1, elem.results[sortItem].Mean, elem.results[sortItem].Mean - elem.results[sortItem]["Quantile0.025"], elem.results[sortItem]["Quantile0.975"] - elem.results[sortItem].Mean];
    });
    errorData.forEach(function (elem, index, array) {
      array[ranking[index]][0] = index + 1;
    });
    var maxYAxis = _niceUpperBound(errorData.map(function (elem) {
      return elem[1] + elem[3];
    }).max(), 50);

    var errorObject = {
      label: "DALY error",
      points: {
        errorbars: "y",
        yerr: {
          show: true,
          asymmetric: true,
          upperCap: "-",
          lowerCap: "-"
        },
        lineWidth: 1
      },
      bars: {
        show: false
      },
      shadowSize: 1,
      lineWidth: 1,
      color: "red",
      data: errorData
    };

    var barData = locData.map(function (elem, index) {
      return {
        label: elem.countryName + " - " + elem.diseaseName + " - " + elem.id,
        color: elem.color,
        data: [
          [index + 1, elem.results[sortItem].Mean]
        ],
        id: elem.id
      };
    });
    barData.forEach(function (elem, index, array) {
      array[ranking[index]].data[0][0] = index + 1;
    });
    var xAxisTicks = barData.map(function (elem, index) {
      return [index + 1, elem.label];
    });
    xAxisTicks.forEach(function (elem, index, array) {
      array[ranking[index]][0] = index + 1;
    });

    var chartData = barData.concat([errorObject]);

    var aggrResults3ChartOptions = {
      bars: {
        show: true,
        align: "center"
      },
      xaxis: {
        min: 0.25,
        max: Math.max(locData.length + 0.75, 7.75),
        ticks: xAxisTicks,
        rotateTicksFont: "11px arial, sans-serif",
        rotateTicks: true
      },
      yaxis: {
        min: 0,
        max: maxYAxis,
        axisLabel: "Disease burden [" + dataLabel + "]",
        tickFormatter: _defaultTickFormatter
      }
    };

    aggrResults3ChartOptions = $.extend(true, {}, _defaultChartOptions, aggrResults3ChartOptions);
    $.plot(chartContainer, chartData, aggrResults3ChartOptions);

    chartContainer.off("plothover");
    chartContainer.tooltip({
      automatic: false
    }).on("plothover", function (event, coords, item) {
      _chartTooltip($(this), event, coords, item, data);
    });

    return null;
  };

  var _plotPieChart = function (chartContainerId, legendContainerId, data) {
    var _getDALYPieChartTooltip = function (elem, event, coords, item, data) {
      // Grab the API reference
      var tooltip = elem.data("tooltip");

      // If we weren't passed the item object, hide the tooltip and remove cached point data
      if (!item) {
        tooltip.cachedPoint = null;
        tooltip.hide();
        return false;
      }

      // Proceed only if the data point has changed
      if (tooltip.cachedPoint !== item.seriesIndex) {
        // Update the cached point data
        tooltip.cachedPoint = item.seriesIndex;

        var contentObject = {
          label: item.series.label + " - " + capitalise(data[item.seriesIndex].acuteType),
          entries: [{
            label: "DALY Mean",
            value: data[item.seriesIndex].daly.Mean
          },
          {
            label: "DALY 2.5%",
            value: data[item.seriesIndex].daly["Quantile0.025"]
          },
          {
            label: "DALY 97.5%",
            value: data[item.seriesIndex].daly["Quantile0.975"]
          },
          ]
        };

        // Update the tooltip content and show
        tooltip.updateContent(tooltipContentFormatter(contentObject));
        tooltip.show({
          left: coords.pageX,
          top: coords.pageY
        });
      }
    };

    var chartContainer = $(chartContainerId);
    var rankDesc = data.map(function (elem) {
      return elem.daly.Mean;
    }).rank(function (a, b) {
      return b - a;
    });
    data = data.order(rankDesc);
    var chartDataSet = data.map(function (elem) {
      return {
        label: elem.name,
        data: elem.daly.Mean
      };
    });

    var chartOptions = {
      series: {
        pie: {
          show: true,
          radius: 1,
          label: {
            show: false
          }
        }
      },
      legend: {
        show: true,
        container: $(legendContainerId)
      }
    };
    $.plot(chartContainer, chartDataSet, $.extend(true, {}, _defaultChartOptions, chartOptions));

    chartContainer.off("plothover");
    chartContainer.tooltip({
      automatic: false
    }).on("plothover", function (event, coords, item) {
      _getDALYPieChartTooltip($(this), event, coords, item, data);
    });
  };

  var _getDALYBarChartTooltip = function (elem, event, coords, item, data) {
    // Grab the API reference
    var tooltip = elem.data("tooltip");

    // If we weren't passed the item object, hide the tooltip and remove cached point data
    if (!item) {
      tooltip.cachedPoint = null;
      tooltip.hide();
      return false;
    }

    // Proceed only if the data point has changed
    if (tooltip.cachedPoint !== item.dataIndex) {

      // Update the cached point data
      tooltip.cachedPoint = item.dataIndex;

      var contentObject = {};
      switch (item.dataIndex) {
        case 0:
          contentObject = {
            label: "Acute",
            entries: [{
              label: "YLD Mean",
              value: data.acute.yld.Mean
            },
            {
              label: "YLL Mean",
              value: data.acute.yll.Mean
            },
            {
              label: "DALY Mean",
              value: data.acute.daly.Mean
            },
            {
              label: "DALY 2.5%",
              value: data.acute.daly["Quantile0.025"]
            },
            {
              label: "DALY 97.5%",
              value: data.acute.daly["Quantile0.975"]
            },
            ]
          };
          break;
        case 1:
          contentObject = {
            label: "Sequela",
            entries: [{
              label: "YLD Mean",
              value: data.sequela.yld.Mean
            },
            {
              label: "YLL Mean",
              value: data.sequela.yll.Mean
            },
            {
              label: "DALY Mean",
              value: data.sequela.daly.Mean
            },
            {
              label: "DALY 2.5%",
              value: data.sequela.daly["Quantile0.025"]
            },
            {
              label: "DALY 97.5%",
              value: data.sequela.daly["Quantile0.975"]
            },
            ]
          };
          break;
        case 2:
          contentObject = {
            label: "Total",
            entries: [{
              label: "YLD Mean",
              value: data.total.yld.Mean
            },
            {
              label: "YLL Mean",
              value: data.total.yll.Mean
            },
            {
              label: "DALY Mean",
              value: data.total.daly.Mean
            },
            {
              label: "DALY 2.5%",
              value: data.total.daly["Quantile0.025"]
            },
            {
              label: "DALY 97.5%",
              value: data.total.daly["Quantile0.975"]
            },
            ]
          };
          break;
      }

      // Update the tooltip content and show
      tooltip.updateContent(tooltipContentFormatter(contentObject));
      tooltip.show({
        left: coords.pageX,
        top: coords.pageY
      });
    }
  };

  var _plotBarChart = function (chartContainerId, legendContainerId, data) {
    var chartContainer = $(chartContainerId);
    var yldData = [
      [1, data.acute.yld.Mean],
      [2, data.sequela.yld.Mean],
      [3, data.total.yld.Mean]
    ];
    var yllData = [
      [1, data.acute.yll.Mean],
      [2, data.sequela.yll.Mean],
      [3, data.total.yll.Mean]
    ];
    var dalyErrorData = [
      [1, data.acute.daly.Mean, data.acute.daly.Mean - data.acute.daly["Quantile0.025"], data.acute.daly["Quantile0.975"] - data.acute.daly.Mean],
      [2, data.sequela.daly.Mean, data.sequela.daly.Mean - data.sequela.daly["Quantile0.025"], data.sequela.daly["Quantile0.975"] - data.sequela.daly.Mean],
      [3, data.total.daly.Mean, data.total.daly.Mean - data.total.daly["Quantile0.025"], data.total.daly["Quantile0.975"] - data.total.daly.Mean],
    ];

    var chartDataSet = [{
      label: "YLD - Mean",
      stack: 1,
      bars: {
        show: true,
        align: "center",
        barWidth: 0.9,
        fill: true,
        lineWidth: 1
      },
      data: yldData
    },
    {
      label: "YLL - Mean",
      stack: 1,
      bars: {
        show: true,
        align: "center",
        barWidth: 0.9,
        fill: true,
        lineWidth: 1
      },
      data: yllData
    },
    {
      label: "DALY error",
      points: {
        errorbars: "y",
        yerr: {
          show: true,
          asymmetric: true,
          upperCap: "-",
          lowerCap: "-"
        },
        lineWidth: 1
      },
      lines: {
        show: false
      },
      shadowSize: 1,
      lineWidth: 1,
      color: "red",
      data: dalyErrorData
    }
    ];

    var maxY = Math.max(data.acute.daly["Quantile0.975"], data.sequela.daly["Quantile0.975"], data.total.daly["Quantile0.975"]);
    var chartOptions = {
      points: {
        shadowSize: 0
      },
      xaxis: {
        ticks: [
          [1, "Acute"],
          [2, "Sequelae"],
          [3, "Total"]
        ],
        font: {
          size: 11
        }
      },
      yaxis: {
        min: 0,
        max: maxY * 1.1,
        font: {
          size: 11
        },
        tickFormatter: _defaultTickFormatter
      },
      legend: {
        show: true,
        container: $(legendContainerId)
      }
    };
    chartOptions = $.extend(true, {}, _defaultChartOptions, chartOptions);
    $.plot(chartContainer, chartDataSet, chartOptions);

    chartContainer.off("plothover");
    chartContainer.tooltip({
      automatic: false
    }).on("plothover", function (event, coords, item) {
      _getDALYBarChartTooltip($(this), event, coords, item, data);
    });
  };

  var _getDALYChartTooltip = function (elem, event, coords, item, data, dataSuffix, dataGender) {

    // Grab the API reference
    var tooltip = elem.data("tooltip");

    // If we weren't passed the item object, hide the tooltip and remove cached point data
    if (!item) {
      tooltip.cachedPoint = null;
      tooltip.hide();
      return false;
    }

    // Proceed only if the data point has changed
    if (tooltip.cachedPoint !== item.dataIndex) {

      // Update the cached point data
      tooltip.cachedPoint = item.dataIndex;

      var yldData = data["yld" + dataSuffix]["array" + dataGender];
      var yllData = data["yll" + dataSuffix]["array" + dataGender];
      var dalyData = data["daly" + dataSuffix]["array" + dataGender];

      // Setup new content
      var contentObject = {
        label: "Age group " + item.series.yaxis.ticks[item.dataIndex].label,
        entries: [{
          label: "YLD Mean",
          value: yldData[item.dataIndex].Mean
        },
        {
          label: "YLL Mean",
          value: yllData[item.dataIndex].Mean
        },
        {
          label: "DALY Mean",
          value: dalyData[item.dataIndex].Mean
        },
        {
          label: "DALY 2.5%",
          value: dalyData[item.dataIndex]["Quantile0.025"]
        },
        {
          label: "DALY 97.5%",
          value: dalyData[item.dataIndex]["Quantile0.975"]
        },
        ]
      };

      // Update the tooltip content and show
      tooltip.updateContent(tooltipContentFormatter(contentObject));
      tooltip.show({
        left: coords.pageX,
        top: coords.pageY
      });

    }
  };

  var _getDALYChartDataMax = function (data, dataSuffix, dataGender) {
    var dalyData = data["daly" + dataSuffix]["array" + dataGender];
    var numValues = dalyData.length; //+ All arrays must have the same number of elements anyway
    var maxX = 0;
    for (var i = 0; i < numValues; ++i) {
      maxX = Math.max(maxX, dalyData[i]["Quantile0.975"]);
    }
    return maxX;
  };

  var _plotDALYChart = function (chartContainerId, legendContainerId, data, dataSuffix, dataGender, extraOptions) {
    var chartContainer = $(chartContainerId);

    var yldData = data["yld" + dataSuffix]["array" + dataGender];
    var yllData = data["yll" + dataSuffix]["array" + dataGender];
    var dalyData = data["daly" + dataSuffix]["array" + dataGender];

    var yldChartData = [];
    var yllChartData = [];
    var dalyErrorChartData = [];
    var chartTicks = [];
    var numValues = dalyData.length; //+ All arrays must have the same number of elements anyway
    for (var i = 0; i < numValues; ++i) {
      yldChartData[i] = [yldData[i].Mean, i + 1];
      yllChartData[i] = [yllData[i].Mean, i + 1];
      var daly = yldChartData[i][0] + yllChartData[i][0];
      dalyErrorChartData[i] = [daly, i + 1.5, daly - dalyData[i]["Quantile0.025"], dalyData[i]["Quantile0.975"] - daly];
      chartTicks[i] = [i + 1.5, dalyData[i].Age];
    }
    var dalyDataSet = [{
      label: "YLD - Mean",
      stack: 1,
      bars: {
        show: true,
        fill: true,
        lineWidth: 1,
        horizontal: true
      },
      data: yldChartData
    },
    {
      label: "YLL - Mean",
      stack: 1,
      bars: {
        show: true,
        fill: true,
        lineWidth: 1,
        horizontal: true
      },
      data: yllChartData
    },
    {
      label: "DALY error",
      points: {
        errorbars: "x",
        xerr: {
          show: true,
          asymmetric: true,
          upperCap: "-",
          lowerCap: "-"
        },
        lineWidth: 1
      },
      lines: {
        show: false
      },
      shadowSize: 1,
      lineWidth: 1,
      color: "red",
      data: dalyErrorChartData
    }
    ];
    var chartOptions = {
      yaxis: {
        ticks: chartTicks,
        min: 1,
        max: numValues + 1,
        tickLength: 0
      },
      xaxis: {
        min: 0,
        tickFormatter: _defaultTickFormatter
      },
      legend: {
        show: true,
        container: $(legendContainerId)
      }
    };
    chartOptions = $.extend(true, {}, _defaultChartOptions, chartOptions);
    chartOptions = $.extend(true, {}, chartOptions, extraOptions);
    $.plot(chartContainer, dalyDataSet, chartOptions);

    chartContainer.off("plothover");
    chartContainer.tooltip({
      automatic: false
    }).on("plothover", function (event, coords, item) {
      _getDALYChartTooltip($(this), event, coords, item, data, dataSuffix, dataGender);
    });
  };

  return {
    defaultChartOptions: _defaultChartOptions,
    defaultTickFormatter: _defaultTickFormatter,
    plotAggrResults3Chart: _plotAggrResults3Chart,
    plotPieChart: _plotPieChart,
    plotBarChart: _plotBarChart,
    getDALYChartDataMax: _getDALYChartDataMax,
    plotDALYChart: _plotDALYChart
  };
};
