/* global
    $, diseaseTemplates, defaultDisease, jsPlumb
*/

/*
 * EDIT MODEL DATA PAGE
 */

var editModelDataModule = function (webGUI, bcode) { // eslint-disable-line no-unused-vars
  "use strict";

  // Utility functions
  var isEmptyObject = function (obj) { return Object.keys(obj).length === 0 && obj.constructor === Object };

  // DOM elements caching
  var editModelData_modelSelect = $("#editModelData_modelSelect");
  var flowChartContainer = $("#flowChartContainer");
  var diseaseReportContainer = $("#diseaseReportContainer");

  var caseBeforeCorrTooltip = "<p>In the table you can specify the number of cases per age group and gender.<br />" +
    "The formula can include arithmetic operators (\"+\", \"-\", \"*\", \"/\"), constants (\"0.3\", \"3.14\", etc.), and the following uncertainty ranges " +
    "that will be evaluated at each iteration of the model sampling by Monte Carlo simulations:" +
    "<ul>" +
    "   <li>RandUniform(a, b) - Uniformly distributed random variable with lower bound &quot;a&quot; and upper bound &quot;b&quot;.<br />" +
    "       <a href=\"https://en.wikipedia.org/wiki/Uniform_distribution_(continuous)\">https://en.wikipedia.org/wiki/Uniform_distribution_(continuous)</a>" +
    "   </li>" +
    "   <li>RandPert(a, b, c) - Pert distributed random variable with lower bound &quot;a&quot;, mode &quot;b&quot; and upper bound &quot;c&quot;.<br />" +
    "       <a href=\"https://www.riskamp.com/beta-pert\">https://www.riskamp.com/beta-pert</a>" +
    "   </li>" +
    "   <li>RandBeta(a, b) - Beta distributed random variable with shape parameters &quot;a&quot; and &quot;b&quot;.<br />" +
    "       <a href=\"https://en.wikipedia.org/wiki/Beta_distribution\">https://en.wikipedia.org/wiki/Beta_distribution</a>" +
    "   </li>" +
    "</ul>" +
    "For instance, the following formula may be specified as a case count parameter:<br /> \"5/3 * RandPert(0.4, 1.4, 4.5) - 0.3\".<br />" +
    "Please note, that \".\" (dot) must be used as decimal separator in parameters values and \",\" (comma) must be used as parameters separator.<br />" +
    "If the same formula is to be used for all age/sex categories, then 1) double-click field in the edited table to start editing its value, 2) type the intended formula," +
    "3) press enter to confirm the formula and end its editing, 4) right-click this formula table cell and select option \"Fill table with this formula\"<br/>" +
    "It is also possible to paste a table already prepared in another software (e.g. Windows Excel&copy;) by copying the table from that software and right-clicking on the table " +
    "below and selecting \"Paste table from clipboard\".</p>";

  var underestimTooltip = "<p>In this table you can input multipliers adjusting for under-estimation either by defining one number or set of numbers (1 table to fill) or " +
    "by adjusting for under-ascertainment and under-reporting separately (2 tables to fill).<br />" +
    "By default the multiplier is 1; if you do not change the table you consider there is no under-estimation of the number of cases you have inputted in the table on the left.</p>";

  var caseBeforeCorrGrid = new bcode.grids.InputTable("#inputCasesTables", $.extend(true, {}, { editable: true, tooltip: caseBeforeCorrTooltip }, { caption: "Average number of cases per year", table: { containerClass: "moduleBox floatLeft firstBox", containerId: "caseBeforeCorr" }, stratified: { aggregatorType: "sum" } }));
  var underestimationGrid = new bcode.grids.InputTable("#correctionTables", $.extend(true, {}, { editable: true, tooltip: underestimTooltip }, { caption: "Underestimation", table: { containerId: "underestimation" } }));
  var underreportingGrid = new bcode.grids.InputTable("#correctionTables", $.extend(true, {}, { editable: true, tooltip: underestimTooltip }, { caption: "Underreporting", table: { containerId: "underreporting" } }));
  var underascertainmentGrid = new bcode.grids.InputTable("#correctionTables", $.extend(true, {}, { editable: true, tooltip: underestimTooltip }, { caption: "Underascertainment", table: { containerId: "underascertainment" } }));

  var ageDistributionGrid = new bcode.grids.InputTable("#populationDataTab", $.extend(true, {}, { editable: true }, { caption: "Age distribution data", table: { containerClass: "moduleBox floatLeft firstBox", containerId: "ageDistrTable" }, stratified: { aggregatorType: "sum" }, simplified: { enabled: false } }));
  var lifeExpectancyGrid = new bcode.grids.InputTable("#populationDataTab", $.extend(true, {}, { editable: true }, { caption: "Life expectancy data", table: { containerId: "lifeExpectancyTable" }, simplified: { enabled: false } }));

  var diseasePageState = {
    jsonDiseaseTemplates: {},
    countryID: "",
    diseaseID: "",
    modelID: "",
    flowChartVisible: false,
    estimType: 0,

    _updateModelSelection: function () {
      var value = editModelData_modelSelect.val();
      var values = value.split(",");
      this.countryID = values[0];
      this.diseaseID = values[1];
      this.modelID = values[2];
    },
    _readSelect: function () {
      var oldDiseaseID = this.diseaseID;
      var oldModelID = this.modelID;

      this._updateModelSelection();
      if (oldModelID !== this.modelID) {
        if (oldDiseaseID !== this.diseaseID) {
          // Refresh outcome tree
          this._showOutcomeTree(this.diseaseID);
        }

        // Refresh Model parameters and Population data
        webGUI.selectModel(this.countryID, this.diseaseID, this.modelID);
      }
    },
    _setUnderestimType: function (type) {
      this.estimType = type;

      switch (type) {
        case 0:
          $("#underreporting, #underascertainment").fadeOut("fast", function () {
            $("#underestimation").fadeIn("slow");
          });
          $("#estimType").prop("checked", true);
          break;
        case 1:
          $("#underestimation").fadeOut("fast", function () {
            $("#underreporting, #underascertainment").fadeIn("slow");
          });
          $("#reportType").prop("checked", true);
          break;
      }
    },
    _showOutcomeTree: function (diseaseID) {
      if (this.flowChartVisible) {

        var templateType = "ORIGINAL";
        var diseaseTemplate = null;
        if (diseaseID !== "") {
          let jsonDiseaseTemplate = diseasePageState.jsonDiseaseTemplates[diseaseID];
          if (!isEmptyObject(jsonDiseaseTemplate)) {
            templateType = "JSON";
            diseaseTemplate = jsonDiseaseTemplate;
          } else if (diseaseTemplates.hasOwnProperty(diseaseID)) {
            diseaseTemplate = diseaseTemplates[diseaseID];
          } else {
            diseaseTemplate = diseaseTemplates["Default"];
          }
        } else {
          diseaseTemplate = diseaseTemplates["Default"];
        }

        if (templateType === "ORIGINAL") {
          jsPlumb.reset();
          flowChartContainer.html(diseaseTemplate.htmlCode);
          jsPlumb.setSuspendDrawing(true);
          diseaseTemplate.jsCode();
        } else if (templateType === "JSON") {
          renderDisease(flowChartContainer, diseaseTemplate);

          jsPlumb.reset();
          jsPlumb.setSuspendDrawing(true);

          var otSpecs = diseaseTemplate.outcomeTrees;
          var otTemplates = "";
          for (let otId in otSpecs) {
            let otSpec = otSpecs[otId];

            let tSpecs = otSpec.transitions;
            for (let tId in tSpecs) {
              let tSpec = tSpecs[tId];

              let sourceId = tSpec.source.id;
              let targetId = tSpec.target.id;
              let otId = tSpec.outcomeTree;
              let tName = tSpec.name;

              createConnection(sourceId, targetId, {}, tId, otId, tName);
            }
          }
        }
        jsPlumb.setSuspendDrawing(false, true);

        jsPlumb.bind("click", function (connection) {
          webGUI.showOutcomeTreeElementDetails(connection.outcomeTreeID, "arrow", connection.name, diseasePageState.countryID, diseasePageState.diseaseID, diseasePageState.modelID);
        });

        $(".healthState:not(.inactive)").on("click", function () {
          var elemType = $(this).parents(".healthOutcome").attr("class").substr(14).split(" ")[0];
          var elemID = $(this).attr("id");
          var outcomeTreeID = $(this).parents(".outcomeTree").attr("id");
          webGUI.showOutcomeTreeElementDetails(outcomeTreeID, elemType, elemID, diseasePageState.countryID, diseasePageState.diseaseID, diseasePageState.modelID);
        });
      }
    },
    setAvailableModels: function (availableModels) {
      Object.assign(
        diseasePageState.jsonDiseaseTemplates,
        availableModels.reduce(function (map, obj) { map[obj.diseaseID] = obj.jsonDiseaseSpec; return map; }, {})
      );
      bcode.state.createdModels.concat(availableModels);
      editModelData_modelSelect.children("option[value=',,']").remove();
      if (availableModels.length > 0) {
        $("#editModelParamChagesBtn").css({ "display": "inline-block" });
        $.each(availableModels, function (index, value) {
          editModelData_modelSelect
            .append($("<option>", { value: value.countryID + "," + value.diseaseID + "," + value.modelID })
              .text(value.countryName + " - " + value.diseaseName + " - " + value.modelID));
        });
      } else {
        $("#editModelParamChagesBtn").css({ "display": "none" });
        editModelData_modelSelect.children().remove();
        editModelData_modelSelect.append("<option value=\",,\">No models created</option>");
      }
      diseasePageState._readSelect();
    },
    displayModelParameters: function (modelInfo, correctionData, caseData, lifeExpData, ageDistrData, diseaseReport) {
      caseBeforeCorrGrid.setData(caseData);
      ageDistributionGrid.setData(ageDistrData);
      lifeExpectancyGrid.setData(lifeExpData);

      underestimationGrid.setData(correctionData.underestimation);
      underreportingGrid.setData(correctionData.underreporting);
      underascertainmentGrid.setData(correctionData.underascertainment);

      diseaseReportContainer.html(diseaseReport);

      diseasePageState._setUnderestimType(correctionData.estimType);
    },
    highlightFormulaErrors: function (highlightData) {
      caseBeforeCorrGrid.highlightErrorCells(highlightData.caseBeforeCorr);
      ageDistributionGrid.highlightErrorCells(highlightData.ageDistr);
      lifeExpectancyGrid.highlightErrorCells(highlightData.lifeExp);
      underestimationGrid.highlightErrorCells(highlightData.underestimation);
      underreportingGrid.highlightErrorCells(highlightData.underreporting);
      underascertainmentGrid.highlightErrorCells(highlightData.underascertainment);
    },
    reinitialize: function () {
      this.jsonDiseaseTemplates = {};
      this.countryID = "";
      this.diseaseID = "";
      this.modelID = "";
      this.flowChartVisible = false;
      this.estimType = 0;
      this.setAvailableModels([]);
      this._showOutcomeTree("");
      $("#editModelData > .contentRow").css("display", "none");
      $("#applyModelParamChagesBtn").hide();
    }
  };

  /*
   * Flow chart
   */

  jsPlumb.importDefaults({
    RenderMode: "svg",
    DefaultDragOptions: { cursor: "pointer", zIndex: 2000 },
    PaintStyle: { lineWidth: 1, strokeStyle: "rgb(150, 150, 150)" },
    Endpoint: "Blank",
    Endpoints: ["Blank", "Blank"],
    HoverPaintStyle: { lineWidth: 2, strokeStyle: "rgb(100, 100, 100)" },
    LabelStyle: { color: "#333", font: "normal 10px", location: 0.5 },
    Connector: ["Straight"],
    Container: flowChartContainer,
    ConnectionsDetachable: false,
    Overlays: [["Arrow", { length: 8, width: 8, location: -1 }]],
    Anchors: ["RightMiddle", "LeftMiddle"]
  });

  /*
   * Local events
   */
  editModelData_modelSelect.change(function () {
    diseasePageState._readSelect();
  });

  $("input[name=underestimTypeSelect]").change(function () {
    var val = parseInt($(this).val());
    diseasePageState._setUnderestimType(val);
  });

  $(document).on("flowChartShown", function () {
    diseasePageState.flowChartVisible = true;
    diseasePageState._showOutcomeTree(diseasePageState.diseaseID);
  });

  /*
   * Hook up local events with C++
   */

  $("#applyModelParamChagesBtn").click(function () {
    var modelInfo = {
      countryID: diseasePageState.countryID,
      diseaseID: diseasePageState.diseaseID,
      modelID: diseasePageState.modelID
    };
    webGUI.applyChangesToModelParams(
      modelInfo, diseasePageState.estimType, underestimationGrid.getData(), underreportingGrid.getData(),
      underascertainmentGrid.getData(), caseBeforeCorrGrid.getData(), lifeExpectancyGrid.getData(),
      ageDistributionGrid.getData());
  });

  /*
   * Hook up C++ engine with JavaScript
   */
  (function (test) {
    if (!test) {
      webGUI.modelsCreated.connect(diseasePageState.setAvailableModels);
      webGUI.modelSelected.connect(diseasePageState.displayModelParameters);
      webGUI.formulaCheckResultsAvailable.connect(diseasePageState.highlightFormulaErrors);
    }
  })(bcode.settings.development);

  return {
    diseasePageState: diseasePageState,
    flowChartContainer: flowChartContainer
  };
};
