//==========================================================================================================================
//  'BCode'
//      BCoDE application main entry point.
//
//  Written by Daniel Lewandowski
//  Copyright @ 2011 NextPage Software
//==========================================================================================================================

#include <QSettings>
#include <iostream>
#include <stdio.h>
#include "ut/globals.h"
#include "ui/BcodeMainWnd.h"
#include "ui/Chief.h"

QCoreApplication* createApplication(int& argc, char* argv[])
{
    if (argc > 1) {
        return new QCoreApplication(argc, argv);
    } else {
        return new QApplication(argc, argv);
    }
}

int main(int argc, char *argv[])
{
    // Enable support for high DPI screens
    QApplication::setAttribute(Qt::AA_EnableHighDpiScaling);
    QApplication::setAttribute(Qt::AA_UseHighDpiPixmaps);
    QGuiApplication::setHighDpiScaleFactorRoundingPolicy(Qt::HighDpiScaleFactorRoundingPolicy::PassThrough);

    QApplication bcode(argc, argv);

    bcode.setOrganizationName(APP_ORGANIZATION_NAME);
    bcode.setOrganizationDomain(APP_ORGANIZATION_DOMAIN);
    bcode.setApplicationName(APP_APPLICATION_NAME);
    bcode.setApplicationVersion(APP_VERSION);

    // Create the main object - Chief
    ui::Chief& chief = ui::Chief::getInstance();
    chief.setAppPath(bcode.applicationDirPath());

    // Check if GUI or CLI mode should be set
    bool guiApp = argc <= 1;

    int retVal = 0;
    if (guiApp) {
        // Start GUI version

        #if (!(defined(FINAL)))
            qputenv("QTWEBENGINE_REMOTE_DEBUGGING", "23654");
        #endif

        // Create the one and only GUI, set it up and show.
        ui::BcodeMainWnd& mainWindow = ui::BcodeMainWnd::getInstance();

        // Set the one and only Chief
        mainWindow.setChief(&chief);

        // Show GUI
        mainWindow.show();

        // Schedule destruction of the main window singleton object when the application finishes.
        QObject::connect(qApp, SIGNAL(aboutToQuit()), &mainWindow, SLOT(deleteLater()));

        try {
            // Enter the message loop and start using the program.
            retVal = bcode.exec();
        } catch (const ut::Exception& e) {
            ut::DisplayException(e);
        } catch (const std::runtime_error& e) {
            QMessageBox::critical(nullptr, "Error", e.what());
        } catch (...) {
            QMessageBox::critical(nullptr, "Error", "An error has occured. The program will terminate now");
        }
    } else {
        // Start CLI version

        QByteArray appString = QString("%1 %2, CLI mode").arg(bcode.applicationName()).arg(bcode.applicationVersion()).toLocal8Bit();

        QCommandLineOption inputFileOption(QStringList() << "i" << "input", "Path to the model specificaction <file>.", "file");
        QCommandLineOption outputFileOption(QStringList() << "o" << "output", "Path to the output results JSON <file>.", "file");
        QCommandLineOption outputClipboardOption(QStringList() << "c" << "clipboard", "Output results to clipboard.");
        QCommandLineOption sampleSizeOption(QStringList() << "s" << "sample-size", "Set number of iterations of Monte-Carlo simulation to <size>.", "size", "1000");
        QCommandLineOption timeDiscountOption(QStringList() << "t" << "time-discount", "Set time discount factor to <timeDiscount>.", "timeDiscount", "0");
        QCommandLineOption randomSeedOption(QStringList() << "r" << "random-seed", "Set random seed for Monte-Carlo simulation to <seed>.", "seed", "0");
        QCommandLineOption countryOption(QStringList() << "p" << "country", "Set <country> to be used for population data.", "country");

        QCommandLineParser parser;
        parser.setApplicationDescription(appString);
        parser.addHelpOption();
        parser.addVersionOption();
        parser.addOption(inputFileOption);
        parser.addOption(outputFileOption);
        parser.addOption(outputClipboardOption);
        parser.addOption(sampleSizeOption);
        parser.addOption(timeDiscountOption);
        parser.addOption(randomSeedOption);
        parser.addOption(countryOption);
        parser.process(bcode);

        QString inputFile = parser.value(inputFileOption);
        QString outputFile = parser.value(outputFileOption);
        bool outputClipboard = parser.isSet(outputClipboardOption);
        uint sampleSize = parser.value(sampleSizeOption).toUInt();
        double timeDiscount = parser.value(timeDiscountOption).toDouble();
        uint randomSeed = parser.value(randomSeedOption).toUInt();
        fn::InputType::Country country = parser.value(countryOption);

        QFileInfo inputFileInfo(inputFile);
        QFileInfo outputFileInfo(outputFile);
        if (inputFileInfo.isRelative()) {
            inputFileInfo.setFile(chief.getAppPath() + inputFile);
        }
        if (outputFileInfo.isRelative()) {
            outputFileInfo.setFile(chief.getAppPath() + outputFile);
        }

        puts(appString);
        puts(QString("Run start time         : %1").arg(QDateTime::currentDateTime().toString()).toLocal8Bit());
        puts(QString("Input file             : %1").arg(inputFileInfo.absoluteFilePath()).toLocal8Bit());
        puts(QString("Output file            : %1").arg(outputFileInfo.absoluteFilePath()).toLocal8Bit());
        puts(QString("Output to clipboard    : %1").arg(outputClipboard ? "Yes" : "No").toLocal8Bit());
        puts(QString("Sample size            : %1").arg(sampleSize).toLocal8Bit());
        puts(QString("Time discounting factor: %1").arg(timeDiscount).toLocal8Bit());
        puts(QString("Random seed            : %1").arg(randomSeed).toLocal8Bit());

        try {
            if (inputFileInfo.completeSuffix().contains("xls")) {
                chief.loadModelsFromExcel(inputFileInfo);
            } else if (inputFileInfo.completeSuffix().contains("json")) {
                puts(QString("Country ID             : %1").arg(country).toLocal8Bit());
                QFileInfoList inputFileInfos;
                inputFileInfos.append(inputFileInfo);
                chief.loadModelsFromJSON(inputFileInfos, country);
            } else {
                puts("Unsupported input file specified");
                return(retVal);
            }

            chief.setNumSamples(sampleSize);
            chief.setTimeDiscountRate(timeDiscount);
            chief.getAppSettings().setRandomSeed(randomSeed);
            chief.runModels(false);
            chief.saveDetailedResultsToJSON(outputFileInfo);
            if (outputClipboard) {
                chief.saveDetailedResultsToClipboard();
            }
            puts(QString("Run finish time        : %1").arg(QDateTime::currentDateTime().toString()).toLocal8Bit());
        } catch (...) {
            puts("Error");
        }
    }

    return retVal;
}
