//==========================================================================================================================
//  Implementation file for "BcodeMainWnd.h"
//  Written by Daniel Lewandowski
//  Copyright @ 2011 NextPage Software
//==========================================================================================================================

#include <QObject>
#include <QAction>
#include <QMenu>
#include <QMenuBar>
#include "BcodeMainWnd.h"
#include "ui_BcodeMainWnd.h"
#include "fn/formulaParserHelper.h"
#include "ui/AboutBcodeDlg.h"
#include "ui/About3rdPartyDlg.h"
#include "ui/WhatsNewDlg.h"
#include "ui/SettingsDlg.h"
#include "ui/ModelCreateDlg.h"

namespace ui {

// Global static pointer used to ensure a single instance of the class.
BcodeMainWnd* BcodeMainWnd::m_mainWnd = nullptr;

BcodeMainWnd& BcodeMainWnd::getInstance()
{
    if (!m_mainWnd) {
        m_mainWnd = new BcodeMainWnd();
    }
    return *m_mainWnd;
}

BcodeMainWnd::BcodeMainWnd(QWidget* parent) :
    QMainWindow(parent),
    m_chief(nullptr),
    m_webGUI(),
    m_modelCreateWebGUI()
{
    setupUi(this);
    connectActions();
    createMenuBar();

    #if (defined(FINAL))
        webContent->setContextMenuPolicy(Qt::PreventContextMenu);
    #else
        webContent->setContextMenuPolicy(Qt::DefaultContextMenu);
    #endif

    m_webGUI.setPage(webContent->page());
}

BcodeMainWnd::~BcodeMainWnd()
{

}

void BcodeMainWnd::closeEvent(QCloseEvent* event)
{
    if (userReallyWantsToQuit()) {
        QApplication::closeAllWindows();
        event->accept();
    } else {
        event->ignore();
    }
}

void BcodeMainWnd::setChief(ui::Chief* chief)
{
    // Store references
    m_chief = chief;
    m_webGUI.setChief(chief);
    m_chief->setWebGUI(&m_webGUI);

    QObject::connect(actImportExcelReferenceData, SIGNAL(triggered()), m_chief, SLOT(importReferenceData()));
    QObject::connect(actImportJSONReferenceData, SIGNAL(triggered()), m_chief, SLOT(importReferenceDataFromJSON()));
    QObject::connect(this, SIGNAL(modelSaveFileSpecified(const QFileInfo&)), m_chief, SLOT(saveModelsToExcel(const QFileInfo&)));
    QObject::connect(this, SIGNAL(modelLoadFileSpecified(const QFileInfo&)), m_chief, SLOT(loadModelsFromExcel(const QFileInfo&)));
}

void BcodeMainWnd::connectActions()
{
    QObject::connect(actSaveModels, SIGNAL(triggered()), this, SLOT(showSaveModelsDialog()));
    QObject::connect(actLoadModels, SIGNAL(triggered()), this, SLOT(showLoadModelsDialog()));
    QObject::connect(actOptions, SIGNAL(triggered()), this, SLOT(showOptionsDialog()));
    QObject::connect(actAboutBCoDE, SIGNAL(triggered()), this, SLOT(showAboutBcodeDialog()));
    QObject::connect(actAbout3rdParty, SIGNAL(triggered()), this, SLOT(showAbout3rdPartyDialog()));
    QObject::connect(actWhatsNew, SIGNAL(triggered()), this, SLOT(showWhatsNewDialog()));
    QObject::connect(actModelCreate, SIGNAL(triggered()), this, SLOT(showModelCreateDialog()));
}

void BcodeMainWnd::createMenuBar()
{
    QMenu* fileMenu = new QMenu("File", this);
    fileMenu->addAction(actSaveModels);
    fileMenu->addAction(actLoadModels);
    fileMenu->addSeparator();
    fileMenu->addAction(actExit);

    QMenu* toolsMenu = new QMenu("Tools", this);
    toolsMenu->addAction(actOptions);
    toolsMenu->addAction(actModelCreate);
    toolsMenu->addSeparator();
    toolsMenu->addAction(actImportExcelReferenceData);
    toolsMenu->addAction(actImportJSONReferenceData);

    QMenu* aboutMenu = new QMenu("About", this);
    aboutMenu->addAction(actWhatsNew);
    aboutMenu->addSeparator();
    aboutMenu->addAction(actAboutBCoDE);
    aboutMenu->addAction(actAbout3rdParty);

    QMainWindow::menuBar()->addMenu(fileMenu);
    QMainWindow::menuBar()->addMenu(toolsMenu);
    QMainWindow::menuBar()->addMenu(aboutMenu);
}

bool BcodeMainWnd::userReallyWantsToQuit()
{
    return QMessageBox::question(this, "Quit confirm", "Do you want to quit the application?",
                                 QMessageBox::Yes | QMessageBox::No, QMessageBox::No) == QMessageBox::Yes;
}

void BcodeMainWnd::showAboutBcodeDialog()
{
    QVariantMap buildInfo;
    buildInfo.insert("buildDate", QString("%1, %2").arg(__DATE__).arg(__TIME__));
    buildInfo.insert("buildVersion", qApp->applicationVersion());

    ui::AboutBcodeDlg* dlg = new ui::AboutBcodeDlg(&m_webGUI, buildInfo);
    dlg->show();
}

void BcodeMainWnd::showAbout3rdPartyDialog()
{
    ui::About3rdPartyDlg* dlg = new ui::About3rdPartyDlg(&m_webGUI);
    dlg->show();
}

void BcodeMainWnd::showWhatsNewDialog()
{
    ui::WhatsNewDlg* dlg = new ui::WhatsNewDlg(&m_webGUI);
    dlg->show();
}

void BcodeMainWnd::showModelCreateDialog()
{
    ui::ModelCreateDlg* dlg = new ui::ModelCreateDlg(m_chief->getAppPath());
    m_modelCreateWebGUI.setPage(dlg->page());

    dlg->show();
}

void BcodeMainWnd::showOptionsDialog()
{
    fn::Settings& settings = m_chief->getAppSettings();
    ui::SettingsDlg settingsDlg(settings, this);
    if (settingsDlg.exec() == QDialog::Accepted) {
        settingsDlg.writeSettings();
        fn::setRandomSeed(settings.getRandomSeed());
    }
}

void BcodeMainWnd::showSaveModelsDialog()
{
    QString filename = QFileDialog::getSaveFileName(nullptr, tr("Export models"), QDir::currentPath(), tr("Excel files (*.xls)"));
    QFileInfo fileInfo(filename);
    filename.remove(".xls");
    filename.replace("/", "\\");
    if (!filename.isNull()) {
        emit modelSaveFileSpecified(fileInfo);
    }
}

void BcodeMainWnd::showLoadModelsDialog()
{
    QString filename = QFileDialog::getOpenFileName(nullptr, tr("Import models"), QDir::currentPath(), tr("Excel files (*.xls)"));
    QFileInfo fileInfo(filename);
    if (fileInfo.exists()) {
        emit modelLoadFileSpecified(fileInfo);
    }
}

} // namespace ui
