//==========================================================================================================================
//  Implementation file for "Chief.h"
//  Written by Daniel Lewandowski
//  Copyright @ 2010 NextPage Software
//==========================================================================================================================

#include <QCoreApplication>
#include <QFileInfo>
#include <QFileDialog>
#include <QThread>
#include "ui/Chief.h"
#include "ut/globals.h"
#include "ut/Excel.h"
#include "fn/globals.h"
#include "fn/ModelRunWorker.h"
#include "fn/ImportExportHelpers.h"
#include "fn/ArrayHelpers.h"
#include "fn/FormulaChecker.h"
#include "ui/globals.h"
#include "ui/BcodeMainWnd.h"
#include "ui/InfectionDataPresenter.h"
#include "ui/YldHealthOutcomeDataPresenter.h"
#include "ui/YllHealthOutcomeDataPresenter.h"
#include "ui/AsymptomaticDataPresenter.h"
#include "ui/TransProbDataPresenter.h"
#include "ui/WebGUI.h"

namespace ui {

Chief& Chief::getInstance()
{
    // This is probably not thread safe, but we don't care about it at this moment.
    // Application initialization is single-threaded anyway.
    static Chief instance;
    return instance;
}

Chief::Chief():
    m_diseaseModelList(),
    m_diseaseModelMap(),
    m_lastModelId(0),
    m_usedModelID(),
    m_inputDataMgr(),
    m_jsonInputDataMgr(),
    m_appPath(""),
    m_webGUI(nullptr),
    m_numSamples(0),
    m_randomSeed(0),
    m_timeDiscountRate(0),
    m_presenterMgr(),
    m_diseaseReportMgr(),
    m_resultsComputer(m_diseaseModelList),
    m_settings(),
    m_tracker()
{
    #if (defined(FINAL))
        m_tracker.setTrackingID(ui::kProdAnalyticsTrackID);
    #else
        m_tracker.setTrackingID(ui::kTestAnalyticsTrackID);
    #endif

    reinitialize();

    QObject::connect(&m_inputDataMgr, SIGNAL(excelSheetImporting(fn::FileOperationType, const QString&, const QString&, const QString&)), this, SLOT(updateDataExportImportStatus(fn::FileOperationType, const QString&, const QString&, const QString&)));
    QObject::connect(&m_inputDataMgr, SIGNAL(showMessage(const QString&, const QString&, const QString&)), this, SIGNAL(showMessage(const QString&, const QString&, const QString&)));
}

Chief::~Chief()
{

}

void Chief::reinitialize()
{
    m_numSamples = m_settings.getNumSamples();
    m_randomSeed = m_settings.getRandomSeed();
    m_timeDiscountRate = m_settings.getTimeDiscountRate();

    m_diseaseModelMap.clear();
    m_diseaseModelList.clear();
    m_usedModelID.clear();
    m_lastModelId = 0;
    m_presenterMgr.closeAllPresenters();

    emit programReinitialized();
}

void Chief::setWebGUI(WebGUI* webGUI)
{
    m_webGUI = webGUI;
}

void Chief::setAppPath(const QString& appPath)
{
    m_appPath = appPath;
    m_inputDataMgr.setPath(appPath);
    m_jsonInputDataMgr.setDirPath(appPath);
    m_diseaseReportMgr.setPath(appPath);
}

const QString& Chief::getAppPath() const
{
    return m_appPath;
}

fn::Settings& Chief::getAppSettings()
{
    return m_settings;
}

const fn::InputDataMgr& Chief::getInputDataMgr() const
{
    return m_inputDataMgr;
}

fn::InputDataMgr& Chief::getInputDataMgr()
{
    return m_inputDataMgr;
}

const fn::JSONInputDataMgr& Chief::getJSONInputDataMgr() const
{
    return m_jsonInputDataMgr;
}

fn::JSONInputDataMgr& Chief::getJSONInputDataMgr()
{
    return m_jsonInputDataMgr;
}

int Chief::getNumSamples() const
{
    return m_numSamples;
}

double Chief::getTimeDiscountRate() const
{
    return m_timeDiscountRate;
}

const fn::ResultsComputer& Chief::getResultsComputer() const
{
    return m_resultsComputer;
}

void Chief::pageOpened(int tabIndex, const QString& pageName)
{
    Q_UNUSED(tabIndex);

    if (pageName != QStringLiteral("")) {
        m_tracker.sendScreenView(pageName);
    }
}

void Chief::importReferenceData()
{
    reinitialize();
    m_inputDataMgr.importReferenceData();
}

void Chief::importReferenceDataFromJSON()
{
    reinitialize();
    m_jsonInputDataMgr.importReferenceData();
}

QVariantList Chief::getAvailableDiseases()
{
    QVariantList diseaseList;
    const QList<fn::InputType::Disease> diseaseIds = m_inputDataMgr.getAvailableDiseases();
    fn::DiseaseTemplateMgr& tmplMgr = m_inputDataMgr.getDiseaseTemplateMgr();
    for (int i = 0; i < diseaseIds.count(); ++i) {
        const fn::InputType::Disease diseaseId = diseaseIds.at(i);
        const QString diseaseName = tmplMgr.getDiseaseTemplate(diseaseId).getCaption();
        QVariantMap diseaseMap;
        diseaseMap.insert("id", diseaseId);
        diseaseMap.insert("disease", diseaseName);
        diseaseList.append(diseaseMap);
    }

    diseaseList.append(m_jsonInputDataMgr.getDiseaseCaptions());

    return(diseaseList);
}

fn::DiseaseModel::ID Chief::allocateNewModelID(const QString candidate)
{
    if (!candidate.isEmpty()) {
        QString modelID = candidate;
        if (m_usedModelID.contains(modelID)) {
            modelID.append(QString("_2"));
        }
        m_usedModelID.append(modelID);
        return modelID;
    } else {
        m_lastModelId++;
        QString modelID = QString("model %1").arg(m_lastModelId);
        m_usedModelID.append(modelID);
        return modelID;
    }
}

fn::DiseaseModel* Chief::createNewDiseaseModel(
    const fn::InputType::Country& country,
    const fn::InputType::Disease& disease,
    const fn::DiseaseModel::ID& modelID
) {
    fn::DiseaseModel* newModel(nullptr);

    // Currenty we create only models for which data exists
    if (
        m_inputDataMgr.countryDiseaseCombinationAvailable(country, disease) ||
        m_jsonInputDataMgr.diseaseDataExist(disease)
    ) {
        const QString countryName = m_inputDataMgr.getCountryName(country);

        if (m_inputDataMgr.diseaseDataExist(disease)) {
            const fn::ConstInputFactorPtrList inputFactorList = m_inputDataMgr.getInputFactorMgr().getInputFactorList(disease);
            const fn::DiseaseTemplate& diseaseTemplate = m_inputDataMgr.getDiseaseTemplateMgr().getDiseaseTemplate(disease);
            m_diseaseModelList.push_back(fn::DiseaseModel(this, modelID, country, countryName, &diseaseTemplate, inputFactorList));
        } else if (m_jsonInputDataMgr.diseaseDataExist(disease)) {
            const fn::JSONDiseaseTemplate& diseaseTemplate = m_jsonInputDataMgr.getDiseaseTemplate(disease);
            m_diseaseModelList.push_back(fn::DiseaseModel(this, modelID, country, countryName, &diseaseTemplate));
        }

        newModel = &m_diseaseModelList.last();
        newModel->initialize();
        m_diseaseModelMap[country][disease].insert(modelID, newModel);
    }
    return newModel;
}

fn::DiseaseModel* Chief::getDiseaseModel(const fn::InputType::Country& country, const fn::InputType::Disease& disease, const QString& modelId)
{
    fn::DiseaseModel* model(nullptr);
    if (m_diseaseModelMap.contains(country)) {
        if (m_diseaseModelMap[country].contains(disease)) {
            if (m_diseaseModelMap[country][disease].contains(modelId)) {
                model = m_diseaseModelMap[country][disease][modelId];
            }
        }
    }

    return model;
}

int Chief::getModelCount() const
{
    return m_diseaseModelList.length();
}

const fn::DiseaseModel& Chief::getModel(const size_t index) const
{
    return m_diseaseModelList.at(static_cast<int>(index));
}

void Chief::changeCountrySelection(const QStringList& countryIDStrs)
{
    m_selCountries = countryIDStrs;
}

void Chief::changeDiseaseSelection(const QStringList& diseaseIDStrs)
{
    m_selDiseases = diseaseIDStrs;
}

void Chief::createModels()
{
    QVariantList newModelsInfo;
    for (int i = 0; i < m_selCountries.size(); ++i) {
        for (int j = 0; j < m_selDiseases.size(); ++j) {
            fn::DiseaseModel* newModel = nullptr;

            const fn::InputType::Country countryId = m_selCountries.at(i);
            const fn::InputType::Disease diseaseId = m_selDiseases.at(j);
            const fn::DiseaseModel::ID modelID = allocateNewModelID();
            newModel = createNewDiseaseModel(countryId, diseaseId, modelID);

            if (newModel != nullptr) {
                newModelsInfo.append(newModel->getVariantModelInfo());
            }
        }
    }

    int totalModelCount = m_diseaseModelList.length();
    int newModelsCount = newModelsInfo.length();
    if (newModelsCount > 0) {
        emit showMessage("Model creation", QString("%1 new model(s) created.<br />Total number of models defined: %2.").arg(newModelsCount).arg(totalModelCount), "info");
    } else {
        emit showMessage("Model creation", "Data is missing for the selected combination of country and disease.", "error");
    }

    emit modelsCreated(newModelsInfo);
}

void Chief::deleteModels()
{
    reinitialize();
    emit showMessage("Model deletion", "All models have been deleted.", "info");
}

void Chief::selectModel(const QString& countryId, const QString& diseaseId, const QString& modelId)
{
    fn::DiseaseModel* model = getDiseaseModel(countryId, diseaseId, modelId);
    if (model != nullptr) {
        QVariantMap modelInfo = model->getVariantModelInfo();

        QVariantMap correctionsData;
        correctionsData["underestimation"] = fn::DataArrayToVariantMap(model->getParameter(fn::DiseaseModel::Underestimation), fn::DataArray::Formula, fn::DataArray::Input);
        correctionsData["underreporting"] = fn::DataArrayToVariantMap(model->getParameter(fn::DiseaseModel::Underreporting), fn::DataArray::Formula, fn::DataArray::Input);
        correctionsData["underascertainment"] = fn::DataArrayToVariantMap(model->getParameter(fn::DiseaseModel::Underascertainment), fn::DataArray::Formula, fn::DataArray::Input);
        correctionsData["estimType"] = model->getUnderestimationType();

        QVariantMap caseBeforeCorrData = fn::DataArrayToVariantMap(model->getParameter(fn::DiseaseModel::RawCasesCount), fn::DataArray::Formula, fn::DataArray::Input);
        QVariantMap lifeExpData = fn::DataArrayToVariantMap(model->getParameter(fn::DiseaseModel::LifeExpectancy), fn::DataArray::Formula, fn::DataArray::Input);
        QVariantMap ageDistrData = fn::DataArrayToVariantMap(model->getParameter(fn::DiseaseModel::AgeDistribution), fn::DataArray::Formula, fn::DataArray::Input);

        QString diseaseReport = m_diseaseReportMgr.getDiseaseReportHtml(diseaseId);

        emit modelSelected(modelInfo, correctionsData, caseBeforeCorrData, lifeExpData, ageDistrData, diseaseReport);
    }
}

void Chief::showDetailedResults(const fn::InputType::Country& countryId, const fn::InputType::Disease& diseaseId, const QString& modelId)
{
    fn::DiseaseModel* model = getDiseaseModel(countryId, diseaseId, modelId);
    if (model != nullptr) {
        const QVariantMap results = model->getDetailedPageResults();
        emit detailedResultsAvailable(results);
    }
}

void Chief::saveDetailedResultsToJSON(const QFileInfo& fileInfo)
{
    QString fileName = fileInfo.absoluteFilePath();
    fileName.remove(".xls");
    fileName.replace("/", "\\");
    if (!fileName.isNull()) {
        QFile file(fileName);
        if (file.open(QIODevice::WriteOnly)) {
            file.write(getDetailedResultsArray());
        }
    }
}

void Chief::saveDetailedResultsToClipboard()
{
    QClipboard* clipboard = QApplication::clipboard();
    clipboard->setText(getDetailedResultsArray());
}

void Chief::applyChangesToModelParams(
        const QVariantMap& modelInfo, const int estimType, const QVariantMap& underestimationData, const QVariantMap& underreportingData,
        const QVariantMap& underascertainmentData, const QVariantMap& caseBeforeCorrData, const QVariantMap& lifeExpData,
        const QVariantMap& ageDistrData)
{
    fn::InputType::Country countryID = static_cast<fn::InputType::Country>(modelInfo["countryID"].toString());
    fn::InputType::Disease diseaseID = static_cast<fn::InputType::Disease>(modelInfo["diseaseID"].toString());
    fn::DiseaseModel::ID modelID = static_cast<fn::DiseaseModel::ID>(modelInfo["modelID"].toString());

    fn::DiseaseModel* model = getDiseaseModel(countryID, diseaseID, modelID);
    if (model != nullptr) {

        QVariantMap formulaCheckVariant;
        fn::BoolBaseArray formulaCheck;
        QStringList paramsSuccess;
        QStringList paramsFail;

        if ((estimType >= 0) && (estimType <= 1)) {
            model->setUnderestimationType(estimType);
        }

        fn::DataArray& rawCasesArray = model->getParameter(fn::DiseaseModel::RawCasesCount);
        const fn::ConstInputFactorPtrList& rawCasesFactors = rawCasesArray.getInputFactorList();
        if (!fn::FormulaChecker::checkFormulas(fn::VariantMapToFormulaArray(caseBeforeCorrData, rawCasesFactors), formulaCheck)) {
            formulaCheckVariant.insert("caseBeforeCorr", fn::BoolArrayToVariantMap(formulaCheck, rawCasesFactors));
            paramsFail.push_back("Average number of cases per year");
        } else {
            fn::StringBaseArray formulaArray = fn::VariantMapToFormulaArray(caseBeforeCorrData, rawCasesFactors);
            fn::StringToStringMap metaData = fn::VariantMapToMetaData(caseBeforeCorrData);
            rawCasesArray.setInputFormulaArray(formulaArray);
            rawCasesArray.setMetaData(metaData);
            paramsSuccess.push_back("Average number of cases per year");
        }

        fn::DataArray& underestimArray = model->getParameter(fn::DiseaseModel::Underestimation);
        const fn::ConstInputFactorPtrList& underestimFactors = underestimArray.getInputFactorList();
        if (!fn::FormulaChecker::checkFormulas(fn::VariantMapToFormulaArray(underestimationData, underestimFactors), formulaCheck)) {
            formulaCheckVariant.insert("underestimation", fn::BoolArrayToVariantMap(formulaCheck, underestimFactors));
            paramsFail.push_back("Underestimation");
        } else {
            fn::StringBaseArray formulaArray = fn::VariantMapToFormulaArray(underestimationData, underestimFactors);
            fn::StringToStringMap metaData = fn::VariantMapToMetaData(underestimationData);
            underestimArray.setInputFormulaArray(formulaArray);
            underestimArray.setMetaData(metaData);
            paramsSuccess.push_back("Underestimation");
        }

        fn::DataArray& underreportArray = model->getParameter(fn::DiseaseModel::Underreporting);
        const fn::ConstInputFactorPtrList& underreportFactors = underreportArray.getInputFactorList();
        if (!fn::FormulaChecker::checkFormulas(fn::VariantMapToFormulaArray(underreportingData, underreportFactors), formulaCheck)) {
            formulaCheckVariant.insert("underreporting", fn::BoolArrayToVariantMap(formulaCheck, underreportFactors));
            paramsFail.push_back("Underreporting");
        } else {
            fn::StringBaseArray formulaArray = fn::VariantMapToFormulaArray(underreportingData, underreportFactors);
            fn::StringToStringMap metaData = fn::VariantMapToMetaData(underreportingData);
            underreportArray.setInputFormulaArray(formulaArray);
            underreportArray.setMetaData(metaData);
            paramsSuccess.push_back("Underreporting");
        }

        fn::DataArray& underascertArray = model->getParameter(fn::DiseaseModel::Underascertainment);
        const fn::ConstInputFactorPtrList& underascertFactors = underascertArray.getInputFactorList();
        if (!fn::FormulaChecker::checkFormulas(fn::VariantMapToFormulaArray(underascertainmentData, underascertFactors), formulaCheck)) {
            formulaCheckVariant.insert("underascertainment", fn::BoolArrayToVariantMap(formulaCheck, underascertFactors));
            paramsFail.push_back("Underascertainment");
        } else {
            fn::StringBaseArray formulaArray = fn::VariantMapToFormulaArray(underascertainmentData, underascertFactors);
            fn::StringToStringMap metaData = fn::VariantMapToMetaData(underascertainmentData);
            underascertArray.setInputFormulaArray(formulaArray);
            underascertArray.setMetaData(metaData);
            paramsSuccess.push_back("Underascertainment");
        }

        fn::DataArray& lifeExpArray = model->getParameter(fn::DiseaseModel::LifeExpectancy);
        const fn::ConstInputFactorPtrList& lifeExpFactors = lifeExpArray.getInputFactorList();
        if (!fn::FormulaChecker::checkFormulas(fn::VariantMapToFormulaArray(lifeExpData, lifeExpFactors), formulaCheck)) {
            formulaCheckVariant.insert("lifeExp", fn::BoolArrayToVariantMap(formulaCheck, lifeExpFactors));
            paramsFail.push_back("Life expectancy");
        } else {
            fn::StringBaseArray formulaArray = fn::VariantMapToFormulaArray(lifeExpData, lifeExpFactors);
            fn::StringToStringMap metaData = fn::VariantMapToMetaData(lifeExpData);
            lifeExpArray.setInputFormulaArray(formulaArray);
            lifeExpArray.setMetaData(metaData);
            paramsSuccess.push_back("Life expectancy");
        }

        fn::DataArray& ageDistrArray = model->getParameter(fn::DiseaseModel::AgeDistribution);
        const fn::ConstInputFactorPtrList& ageDistrFactors = ageDistrArray.getInputFactorList();
        if (!fn::FormulaChecker::checkFormulas(fn::VariantMapToFormulaArray(ageDistrData, ageDistrFactors), formulaCheck)) {
            formulaCheckVariant.insert("ageDistr", fn::BoolArrayToVariantMap(formulaCheck, ageDistrFactors));
            paramsFail.push_back("Age distribution");
        } else {
            fn::StringBaseArray formulaArray = fn::VariantMapToFormulaArray(ageDistrData, ageDistrFactors);
            fn::StringToStringMap metaData = fn::VariantMapToMetaData(ageDistrData);
            ageDistrArray.setInputFormulaArray(formulaArray);
            ageDistrArray.setMetaData(metaData);
            paramsSuccess.push_back("Age distribution");
        }

        if (paramsFail.length() > 0) {
            emit showMessage("Model parameters",
                             "Changes have NOT been applied to:<br /><ul><li>" + paramsFail.join("</li><li>") + "</li></ul>",
                             "error");
        }

        if (paramsSuccess.length() > 0) {
            model->propagateChangesInModelParameters();
            emit showMessage("Model parameters",
                             "Changes have been applied to:<br /><ul><li>" + paramsSuccess.join("</li><li>") + "</li></ul>",
                             "info");
        }

        emit formulaCheckResultsAvailable(formulaCheckVariant);
    }
}

void Chief::applyChangesToYLDParams(
        const QVariantMap& modelInfo, const QVariantMap& severityData, const QVariantMap& durationData, const QVariantMap& proportionData)
{
    fn::InputType::Country countryID = static_cast<fn::InputType::Country>(modelInfo["countryID"].toString());
    fn::InputType::Disease diseaseID = static_cast<fn::InputType::Disease>(modelInfo["diseaseID"].toString());
    fn::DiseaseModel::ID modelID = static_cast<fn::DiseaseModel::ID>(modelInfo["modelID"].toString());
    fn::OutcomeTree::ID outcomeTreeID = static_cast<fn::DiseaseModel::ID>(modelInfo["outcomeTreeID"].toString());
    fn::OutcomeTreeNode::ID healthStateID = static_cast<fn::DiseaseModel::ID>(modelInfo["outcomeTreeNodeID"].toString());

    fn::DiseaseModel* model = getDiseaseModel(countryID, diseaseID, modelID);

    if (model != nullptr) {

        QVariantMap formulaCheckVariant;
        fn::BoolBaseArray formulaCheck;
        QStringList paramsSuccess;
        QStringList paramsFail;

        fn::HealthState& healthState = model->getOutcomeTree(outcomeTreeID).getHealthState(healthStateID);

        fn::DataArray& severityArray = healthState.getParameter(fn::HealthState::Severity);
        const fn::ConstInputFactorPtrList& severityFactors = severityArray.getInputFactorList();
        if (!fn::FormulaChecker::checkFormulas(fn::VariantMapToFormulaArray(severityData, severityFactors), formulaCheck)) {
            formulaCheckVariant.insert("severity", fn::BoolArrayToVariantMap(formulaCheck, severityFactors));
            paramsFail.push_back("Severity");
        } else {
            fn::StringBaseArray formulaArray = fn::VariantMapToFormulaArray(severityData, severityFactors);
            fn::StringToStringMap metaData = fn::VariantMapToMetaData(severityData);
            severityArray.setInputFormulaArray(formulaArray);
            severityArray.setMetaData(metaData);
            paramsSuccess.push_back("Severity");
        }

        fn::DataArray& durationArray = healthState.getParameter(fn::HealthState::Duration);
        const fn::ConstInputFactorPtrList& durationFactors = durationArray.getInputFactorList();
        if (!fn::FormulaChecker::checkFormulas(fn::VariantMapToFormulaArray(durationData, durationFactors), formulaCheck)) {
            formulaCheckVariant.insert("duration", fn::BoolArrayToVariantMap(formulaCheck, durationFactors));
            paramsFail.push_back("Duration");
        } else {
            fn::StringBaseArray formulaArray = fn::VariantMapToFormulaArray(durationData, durationFactors);
            fn::StringToStringMap metaData = fn::VariantMapToMetaData(durationData);
            durationArray.setInputFormulaArray(formulaArray);
            durationArray.setMetaData(metaData);
            paramsSuccess.push_back("Duration");
        }

        fn::DataArray& proportionArray = healthState.getParameter(fn::HealthState::Proportion);
        const fn::ConstInputFactorPtrList& proportionFactors = proportionArray.getInputFactorList();
        if (!fn::FormulaChecker::checkFormulas(fn::VariantMapToFormulaArray(proportionData, proportionFactors), formulaCheck)) {
            formulaCheckVariant.insert("proportion", fn::BoolArrayToVariantMap(formulaCheck, proportionFactors));
            paramsFail.push_back("Proportion");
        } else {
            fn::StringBaseArray formulaArray = fn::VariantMapToFormulaArray(proportionData, proportionFactors);
            fn::StringToStringMap metaData = fn::VariantMapToMetaData(proportionData);
            proportionArray.setInputFormulaArray(formulaArray);
            proportionArray.setMetaData(metaData);
            paramsSuccess.push_back("Proportion");
        }

        if (paramsFail.length() > 0) {
            emit showMessageInYLDPage("YLD parameters",
                                      "Changes have NOT been applied to:<br /><ul><li>" + paramsFail.join("</li><li>") + "</li></ul>",
                                      "error");
        }

        if (paramsSuccess.length() > 0) {
            model->propagateChangesInModelParameters();
            emit showMessageInYLDPage("YLD parameters",
                                      "Changes have been applied to:<br /><ul><li>" + paramsSuccess.join("</li><li>") + "</li></ul>",
                                      "info");
        }

        emit formulaCheckResultsAvailableForYLDPage(formulaCheckVariant);
    }
}

void Chief::applyChangesToYLLParams(
        const QVariantMap& modelInfo, const QVariantMap& proportionData)
{
    fn::InputType::Country countryID = static_cast<fn::InputType::Country>(modelInfo["countryID"].toString());
    fn::InputType::Disease diseaseID = static_cast<fn::InputType::Disease>(modelInfo["diseaseID"].toString());
    fn::DiseaseModel::ID modelID = static_cast<fn::DiseaseModel::ID>(modelInfo["modelID"].toString());
    fn::OutcomeTree::ID outcomeTreeID = static_cast<fn::DiseaseModel::ID>(modelInfo["outcomeTreeID"].toString());
    fn::OutcomeTreeNode::ID healthStateID = static_cast<fn::DiseaseModel::ID>(modelInfo["outcomeTreeNodeID"].toString());

    fn::DiseaseModel* model = getDiseaseModel(countryID, diseaseID, modelID);

    if (model != nullptr) {

        QVariantMap formulaCheckVariant;
        fn::BoolBaseArray formulaCheck;
        QStringList paramsSuccess;
        QStringList paramsFail;

        fn::HealthState& healthState = model->getOutcomeTree(outcomeTreeID).getHealthState(healthStateID);

        fn::DataArray& proportionArray = healthState.getParameter(fn::HealthState::Proportion);
        const fn::ConstInputFactorPtrList& proportionFactors = proportionArray.getInputFactorList();
        if (!fn::FormulaChecker::checkFormulas(fn::VariantMapToFormulaArray(proportionData, proportionFactors), formulaCheck)) {
            formulaCheckVariant.insert("proportion", fn::BoolArrayToVariantMap(formulaCheck, proportionFactors));
            paramsFail.push_back("Proportion");
        } else {
            fn::StringBaseArray formulaArray = fn::VariantMapToFormulaArray(proportionData, proportionFactors);
            fn::StringToStringMap metaData = fn::VariantMapToMetaData(proportionData);
            proportionArray.setInputFormulaArray(formulaArray);
            proportionArray.setMetaData(metaData);
            paramsSuccess.push_back("Proportion");
        }

        if (paramsFail.length() > 0) {
            emit showMessageInYLLPage("YLL parameters",
                                      "Changes have NOT been applied to:<br /><ul><li>" + paramsFail.join("</li><li>") + "</li></ul>",
                                      "error");
        }

        if (paramsSuccess.length() > 0) {
            model->propagateChangesInModelParameters();
            emit showMessageInYLLPage("YLL parameters",
                                      "Changes have been applied to:<br /><ul><li>" + paramsSuccess.join("</li><li>") + "</li></ul>",
                                      "info");
        }

        emit formulaCheckResultsAvailableForYLLPage(formulaCheckVariant);
    }
}

void Chief::applyChangesToTransParams(
        const QVariantMap& modelInfo, const QVariantMap& probabilityData, const QVariantMap& redistributionData)
{
    fn::InputType::Country countryID = static_cast<fn::InputType::Country>(modelInfo["countryID"].toString());
    fn::InputType::Disease diseaseID = static_cast<fn::InputType::Disease>(modelInfo["diseaseID"].toString());
    fn::DiseaseModel::ID modelID = static_cast<fn::DiseaseModel::ID>(modelInfo["modelID"].toString());
    fn::OutcomeTree::ID outcomeTreeID = static_cast<fn::DiseaseModel::ID>(modelInfo["outcomeTreeID"].toString());
    fn::TransitionProbabilityBase::ID transProbID = static_cast<fn::DiseaseModel::ID>(modelInfo["outcomeTreeNodeID"].toString());

    fn::DiseaseModel* model = getDiseaseModel(countryID, diseaseID, modelID);

    if (model != nullptr) {

        QVariantMap formulaCheckVariant;
        fn::BoolBaseArray formulaCheck;
        QStringList paramsSuccess;
        QStringList paramsFail;

        fn::TransitionProbability& transProb = static_cast<fn::TransitionProbability&>(model->getOutcomeTree(outcomeTreeID).getTransitionProbability(transProbID));

        fn::DataArray& probabilityArray = transProb.getProbability();
        const fn::ConstInputFactorPtrList& probabilityFactors = probabilityArray.getInputFactorList();
        if (!fn::FormulaChecker::checkFormulas(fn::VariantMapToFormulaArray(probabilityData, probabilityFactors), formulaCheck)) {
            formulaCheckVariant.insert("probability", fn::BoolArrayToVariantMap(formulaCheck, probabilityFactors));
            paramsFail.push_back("Probability");
        } else {
            fn::StringBaseArray formulaArray = fn::VariantMapToFormulaArray(probabilityData, probabilityFactors);
            fn::StringToStringMap metaData = fn::VariantMapToMetaData(probabilityData);
            probabilityArray.setInputFormulaArray(formulaArray);
            transProb.setMetaData(metaData);
            paramsSuccess.push_back("Probability");
        }

        fn::DataArray& redistributionArray = transProb.getRedistribution();
        const fn::ConstInputFactorPtrList& redistributionFactors = redistributionArray.getInputFactorList();
        if (!fn::FormulaChecker::checkFormulas(fn::VariantMapToFormulaArray(redistributionData, redistributionFactors), formulaCheck)) {
            formulaCheckVariant.insert("redistribution", fn::BoolArrayToVariantMap(formulaCheck, redistributionFactors));
            paramsFail.push_back("Redistribution");
        } else {
            fn::StringBaseArray formulaArray = fn::VariantMapToFormulaArray(redistributionData, redistributionFactors);
            fn::StringToStringMap metaData = fn::VariantMapToMetaData(redistributionData);
            redistributionArray.setInputFormulaArray(formulaArray);
            redistributionArray.setMetaData(metaData);
            paramsSuccess.push_back("Redistribution");
        }

        if (paramsFail.length() > 0) {
            emit showMessageInTransPage("Transition parameters",
                                        "Changes have NOT been applied to:<br /><ul><li>" + paramsFail.join("</li><li>") + "</li></ul>",
                                        "error");
        }

        if (paramsSuccess.length() > 0) {
            emit showMessageInTransPage("Transition parameters",
                                        "Changes have been applied to:<br /><ul><li>" + paramsSuccess.join("</li><li>") + "</li></ul>",
                                        "info");
        }

        emit formulaCheckResultsAvailableForTransPage(formulaCheckVariant);
    }
}

void Chief::showOutcomeTreeElementDetails(const fn::OutcomeTree::ID& outcomeTreeId, const QString& elementType, const fn::OutcomeTreeNode::ID& elementId,
                                          const fn::InputType::Country& countryId, const fn::InputType::Disease& diseaseId, const QString& modelId)
{
    // Get model
    fn::DiseaseModel* model = getDiseaseModel(countryId, diseaseId, modelId);

    // Show outcome tree data in appropriate format
    if ((model != nullptr) && (!elementId.isEmpty())) {
        if (!m_presenterMgr.presenterExists(modelId, outcomeTreeId, elementId)) {
            ui::DataPresenter* presenter = nullptr;

            if (elementType == "infection") {
                presenter = showInfectionData(model, outcomeTreeId);
            } else if ((elementType == "symptomatic") || (elementType == "sequela")) {
                presenter = showYLDHealthStateData(model, outcomeTreeId, elementId);
            } else if (elementType == "death") {
                presenter = showYLLHealthStateData(model, outcomeTreeId, elementId);
            } else if (elementType == "asymptomatic") {
                presenter = showAsymptomaticData(model, outcomeTreeId, elementId);
            } else if (elementType == "arrow") {
                presenter = showTransProbData(model, outcomeTreeId, elementId);
            }

            if (presenter != nullptr) {
                m_presenterMgr.registerPresenter(presenter);
                connect(presenter, SIGNAL(gettingDestroyed(ui::DataPresenter*)),
                        &m_presenterMgr, SLOT(unregisterPresenter(ui::DataPresenter*)));
            }
        } else {
            emit showMessage("Outcome tree", "Pop-up window for this element is already displayed", "info");
        }
    }
}

ui::DataPresenter* Chief::showInfectionData(const fn::DiseaseModel* diseaseModel, const fn::OutcomeTree::ID& outcomeTreeID)
{
    // Cache references
    const fn::Infection& infection = diseaseModel->getOutcomeTree(outcomeTreeID).getInfection();
    const fn::ConstInputFactorPtrList& inputFactors = diseaseModel->getInputFactorsList();

    // Meta data
    QString caption = infection.getCaption();
    QVariantMap modelInfo = diseaseModel->getVariantModelInfo();
    modelInfo["outcomeTreeID"] = outcomeTreeID;
    modelInfo["outcomeTreeNodeID"] = infection.getId();

    // Input data
    QVariantMap lifeExpData = fn::DataArrayToVariantMap(infection.getLifeExpectancy(), fn::DataArray::Formula, fn::DataArray::Run);

    // Results (if ready)
    bool resultsAvailable = diseaseModel->isModelSampled();

    QVariantMap sampleMeanOutputCasesCount = fn::NumericArrayToVariantMap(infection.getSampleMeanCasesCount(), inputFactors);

    ui::InfectionDataPresenter* presenter = new ui::InfectionDataPresenter(modelInfo, caption, resultsAvailable, lifeExpData,
                                                                           sampleMeanOutputCasesCount, m_webGUI);

    presenter->show();
    return presenter;
}

ui::DataPresenter* Chief::showYLDHealthStateData(const fn::DiseaseModel* diseaseModel, const fn::OutcomeTree::ID& outcomeTreeID, const fn::OutcomeTreeNode::ID& healthStateID)
{
    // Cache references
    const fn::HealthState& healthState = diseaseModel->getOutcomeTree(outcomeTreeID).getHealthState(healthStateID);
    const fn::ConstInputFactorPtrList& inputFactors = diseaseModel->getInputFactorsList();

    // Meta data
    QString caption = healthState.getCaption();
    QVariantMap modelInfo = diseaseModel->getVariantModelInfo();
    modelInfo["outcomeTreeID"] = outcomeTreeID;
    modelInfo["outcomeTreeNodeID"] = healthStateID;

    // Input data
    QVariantMap severityData = fn::DataArrayToVariantMap(healthState.getParameter(fn::HealthState::Severity), fn::DataArray::Formula, fn::DataArray::Input);
    QVariantMap durationData = fn::DataArrayToVariantMap(healthState.getParameter(fn::HealthState::Duration), fn::DataArray::Formula, fn::DataArray::Input);
    QVariantMap proportionData = fn::DataArrayToVariantMap(healthState.getParameter(fn::HealthState::Proportion), fn::DataArray::Formula, fn::DataArray::Input);

    // Results (if ready)
    bool resultsAvailable = diseaseModel->isModelSampled();

    QVariantMap casesCountData = fn::NumericArrayToVariantMap(healthState.getSampleMeanCasesCount(), inputFactors);
    QVariantMap burdenData = fn::NumericArrayToVariantMap(healthState.getSampleMeanBurden(), inputFactors);
    QVariantMap remLifeExpData = fn::NumericArrayToVariantMap(healthState.getOutputRemLifeExp(), inputFactors);

    ui::YLDHealthOutcomeDataPresenter* presenter =
            new ui::YLDHealthOutcomeDataPresenter(modelInfo, caption, resultsAvailable, severityData,
                                                  durationData, proportionData, casesCountData, burdenData,
                                                  remLifeExpData, m_webGUI);
    presenter->show();
    return presenter;
}


ui::DataPresenter* Chief::showYLLHealthStateData(const fn::DiseaseModel* diseaseModel, const fn::OutcomeTree::ID& outcomeTreeID, const fn::OutcomeTreeNode::ID& healthStateID)
{
    ui::YLLHealthOutcomeDataPresenter* presenter = nullptr;

    // Cache references
    const fn::HealthState& healthState = diseaseModel->getOutcomeTree(outcomeTreeID).getHealthState(healthStateID);
    const fn::ConstInputFactorPtrList& inputFactors = diseaseModel->getInputFactorsList();

    // Meta data
    QString caption = healthState.getCaption();
    QVariantMap modelInfo = diseaseModel->getVariantModelInfo();
    modelInfo["outcomeTreeID"] = outcomeTreeID;
    modelInfo["outcomeTreeNodeID"] = healthStateID;

    // Results (if ready)
    bool resultsAvailable = diseaseModel->isModelSampled();

    if (resultsAvailable) {
        // Input data
        QVariantMap proportionData = fn::DataArrayToVariantMap(healthState.getParameter(fn::HealthState::Proportion), fn::DataArray::Formula, fn::DataArray::Run);

        QVariantMap casesCountData = fn::NumericArrayToVariantMap(healthState.getSampleMeanCasesCount(), inputFactors);
        QVariantMap burdenData = fn::NumericArrayToVariantMap(healthState.getSampleMeanBurden(), inputFactors);
        QVariantMap remLifeExpData = fn::NumericArrayToVariantMap(healthState.getOutputRemLifeExp(), inputFactors);

        presenter = new ui::YLLHealthOutcomeDataPresenter(modelInfo, caption, resultsAvailable,
                                                          proportionData, casesCountData,
                                                          burdenData, remLifeExpData, m_webGUI);
        presenter->show();
    } else {
        showMessage("YLL health outcome",
                    "There are no parameters defined for this health outcome.<br />Results are available only after model run.",
                    "error");
    }

    return presenter;
}

ui::DataPresenter* Chief::showAsymptomaticData(const fn::DiseaseModel* diseaseModel, const fn::OutcomeTree::ID& outcomeTreeID, const fn::OutcomeTreeNode::ID& healthStateID)
{
    ui::AsymptomaticDataPresenter* presenter = nullptr;

    // Cache references
    const fn::HealthState& healthState = diseaseModel->getOutcomeTree(outcomeTreeID).getHealthState(healthStateID);
    const fn::ConstInputFactorPtrList& inputFactors = diseaseModel->getInputFactorsList();

    // Meta data
    QString caption = healthState.getCaption();
    QVariantMap modelInfo = diseaseModel->getVariantModelInfo();
    modelInfo["outcomeTreeID"] = outcomeTreeID;
    modelInfo["outcomeTreeNodeID"] = healthStateID;

    // Results (if ready)
    bool resultsAvailable = diseaseModel->isModelSampled();

    if (resultsAvailable) {
        QVariantMap casesCountData = fn::NumericArrayToVariantMap(healthState.getSampleMeanCasesCount(), inputFactors);

        presenter = new ui::AsymptomaticDataPresenter(modelInfo, caption, resultsAvailable, casesCountData, m_webGUI);
        presenter->show();
    } else {
        showMessage("Asymptomatic health outcome",
                    "There are no parameters defined for this health outcome.<br />Results are available only after model run.",
                    "error");
    }
    return presenter;
}

ui::DataPresenter* Chief::showTransProbData(const fn::DiseaseModel* diseaseModel, const fn::OutcomeTree::ID& outcomeTreeID, const fn::TransitionProbability::ID& transProbID)
{
    ui::TransProbDataPresenter* presenter = nullptr;

    // Cache references
    const fn::TransitionProbabilityBase& transProb = diseaseModel->getOutcomeTree(outcomeTreeID).getTransitionProbability(transProbID);
    const fn::ConstInputFactorPtrList& inputFactors = diseaseModel->getInputFactorsList();

    // Meta data
    QString caption = transProb.getCaption();
    QVariantMap modelInfo = diseaseModel->getVariantModelInfo();
    modelInfo["outcomeTreeID"] = outcomeTreeID;
    modelInfo["outcomeTreeNodeID"] = transProbID;

    // Base transition
    bool isBaseTransProb = transProb.getTransProbSourceType() == fn::TransitionProbabilityBase::Base;

    // Input data
    QVariantMap inputProbability;
    QVariantMap inputRedistribution;
    if (isBaseTransProb) {
        inputProbability = fn::DataArrayToVariantMap(static_cast<const fn::TransitionProbability&>(transProb).getProbability(), fn::DataArray::Formula, fn::DataArray::Input);
        inputRedistribution = fn::DataArrayToVariantMap(static_cast<const fn::TransitionProbability&>(transProb).getRedistribution(), fn::DataArray::Formula, fn::DataArray::Input);
    }

    // Results (if ready)
    bool resultsAvailable = diseaseModel->isModelSampled();

    if (isBaseTransProb || resultsAvailable) {
        QVariantMap outputProbabilityData = fn::NumericArrayToVariantMap(transProb.getSampleMeanProbability(), inputFactors);
        QVariantMap outputRedistributionData = fn::NumericArrayToVariantMap(transProb.getSampleMeanRedistribution(), inputFactors);
        QVariantMap casesCountData = fn::NumericArrayToVariantMap(transProb.getSampleMeanOutputCasesCount(), inputFactors);

        bool isRedistributeType = transProb.isRedistributeType();

        presenter = new ui::TransProbDataPresenter(modelInfo, caption, resultsAvailable, isBaseTransProb, isRedistributeType,
                                                   inputProbability, inputRedistribution,
                                                   outputProbabilityData, outputRedistributionData, casesCountData, m_webGUI);
        presenter->show();
    } else {
        showMessage("Transition probability",
                    "Please, browse parameters of the source transition probability.<br />Results for this transition are available only after model run.",
                    "error");
    }

    return presenter;
}

void Chief::updateDataExportImportStatus(
        const fn::FileOperationType& operationType,
        const QString& dataType,
        const QString& text1,
        const QString& text2)
{
    qApp->processEvents(QEventLoop::ExcludeUserInputEvents);

    QString header = "export";
    if (operationType == fn::FileOperationType::Import) {
        header = "import";
    }

    QString title = QString("%1 %2").arg(dataType).arg(header);
    if (text1 == QString("%2Started").arg(header)) {
        QString message = QString("%1 %2 started").arg(dataType).arg(header);
        emit createUpdatableMessage(title, message);
    } else if (text1 == QString("%2Finished").arg(header)) {
        QString message = QString("%1 %2 finished").arg(dataType).arg(header);
        emit destroyUpdatableMessage();
        emit showMessage(title, message, "info");
    } else {
        QString message = QString("Exporting file: %1<br />Sheet: %2").arg(text1).arg(text2);
        if (operationType == fn::FileOperationType::Import) {
            message = QString("Importing file: %1<br />Sheet: %2").arg(text1).arg(text2);
        }
        emit updateMessage(title, message);
    }
}

void Chief::setNumSamples(const int numSamples)
{
    m_numSamples = numSamples;
}

void Chief::setTimeDiscountRate(const double timeDiscountRate)
{
    m_timeDiscountRate = timeDiscountRate;
}

void Chief::runModels(const bool runInThread = true)
{
    try {
        m_randomSeed = m_settings.getRandomSeed();
        m_resultsComputer.initializeSample(m_numSamples);

        fn::ModelRunWorker* modelRunWorker = new fn::ModelRunWorker(m_numSamples, m_randomSeed, m_timeDiscountRate, m_diseaseModelList);

        // Signal worker process start
        connect(modelRunWorker, SIGNAL(started()), this, SLOT(onRunStarted()));

        // Signal worker process end
        connect(modelRunWorker, SIGNAL(finished()), this, SLOT(onRunFinished()));

        // Signal worker process progress info
        connect(modelRunWorker, SIGNAL(progressInfo(float, int, int, QString)),
                this, SIGNAL(runProgressUpdated(float, int, int, QString)));

        // Clean up
        connect(modelRunWorker, SIGNAL(finished()), modelRunWorker, SLOT(deleteLater()));

        if (runInThread) {
            QThread* modelRunThread = new QThread;

            modelRunWorker->moveToThread(modelRunThread);

            // Start worker process
            connect(modelRunThread, SIGNAL(started()), modelRunWorker, SLOT(process()));

            // Clean up
            connect(modelRunWorker, SIGNAL(finished()), modelRunThread, SLOT(quit()));
            connect(modelRunThread, SIGNAL(finished()), modelRunThread, SLOT(deleteLater()));

            // Start calculations
            modelRunThread->start();
        } else {
            modelRunWorker->process();
        }
    } catch(...) {
        UT_THROW("Problem occured while running calculations");
    }
}

void Chief::onRunStarted()
{
    emit runStarted();
}

void Chief::computeAggregatedResults()
{
    // Compute results
    for (int i = 0; i < m_diseaseModelList.length(); ++i) {
        fn::DiseaseModel& diseaseModel = m_diseaseModelList[i];
        m_resultsComputer.addTotalPopulation(diseaseModel.getTotalPopulation());
        m_resultsComputer.addYLDSample(diseaseModel.getSampleYLD());
        m_resultsComputer.addYLLSample(diseaseModel.getSampleYLL());
        m_resultsComputer.addAgeDistrSample(diseaseModel.getSampleAgeDistr());
        m_resultsComputer.addNumSymptomaticCasesCountSample(diseaseModel.getSampleNumSymptomaticCasesCount());
        m_resultsComputer.addDeathCasesCountSample(diseaseModel.getSampleDeathCasesCount());
        diseaseModel.clearSample();
    }
    m_resultsComputer.computeAggregatedResults();
}

void Chief::showAggregatedResults()
{
    QVariantMap result = m_resultsComputer.getAggregatedResults();
    emit aggregatedResultsAvailable(result);
}

QByteArray Chief::getDetailedResultsArray()
{
    QJsonArray jsonArray;
    for (size_t i = 0; i < getModelCount(); ++i) {
        const fn::DiseaseModel& model = getModel(i);
        const QVariantMap results = model.getDetailedPageResults();
        QJsonObject jsonResults = QJsonObject::fromVariantMap(results);
        jsonResults["name"] = model.getCaption();
        jsonArray.append(jsonResults);
    }

    QJsonDocument jsonDocument(jsonArray);

    return jsonDocument.toJson();
}

void Chief::onRunFinished()
{
    computeAggregatedResults();
    showAggregatedResults();

    // Notify UI which models have been computed
    QVariantList modelInfo;
    for (int i = 0; i < m_diseaseModelList.length(); ++i) {
        const fn::DiseaseModel& diseaseModel = m_diseaseModelList[i];
        QVariantMap map;
        map["countryID"] = diseaseModel.getCountry();
        map["diseaseID"] = diseaseModel.getDisease();
        map["modelID"] = diseaseModel.getId();
        map["countryName"] = diseaseModel.getCountryName();
        map["diseaseName"] = diseaseModel.getDiseaseName();
        modelInfo.append(map);
    }
    emit runFinished(modelInfo);
}

void Chief::saveModelsToExcel(const QFileInfo& fileInfo)
{
    if (m_diseaseModelList.size() == 0) {
        emit showMessage("Export models", "There are no models to export.\nCreate models first and then repeat this action.", "error");
        return;
    }

    QString fileName = fileInfo.absoluteFilePath();
    fileName.remove(".xls");
    fileName.replace("/", "\\");
    if (!fileName.isNull()) {

         updateDataExportImportStatus(fn::FileOperationType::Export, "Models", "exportStarted", "");

        // Run Excel application
        ut::Excel excelApp(fileName);

        int initSheetCount = excelApp.getWorksheetCount();

        for (int i = 0; i < m_diseaseModelList.size(); ++i) {

            fn::DiseaseModel* diseaseModel = &m_diseaseModelList[i];

            updateDataExportImportStatus(fn::FileOperationType::Export, "Models", fileInfo.baseName(), diseaseModel->getId());

            QAxObject* sheet = nullptr;
            if (i < initSheetCount) {
                sheet = excelApp.setCurrentSheet(i + 1);
            } else {
                sheet = excelApp.appendWorksheet(QString("Export %1").arg(i + 1));
            }

            ut::Excel::setWorksheetName(sheet, diseaseModel->getId());

            int startRow = 0;

            // Set number format of the first column to "TEXT"
            ut::Excel::setProperty(excelApp.getRange("A:A"), "NumberFormat", "@");

            // Set source information
            QAxObject* headerCell = excelApp.getCell(++startRow, 1);
            ut::Excel::setCellValue(headerCell, QVariant("File info"));

            excelApp.setCellValue(++startRow + 1, 1, QVariant("Version"));
            excelApp.setCellValue(++startRow, 2, QVariant(6));

            // Set model info header text
            startRow++;
            headerCell = excelApp.getCell(++startRow, 1);
            ut::Excel::setCellValue(headerCell, QVariant("Model info"));

            // Save model info values
            startRow++;
            excelApp.setCellValue(++startRow, 1, QVariant("CountryID"));
            excelApp.setCellValue(startRow, 2, QVariant(diseaseModel->getCountry()));
            excelApp.setCellValue(++startRow, 1, QVariant("DiseaseID"));
            excelApp.setCellValue(startRow, 2, QVariant(diseaseModel->getDisease()));
            excelApp.setCellValue(++startRow, 1, QVariant("ModelID"));
            excelApp.setCellValue(startRow, 2, QVariant(diseaseModel->getId()));

            // Save outcome tree mapping data
            startRow++;
            fn::writeInputDataToExcel(excelApp.getCell(++startRow, 1), excelApp, diseaseModel->getParameter(fn::DiseaseModel::OutcomeTreeMapping), startRow);

            // Save notification data
            fn::writeInputDataToExcel(excelApp.getCell(++startRow, 1), excelApp, diseaseModel->getParameter(fn::DiseaseModel::RawCasesCount), startRow);

            // Set corrections header text
            headerCell = excelApp.getCell(++startRow, 1);
            ut::Excel::setCellValue(headerCell, QVariant("Corrections"));

            // Save coverage parameters
            startRow++;
            QVariantList variantArray;
            QVariant variantEmptyRow = QVariant(QVariantList() << QVariant() << QVariant() << QVariant());

            variantArray.clear();
            variantArray.append(fn::DataArrayToVariantList(diseaseModel->getParameter(fn::DiseaseModel::Coverage)));
            variantArray.append(variantEmptyRow);
            variantArray.append(fn::DataArrayToVariantList(diseaseModel->getParameter(fn::DiseaseModel::Underestimation)));
            variantArray.append(variantEmptyRow);
            variantArray.append(fn::DataArrayToVariantList(diseaseModel->getParameter(fn::DiseaseModel::Underreporting)));
            variantArray.append(variantEmptyRow);
            variantArray.append(fn::DataArrayToVariantList(diseaseModel->getParameter(fn::DiseaseModel::Underascertainment)));
            excelApp.setRangeValues(QPoint(++startRow, 1), variantArray, 3, startRow);

            // Set underestimation type
            int underEstimType = diseaseModel->getUnderestimationType();
            headerCell = excelApp.getCell(++startRow, 1);
            ut::Excel::setCellValue(headerCell, QVariant("underestimation type"));
            startRow++;
            excelApp.setCellValue(++startRow, 1, "Type");
            excelApp.setCellValue(startRow, 2, QString("%1").arg(underEstimType));

            // Set population data header text
            startRow++;
            headerCell = excelApp.getCell(++startRow, 1);
            ut::Excel::setCellValue(headerCell, QVariant("Population data"));

            // Save life expectancy and age distribution parameters
            startRow++;
            variantArray.clear();
            variantArray.append(fn::DataArrayToVariantList(diseaseModel->getParameter(fn::DiseaseModel::LifeExpectancy)));
            variantArray.append(variantEmptyRow);
            variantArray.append(fn::DataArrayToVariantList(diseaseModel->getParameter(fn::DiseaseModel::AgeDistribution)));
            excelApp.setRangeValues(QPoint(++startRow, 1), variantArray, 3, startRow);

            //Go over outcome trees and export health state and transition probability data
            for (int j = 0; j < diseaseModel->getOutcomeTreeCount(); ++j) {
                const fn::OutcomeTree& outcomeTree = diseaseModel->getOutcomeTree(j);
                // Set outcome tree header text
                headerCell = excelApp.getCell(++startRow, 1);
                ut::Excel::setCellValue(headerCell, QVariant(outcomeTree.getId()));

                // Set parameters header text
                startRow++;
                headerCell = excelApp.getCell(++startRow, 1);
                ut::Excel::setCellValue(headerCell, QVariant(QString("Parameters - %1").arg(outcomeTree.getId())));

                startRow++;
                for (int k = 0; k < outcomeTree.getHealthStateCount(); ++k) {
                    const fn::HealthState& healthState = outcomeTree.getHealthState(k);
                    if (healthState.getType() == fn::InputType::HealthState::YLD) {
                        headerCell = excelApp.getCell(++startRow, 1);
                        ut::Excel::setCellValue(headerCell, QVariant(healthState.getId()));

                        startRow++;
                        variantArray.clear();
                        variantArray.append(fn::DataArrayToVariantList(healthState.getParameter(fn::HealthState::Severity)));
                        variantArray.append(variantEmptyRow);
                        variantArray.append(fn::DataArrayToVariantList(healthState.getParameter(fn::HealthState::Duration)));
                        variantArray.append(variantEmptyRow);
                        variantArray.append(fn::DataArrayToVariantList(healthState.getParameter(fn::HealthState::Proportion)));
                        excelApp.setRangeValues(QPoint(++startRow, 1), variantArray, 3, startRow);
                    }
                }

                // Set transition probabilities header
                headerCell = excelApp.getCell(++startRow, 1);
                ut::Excel::setCellValue(headerCell, QVariant(QString("Transitions - %1").arg(outcomeTree.getId())));

                startRow++;
                for (int k = 0; k < outcomeTree.getTransProbCount(); ++k) {
                    const fn::TransitionProbability& transProb = outcomeTree.getTransitionProbability(k);

                    headerCell = excelApp.getCell(++startRow, 1);
                    ut::Excel::setCellValue(headerCell, QVariant(transProb.getId()));

                    startRow++;
                    variantArray.clear();
                    variantArray.append(fn::DataArrayToVariantList(transProb.getProbability()));
                    variantArray.append(variantEmptyRow);
                    variantArray.append(fn::DataArrayToVariantList(transProb.getRedistribution()));
                    excelApp.setRangeValues(QPoint(++startRow, 1), variantArray, 3, startRow);
                }
            }
        }

        // Remove unused sheets
        for (int i = initSheetCount; i > m_diseaseModelList.size(); --i) {
            excelApp.deleteWorksheet(i);
        }

        // Close model data workbook
        excelApp.saveWorkbook(fileName);

        excelApp.closeAll();

        updateDataExportImportStatus(fn::FileOperationType::Export, "Models", "exportFinished", "");
    }
}

void Chief::loadModelsFromExcel(const QFileInfo& fileInfo)
{
    const QString kModelInfoHeaderText = "Model info";
    const QString kOutcomeTreeMappingDataHeaderText = "Outcome tree mapping";
    const QString kNotificationDataHeaderText = "Notification data";
    const QString kCoverageDataString = "coverage";
    const QString kUnderestimationDataString = "underestimation";
    const QString kUnderreportingDataString = "underreporting";
    const QString kUnderascertainmentDataString = "underascertainment";
    const QString kUnderEstimationType = "underestimation type";
    const QString kLifeExpectancyDataString = "life expectancy";
    const QString kAgeDistributionDataString = "age distribution";

    QVariantList newModelsInfo;
    if (fileInfo.exists()) {

        updateDataExportImportStatus(fn::FileOperationType::Import, "Models", "importStarted", "");

        // Run Excel application
        ut::Excel excelApp(fileInfo.absoluteFilePath());

        int sheetCount = excelApp.getWorksheetCount();

        for (int i = 0; i < sheetCount; ++i) {
            updateDataExportImportStatus(fn::FileOperationType::Import, "Models", fileInfo.baseName(), excelApp.getCurrentSheetName());

            int headerRowIdx = 0;
            excelApp.setCurrentSheet(i + 1);

            // Load version number
            int fileVersion = excelApp.getCellValue(3, 2).toInt();
            if (fileVersion != 6) {
                emit showMessage("Model import",
                                 "Sheet " + excelApp.getCurrentSheetName() + " contains model created by the previous, unsupported version of the tool. It will not be imported.",
                                 "error");
                continue;
            }

            // Load meta data
            QAxObject* headerCell = excelApp.getCellWithText(kModelInfoHeaderText, "A");
            headerRowIdx = ut::Excel::getCellRow(headerCell);
            fn::InputType::Country country = excelApp.getCellValue(headerRowIdx + 2, 2).toString();
            fn::InputType::Disease disease = excelApp.getCellValue(headerRowIdx + 3, 2).toString();
            fn::DiseaseModel::ID modelID = allocateNewModelID(excelApp.getCellValue(headerRowIdx + 4, 2).toString());

            // Create a new model
            fn::DiseaseModel* diseaseModel = createNewDiseaseModel(country, disease, modelID);
            if (diseaseModel == nullptr) {
                emit showMessage("Model import",
                                 "Sheet " + excelApp.getCurrentSheetName() + " contains model not supported by the reference data loaded. It will not be imported.",
                                 "error");
                continue;
            }

            int lastRow = 0;

            // Load outcome tree mapping data
            headerCell = excelApp.getCellWithText(kOutcomeTreeMappingDataHeaderText, "A");
            fn::DataArray outcomeTreeMapping = fn::readInputDataFromExcel(headerCell, excelApp, lastRow);

            // Load notification data
            headerCell = excelApp.getCellWithText(kNotificationDataHeaderText, "A");
            fn::DataArray notificationData = fn::readInputDataFromExcel(headerCell, excelApp, lastRow);

            // Load correction data
            headerCell = excelApp.getCellWithText(kCoverageDataString, "A");
            fn::DataArray coverage = fn::readInputDataFromExcel(headerCell, excelApp, lastRow);

            headerCell = excelApp.getCellWithText(kUnderestimationDataString, "A");
            fn::DataArray underestimation = fn::readInputDataFromExcel(headerCell, excelApp, lastRow);

            headerCell = excelApp.getCellWithText(kUnderreportingDataString, "A");
            fn::DataArray underreporting = fn::readInputDataFromExcel(headerCell, excelApp, lastRow);

            headerCell = excelApp.getCellWithText(kUnderascertainmentDataString, "A");
            fn::DataArray underascertainment = fn::readInputDataFromExcel(headerCell, excelApp, lastRow);

            headerRowIdx = ut::Excel::getCellRow(excelApp.getCellWithText(kUnderEstimationType, "A"));
            int underestimType = excelApp.getCellValue(headerRowIdx + 2, 2).toInt();

            headerCell = excelApp.getCellWithText(kLifeExpectancyDataString, "A");
            fn::DataArray lifeExp = fn::readInputDataFromExcel(headerCell, excelApp, lastRow);

            headerCell = excelApp.getCellWithText(kAgeDistributionDataString, "A");
            fn::DataArray ageDistr = fn::readInputDataFromExcel(headerCell, excelApp, lastRow);

            diseaseModel->setParameter(fn::DiseaseModel::OutcomeTreeMapping, outcomeTreeMapping);
            diseaseModel->setParameter(fn::DiseaseModel::RawCasesCount, notificationData);
            diseaseModel->setParameter(fn::DiseaseModel::Coverage, coverage);
            diseaseModel->setParameter(fn::DiseaseModel::Underestimation, underestimation);
            diseaseModel->setParameter(fn::DiseaseModel::Underreporting, underreporting);
            diseaseModel->setParameter(fn::DiseaseModel::Underascertainment, underascertainment);
            diseaseModel->setParameter(fn::DiseaseModel::LifeExpectancy, lifeExp);
            diseaseModel->setParameter(fn::DiseaseModel::AgeDistribution, ageDistr);

            diseaseModel->setUnderestimationType(underestimType);

            //Go over outcome trees and import health state and transition probability data
            for (int j = 0; j < diseaseModel->getOutcomeTreeCount(); ++j) {
                fn::OutcomeTree& outcomeTree = diseaseModel->getOutcomeTree(j);
                int lastRow = 0;

                // Health states
                for (int k = 0; k < outcomeTree.getHealthStateCount(); ++k) {
                    fn::HealthState& healthState = outcomeTree.getHealthState(k);

                    if (healthState.getType() == fn::InputType::HealthState::YLD) {
                        const fn::OutcomeTreeNode::ID healthStateId = healthState.getId();
                        headerCell = excelApp.getCellWithText(healthStateId);

                        QAxObject* severityHeaderCell = excelApp.getCell(ut::Excel::getCellRow(headerCell) + 2, 1);
                        fn::DataArray severity = fn::readInputDataFromExcel(severityHeaderCell, excelApp, lastRow);
                        healthState.setParameter(fn::HealthState::Severity, severity);

                        QAxObject* durationHeaderCell = excelApp.getCell(lastRow + 2, 1);
                        fn::DataArray duration = fn::readInputDataFromExcel(durationHeaderCell, excelApp, lastRow);
                        healthState.setParameter(fn::HealthState::Duration, duration);

                        QAxObject* proportionHeaderCell = excelApp.getCell(lastRow + 2, 1);
                        fn::DataArray proportion = fn::readInputDataFromExcel(proportionHeaderCell, excelApp, lastRow);
                        healthState.setParameter(fn::HealthState::Proportion, proportion);
                    }
                }

                // Transition probabilities
                for (int k = 0; k < outcomeTree.getTransProbCount(); ++k) {
                    fn::TransitionProbability& transProb = outcomeTree.getTransitionProbability(k);
                    const fn::TransitionProbabilityBase::ID transProbId = transProb.getId();
                    headerCell = excelApp.getCellWithText(transProbId);

                    QAxObject* probabilityHeaderCell = excelApp.getCell(ut::Excel::getCellRow(headerCell) + 2, 1);
                    fn::DataArray probability = fn::readInputDataFromExcel(probabilityHeaderCell, excelApp, lastRow);
                    transProb.setProbability(probability);

                    QAxObject* redistributionHeaderCell = excelApp.getCell(lastRow + 2, 1);
                    fn::DataArray redistribution = fn::readInputDataFromExcel(redistributionHeaderCell, excelApp, lastRow);
                    transProb.setRedistribution(redistribution);
                }
            }

            diseaseModel->propagateChangesInModelParameters();

            if (diseaseModel != nullptr) {
                newModelsInfo.append(diseaseModel->getVariantModelInfo());
            }
        }

        excelApp.closeAll();

        if (newModelsInfo.size() > 0) {
            updateDataExportImportStatus(fn::FileOperationType::Import, "Models", "importFinished", "");
            emit modelsCreated(newModelsInfo);
        } else {
            emit showMessage("Model import", "No valid models found in the input file", "error");
        }

    } else {
        emit showMessage("Model import", "Invalid input file", "error");
    }
}

void Chief::loadModelsFromJSON(const QFileInfoList& fileInfos, const fn::InputType::Country& country)
{
    m_jsonInputDataMgr.setFilePaths(fileInfos);

    QStringList diseaseIds = m_jsonInputDataMgr.getDiseaseIds();
    for (int i = 0; i < diseaseIds.size(); ++i) {
        fn::InputType::Disease disease = diseaseIds[i];
        fn::DiseaseModel::ID modelID = allocateNewModelID();
        fn::DiseaseModel* diseaseModel = createNewDiseaseModel(country, disease, modelID);
        diseaseModel->propagateChangesInModelParameters();
    }
}

} // namespace ui
