//==========================================================================================================================
//  'Chief' class
//      Top-level management class responsible for operation of the entire application.
//
//  Written by Daniel Lewandowski
//  Copyright @ 2010 NextPage Software
//==========================================================================================================================

#ifndef _ui_Chief_
#define _ui_Chief_

#include <stdio.h>
#include <QObject>
#include <QMap>
#include <QVariantList>
#include "external/ganalytics/ganalytics.h"
#include "fn/globals.h"
#include "fn/DiseaseReportMgr.h"
#include "fn/DiseaseModel.h"
#include "fn/InputDataMgr.h"
#include "fn/JSONInputDataMgr.h"
#include "fn/ResultsComputer.h"
#include "ui/PresenterMgr.h"
#include "fn/Settings.h"

namespace ut {
    class Excel;
}

namespace ui {

class WebGUI;

class Chief : public QObject
{

    Q_OBJECT

public:

    // Create instance.
    static Chief& getInstance();

    // Create new disease model.
    //
    // 'country' - [in] Country the model is specified for.
    // 'disease' - [in] Disease the model is specified for.
    // 'modelID' - [in] Model ID.
    //
    // Return - TRUE if model created successfully
    fn::DiseaseModel* createNewDiseaseModel(const fn::InputType::Country& country, const fn::InputType::Disease& disease, const fn::DiseaseModel::ID& modelID);

    // Get disease model
    //
    // 'country' - [in] Country index
    // 'disease' - [in] Disease index
    // 'modelID' - [in] Model index
    //
    // Return - pointer to disease model if found, NULL otherwise
    fn::DiseaseModel* getDiseaseModel(const fn::InputType::Country& country, const fn::InputType::Disease& disease, const QString& modelID);

    // Set application path
    //
    // 'appPath' - [in] New application path
    //
    void setAppPath(const QString& appPath);
    const QString& getAppPath() const;

    // Get application settings object
    fn::Settings& getAppSettings();

    // Get input data manager
    //
    // Return - Reference to the input data manager
    const fn::InputDataMgr& getInputDataMgr() const;
          fn::InputDataMgr& getInputDataMgr();

    // Get JSON input data manager
    //
    // Return - Reference to the input data manager
    const fn::JSONInputDataMgr& getJSONInputDataMgr() const;
          fn::JSONInputDataMgr& getJSONInputDataMgr();

    // Set web based interface
    //
    // 'webGUI' - [in] Web interface
    //
    void setWebGUI(ui::WebGUI* webGUI);

    // Get list of available diseases
    //
    QVariantList getAvailableDiseases();

    // Get the number of samples per variable
    //
    int getNumSamples() const;

    // Get the time discount rate
    //
    double getTimeDiscountRate() const;

    // Get model count
    //
    int getModelCount() const;

    // Get model by index
    //
    const fn::DiseaseModel& getModel(const size_t index) const;

    // Get results computer
    //
    const fn::ResultsComputer& getResultsComputer() const;

public slots:

    // Import reference data.
    //
    void importReferenceData();

    // Import reference data from JSON.
    //
    void importReferenceDataFromJSON();

    // JavaScript to C++ notifications
    void pageOpened(int tabIndex, const QString& pageName);
    void changeCountrySelection(const QStringList& countryIDStrs);
    void changeDiseaseSelection(const QStringList& diseaseIDStrs);
    void createModels();
    void deleteModels();
    void selectModel(const QString& countryId, const QString& diseaseId, const QString& modelId);
    void showOutcomeTreeElementDetails(const fn::OutcomeTree::ID& outcomeTreeId, const QString& elementType, const fn::OutcomeTreeNode::ID& elementId,
                                       const fn::InputType::Country& countryId, const fn::InputType::Disease& diseaseId, const QString& modelId);
    void applyChangesToModelParams(const QVariantMap& modelInfo, const int estimType, const QVariantMap& underestimationData,
                                   const QVariantMap& underreportingData, const QVariantMap& underascertainmentData,
                                   const QVariantMap& caseBeforeCorrData, const QVariantMap& lifeExpData, const QVariantMap& ageDistrData);
    void applyChangesToYLDParams(const QVariantMap& modelInfo, const QVariantMap& severityData, const QVariantMap& durationData,
                                 const QVariantMap& proportionData);
    void applyChangesToYLLParams(const QVariantMap& modelInfo, const QVariantMap& proportionData);
    void applyChangesToTransParams(const QVariantMap& modelInfo, const QVariantMap& probabilityData, const QVariantMap& redistributionData);
    void updateDataExportImportStatus(const fn::FileOperationType& operationType, const QString& dataType, const QString& text1, const QString& text2);
    void setNumSamples(const int numSamples);
    void setTimeDiscountRate(const double timeDiscountRate);
    void runModels(const bool runInThread);
    void showDetailedResults(const fn::InputType::Country& countryId, const fn::InputType::Disease& diseaseId, const QString& modelId);

    // Save disease models to Excel
    void saveModelsToExcel(const QFileInfo& fileInfo);

    // Load disease models from Excel
    void loadModelsFromExcel(const QFileInfo& fileInfo);

    // Load disease models from JSON
    void loadModelsFromJSON(const QFileInfoList& fileInfos, const fn::InputType::Country& country);

    // Save detailed results to text file in JSON format
    void saveDetailedResultsToJSON(const QFileInfo& fileInfo);

    // Saved detailed results to clipboard in JSON format
    void saveDetailedResultsToClipboard();

private slots:
    void onRunStarted();
    void onRunFinished();

signals:

    void programReinitialized();
    void modelsCreated(const QVariantList&);
    void modelSelected(const QVariantMap&, const QVariantMap&, const QVariantMap&, const QVariantMap&, const QVariantMap&, const QString&);
    void formulaCheckResultsAvailable(const QVariantMap&);
    void formulaCheckResultsAvailableForYLDPage(const QVariantMap&);
    void formulaCheckResultsAvailableForYLLPage(const QVariantMap&);
    void formulaCheckResultsAvailableForTransPage(const QVariantMap&);
    void runStarted();
    void runFinished(const QVariantList&);
    void detailedResultsAvailable(const QVariantMap&);
    void aggregatedResultsAvailable(const QVariantMap&);
    void runProgressUpdated(const float, const int, const int, const QString&);
    void showMessage(const QString&, const QString&, const QString&);
    void showMessageInYLDPage(const QString&, const QString&, const QString&);
    void showMessageInYLLPage(const QString&, const QString&, const QString&);
    void showMessageInTransPage(const QString&, const QString&, const QString&);
    void createUpdatableMessage(const QString&, const QString&);
    void updateMessage(const QString&, const QString&);
    void destroyUpdatableMessage();

private:

    // Map of disease models
    typedef QMap<fn::DiseaseModel::ID, fn::DiseaseModel*> IDModelMap;
    typedef QMap<fn::InputType::Disease, IDModelMap> DiseaseModelMap;
    typedef QMap<fn::InputType::Country, DiseaseModelMap> CountryModelMap;

    // Constructor.
    //
    explicit Chief();

    // Destructor.
    //
    virtual ~Chief();

    // Reinitialize
    void reinitialize();

    // Allocate new ID
    fn::DiseaseModel::ID allocateNewModelID(const QString candidate = "");

    // Show infection data
    ui::DataPresenter* showInfectionData(const fn::DiseaseModel* diseaseModel, const fn::OutcomeTree::ID& outcomeTreeID);

    // Show sequela health state data
    ui::DataPresenter* showYLDHealthStateData(const fn::DiseaseModel* diseaseModel, const fn::OutcomeTree::ID& outcomeTreeID, const fn::OutcomeTreeNode::ID& healthStateID);

    // Show death health state data
    ui::DataPresenter* showYLLHealthStateData(const fn::DiseaseModel* diseaseModel, const fn::OutcomeTree::ID& outcomeTreeID, const fn::OutcomeTreeNode::ID& healthStateID);

    // Show asymptomatic health state data
    ui::DataPresenter* showAsymptomaticData(const fn::DiseaseModel* diseaseModel, const fn::OutcomeTree::ID& outcomeTreeID, const fn::OutcomeTreeNode::ID& healthStateID);

    // Show transition data
    ui::DataPresenter* showTransProbData(const fn::DiseaseModel* diseaseModel, const fn::OutcomeTree::ID& outcomeTreeID, const fn::TransitionProbability::ID& transProbID);

    // Compute aggregated results
    void computeAggregatedResults();

    // Push aggregated results to GUI
    void showAggregatedResults();

    QByteArray getDetailedResultsArray();

    // Copy and assignment operators. Intentionally not implemented.
    Chief(const Chief&);
    Chief& operator=(const Chief&);

    // Disease model map
    fn::DiseaseModelList m_diseaseModelList;
    CountryModelMap m_diseaseModelMap;

    // Last used disease model ID
    size_t m_lastModelId;
    QVector<QString> m_usedModelID;

    // Reference data
    fn::InputDataMgr m_inputDataMgr;

    // Reference data
    fn::JSONInputDataMgr m_jsonInputDataMgr;

    // Directory of current executable
    QString m_appPath;

    // Keep list of selected countries
    QStringList m_selCountries;

    // Keep list of selected diseases
    QStringList m_selDiseases;

    // Web gui
    ui::WebGUI* m_webGUI;

    // Number of samples per variable
    int m_numSamples;

    // Random seed
    int m_randomSeed;

    // Time discount rate
    double m_timeDiscountRate;

    // Presenter manager
    ui::PresenterMgr m_presenterMgr;

    // Disease report manager
    fn::DiseaseReportMgr m_diseaseReportMgr;

    // Results computer
    fn::ResultsComputer m_resultsComputer;

    // Application settings
    fn::Settings m_settings;

    // Google Analytics
    GAnalytics m_tracker;
};

} // namespace ui

#endif // _ui_Chief_
