//==========================================================================================================================
//  'WebGUI' class
//			This class implements graphical user interface pages
//
//  Written by Daniel Lewandowski
//  Copyright @ 2011 NextPage Software
//==========================================================================================================================

#ifndef _ui_WebGUI_
#define _ui_WebGUI_

#include <QObject>
#include <QVariantList>
#include <QWebEnginePage>

#include "fn/OutcomeTree.h"

namespace fn {
    class DataArray;
}

namespace ui {

class Chief;
class YLDHealthOutcomeDataPresenter;
class YLLHealthOutcomeDataPresenter;
class TransProbDataPresenter;

class WebGUI : public QObject
{
    Q_OBJECT

public:

    // Constructor.
    //
    WebGUI();

    // Destructor.
    //
    virtual ~WebGUI();

    // Set Chief
    //
    // 'chief' - [in] Chief
    //
    void setChief(ui::Chief* chief);

    // Set page
    //
    // 'page' - [in] Web page
    //
    void setPage(QWebEnginePage* page);

    // JavaScript to C++ methods

    // Indicate that load of the page has finished and data can be requested
    Q_INVOKABLE void dataRequested();

    // Indicate that weblink has been clicked.
    Q_INVOKABLE void webLinkClicked(const QString& webLink);

    // Indicate that a page has opened.
    Q_INVOKABLE void pageOpened(int tabIndex, const QString& pageName);

    // Get list of available diseases
    Q_INVOKABLE QVariantList getAvailableDiseases() const;

    // Retrieve selected countries
    Q_INVOKABLE void setSelectedCountries(const QStringList& countryIDs);

    // Retrieve selected diseases
    Q_INVOKABLE void setSelectedDiseases(const QStringList& diseaseIDs);

    // Create models (selected countries and diseases are already submitted)
    Q_INVOKABLE void createModels();

    // Create models (selected countries and diseases are already submitted)
    Q_INVOKABLE void deleteModels();

    // Select model
    Q_INVOKABLE void selectModel(
            const QString& countryID,
            const QString& diseaseID,
            const QString& modelID);

    // Detailed results page selection changed
    Q_INVOKABLE void showDetailedResults(
            const QString& countryIdStr,
            const QString& diseaseIdStr,
            const QString& modelId);

    // Show details of an outcome tree element
    Q_INVOKABLE void showOutcomeTreeElementDetails(
            const QString& outcomeTreeIdStr,
            const QString& elementType,
            const QString& elementId,
            const QString& countryIdStr,
            const QString& diseaseIdStr,
            const QString& modelId);

    // Apply changes to correction parameters in model edit page
    //
    Q_INVOKABLE void applyChangesToModelParams(
            const QVariantMap& modelInfo,
            const int estimType,
            const QVariantMap& underestimationData,
            const QVariantMap& underreportingData,
            const QVariantMap& underascertainmentData,
            const QVariantMap& caseBeforeCorrData,
            const QVariantMap& lifeExpData,
            const QVariantMap& ageDistrData);

    // Apply changes to correction parameters in YLD dialog
    //
    Q_INVOKABLE void applyChangesToYLDParams(
            const QVariantMap& modelInfo,
            const QVariantMap& severityData,
            const QVariantMap& durationData,
            const QVariantMap& proportionData);

    // Apply changes to correction parameters in YLD dialog
    //
    Q_INVOKABLE void applyChangesToYLLParams(
            const QVariantMap& modelInfo,
            const QVariantMap& proportionData);

    // Apply changes to correction parameters in Trans Prob dialog
    //
    Q_INVOKABLE void applyChangesToTransParams(
            const QVariantMap& modelInfo,
            const QVariantMap& probabilityData,
            const QVariantMap& redistributionData);

    // Get the number of samples from the engine
    //
    Q_INVOKABLE int getNumSamples();

    // Get the time discount rate from the engine
    //
    Q_INVOKABLE double getTimeDiscountRate();

    // Set the number of samples in the engine
    //
    Q_INVOKABLE void setNumSamples(const int numSamples);

    // Set the time discount rate in the engine
    //
    Q_INVOKABLE void setTimeDiscountRate(const double timeDiscountRate);

    // Start models run
    //
    Q_INVOKABLE void runModels();

    // Copy text to clipboard
    //
    Q_INVOKABLE void copyTextToClipboard(const QString& text);

    // Paste from clipboard to stratified grid
    //
    Q_INVOKABLE QVariantMap pasteFromClipboardToGrid(
            const QStringList& rowNames,
            const QStringList& colNames,
            const QString& distrType);

    // Copy image to clipboard
    //
    Q_INVOKABLE void copyPngImageToClipboard(const QString& base64Str);

    // Save image to file
    //
    Q_INVOKABLE void savePngImageToFile(const QString& base64Str);

    // Print page
    //
    Q_INVOKABLE void printPage();

    // Save page as pdf
    //
    Q_INVOKABLE void exportPageToPdf();

signals:

    // Relay JavaScript to C++ notifications
    void pageOpenNotified(int, const QString&);
    void countrySelectionChangeRequested(const QStringList&);
    void diseaseSelectionChangeRequested(const QStringList&);
    void createModelsRequested();
    void deleteModelsRequested();
    void modelSelectionRequested(const QString&, const QString&, const QString&);
    void showingOutcomeTreeElementDetailsRequested(const fn::OutcomeTree::ID&, const QString&, const fn::OutcomeTreeNode::ID&, const fn::InputType::Country&, const fn::InputType::Disease&, const QString&);
    void applyingChangesToModelParamsRequested(const QVariantMap&, const int, const QVariantMap&, const QVariantMap&,
                                               const QVariantMap&, const QVariantMap&, const QVariantMap&, const QVariantMap&);
    void applyingChangesToYLDParamsRequested(const QVariantMap&, const QVariantMap&, const QVariantMap&, const QVariantMap&);
    void applyingChangesToYLLParamsRequested(const QVariantMap&, const QVariantMap&);
    void applyingChangesToTransParamsRequested(const QVariantMap&, const QVariantMap&, const QVariantMap&);
    void settingNumSamplesRequested(const int);
    void settingTimeDiscountRateRequested(const double);
    void runningModelsRequested();
    void showingDetailedResultsRequested(const fn::InputType::Country&, const fn::InputType::Disease&, const QString&);

    // C++ to JavaScript notifications
    void programReinitialized();
    void showMessage(const QString&, const QString&, const QString&);
    void showMessageInYLDPage(const QString&, const QString&, const QString&);
    void showMessageInYLLPage(const QString&, const QString&, const QString&);
    void showMessageInTransPage(const QString&, const QString&, const QString&);
    void modelsCreated(const QVariantList&);
    void modelSelected(const QVariantMap&, const QVariantMap&, const QVariantMap&, const QVariantMap&, const QVariantMap&, const QString&);
    void detailedResultsAvailable(const QVariantMap&);
    void aggregatedResultsAvailable(const QVariantMap&);
    void formulaCheckResultsAvailable(const QVariantMap&);
    void formulaCheckResultsAvailableForYLDPage(const QVariantMap&);
    void formulaCheckResultsAvailableForYLLPage(const QVariantMap&);
    void formulaCheckResultsAvailableForTransPage(const QVariantMap&);
    void createUpdatableMessage(const QString&, const QString&);
    void updateMessage(const QString&, const QString&);
    void destroyUpdatableMessage();
    void runStarted();
    void runFinished(const QVariantList&);
    void runProgressUpdated(const float, const int, const int, const QString&);

private slots:

    void showPrintPreview(QPrinter* printer);

private:

    // Convert base64 encoded string to PNG image
    //
    QImage getPngImageFromBase64Str(const QString& base64Str);

    ui::Chief* m_chief;
    QWebEnginePage* m_page;
    QPrinter* m_printer;
};

}  // namespace ui

#endif // _ui_WebGUI_
