//==========================================================================================================================
//  Implementation file for "Excel.h"
//  Written by Daniel Lewandowski
//  Copyright @ 2011 NextPage Software
//==========================================================================================================================

#include <QTextStream>
#include <QFileInfo>
#include <QDir>
#include "ut/globals.h"
#include "ut/Excel.h"

namespace ut {

Excel::Excel(const QString& fileName) :
    m_excelApp(nullptr),
    m_workbooks(nullptr),
    m_workbook(nullptr),
    m_worksheets(nullptr),
    m_worksheet(nullptr)
{
    try {
        m_excelApp = new QAxObject("Excel.Application");
        m_excelApp->setProperty("Visible", false);
        m_excelApp->setProperty("DisplayAlerts", false);
        if (m_excelApp) {
            m_workbooks = m_excelApp->querySubObject("Workbooks");
            if (m_workbooks) {
                QFileInfo file(fileName);
                if (file.exists()) {
                    QVariantList params;
                    params.append(QVariant(QDir::toNativeSeparators(file.canonicalFilePath()))); // file name
                    params.append(QVariant());                                                   // update links
                    params.append(QVariant(true));                                               // read-only
                    m_workbook = m_workbooks->querySubObject("Open(QVariant, QVariant, QVariant)", params);
                } else {
                    m_workbook = m_workbooks->querySubObject("Add");
                }
                if (m_workbook) {
                    m_worksheets = m_workbook->querySubObject("Worksheets");
                    m_worksheet = setCurrentSheet(1);
                }
            } else
                QMessageBox::information(nullptr, "MS Office - Excel", "Excel application is not available");
        } else
            QMessageBox::information(nullptr, "MS Office - Excel", "Excel application is not available");
    } catch (...) {}
}

Excel::Excel() :
    m_excelApp(nullptr),
    m_workbooks(nullptr),
    m_workbook(nullptr),
    m_worksheets(nullptr),
    m_worksheet(nullptr)
{
    try {
        m_excelApp = new QAxObject("Excel.Application");
        m_excelApp->setProperty("Visible", false);
        m_excelApp->setProperty("DisplayAlerts", false);
        m_excelApp->setProperty("ScreenUpdating", false);
        if (m_excelApp) {
            m_workbooks = m_excelApp->querySubObject("Workbooks");
            if (m_workbooks) {
                m_workbook = m_workbooks->querySubObject("Add");
                m_worksheets = m_workbook->querySubObject("Worksheets");
                m_worksheet = setCurrentSheet(1);
            } else
                QMessageBox::information(nullptr, "MS Office - Excel", "Excel application is not available");
        } else
            QMessageBox::information(nullptr, "MS Office - Excel", "Excel application is not available");
    } catch (...) {}
}

Excel::~Excel()
{
    closeAll();
}

void Excel::openWorkbook(const QString& fileName)
{
    if (m_workbooks) {
        closeWorkbook();
        QFileInfo file(fileName);
        if (file.exists()) {
            QVariantList params;
            params.append(QVariant(QDir::toNativeSeparators(file.canonicalFilePath()))); // file name
            params.append(QVariant());                                   // update links
            params.append(QVariant(true));                               // read-only
            m_workbook = m_workbooks->querySubObject("Open(QVariant, QVariant, QVariant)", params);
            if (m_workbook) {
                m_worksheets = m_workbook->querySubObject("Worksheets");
                m_worksheet = setCurrentSheet(1);
            }
        }
    }
}

void Excel::closeWorkbook()
{
    if (m_workbook) {
        m_workbook->dynamicCall("Close()");
        m_workbook = nullptr;
        m_worksheets = nullptr;
        m_worksheet = nullptr;
    }
}

void Excel::saveWorkbook(const QString& fileName)
{
    if (m_workbook) {
        m_workbook->dynamicCall("SaveAs(const QString&, int)", fileName, 56); // Excel 97-2003 format
    }
}

int Excel::getWorksheetCount() const
{
    int result = 0;
    if (m_worksheets) {
        result = m_worksheets->property("Count").toInt();
    }
    return result;
}

QAxObject* Excel::appendWorksheet(const QString& sheetName)
{
    int sheetCount = getWorksheetCount();
    QAxObject* lastSheet = m_worksheets->querySubObject("Item(int)", sheetCount);
    QVariantList params;
    params.append(QVariant());
    params.append(lastSheet->asVariant());
    m_worksheet = m_worksheets->querySubObject("Add(QVariant, QVariant)", params);
    setWorksheetName(m_worksheet, sheetName);
    return m_worksheet;
}

void Excel::deleteWorksheet(const int index)
{
    if (index <= getWorksheetCount()) {
        m_worksheets->querySubObject("Item(int)", index)->dynamicCall("Delete");
    }
}

void Excel::setRangeValues(const QPoint& topLeft, const QVariant& value, const int numCols, int& lastRow)
{
    const int numRows = value.toList().length();
    const int startRow = topLeft.x();
    const int startCol = topLeft.y();
    const int endRow = startRow + numRows - 1;
    const int endCol = startCol + numCols - 1;

    QAxObject* topLeftCell = m_worksheet->querySubObject("Cells(int, int)", startRow, startCol);
    QAxObject* bottomRightCell = m_worksheet->querySubObject("Cells(int, int)", endRow, endCol);
    QAxObject* range = m_worksheet->querySubObject("Range(QVariant, QVariant)",
                                                   topLeftCell->asVariant(),
                                                   bottomRightCell->asVariant());
    range->setProperty("Value", value);

    lastRow = endRow + 1;
}

QVariant Excel::getRangeValues(const QAxObject* range)
{
    return range->property("Value");
}

void Excel::setCellValue(int row, int column, const QVariant& value)
{
    QAxObject* cell = getCell(row, column);
    if (cell) {
        cell->setProperty("Value", value);
    }
}

QVariant Excel::getCellValue(int row, int column)
{
    QAxObject* cell = getCell(row, column);
    return getCellValue(cell);
}

QString Excel::getStringCellValue(int row, int column)
{
    return getCellValue(row, column).toString().trimmed();
}

void Excel::setNumberFormat(QAxObject* range, const QString& format)
{
    if (!format.isEmpty()) {
        range->setProperty("NumberFormat", format);
    }
}

void Excel::setFontWeight(QAxObject* range, const QString& fontWeight)
{
    if (!fontWeight.isEmpty()) {
        QAxObject* rangeFont = range->querySubObject("Font");
        rangeFont->setProperty("Bold", fontWeight.toLower() == "bold");
    }
}

void Excel::setFontSize(QAxObject* range, int fontSize)
{
    if (fontSize > 0) {
        QAxObject* rangeFont = range->querySubObject("Font");
        rangeFont->setProperty("Size", fontSize);
    }
}

void Excel::setProperty(QAxObject* range, const QString& propertyName, const QString& propertyValue)
{
    if (!propertyName.isEmpty() & !propertyValue.isEmpty()) {
        range->setProperty(propertyName.toStdString().c_str(), propertyValue);
    }
}

void Excel::setWorksheetName(QAxObject* sheet, const QString& sheetName)
{
    sheet->setProperty("Name", sheetName);
}

QVariant Excel::getCellValue(QAxObject *cell)
{
    QVariant result;
    if (cell) {
        result = cell->property("Value");
    }
    return result;
}

QString Excel::getStringCellValue(QAxObject *cell)
{
    return getCellValue(cell).toString().trimmed();
}

void Excel::setCellValue(QAxObject* cell, const QVariant& value)
{
    cell->setProperty("Value", value);
}

int Excel::getCellRow(QAxObject* cell)
{
    return cell ? cell->property("Row").toInt() : -1;
}

void Excel::closeAll()
{
    if (m_excelApp) {
        if (m_workbook) {
            m_workbook->dynamicCall("Close(Boolean)", false);
            UT_DELETE(m_worksheet);
            UT_DELETE(m_worksheets);
            UT_DELETE(m_workbook);
        }
        m_excelApp->dynamicCall("Quit()");
        UT_DELETE(m_workbooks);
        UT_DELETE(m_excelApp);
    }
}

QAxObject* Excel::getCell(int row, int column)
{
    if (m_worksheet) {
        return m_worksheet->querySubObject("Cells(int, int)", row, column);
    } else {
        return nullptr;
    }
}

QAxObject* Excel::getCellWithText(const QString& text, const QString colID, const bool matchCase)
{
    QAxObject* source;
    if (!colID.isEmpty()) {
        source = m_worksheet->querySubObject("Range(const QString&)", colID + ":" + colID);
    } else {
        source = m_worksheet;
    }
    QVariantList params;
    params.append(QVariant(text));      // What
    params.append(QVariant());          // After
    params.append(QVariant());          // LookIn
    params.append(QVariant(1));         // LookAt
    params.append(QVariant());          // SearchOrder
    params.append(QVariant());          // SearchDirection
    params.append(QVariant(matchCase)); // MatchCase

    return source->querySubObject("Find(QVariant, QVariant, QVariant, QVariant, QVariant, QVariant, QVariant)", params);
}

QAxObject* Excel::getRange(int top, int left, int bottom, int right)
{
    if (m_worksheet) {
        QAxObject* topLeftCell = m_worksheet->querySubObject("Cells(int, int)", top, left);
        QAxObject* bottomRightCell = m_worksheet->querySubObject("Cells(int, int)", bottom, right);
        return m_worksheet->querySubObject("Range(QVariant, QVariant)", topLeftCell->asVariant(), bottomRightCell->asVariant());
    } else {
        return nullptr;
    }
}

QAxObject* Excel::getRange(int top, int left)
{
    if (m_worksheet) {
        QAxObject* topLeftCell = m_worksheet->querySubObject("Cells(int, int)", top, left);
        int bottom = topLeftCell->querySubObject("End(xlDown)")->property("Row").toInt();
        int right = topLeftCell->querySubObject("End(xlToRight)")->property("Column").toInt();
        QAxObject* bottomRightCell = m_worksheet->querySubObject("Cells(int, int)", bottom, right);
        return m_worksheet->querySubObject("Range(QVariant, QVariant)", topLeftCell->asVariant(), bottomRightCell->asVariant());
    } else {
        return nullptr;
    }
}

QAxObject* Excel::getRange(const QString& string)
{
    if (m_worksheet) {
        return m_worksheet->querySubObject("Range(const QString&)", string);
    } else {
        return nullptr;
    }
}

QAxObject* Excel::setCurrentSheet(int index)
{
    if (m_worksheets) {
        try {
            m_worksheet = m_worksheets->querySubObject("Item(int)", index);
        } catch (...) {
            return nullptr;
        }
    }
    return m_worksheet;
}

QAxObject* Excel::getCurrentSheet() const
{
    return m_worksheet;
}

const QString Excel::getCurrentSheetName() const
{
    return m_worksheet->dynamicCall("Name()").toString().trimmed();
}


}  // namespace ut
