//==========================================================================================================================
//  'Excel' class
//			This class implements interface to Excel
//
//  Written by Daniel Lewandowski
//  Copyright @ 2011 NextPage Software
//==========================================================================================================================

#ifndef _ut_Excel_
#define _ut_Excel_

#include <QAxObject>

namespace ut {

class Excel
{
public:

    // Constructor.
    //
    // 'fileName' - [in] Name of the Excel file to open
    //
    Excel(const QString& fileName);
    Excel();

    // Destructor.
    //
    virtual ~Excel();

    // Open workbook
    //
    // 'fileName' - [in] Name of the file to open
    //
    void openWorkbook(const QString& fileName);

    // Close current workbook
    //
    void closeWorkbook();

    // Save workbook
    //
    // 'fileName' - [in] Name of the file
    //
    void saveWorkbook(const QString& fileName);

    // Get the count of worksheets
    //
    // Return - count
    int getWorksheetCount() const;

    // Append a worksheet at the end
    //
    // 'sheetName' - [in] Name to be assigned to the newly created sheet
    //
    // Return - worksheet object
    QAxObject* appendWorksheet(const QString& sheetName);

    // Delete worksheet
    //
    // 'index' - [in] Index of the worksheet
    //
    void deleteWorksheet(const int index);

    // Get range object
    //
    // 'top' - [in] Row index of the top left cell of the range
    // 'left' - [in] Column index of the top left cell of the range
    // 'bottom' - [in] Row index of the bottom right cell of the range
    // 'right' - [in] Column index of the bottom right cell of the range
    //
    // Return - range object
    QAxObject* getRange(int top, int left, int bottom, int right);

    // Get range object with top left cell. Determine bottom right cell automatically.
    //
    // 'top' - [in] Row index of the top left cell of the range
    // 'left' - [in] Column index of the top left cell of the range
    //
    // Return - range object
    QAxObject* getRange(int top, int left);

    // Get range object with cell reference string
    //
    // 'string' - [in] Excel range string
    //
    // Return - range object
    QAxObject* getRange(const QString& string);

    // Get cell object
    //
    // 'row' - [in] Row
    // 'column' - [in] Column
    //
    // Return - cell object
    QAxObject* getCell(int row, int column);

    // Get first cell with text
    //
    // 'text' - [in] Text
    // 'colID' - [in] (Optional) Limit search to the specified column
    // 'matchCase' - [in] (Optional) Case sensitive search
    //
    // Return - cell
    QAxObject* getCellWithText(const QString& text, const QString colID = "A", const bool matchCase = false);

    // Set values in range
    //
    // 'topLeft' - [in] Coordinates of the top left cell of the range to be filled with values.
    // 'value' - [in] List of arrays of values. Each array corresponds to a single row in the Excel file.
    // 'numCols' - [in] Number of colums in the range to set.
    // 'lastRow' - [out] Row index of the last cell of the range filled with values.
    //
    void setRangeValues(const QPoint& topLeft, const QVariant& value, const int numCols, int& lastRow);

    // Get values from range
    //
    // 'rangeRect' - [in] Coordinates of the top left and bottom right cell of the range with values.
    //
    // Return - Value
    static QVariant getRangeValues(const QAxObject* range);

    // Set value in cell
    //
    // 'row' - [in] Row index of the cell.
    // 'column' - [in] Column index of the cell.
    // 'value' - [in] Value
    //
    void setCellValue(int row, int column, const QVariant& value);

    // Get value from cell
    //
    // 'row' - [in] Row index of the cell.
    // 'column' - [in] Column index of the cell.
    //
    // Return - Value
    QVariant getCellValue(int row, int column);

    // Get value as string from cell
    //
    // 'row' - [in] Row index of the cell.
    // 'column' - [in] Column index of the cell.
    //
    // Return - Value
    QString getStringCellValue(int row, int column);

    // Set current sheet
    //
    // 'index' - [in] Index of the sheet to be set as current (1-based).
    //
    // Return - current sheet
    QAxObject* setCurrentSheet(int index);

    // Get current sheet
    //
    // Return - current sheet
    QAxObject* getCurrentSheet() const;

    // Get current sheet name
    //
    // Return - current sheet
    const QString getCurrentSheetName() const;

    // Set number format in range.
    //
    // 'range' - [in] Range to be formatted
    // 'format' - [in] String containing number format (for instance "@" denotes Text format")
    //
    static void setNumberFormat(QAxObject* range, const QString& format);

    // Set font weight in range
    //
    // 'range' - [in] Range where font weight is to be set
    // 'fontWeight' - [in] String containing font weight (for instance "bold")
    //
    static void setFontWeight(QAxObject* range, const QString& fontWeight);

    // Set font size in range
    //
    // 'range' - [in] Range where font size is to be set
    // 'fontSize' - [in] Font size
    //
    static void setFontSize(QAxObject* range, int fontSize);

    // Set a property of range
    //
    // 'range' - [in] Range where property is to be set
    // 'propertyName' - [in] Property name
    // 'propertyValue' - [in] Property value
    //
    static void setProperty(QAxObject* range, const QString& propertyName, const QString& propertyValue);

    // Set worksheet name
    //
    // 'sheet' - [in] Worksheet object
    // 'sheetName' - [in] Worksheet name
    //
    static void setWorksheetName(QAxObject* sheet, const QString& sheetName);

    // Get value in cell
    //
    // 'cell' - [in] Cell.
    //
    // Return - Cell value
    static QVariant getCellValue(QAxObject* cell);

    // Get value in cell as string
    //
    // 'cell' - [in] Cell.
    //
    // Return - Cell value as string
    static QString getStringCellValue(QAxObject* cell);

    // Set value in cell
    //
    // 'cell' - [in] Cell.
    // 'value' - [in] Value
    //
    static void setCellValue(QAxObject* cell, const QVariant& value);

    // Get row index of the cell
    //
    // Return - row
    static int getCellRow(QAxObject* cell);

    // Close the file if any and release all object
    //
    void closeAll();

private:

    QAxObject* m_excelApp;
    QAxObject* m_workbooks;
    QAxObject* m_workbook;
    QAxObject* m_worksheets;
    QAxObject* m_worksheet;
};

}  // namespace ut

#endif  // _ut_Excel_



