//==========================================================================================================================
//  Implementation file for "SmtpClient.h"
//  Copyright @ 2019 Daniel Lewandowski
//==========================================================================================================================

#include "ut/SmtpClient.h"

namespace ut {

SmtpClient::SmtpClient(const QString& user, const QString& pass, const QString& host, int port, int timeout)
{
    socket = new QSslSocket(this);

    connect(socket, SIGNAL(readyRead()), this, SLOT(readyRead()));
    connect(socket, SIGNAL(connected()), this, SLOT(connected() ) );
    connect(socket, SIGNAL(error(QAbstractSocket::SocketError)), this,SLOT(errorReceived(QAbstractSocket::SocketError)));
    connect(socket, SIGNAL(stateChanged(QAbstractSocket::SocketState)), this, SLOT(stateChanged(QAbstractSocket::SocketState)));
    connect(socket, SIGNAL(disconnected()), this,SLOT(disconnected()));

    this->user = user;
    this->pass = pass;

    this->host = host;
    this->port = port;
    this->timeout = timeout;
}

void SmtpClient::sendMail(const QString& from, const QString& to, const QString& subject, const QString& body, const QString& fileName, const QByteArray& fileContent)
{
    message = "To: " + to + "\n";
    message.append("From: " + from + "\n");
    message.append("Subject: " + subject + "\n");

    //Let's intitiate multipart MIME with cutting boundary "frontier"
    message.append("MIME-Version: 1.0\n");
    message.append("Content-Type: multipart/mixed; boundary=frontier\n\n");

    message.append( "--frontier\n" );
    //message.append( "Content-Type: text/html\n\n" );  //Uncomment this for HTML formating, coment the line below
    message.append( "Content-Type: text/plain\n\n" );
    message.append(body);
    message.append("\n\n");

    if (!fileName.isEmpty()) {
        message.append( "--frontier\n" );
        message.append( "Content-Type: application/octet-stream\nContent-Disposition: attachment; filename=" + fileName +";\nContent-Transfer-Encoding: base64\n\n");
        message.append(fileContent.toBase64());
        message.append("\n");
    }

    message.append( "--frontier--\n" );

    message.replace(QString::fromLatin1("\n"),
                    QString::fromLatin1("\r\n"));
    message.replace(QString::fromLatin1("\r\n.\r\n"),
                    QString::fromLatin1("\r\n..\r\n"));

    this->from = from;
    rcpt = to;
    state = Init;
    socket->connectToHostEncrypted(host, static_cast<quint16>(port)); //"smtp.gmail.com" and 465 for gmail TLS
    if (!socket->waitForConnected(timeout)) {
         qInfo() << socket->errorString();
    }

    t = new QTextStream( socket );
}

SmtpClient::~SmtpClient()
{
    delete t;
    delete socket;
}
void SmtpClient::stateChanged(QAbstractSocket::SocketState socketState)
{
    qInfo() << "stateChanged " << socketState;
}

void SmtpClient::errorReceived(QAbstractSocket::SocketError socketError)
{
    qInfo() << "error " << socketError;
}

void SmtpClient::disconnected()
{
    qInfo() << "disconneted";
    qInfo() << "error " << socket->errorString();
}

void SmtpClient::connected()
{
    qInfo() << "Connected ";
}

void SmtpClient::readyRead()
{
     qInfo() << "readyRead";
    // SMTP is line-oriented

    QString responseLine;
    do {
        responseLine = socket->readLine();
        response += responseLine;
    }
    while (socket->canReadLine() && responseLine[3] != ' ');

    responseLine.truncate(3);

    qInfo() << "Server response code:" <<  responseLine;
    qInfo() << "Server response: " << response;
    qInfo() << QSslSocket::sslLibraryBuildVersionString();

    if (state == Init && responseLine == "220") {
        // banner was okay, let's go on
        *t << "EHLO localhost" <<"\r\n";
        t->flush();

        state = HandShake;
    } else if (state == HandShake && responseLine == "250") {
        socket->startClientEncryption();
        if (!socket->waitForEncrypted(timeout)) {
            qInfo() << socket->errorString();
            state = Close;
        }

        //Send EHLO once again but now encrypted

        *t << "EHLO localhost" << "\r\n";
        t->flush();
        state = Auth;
    } else if (state == Auth && responseLine == "250") {
        // Trying AUTH
        qInfo() << "Auth";
        *t << "AUTH LOGIN" << "\r\n";
        t->flush();
        state = User;
    } else if (state == User && responseLine == "334") {
        //Trying User
        qInfo() << "Username";
        //GMAIL is using XOAUTH2 protocol, which basically means that password and username has to be sent in base64 coding
        //https://developers.google.com/gmail/xoauth2_protocol
        *t << QByteArray().append(user).toBase64()  << "\r\n";
        t->flush();

        state = Pass;
    } else if (state == Pass && responseLine == "334") {
        //Trying pass
        qInfo() << "Pass";
        *t << QByteArray().append(pass).toBase64() << "\r\n";
        t->flush();

        state = Mail;
    } else if (state == Mail && responseLine == "235") {
        // HELO response was okay (well, it has to be)

        //Apperantly for Google it is mandatory to have MAIL FROM and RCPT email formated the following way -> <email@gmail.com>
        qInfo() << "MAIL FROM:<" << from << ">";
        *t << "MAIL FROM:<" << from << ">\r\n";
        t->flush();
        state = Rcpt;
    } else if (state == Rcpt && responseLine == "250") {
        //Apperantly for Google it is mandatory to have MAIL FROM and RCPT email formated the following way -> <email@gmail.com>
        *t << "RCPT TO:<" << rcpt << ">\r\n"; //r
        t->flush();
        state = Data;
    } else if (state == Data && responseLine == "250") {
        *t << "DATA\r\n";
        t->flush();
        state = Body;
    } else if (state == Body && responseLine == "354") {
        *t << message << "\r\n.\r\n";
        t->flush();
        state = Quit;
    } else if (state == Quit && responseLine == "250") {
        *t << "QUIT\r\n";
        t->flush();
        // here, we just close.
        state = Close;
        emit status(tr("Message sent"));
    } else if (state == Close) {
        deleteLater();
        return;
    } else {
        // something broke.
        QMessageBox::warning(nullptr, tr("Qt Simple SMTP client"), tr("Unexpected reply from SMTP server:\n\n") + response);
        state = Close;
        emit status(tr("Failed to send message"));
    }
    response = "";
}

} // namespace ut

